{ program LQTBareBoneExample
  a bare bone example of the QuadTree Usage
  Copyright (C) 2024 Ekkehard Domning (www.domis.de) ekkehard@domning.eu

  License: modified LGPL with linking exception (like RTL, FCL and LCL)

  See the file COPYING.modifiedLGPL.txt, included in the Lazarus distribution,
  for details about the license.
}
program LQTBareBoneExample;

{$mode objfpc}{$H+}

uses
  {$IFDEF UNIX}
  cthreads,
  {$ENDIF}
  Classes,SysUtils,
  ulazQuadTreeGeometry, ulazQuadTree;
type

  { TEventWrapper
    The elementary Events/CallBacks are defined as "of object", since this is the
    most convinient way for the most applications.
    But here we have to implement a wrapper class to allow the usage in this
    bare bone console application.
  }

  TEventWrapper = class(TObject)
  public
    function OnTestQuadTreeGetItemCoord(Sender : TObject; AItem : TObject; var AX, AY : Double) : Boolean;
    function OnTestQuadTreeGetItemCaption(Sender : TObject; AItem : TObject; var AItemCaption : String) : Boolean;
    function TreeItemEnumerationCallBack(Sender : TObject;
                                         ANode : TLQTNode;
                                         ANodeItem : TObject;
                                         ANodeItemIndex : Integer;
                                         AUserData : Pointer) : Boolean;
  end;

  { The POI (= Point of Interest) will be hold in an array of records
    The data are simply the coordinates and a textual description  }
  TPOIRec = record
    Lon : Double;
    Lat : Double;
    Descr : String;
  end;


var
  // Here the POIs are populated.
  // I choosed 10 POIs in the four quadrants of the globe
  // Caution: The array starts at the index value 1, to overcome a problem explained below
  POIs : array[1..10] of TPOIRec = (
    (Lon:    2.2945500; Lat:  48.8582300; Descr: 'Tour d''Eiffel, Paris'),          // https://en.wikipedia.org/wiki/Eiffel_Tower
    (Lon:  -79.3884000; Lat:  43.6439500; Descr: 'CN Tower, Toronto'),              // https://en.wikipedia.org/wiki/CN_Tower
    (Lon: -157.9748600; Lat:  21.3445700; Descr: 'Pearl Harbor, USA'),              // https://en.wikipedia.org/wiki/Pearl_Harbor
    (Lon:  114.1438890; Lat:  22.2755560; Descr: 'Victoria Peak, Hong Kong'),       // https://en.wikipedia.org/wiki/Victoria_Peak
    (Lon:   13.0479700; Lat:  52.3995200; Descr: 'Brandenburger Tor, Berlin'),      // https://en.wikipedia.org/wiki/Brandenburg_Gate
    (Lon:  -58.3722400; Lat: -34.6084700; Descr: 'Piramide de Mayo, Buenos Aires'), // https://en.wikipedia.org/wiki/Pir%C3%A1mide_de_Mayo
    (Lon:  151.2107600; Lat: -33.8519800; Descr: 'Harbour Bridge, Sydney'),         // https://en.wikipedia.org/wiki/Sydney_Harbour_Bridge
    (Lon:  139.7021800; Lat:  35.6787500; Descr: 'Meiji Jingu Shrine, Tokyo'),      // https://en.wikipedia.org/wiki/Meiji_Shrine
    (Lon:   18.4124600; Lat: -33.9634700; Descr: 'Table Mountain, South Africa'),   // https://en.wikipedia.org/wiki/Table_Mountain
    (Lon:   31.1338600; Lat:  29.9788100; Descr: 'Cheops Pyramid, Cairo')           // https://en.wikipedia.org/wiki/Great_Pyramid_of_Giza
  );

// WriteLnPOI will write the textual representation of a single POI on one line on the Console
procedure WriteLnPOI(const APOIIndex : Integer);
begin
  if (APOIIndex >= Low(POIs)) and (APOIIndex <= High(POIs)) then
    WriteLn(Format('Lon/X: %1.6f, Lat/Y: %1.6f ''%s''',[POIs[APOIIndex].Lon,POIs[APOIIndex].Lat,POIs[APOIIndex].Descr]));
end;

{ TEventWrapper }

// OnTestQuadTreeGetItemCoord is called, when the Tree needs the
// Coordinates (X/Y) of a certain item
function TEventWrapper.OnTestQuadTreeGetItemCoord(Sender: TObject;
  AItem: TObject; var AX, AY: Double): Boolean;
var
  // This declaration is a kind of typeCast to the AItem parameter
  // It overlays AItem and allows to read the Value a a NativeInt
  // which has the same size a a Pointer or a TObject
  Ndx : NativeInt absolute AItem;
begin
  // Check if the Ndx is insiode the array above
  Result := (Ndx >= Low(POIs)) and (Ndx <= High(POIs));
  if Result then
  begin // If True, return the value of the given POI
    AX := POIs[Ndx].Lon;
    AY := POIs[Ndx].Lat;
  end;
end;

// Same as OnTestQuadTreeGetItemCoord but for the Description String
function TEventWrapper.OnTestQuadTreeGetItemCaption(Sender: TObject;
  AItem: TObject; var AItemCaption: String): Boolean;
var
  Ndx : NativeInt absolute AItem;
begin
  Result := (Ndx >= Low(POIs)) and (Ndx <= High(POIs));
  if Result then
    AItemCaption := POIs[Ndx].Descr;
end;

// This callback is called for every Item in an enumeration process
function TEventWrapper.TreeItemEnumerationCallBack(Sender: TObject;
  ANode: TLQTNode; ANodeItem: TObject; ANodeItemIndex: Integer;
  AUserData: Pointer): Boolean;
var
  Ndx : NativeInt absolute ANodeItem;
begin
  Result := (Ndx >= Low(POIs)) and (Ndx <= High(POIs));
  if Result then
    WriteLnPOI(Ndx);
end;

// Will print the content of the QuadTree to the Console
procedure WriteLnQuadTree(const AQuadTree : TLazQuadTree);
var
  i : Integer;
  sl : TStringList;
begin
  sl := TStringList.Create;
  try
    // TreeAsString will return a verbose textual description of the tree content
    // For larger trees this method is more or less useless, because the list will be
    // very large.
    // But for debugging purpose, it might be very helpful
    // One could activate the TreeAsStrings call above after the insertion of every item to
    // see how the tree nodes are created and populated.
    // It is possible to perform the output for subparts of the tree using
    // the TLazQuadNode.NodeAsStrings method

    AQuadTree.TreeAsStrings(sl);
    for i := 0 to sl.Count-1 do
      WriteLn(sl[i]);
  finally
    sl.Free;
  end;
end;

// The main work procedure
procedure DoLQTWork;
var
  TestQuadTree : TLazQuadTree; // The QuadTree
  lqtEventWrapper : TEventWrapper; // The Eventwrapper
  lqtrect : TLQTRect; // a temporary rectangle
  i : Integer;
  d : Double;
  cnt : Integer;
  lList : TList; // A list for the listing of some items
begin
  lqtEventWrapper := TEventWrapper.Create; // Create the eventwrapper
  try
    // Rectangles are crucial for the function of the QuadTree, see the LQTGeometryTest-Example for all gory details
    lqtrect.SetFrom(-180.0,90.0,180.0,-90.0); // Define the world (Y-Axis is increasing upwards, like a Map
    TestQuadTree := TLazQuadTree.Create(lqtrect,lqtwmCylindrical); // Create the Tree. The World wraps around cylindrical. The Tree will not store objects
    try
      // Assign the two Events
      TestQuadTree.OnGetItemCoord := @lqtEventWrapper.OnTestQuadTreeGetItemCoord; // will be called, if the tree needs coordinates of the inserted items
      TestQuadTree.OnGetItemCaption := @lqtEventWrapper.OnTestQuadTreeGetItemCaption; // will be called if the tree needs the caption of the item
      WriteLn('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-');
      WriteLn('Inserting Items into the QuadTree');
      WriteLn('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-');
      // Insert the Items.
      // The InsertItem expect a value from the type TObject and will refuse the
      // insertion of a Nil Pointer/Class
      // Since we will use the defined array above and not create descendants of TObject
      // we could not insert a cast of an index with the value 0, which would be casted to Nil.
      // Thus the array is defined starting with the Index value of 1 (instead commonly 0).
      for i := Low(POIs) to High(POIs) do
      begin
        Write('Inserting Item: ');
        WriteLnPOI(i); // Print the POI to the console
        TestQuadTree.InsertItem(TObject(NativeInt(i))); // NativeInt will have the same size as a Pointer/Class
        // You may uncomment the following line for an interesting view on how the QuadTree becomes populated.
        // WriteLnQuadTree(TestQuadTree);
      end;
      WriteLn('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-');
      WriteLn('A human readable version of the QuadTree');
      WriteLn('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-');
      WriteLnQuadTree(TestQuadTree);

      // Here we will retrieve Items with the most used functions

      // First case, a regular rectangle on the northern hemisphere
      lqtrect.SetFrom(-80.0,60.0,20.0,30.0);
      // Optional, query the amount of Items in a specified area
      d := TestQuadTree.EstimatedCountOfAssignedItemsInRect(lqtrect,100); // ask for the estimated number of Items in the given rectangle (very fast)
      cnt := TestQuadTree.CountAssignedItemsInRect(lqtrect); //ask for the exact count of Items in the given rectangle (fast)
      WriteLn('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-');
      WriteLn(Format('Enumerating %d (estimated %1.1f) Items in the rectangle Left/Lon/X %1.6f; Top/Lat/Y %1.6f; Right/Lon/X %1.6f; Bottom/Lat/Y %1.6f',
                     [cnt, d, lqtrect.Left,lqtrect.Top,lqtrect.Right,lqtrect.Bottom]));
      // Enumerate the items in the rectangle using the call back function
      TestQuadTree.EnumerateItemsInRect(lqtrect,@lqtEventWrapper.TreeItemEnumerationCallBack);

      // Second case, a wrap around rectangle on the southern hemisphere, starting in Australia, ending in South America
      // South Africa is not included!
      lqtrect.SetFrom(140.0,20.0,-40.0,-50.0);
      d := TestQuadTree.EstimatedCountOfAssignedItemsInRect(lqtrect,100);
      cnt := TestQuadTree.CountAssignedItemsInRect(lqtrect);
      WriteLn('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-');
      WriteLn(Format('Enumerating %d (estimated %1.1f) Items in the rectangle Left/Lon/X %1.6f; Top/Lat/Y %1.6f; Right/Lon/X %1.6f; Bottom/Lat/Y %1.6f',
                     [cnt, d, lqtrect.Left,lqtrect.Top,lqtrect.Right,lqtrect.Bottom]));
      // Second method, instead using the enumeration of items, the summing up in a list is used
      lList := TList.Create;
      try
        TestQuadTree.ListItemsInRect(lqtrect,lList);
        for i := 0 to lList.Count-1 do
          WriteLnPOI(NativeInt(lList.Items[i]));
      finally
        lList.Free;
      end;

    finally
      TestQuadTree.Free; // Free the QuadTree
    end;
  finally
    lqtEventWrapper.Free; // Free the EventWrapper
  end;
end;

begin // of Program
  WriteLn('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-');
  WriteLn('A bare bone example of the LazQuadTree');
  WriteLn('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-');
  WriteLn('');
  DoLQTWork;
  WriteLn('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-');
  WriteLn('Done');
  WriteLn('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-');
end.

