{ Some common declarations for the QuadTree demo
  Copyright (C) 2023,2024 Ekkehard Domning (www.domis.de)

  License: modified LGPL with linking exception (like RTL, FCL and LCL)

  See the file COPYING.modifiedLGPL.txt, included in the Lazarus distribution,
  for details about the license.

  Version Date        Change
  0.0.1   2024-08-23  First release
}
unit uBasicTestCommons;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils,
  ulazQuadTreeGeometry, ulazQuadTree;

type
  { TTestPOI the class which will represent one single POI, which will be
    hold in the QuadTree  }
  TTestPOI = class(TObject)
  private
    FTag : Integer;
    FLat : Double;
    FLon : Double;
    FCaption : String;
  public
    property Tag : Integer read FTag write FTag;
    property Lat : Double read FLat;
    property Lon : Double read FLon;
    property Caption : String read FCaption;
    constructor Create(const ALat, ALon : Double; const ACaption : String);
    function SetPOIFromCSVLine(const ACSVLine : String; out AErrorMessage : String) : Boolean;
  end;

  { TTestShape hold shapes for one country}

  TTestShape = class(TTestPOI)
  private
    // The shape is stored in a single array of double, alternating x and y values
    FShape : array of Double;
    // The bounding box is created during loading
    FShapeBounds : TLQTRect;
    function GetCoordsCount : Integer;
    function GetCoordsLats(CoordIndex : Integer) : Double;
    function GetCoordsLons(CoordIndex : Integer) : Double;
    function GetShapeBounds : TLQTRect;
  public
    property CoordsCount : Integer read GetCoordsCount;
    property CoordsLats[CoordIndex : Integer] : Double read GetCoordsLats;
    property CoordsLons[CoordIndex : Integer] : Double read GetCoordsLons;
    property ShapeBounds : TLQTRect read GetShapeBounds;
    function ReadShapesFromString(const AShapeString : String) : Boolean;
    procedure SetShapesFromArray(const ALonLatArray : array of Double);
    constructor Create(const ACaption : String);
  end;

  { TTestItemStreamHelper }

  TTestItemStreamHelper = class(TObject)
  public
    class function OnTestQuadTreeReadStreamItem(AOwner : TLazQuadTree; ANode : TLQTNode;
                                          AItemIndex : Integer;
                                          AStream : TStream) : TObject;
    class procedure OnTestQuadTreeWriteStreamItem(AOwner : TLazQuadTree; ANode : TLQTNode;
                                            AItem : TObject; AItemIndex : Integer;
                                            AStream : TStream);
  end;


const
  // Flat world, like an ordinary bitmap, zero is top-left
  // Do not change!
  FlatWorld : TLQTRect = (
    Left :     0.0;
    Top :      0.0;
    Right :  360.0;
    Bottom : 180.0;
  );
  EarthWorld : TLQTRect = (
    Left :   -180.0;
    Top :      90.0;
    Right :   180.0;
    Bottom :  -90.0;
  );

  // Tokens for the Stream Read/Write
  TestPOIToken   = $ED504F49; // ED+'POI'
  TestShapeToken = $ED534850; // ED+'SHP'

  ExampleDataPath = '..\..\..\LQTExampleData\';

implementation

{ TTestPOI }

constructor TTestPOI.Create(const ALat, ALon: Double; const ACaption: String);
begin
  inherited Create;
  FLat := ALat;
  FLon := ALon;
  FCaption := ACaption;
end;

function TTestPOI.SetPOIFromCSVLine(const ACSVLine: String; out
  AErrorMessage: String): Boolean;
var
  line : TStringList;
  llat, llon : Double;
  fs : TFormatSettings;
  s : String;
  cnt : Integer;
begin
  Result := False;
  AErrorMessage := '';
  // One Line consists of three data fields, Longitude, Latitude and a descriptive string
  fs.ThousandSeparator:= #0;
  fs.DecimalSeparator:='.';
  line := TStringList.Create;
  try
    line.Delimiter:=',';
    line.StrictDelimiter := True;
    line.QuoteChar:='"';

    // Split the line into pieces
    line.DelimitedText := ACSVLine;
    cnt := line.Count;
    if cnt < 3 then // if less than three items, error and out
    begin
      AErrorMessage := 'Less than three items!';
      Exit;
    end;
    // Read the Longitude and Latitude from the line, the formatsettings variable ensure
    // the decimal point is used independend from the local windows setting (e.g. Germany uses a ",")
    s := line[0];
    llon := StrToFloatDef(s,-181,fs);
    s := line[1];
    llat := StrToFloatDef(s,-181,fs);
    if (llat <= -90) or (llat > 90) or
       (llon < -180) or (llon >= 180) then
    begin // The coords must be located within the earth borders
      AErrorMessage := 'Coords out of range!';
      Exit;
    end;
    s := line[2];
    FLat := llat;
    FLon := llon;
    FCaption := s;
    Result := True;
  finally
    line.Free;
  end;
end;


{ TTestShape }

function TTestShape.GetCoordsCount: Integer;
begin
  Result := Length(FShape) div 2;
end;

function TTestShape.GetCoordsLats(CoordIndex: Integer): Double;
begin
  Result := FShape[(CoordIndex shl 1)+1];
end;

function TTestShape.GetCoordsLons(CoordIndex: Integer): Double;
begin
  Result := FShape[(CoordIndex shl 1)];
end;

function TTestShape.GetShapeBounds: TLQTRect;
begin
  Result := FShapeBounds;
end;

function TTestShape.ReadShapesFromString(const AShapeString: String): Boolean;
var
  sl1 : TStringList;
  j : Integer;
  s : String;
  posi : Integer;
  fs : TFormatSettings;
  lats,lons : String;
  llat, llon : Double;
begin
  Result := False;
  fs.ThousandSeparator:= #0;
  fs.DecimalSeparator := '.';
  sl1 := TStringList.Create;
  try
    sl1.Delimiter:=';';
    sl1.StrictDelimiter := True;
    sl1.QuoteChar := #0;
    sl1.DelimitedText := AShapeString;
    SetLength(FShape,sl1.Count*2);
    for j := 0 to sl1.Count-1 do
    begin
      s := sl1[j];
      posi := Pos(',',s);
      lons := Copy(s,1,posi-1);
      lats := Copy(s,posi+1,MaxInt);
      llat := StrToFloat(lats,fs);
      llon := StrToFloat(lons,fs);

      // Simply limit the lat and lon to the earth extent
      // You may notice that the coords of shapes include the
      // right and bottom border of the EarthWorld.
      // This is possible since areas (aka this shape) does not include this two borders.
      // See further description in the geometric unit ulazQuadTreeGeometry.pas
      if llat > EarthWorld.Top then
        llat := EarthWorld.Top
      else if llat < EarthWorld.Bottom then
        llat := EarthWorld.Bottom;
      if llon < EarthWorld.Left then
        llon := EarthWorld.Left
      else if llon > EarthWorld.Right then
        llon := EarthWorld.Right;
      FShape[j*2] := llon;
      FShape[j*2+1] := llat;
      if j = 0 then
      begin // set the shape bounding box to the first lat/lon
        FShapeBounds.Left:=llon;
        FShapeBounds.Right:=llon;
        FShapeBounds.Top:=llat;
        FShapeBounds.Bottom:=llat;
      end
      else
      begin // Update the bounding box if the bnew lat/lon is outside the bounding box
        if FShapeBounds.Left > llon then
          FShapeBounds.Left := llon
        else if FShapeBounds.Right < llon then
          FShapeBounds.Right := llon;
        if FShapeBounds.Top < llat then
          FShapeBounds.Top := llat
        else if FShapeBounds.Bottom > llat then
          FShapeBounds.Bottom := llat;
      end;
      Result := True;
      FLat := (FShapeBounds.Top + FShapeBounds.Bottom) / 2.0;
      FLon := (FShapeBounds.Left + FShapeBounds.Right) / 2.0;

    end;
  finally
    if Assigned(sl1) then
      sl1.Free;
  end;
end;

procedure TTestShape.SetShapesFromArray(const ALonLatArray: array of Double);
var
  i : Integer;
  llon, llat : Double;
begin
  SetLength(FShape,Length(ALonLatArray));
  for i := 0 to High(ALonLatArray) do
    FShape[i] := ALonLatArray[i];
  FShapeBounds.Clear;
  for i := 0 to CoordsCount-1 do
  begin
    llon := CoordsLons[i];
    llat := CoordsLats[i];

    if i = 0 then
    begin // set the shape bounding box to the first lat/lon
      FShapeBounds.Left := llon;
      FShapeBounds.Right := llon;
      FShapeBounds.Top := llat;
      FShapeBounds.Bottom :=llat;
    end
    else
    begin // Update the bounding box if the bnew lat/lon is outside the bounding box
      if FShapeBounds.Left > llon then
        FShapeBounds.Left := llon
      else if FShapeBounds.Right < llon then
        FShapeBounds.Right := llon;
      if FShapeBounds.Top < llat then
        FShapeBounds.Top := llat
      else if FShapeBounds.Bottom > llat then
        FShapeBounds.Bottom := llat;
    end;
  end;
end;

constructor TTestShape.Create(const ACaption: String);
begin
  inherited Create(0,0,ACaption);
end;

{ TTestItemStreamHelper }
{ OnTestQuadTreeReadStreamItem
   This is the implementation for reading a user item (back) from stream.
   The Event/CallBack distinguish between the two different classes
   used in this demo: The POI and the derived Shapes.
}
class function TTestItemStreamHelper.OnTestQuadTreeReadStreamItem(AOwner: TLazQuadTree;
  ANode: TLQTNode; AItemIndex: Integer; AStream: TStream): TObject;
var
  c : Cardinal;
  poi : TTestPoi = Nil;
  llat, llon : Double;
  s : String;
  len : Integer;
  ds : array of Double = Nil;
  i : Integer;
begin
  s := ''; // Keep the compiler calm
  c := AStream.ReadDWord; // Read the token
  if (c <> TestPOIToken) and
     (c <> TestShapeToken) then // if not our token, than rais an exception
    raise ELQTStreamException.Create('Invalid token read from Stream!');
  AStream.ReadDWord; // Consume the length field (we dont do anything with it here)
  // Read the common data
  llon := ReadDoubleFromStream(AStream); // Longitude
  llat := ReadDoubleFromStream(AStream); // Latitude
  s := ReadStringFromStream(AStream); // Caption
  // Create the objects, depending of the token read above
  try
    case c of
      TestPOIToken : poi := TTestPOI.Create(llat,llon,s); // The simple POI
      TestShapeToken :
        begin
          // The more complicated Shape
          poi := TTestShape.Create(s);
          // Read the additional shape coords
          len := AStream.ReadDWord; // read the count of coord pairs(!)
          len := len * 2; // develop pairs to numbers of double
          SetLength(ds,len); //set the length of the local array
          for i := 0 to len-1 do
            ds[i] := ReadDoubleFromStream(AStream); // read the doubles from the stream
          TTestShape(poi).SetShapesFromArray(ds); // drop the values into the object
        end;
    end;
    Result := poi; // return the created class
  except
    // In the case of exception (for example data corruption, read beyond the file end
    // prevent meomory leaks and free the created object.
    // We clear the Result, but this is not really necesarry, because the raise will
    // not allow the returning of any value
    Result := Nil;
    if Assigned(poi) then
      poi.Free;
    raise;
  end;
end;

{ OnTestQuadTreeWriteStreamItem implements the writing of the user items to the Stream.
   The Event/Callback takes care about the two different classes we are using
   here: POI and Shapes
}
class procedure TTestItemStreamHelper.OnTestQuadTreeWriteStreamItem(AOwner: TLazQuadTree;
  ANode: TLQTNode; AItem: TObject; AItemIndex: Integer; AStream: TStream
  );
var
  poi : TTestPOI absolute AItem; // A kind of type cast
  shp : TTestShape absolute AItem; // A kind of type cast
  cnt : Integer;
  i : Integer;
begin
  // decide what we have, a POI or a Shape
  // and write the Token into the Stream
  // The Token is an arbitrary value, that can be choosen free, but must
  // be different from 0.
  if poi is TTestShape then
    AStream.WriteDWord(TestShapeToken) // = $ED534850; // ED+'SHP'
  else
    AStream.WriteDWord(TestPOIToken); // = $ED504F49; // ED+'POI'
  // Write Length DWord for the data, the actual size will be patched by the Caller, so 0 here
  AStream.WriteDWord(0);
  // Common data for both classes
  WriteDoubleToStream(AStream,poi.Lon);
  WriteDoubleToStream(AStream,poi.Lat);
  WriteStringToStream(AStream,poi.Caption);
  // Extra Data for the country shapes
  if poi is TTestShape then
  begin
    // The shapes have the additionally coords
    cnt := shp.CoordsCount;
    AStream.WriteDWord(cnt); // we save the number of coords. 1 means two doubles are stored!
    for i := 0 to cnt-1 do
    begin // run through the coords and write them to the stream
      WriteDoubleToStream(AStream,shp.CoordsLons[i]);
      WriteDoubleToStream(AStream,shp.CoordsLats[i]);
    end;
  end;
  // done
end;

end.

