{ A simple Earth-Map Frame used in the QuadTree demo
  Copyright (C) 2023,2024 Ekkehard Domning (www.domis.de)

  License: modified LGPL with linking exception (like RTL, FCL and LCL)

  See the file COPYING.modifiedLGPL.txt, included in the Lazarus distribution,
  for details about the license.

  Version Date        Change
  0.0.1   2024-08-23  First release
}

unit uFrameLqtEarthmap;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ComCtrls,
  ExtCtrls, LCLType,
  lazFileUtils,
  ulazQuadTreeGeometry, // The underlying rectangkle geometry
  ulazQuadTree, // The QuadTree
  uMouseRectShifter, // A helper for the shifting rectangle
  uBasicTestCommons; // The Eart and POI definition

type

  { TfrmLQTEarthMap }

  TfrmLQTEarthMap = class(TFrame)
    btnRectToText: TButton;
    cbShowNodeRect: TCheckBox;
    cbShowMouseNodes: TCheckBox;
    cbShowItems: TCheckBox;
    edtABottom: TEdit;
    edtALeft: TEdit;
    edtARight: TEdit;
    edtATop: TEdit;
    gbOptions: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    lblEstimatedItemsThreshold: TLabel;
    memoTreeText: TMemo;
    Panel2: TPanel;
    pbEarthMap: TPaintBox;
    pnlEarthMap: TPanel;
    pnlRectSettings: TPanel;
    pnlSettings: TPanel;
    pnlTreeText: TPanel;
    sbMapStatus: TStatusBar;
    Splitter1: TSplitter;
    tbEstimatedItemsThreshold: TTrackBar;
    procedure btnRectToTextClick(Sender: TObject);
    procedure cbShowItemsClick(Sender: TObject);
    procedure cbShowMouseNodesChange(Sender: TObject);
    procedure cbShowNodeRectClick(Sender: TObject);
    procedure edtALeftChange(Sender: TObject);
    procedure pbEarthMapMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbEarthMapMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure pbEarthMapMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbEarthMapPaint(Sender: TObject);
    procedure pnlEarthMapChangeBounds(Sender: TObject);
    procedure tbEstimatedItemsThresholdChange(Sender: TObject);

  private
    FMapTilePath : String;
    FRectangleMouseMoveActive : Boolean;
    FQuadTree : TLazQuadTree;
    FWorldmap : TBitmap; // The loaded World map from the file (720x360 px)
    FWorldmapLevel1 : TBitmap; // The loaded World map from the file (1440x720 px)
    FWorldmapChached : TBitmap; // The cached version in the size of the size of the pbEarthMap, including the marked node areas
    FTestRect : TLQTRect; // The user rectangle for examination
    FPaintBoxWorld : TLQTRect; // Definttion of the PaintBox World, size of the PaintBox, Y-Axins inc. top -> bottom
    FRectShifter : TMouseRectShifter; // The helper of the rectangle shifter
    FShifterYInverseMode : Boolean; // A needed variable to increase the user expierence :-)
    FGUIUpdateInProgress : Boolean; // Flag to prevent looping due to update changes
    FHighLightNodes : Boolean;
    FHighLightNodesX, FHighLightNodesY : Double;
    FEstimatedItemsThreshold : Integer;
    procedure SetupRectShifter; // Prepares the shifter for the mouse with the current rdctangle data
    procedure SetupCachedWorldMap; // Paints the EarthMap and the node rectangles to the stretched map
    procedure UpdateItemCountInRectangle; // Estimates or Counts the items in the rectangle
    procedure RectToGUI; // Fills the 4 Edits with the Rectangle Coords
    // A call back to fetch the date of the very first POI in the selected rectangle
    function LQTFirstItemEnumerationCallBack(Sender : TObject;
                                             ANode : TLQTNode;
                                             ANodeItem : TObject;
                                             ANodeItemIndex : Integer;
                                             AUserData : Pointer) : Boolean;
    procedure SetQuadTree(Value : TLazQuadTree);
    function WorldItemEnumerationCB(Sender : TObject; ANode : TLQTNode;
                                    ANodeItem : TObject; ANodeItemIndex : Integer;
                                    AUserData : Pointer) : Boolean;
    function WorldAgglomerationEnumerationCB(ANode : TLQTNode;
                                            const ATreeAgglomerationRec : TLQTAgglomerationRec;
                                            AUserData : Pointer) : Boolean;
  public
    property QuadTree : TLazQuadTree read FQuadTree write SetQuadTree;
    procedure TreeContentChanged;
    procedure CallMeAfterFormActivated;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;override;
  end;

implementation

{$R *.lfm}
const
  EarthMapTileFolder = 'EarthMapTiles\';

  ColorSet : array[0..15] of TColor = (
{
*** Primary color:

   shade 0 = #62CCF5 = rgb( 98,204,245) = rgba( 98,204,245,1) = rgb0(0.384,0.8,0.961)
   shade 1 = #E0F5FC = rgb(224,245,252) = rgba(224,245,252,1) = rgb0(0.878,0.961,0.988)
   shade 2 = #96DCF6 = rgb(150,220,246) = rgba(150,220,246,1) = rgb0(0.588,0.863,0.965)
   shade 3 = #32C0F7 = rgb( 50,192,247) = rgba( 50,192,247,1) = rgb0(0.196,0.753,0.969)
   shade 4 = #07B7FA = rgb(  7,183,250) = rgba(  7,183,250,1) = rgb0(0.027,0.718,0.98)
}
$F5CC62,$FCF5E0,$F6DC96,$F7C032,

{
  *** Secondary color (1):

     shade 0 = #FD5E86 = rgb(253, 94,134) = rgba(253, 94,134,1) = rgb0(0.992,0.369,0.525)
     shade 1 = #FEE1E8 = rgb(254,225,232) = rgba(254,225,232,1) = rgb0(0.996,0.882,0.91)
     shade 2 = #FD97B0 = rgb(253,151,176) = rgba(253,151,176,1) = rgb0(0.992,0.592,0.69)
     shade 3 = #FD2B5F = rgb(253, 43, 95) = rgba(253, 43, 95,1) = rgb0(0.992,0.169,0.373)
     shade 4 = #FE0040 = rgb(254,  0, 64) = rgba(254,  0, 64,1) = rgb0(0.996,0,0.251)
}
$865EFD,$E8E1FE,$B097FD,$5F2BFD,

{
  *** Secondary color (2):

     shade 0 = #9CFC5E = rgb(156,252, 94) = rgba(156,252, 94,1) = rgb0(0.612,0.988,0.369)
     shade 1 = #ECFEE1 = rgb(236,254,225) = rgba(236,254,225,1) = rgb0(0.925,0.996,0.882)
     shade 2 = #BEFC96 = rgb(190,252,150) = rgba(190,252,150,1) = rgb0(0.745,0.988,0.588)
     shade 3 = #7DFC2A = rgb(125,252, 42) = rgba(125,252, 42,1) = rgb0(0.49,0.988,0.165)
     shade 4 = #64FD00 = rgb(100,253,  0) = rgba(100,253,  0,1) = rgb0(0.392,0.992,0)
}

  $5EFC9C,$E1FEEC,$96FCBE,$2AFC7D,

{
  *** Complement color:

     shade 0 = #FFB25F = rgb(255,178, 95) = rgba(255,178, 95,1) = rgb0(1,0.698,0.373)
     shade 1 = #FFF1E2 = rgb(255,241,226) = rgba(255,241,226,1) = rgb0(1,0.945,0.886)
     shade 2 = #FFCE98 = rgb(255,206,152) = rgba(255,206,152,1) = rgb0(1,0.808,0.596)
     shade 3 = #FF992B = rgb(255,153, 43) = rgba(255,153, 43,1) = rgb0(1,0.6,0.169)
     shade 4 = #FF8500 = rgb(255,133,  0) = rgba(255,133,  0,1) = rgb0(1,0.522,0)
}
  $5FB2FF,$E2F1FF,$98CEFF,$2B99FF);

{ TfrmLQTEarthMap }

{ This btnRectToTextClick event is called, when the user click to the
  button "Get Items in Rect.".
  The method will then collect all items in the rectangle and drop a textual
  expression into the memoTreeText-field.
  This is an example of the powerfull tree method ListItemsInRect
}
procedure TfrmLQTEarthMap.btnRectToTextClick(Sender: TObject);
var
  lList : TList;
  i : Integer;
  s : String;
  poi : TTestPOI;
  shp : TTestShape;
  cnt : Integer;
  rc : Integer;
  lOldCursor : TCursor;
begin
  memoTreeText.Clear; // clear the memo field
  lList := TList.Create; // Create a TList to store the collected items
  try
    // Ask the tree to list all items within the rectangle.
    // The FTestRect is the rectangle that the user can modify with the mouse within the
    // World map, even acoss the date-border
    FQuadTree.ListItemsInRect(FTestRect,lList);
    // Get the amount of found items
    cnt := lList.Count;
    // If there are more than 1000 items found, it's a good idea to ask the
    // user how to proceed.
    if cnt > 1000 then
    begin
      // We simply use the thee button standard MessageBox with yes, no and cancel button on it
      // Cancel = abort (= print nothing)
      // Yes = print all (may take a longer time)
      // No = print the first 1000 items
      rc := Application.MessageBox(
              PChar(Format('YES = print all %d items, NO = print first 1000, CANCEL = abort',[cnt])),
              'Many items found!',
              MB_YESNOCANCEL);
      if rc = IDCANCEL then Exit;
      if rc = IDNO then
        cnt := 1000;
    end;
    lOldCursor := Screen.Cursor;  // Save the current mouse pointer
    Screen.Cursor := crHourGlass; // set it to the wainting one
    try
      memoTreeText.Lines.BeginUpdate; //Block the memo field
      try
        for i := 0 to cnt-1 do // run through all / first 1000 items in the list
        begin
          poi := TTestPOI(lList.Items[i]);
          // There are possibly two different items in the list
          // POI or Country shapes
          if poi is TTestShape then
          begin // If we have a country shape, print the bounding box and the country name
            shp := TTestShape(poi);
            s := Format('[%d] L: %1.6f, T: %1.6f, R: %1.6f, B: %1.6f %s',
                        [i,
                         shp.ShapeBounds.Left, shp.ShapeBounds.Top, shp.ShapeBounds.Right, shp.ShapeBounds.Bottom,
                         shp.Caption])
          end
          else // else the coordinate and the Caption
            s := Format('[%d] Lon %1.1f, Lat %1.1f %s',[i, poi.Lon,poi.Lat,poi.Caption]);
          memoTreeText.Lines.Add(s); // Add the text
        end;
      finally
        memoTreeText.SelStart := 0; // Position the cursor at the beginning of the text
        memoTreeText.SelLength := 0;
        memoTreeText.Lines.EndUpdate; // Update the memo field
      end;
    finally
      Screen.Cursor := lOldCursor; // reinstall the previous mouse pointer
    end;
  finally
    lList.Free; // Free the list (not the items, they belong to the Tree!)
  end;
end;

procedure TfrmLQTEarthMap.cbShowItemsClick(Sender: TObject);
begin
  SetupCachedWorldMap; // Paint the rectangles into the cached world map
  pbEarthMap.Invalidate; // redraw the map
  pbEarthMap.Repaint;
end;

procedure TfrmLQTEarthMap.cbShowMouseNodesChange(Sender: TObject);
begin
  SetupCachedWorldMap; // Paint the rectangles into the cached world map
  pbEarthMap.Invalidate; // redraw the map
  pbEarthMap.Repaint;
end;

procedure TfrmLQTEarthMap.cbShowNodeRectClick(Sender: TObject);
begin
  SetupCachedWorldMap; // Paint the rectangles into the cached world map
  pbEarthMap.Invalidate; // redraw the map
  pbEarthMap.Repaint;
end;

procedure TfrmLQTEarthMap.edtALeftChange(Sender: TObject);
  function ProcessConversion(const AEdit : TEdit; out AValue : Double) : Boolean;
  var
    s : String;
    fs : TFormatSettings;
    d : Double;
  begin
    Result := False;
    s := AEdit.Text;
    while Pos(',',s) > 0 do
      s[Pos(',',s)] := '.';
    fs.ThousandSeparator:= #0;
    fs.DecimalSeparator:='.';
    d := StrToFloatDef(s,-1e20,fs);
    Result := d > -1e20;
    if not Result then
      AEdit.Color := clRed
    else
    begin
      AEdit.Color := clDefault;
      AValue := d;
    end;
  end;

var
  edtName : String;
  d : Double;
begin
  if not (Sender is TEdit) then Exit;
  edtName := TEdit(Sender).Name;
  if edtName = 'edtALeft' then
  begin
    if not ProcessConversion(TEdit(Sender),d) then Exit;
    FTestRect.Left := d;
  end
  else if edtName = 'edtATop' then
  begin
    if not ProcessConversion(TEdit(Sender),d) then Exit;
    FTestRect.Top := d;
  end
  else if edtName = 'edtARight' then
  begin
    if not ProcessConversion(TEdit(Sender),d) then Exit;
    FTestRect.Right := d;
  end
  else if edtName = 'edtABottom' then
  begin
    if not ProcessConversion(TEdit(Sender),d) then Exit;
    FTestRect.Bottom := d;
  end
  else
    Exit;
  UpdateItemCountInRectangle; // do the calculation of the number of items
  pbEarthMap.Invalidate; // redraw the map
  pbEarthMap.Repaint;
end;

// The MouseDown, MouseMove and MouseUp events control mainly the rectangle shifting
// The MouseMove also does some addition display things on hovering around
procedure TfrmLQTEarthMap.pbEarthMapMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  // On MouseDown the helper stores the last position
  FRectShifter.OnMouseDown(Sender,Button,Shift,X,Y);
  // Reset the inverter
  FShifterYInverseMode := False;
  FRectangleMouseMoveActive := True;
end;

procedure TfrmLQTEarthMap.pbEarthMapMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
  // The following "PerformXXX" procedures, do the actual adjusting of the
  // side of the active rectangle.
  procedure PerformRightSide(const ANewX : Double; var ARect : TLQTRect);
  begin
    ARect.Right := ANewX;
  end;
  procedure PerformLeftSide(const ANewX : Double; var ARect : TLQTRect);
  begin
    ARect.Left := ANewX;
  end;
  procedure PerformTopSide(const ANewY : Double; var ARect : TLQTRect);
  begin
    if FShifterYInverseMode then
    begin
      if (ANewY < ARect.Top) then
        FShifterYInverseMode := False;
    end
    else
    begin
      if (ANewY > ARect.Bottom) then
        FShifterYInverseMode := True;
    end;
    if FShifterYInverseMode then
      ARect.Bottom := ANewY
    else
      ARect.Top := ANewY;
  end;
  procedure PerformBottomSide(const ANewY : Double; var ARect : TLQTRect);
  begin
    if FShifterYInverseMode then
    begin
      if (ANewY > ARect.Bottom) then
        FShifterYInverseMode := False;
    end
    else
    begin
      if (ANewY < ARect.Top) then
        FShifterYInverseMode := True;
    end;
    if FShifterYInverseMode then
      ARect.Top := ANewY
    else
      ARect.Bottom := ANewY;
  end;

var
  lHitItem : TMouseHitItem;
  lx, ly, lx0, ly0 : Double;
  w, h : Double;
  lRect0 : TLQTRect;
  lTestPOI : TTestPOI = Nil; // A Variable to be filled by a callback function, see details below
  s : String;
begin
  // Pass the parameters to the HelperClass.
  FRectShifter.OnMouseMove(Sender,Shift,X,Y,lHitItem);

  // The normated mouse coords are translated into the
  // coords of the current world
  TransformPoint(X,Y,
                 FPaintBoxWorld, EarthWorld,
                 lx,ly);
  // Display the currend coords in the statusbar
  NormatePointToWorld(lx,ly,
                      EarthWorld,lqtwmCylindrical,
                      lx,ly);
  sbMapStatus.Panels[1].Text := Format('%1.1f',[ly]);
  sbMapStatus.Panels[3].Text := Format('%1.1f',[lx]);

  if (not Assigned(lHitItem)) or (not lHitItem.MouseDownFlag) then
  begin
    // No rectangle border/corner/area is selected
    // Do some other fancy stuff here

    // First, initiate a redraw of the map, to impress the "touched" Nodes
    FHighLightNodes := True;
    FHighLightNodesX := lx;
    FHighLightNodesY := ly;
    pbEarthMap.Invalidate; // redraw the map
    pbEarthMap.Repaint;

    // We will pick the very first POI Item in a small rectangle around the mouse (if one)
    // and print the content into the StatusBar
    // Get the height (and width) in degree for three pixels
    TransformPoint(3,3,
                   FPaintBoxWorld, EarthWorld,
                   lx0,ly0);
    ly0 := 90.0-Abs(ly0);
    lx0 := ly0; // Map has squared pixels and degrees, so ignore the returned Y-value
    // Setup the rectangle around the mouse (in total the degree covered by a 4x4px rectangle)
    lRect0.Left := lx-lx0;
    lRect0.Top := ly+ly0;
    lRect0.Right := lx+lx0;
    lRect0.Bottom := ly-ly0;
    // To avoid collecting a very long list of POIs and discarding all but the first item,
    // by using the Tree-Method ListItemsInRect, we using the CallBack functionallity here.
    // For each found item the callback becomes called once.
    // The repetition is terminated as soon as the callback function return false.
    //
    // The storage of the retreived value is very simple, by passing the address
    // of the local variable in the Pointer of the user data.
    FQuadTree.EnumerateItemsInRect(lRect0, // The rectangle to search in
                                   @LQTFirstItemEnumerationCallBack, // The address of the callback function
                                   @lTestPOI); // The address of the data where the cb-function should store the result
    if Assigned(lTestPOI) then // The assigned item could be a POI or a country Shape
      s := Format('Lon %1.1f, Lat %1.1f %s',[lTestPOI.Lon,lTestPOI.Lat,lTestPOI.Caption])
    else
      s := '';
    sbMapStatus.Panels[7].Text := s;
  end
  else
    FHighLightNodes := False;

  if Assigned(lHitItem) then
  begin
    // lHitItem is assigned if one item of the ShiftRectangle (Area, Border, Corner) is selected.
    pbEarthMap.Cursor := lHitItem.Cursor;  // Set the cursor to the hit items kind
    if lHitItem.MouseDownFlag then // if the mouse is down
    begin
      // Copy the user User-Rectangle (in world coords)
      lRect0.Assign(FTestRect);
      // and transform it to the Flat world (same as world cords but Y-Axis is inverse
      lRect0.Transform(EarthWorld, FlatWorld);
      // Transform the shifting point of the mouse also in the flat world
      TransformPoint(lHitItem.OrgX+lHitItem.MouseDeltaX,
                     lHitItem.OrgY+lHitItem.MouseDeltaY,
                     FPaintBoxWorld, FlatWorld,
                     lx,ly);

      // case of the several actions the "Tag" contains in the lowest four bits, what actually is hit
      case (lHitItem.Tag and $F) of
        tagRectArea :
          begin
            // Special operation for the movement of the full rectangle.
            // Just locate the rectangle at the new position ...
            w := lRect0.Width(FlatWorld,lqtwmCylindrical);
            h := lRect0.Height(FlatWorld,lqtwmCylindrical);
            lRect0.Left := lx;
            lRect0.Top := ly;
            lRect0.Right := lRect0.Left + w;
            lRect0.Bottom := lRect0.Top + h;
            // In the cylindrical world model top and bottom are fixed borders
            lRect0.Bottom := lRect0.Top + h;
            if lRect0.Bottom >= FlatWorld.Bottom then
              lRect0.Top := FlatWorld.Bottom - h;
            if lRect0.Top < FlatWorld.Top then
              lRect0.Bottom := FlatWorld.Top + h;
          end;
        tagRectLeft :
          begin
            PerformLeftSide(lx,lRect0);
          end;
        tagRectTop :
          begin
            PerformTopSide(ly,lRect0);
          end;
        tagRectRight :
          begin
            PerformRightSide(lx,lRect0);
          end;
        tagRectBottom :
          begin
            PerformBottomSide(ly,lRect0);
          end;
        tagRectTopLeft :
          begin
            PerformTopSide(ly,lRect0);
            PerformLeftSide(lx,lRect0);
          end;
        tagRectTopRight :
          begin
            PerformTopSide(ly,lRect0);
            PerformRightSide(lx,lRect0);
          end;
        tagRectBottomRight :
          begin
            PerformBottomSide(ly,lRect0);
            PerformRightSide(lx,lRect0);
          end;
        tagRectBottomLeft :
          begin
            PerformBottomSide(ly,lRect0);
            PerformLeftSide(lx,lRect0);
          end;
      end;
      // Now return and adjust the changed coords and transform to the world coordinates
      lRect0.NormateToWorld(FlatWorld, lqtwmCylindrical);
      lRect0.Transform(FlatWorld, EarthWorld);
      // And return into the active TestRect rectangle
      FTestRect.Assign(lRect0);
      RectToGUI; // Drop the values into the edit fields
      UpdateItemCountInRectangle; // do the calculation of the number of items
      pbEarthMap.Invalidate; // redraw the map
      pbEarthMap.Repaint;
    end;
  end
  else
    pbEarthMap.Cursor := crDefault;  // no hit, set the default cursor
end;

procedure TfrmLQTEarthMap.pbEarthMapMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FRectShifter.OnMouseUp(Sender,Button,Shift,X,Y); // The shifting ends
  FRectangleMouseMoveActive := False;
  SetupRectShifter;  // Setup the HelperClass for the current setting
  UpdateItemCountInRectangle;
end;

procedure TfrmLQTEarthMap.pbEarthMapPaint(Sender: TObject);
  // PaintRectangle will draw the selection rectangle
  // It takes care if the rectangle is wrapping around.
  procedure PaintRectangle(const ALQTRect : TLQTRect;
                           const AColor : TColor;
                           const ALineWidth : Integer
                           );
  var
    lLQTRect,lLQTRect0 : TLQTRect;
    lNormalizedRectangles : TLQTRectArray; // The normalized Rectangles
    lNormalizedWorldRect : TLQTRect;
    i : Integer;
    r : TRect;
  begin
    pbEarthMap.Canvas.Pen.Width := ALineWidth;

    // Prepare the canvas
    pbEarthMap.Canvas.Pen.Style := psSolid;
    pbEarthMap.Canvas.Pen.Color := AColor;
    pbEarthMap.Canvas.Brush.Style := bsClear;
    // The flat world is our canvas world
    // Calculate the actual width and height in pixel
    lLQTRect.Assign(ALQTRect);
    lLQTRect.NormateToWorld(EarthWorld,lqtwmCylindrical);
    // NormalizeRectangle will split wrapping around rectangles
    // into several ones, laying within the map.
    // The lNormalizedWorldRect contains an upright version
    // of the used world, so that X increases to the right and Y downwards.
    // lNormalizedRectangles contains the rectangles (1, 2 or 4)
    NormalizeRectangle(lLQTRect,
                       EarthWorld,lqtwmCylindrical,
                       lNormalizedRectangles,
                       lNormalizedWorldRect);
    for i := 0 to High(lNormalizedRectangles) do
    begin
      // Now we transform the rectangles to our drawing surface
      lLQTRect0.Assign(lNormalizedRectangles[i]);
      lLQTRect0.Transform(lNormalizedWorldRect,FPaintBoxWorld);
      // do the stretch and ofset
      r.Left := Round(lLQTRect0.Left);
      r.Top := Round(lLQTRect0.Top);
      r.Right := Round(lLQTRect0.Right);
      r.Bottom := Round(lLQTRect0.Bottom);
      // Make sure the rectangle is displayed
      if r.Left = r.Right then
      begin
        if r.Left > 0 then
          Dec(r.Left)
        else
          Inc(r.Right);
      end;
      if r.Top = r.Bottom then
      begin
        if r.Top > 0 then
          Dec(r.Top)
        else
          Inc(r.Bottom);
      end;

      // end finally paint to the canvas
      pbEarthMap.Canvas.Rectangle(r);
    end;
  end;

  procedure DrawNodeRectangles(const AX, AY : Double);
  // This sub procedure highlights the "path" of node rectangle to
  // the mouse position
  var
    n : TLQTNode;
    r : TRect;
    lRect0 : TLQTRect;
    w : Integer;
    clrStep : Double;
    vi : Integer;
    clr : TColor;
  begin
    pbEarthMap.Canvas.Pen.Style := psSolid;
    if cbShowNodeRect.Checked then
      pbEarthMap.Canvas.Pen.Width := 2
    else
      pbEarthMap.Canvas.Pen.Width := 1;
    pbEarthMap.Canvas.Brush.Style := bsClear;

    clrStep := 256.0;

    // Find the node on the Mouse location (AX,AY = Longitude/Latitude)
    n := FQuadTree.FindCoordNode(AX,AY);
    // Step the colors
    if Assigned(n) and Assigned(n.ParentNode)  then
    begin
      if n.NodeLevel > 0 then
        clrStep := 256.0 / n.NodeLevel;
    end;
    // Now loop upwards from the node to the root node
    while Assigned(n) and Assigned(n.ParentNode) do
    begin
      // Get the rectangle that covers the found node
      lRect0 := n.NodeRect;
      // Transform the rectangle to the PaintBox coordinates
      lRect0.Transform(EarthWorld, FPaintBoxWorld);
      r.Left := Round(lRect0.Left);
      r.Right := Round(lRect0.Right);
      // Calculate the width
      w := Round(r.Right-r.Left);
      // Draw only visible rectangle (ommit if equal or smaller than 2 pixels)
      if w > 2 then
      begin
        r.Top := Round(lRect0.Top);
        r.Bottom := Round(lRect0.Bottom);
        // Calculate the color
        vi := Round(255.0 - (n.NodeLevel * clrStep));
        if vi < 0 then
          vi := 0
        else if vi > 255 then
          vi := 255;
        clr := TColor(vi);
        // Draw the Nodes rectangle
        pbEarthMap.Canvas.Pen.Color := clr;
        pbEarthMap.Canvas.Rectangle(r);
      end;
      // Use the parent of this node as the next node
      n := n.ParentNode;
    end;
  end;
  procedure DrawNodeAncestorLines(const AX, AY : Double);
  // This sub procedure highlights the "path" of node rectangle to
  // the mouse position
  var
    n : TLQTNode;
    lRect0 : TLQTRect;
    w : Integer;
    clrStep : Double;
    vi : Integer;
    clr : TColor;
    cx,cy : Integer;
    firststroke : Boolean = True;
  begin
    pbEarthMap.Canvas.Pen.Style := psSolid;
    if cbShowNodeRect.Checked then
      pbEarthMap.Canvas.Pen.Width := 2
    else
      pbEarthMap.Canvas.Pen.Width := 1;
    pbEarthMap.Canvas.Brush.Style := bsClear;

    clrStep := 256.0;

    // Find the node on the Mouse location (AX,AY = Longitude/Latitude)
    n := FQuadTree.FindCoordNode(AX,AY);
    // Step the colors
    if Assigned(n) and Assigned(n.ParentNode)  then
    begin
      if n.NodeLevel > 0 then
        clrStep := 256.0 / n.NodeLevel;
    end;
    // Now loop upwards from the node to the root node
    while Assigned(n) do
    begin
      // Get the rectangle that covers the found node
      lRect0 := n.NodeRect;
      // Transform the rectangle to the PaintBox coordinates
      lRect0.Transform(EarthWorld, FPaintBoxWorld);
      cx := Round((lRect0.Left+lRect0.Right)/2);
      cy := Round((lRect0.Top+lRect0.Bottom)/2);

      // Calculate the width
      w := Round(lRect0.Right-lRect0.Left);

      if firststroke then
      begin
        firststroke := False;
        pbEarthMap.Canvas.MoveTo(cx,cy);
      end;

      // Draw only visible rectangle (ommit if equal or smaller than 2 pixels)
      if w > 2 then
      begin
        // Calculate the color
        vi := Round(255.0 - (n.NodeLevel * clrStep));
        if vi < 0 then
          vi := 0
        else if vi > 255 then
          vi := 255;
        clr := TColor(vi);
        // Draw the Nodes Line
        pbEarthMap.Canvas.Pen.Color := clr;
        pbEarthMap.Canvas.LineTo(cx,cy);
      end;
      // Use the parent of this node as the next node
      n := n.ParentNode;
    end;
  end;


var
  r : TRect;
begin
  // On startup of the program some parts are not loaded, while paint is called
  // so try to update the rectangles
  if FPaintBoxWorld.IsNotArea then
    pnlEarthMapChangeBounds(Self);
  r.Left := 0;
  r.Top := 0;
  r.Right := pbEarthMap.Width;;
  r.Bottom := pbEarthMap.Height;
  // Draw the cached world map as the "background".
  // This map may contain the NodeRectangles of the tree.
  // And some accumulated items and some independend items
  pbEarthMap.Canvas.StretchDraw(r,FWorldmapChached);
  // If the PaintBoxWorld is still not initialized, exit here
  if FPaintBoxWorld.IsNotArea then Exit;
  // Depending of the users wish, we draw some fancy stuff
  if FHighLightNodes and cbShowMouseNodes.Checked then
  begin
    DrawNodeAncestorLines(FHighLightNodesX, FHighLightNodesY);
    // DrawNodeRectangles(FHighLightNodesX, FHighLightNodesY);
  end;
  // And on top, the pink user rectangle
  PaintRectangle(FTestRect,clFuchsia,3);
end;

procedure TfrmLQTEarthMap.SetupRectShifter;
  // AddRectangle does exactly this, placing all elements for the
  //   userrectangle into the helper class
  //   The method is fairly the same as the painting does.
  //   But instead of "normalization" the rectangles are duplicated around
  //   giving most of the work to windows instead of doing the cropping by our self
  procedure AddRectangle(const ALQTRect : TLQTRect);
  var
    lLQTRect : TLQTRect;

    k : Integer;
    r,r0 : TRect;
    ww,wh : Integer;
  begin
    // Calculate the actual width and height in pixel
    ww := Round(FPaintBoxWorld.Width);
    wh := Round(FPaintBoxWorld.Height);
    lLQTRect.Assign(ALQTRect);
    lLQTRect.NormateToWorld(EarthWorld,lqtwmCylindrical);
    lLQTRect.Transform(EarthWorld,FPaintBoxWorld);
    r.Left := Round(lLQTRect.Left);
    r.Top := Round(lLQTRect.Top);
    r.Right := Round(lLQTRect.Right);
    r.Bottom := Round(lLQTRect.Bottom);
    if r.Left > r.Right then
      r.Left := r.Left - ww;
    if r.Bottom < r.Top then
      r.Bottom := r.Bottom + wh;
    FRectShifter.AddRectangle(r,3,10);
    // Duplicate the rectangles to the dimmed doubled picture areas on cylindrical world
    for k := -1 to 1 do // X = (left, center, right)
    begin
      if (k = 0) then Continue; // The center is allready created
      r0 := r; // Take the last painted rectangle
      // Simply subtract or add the width and height of the world to the coords of the rectangle.
      r0.Right := r0.Right + (ww*k);
      r0.Left := r0.Left + (ww*k);
      FRectShifter.AddRectangle(r0,3,10);
    end;
  end;

begin
  FRectShifter.Clear;
  AddRectangle(FTestRect);
end;
{ WorldItemEnumerationCB will be called for every item that needs to be displayed on the map.
  It will be called by the Tree's method AgglomerateNodeAsEnumeration, which is callen within the
  SetupWorldMap methond downwards in this Frame}
function TfrmLQTEarthMap.WorldItemEnumerationCB(Sender : TObject; ANode : TLQTNode;
                                ANodeItem : TObject; ANodeItemIndex : Integer;
                                AUserData : Pointer) : Boolean;
var
  ptx, pty : Double;
  x,y : Integer;
  w : Integer;
  poi : TTestPOI absolute ANodeItem; // This is a kind of type cast
  shp : TTestShape absolute ANodeItem; // This is a kind of type cast
  i : Integer;
begin
  Result := True; // We return true, since we want to be called again, with the next item
  if poi is TTestShape then
  begin // In the case that the Item is an Country Shape
    FWorldmapChached.Canvas.Brush.Style := bsClear;
    FWorldmapChached.Canvas.Pen.Style := psSolid;
    FWorldmapChached.Canvas.Pen.Width := 1;
    FWorldmapChached.Canvas.Pen.Color := clRed;
    // we draw all the segments on the map.
    // We could assume here, that no segment is passing the borders of the world!
    for i := 0 to shp.CoordsCount-1 do
    begin
      TransformPoint(shp.CoordsLons[i],
                     shp.CoordsLats[i],
                     EarthWorld, FPaintBoxWorld,
                     ptx,pty);
      x := Round(ptx);
      y := Round(pty);
      if i = 0 then
        FWorldmapChached.Canvas.MoveTo(x,y)
      else
        FWorldmapChached.Canvas.LineTo(x,y);
    end;
  end
  else // in all other cases we assume that the item is an POI
  begin
    TransformPoint(poi.Lon, poi.Lat,
                   EarthWorld, FPaintBoxWorld,
                   ptx, pty);
    x := Round(ptx);
    y := Round(pty);
    FWorldmapChached.Canvas.Pen.Style := psClear;
    FWorldmapChached.Canvas.Brush.Style := bsSolid;
    FWorldmapChached.Canvas.Brush.Color := clBlue;
    // and simple draw a 4*4 circle at the place
    w := 2;
    FWorldmapChached.Canvas.Ellipse(x-w,y-w,x+w,y+w);
  end;
end;

{ WorldAgglomerationEnumerationCB will be called for every Agglomeration that needs to be displayed on the map.
  It will be called by the Tree's method AgglomerateNodeAsEnumeration, which is callen within the
  SetupWorldMap methond downwards in this Frame.
  An "Agglomeration" is a part of the map, where so many Items are so tight together that they
  could not be separated by the user (e.g. sharing same pixel).}
function TfrmLQTEarthMap.WorldAgglomerationEnumerationCB(ANode : TLQTNode;
                                        const ATreeAgglomerationRec : TLQTAgglomerationRec;
                                        AUserData : Pointer) : Boolean;
var
  ptx, pty : Double;
  x,y : Integer;
  w,h : Integer;
begin
  Result := True; // Return true, because we want to be called back
  TransformPoint(ATreeAgglomerationRec.Point.X, ATreeAgglomerationRec.Point.Y,
                 EarthWorld, FPaintBoxWorld,
                 ptx, pty);
  x := Round(ptx);
  y := Round(pty);

  // We colorize the agglomerations according to there population
  if ATreeAgglomerationRec.Count > 1000 then
    FWorldmapChached.Canvas.Brush.Color := clRed
  else if ATreeAgglomerationRec.Count > 100 then
    FWorldmapChached.Canvas.Brush.Color := $00A5FF // Orange
  else if ATreeAgglomerationRec.Count > 50 then
    FWorldmapChached.Canvas.Brush.Color := clRed
  else
    FWorldmapChached.Canvas.Brush.Color := clLime;
  // The UserData Pointer is used to tell us the size/diameter of the Agglomeration
  if Assigned(AUserData) then
    w := Integer(AUserData^)
  else
    w := 4;
  w := w div 2;
  if w < 2 then
    w := 2;
  h := w; // div 2;
  // Draw the Ellipse
  FWorldmapChached.Canvas.Ellipse(x-w,y-h,x+w,y+h);
end;
{ SetupCachedWorldMap is used to paint the longer and complicated things once and use the
  Cached image to speed up the repaint later.
  This method is called manually when the size of the display or the tree contant changed}
procedure TfrmLQTEarthMap.SetupCachedWorldMap;
var
  n : TLQTNode;
  r : TRect;
  ndx : Integer;
  lRect0 : TLQTRect;
  w,h : Integer;
  lvl : Integer;
begin
  w := Round(FPaintBoxWorld.Width);
  h := Round(FPaintBoxWorld.Height);
  FWorldmapChached.SetSize(w,h); // The bitmap has the same size as the paintbox
  r.Left := 0;
  r.Top := 0;
  r.Right := w;
  r.Bottom := h;
  // At first (for the background) we paint the world map.
  // If the display is larger than 720 pixel, we draw the one with the higher resolution
  if w > 720 then
    FWorldmapChached.Canvas.StretchDraw(r,FWorldmapLevel1)
  else
    FWorldmapChached.Canvas.StretchDraw(r,FWorldmap);
  // User checkboxes
  // If the NodeRectangle should be display, do so
  if cbShowNodeRect.Checked then
  begin
    FWorldmapChached.Canvas.Pen.Style := psSolid;
    FWorldmapChached.Canvas.Pen.Width := 1;
    FWorldmapChached.Canvas.Brush.Style := bsClear;
    // Starting with the Root Node
    n := FQuadTree.NextNode(FQuadTree.RootNode);
    while Assigned(n) do
    begin
      lRect0 := n.NodeRect;
      lRect0.Transform(EarthWorld, FPaintBoxWorld);
      r.Left := Round(lRect0.Left);
      r.Right := Round(lRect0.Right);
      w := Round(r.Right-r.Left);
      // Calculate how large the rectangle of the current node is
      // if its equal or smaller than two pixel, skip the painting
      if w > 2 then
      begin
        r.Top := Round(lRect0.Top);
        r.Bottom := Round(lRect0.Bottom);
        // Calculate the colors regarding the level and the node orientationn to
        // avoid same coloring
        ndx := (n.NodeLevel mod 4);
        if n = n.ParentNode.ChildNodeNE then
          ndx := ndx + 0
        else if n = n.ParentNode.ChildNodeSE then
          ndx := ndx + 4
        else if n = n.ParentNode.ChildNodeSW then
          ndx := ndx + 8
        else if n = n.ParentNode.ChildNodeNW then
          ndx := ndx + 12;
        FWorldmapChached.Canvas.Pen.Color := ColorSet[ndx];
        FWorldmapChached.Canvas.Rectangle(r);
      end;
      // Select the next node (Result is Nil, when we done)
      n := FQuadTree.NextNode(n);
    end;
  end;
  // Show Items (and Agglomerations)
  if cbShowItems.Checked then
  begin
    FWorldmapChached.Canvas.Pen.Style := psClear;
    FWorldmapChached.Canvas.Pen.Width := 1;
    FWorldmapChached.Canvas.Brush.Style := bsSolid;

    // Calculate the needed depth for the agglomeration
    // we go down to > 10 pixels
    // Since the width of the paintbox is the width of to RootNode
    // we can simply divide by two and increase the level
    lvl := 0;
    w := Round(FPaintBoxWorld.Width);
    while w > 10 do
    begin
      w := w div 2;
      Inc(lvl);
    end;
    // Let's do the Tree the Agglomeration. The two callbacks (described above)
    // will paint the Items and the Agglomerations
    FQuadTree.AgglomerateNodeAsEnumeration(FQuadTree.RootNode,lvl,8,
                @WorldItemEnumerationCB, Nil,
                @WorldAgglomerationEnumerationCB, @w);
  end;

end;

procedure TfrmLQTEarthMap.UpdateItemCountInRectangle;
var
  d,dx : Double;
begin
  // Here we "count" the items located in the rectangle
  // The counting has been done within the tree filling
  // so every node "knows" how many items are stored within his area.
  // The problem is, that the user rectangle is random and does not
  // fit (exact) to the splitted rectangles in the tree.
  // To overcome this limitation, the tree offers an "Estimated" count of items
  // doing some counting an extrapolation
  // This is faster (in contrast to counting), but less precise.
  // Since in 99% of all cases on large scales an approximation is sufficent
  // this is a good method to indicate the need for a bigger magnification in a
  // specific area.
  d := FQuadTree.EstimatedCountOfAssignedItemsInRect(FTestRect,FEstimatedItemsThreshold);
  // If the estimation indicates a small amount of items
  // we do a pure count. This is also fast, because
  // for all fully covered node-areas the stored sum is taken
  // and only from the remaining items the query "Is Item in FTestRect?"
  // is performed.
  dx := FQuadTree.CountAssignedItemsInRect(FTestRect);
  sbMapStatus.Panels[5].Text := Format('%1.0f (≈ %1.1f)',[dx, d]);
end;

{ RectToGUI fills the Edit fields with the values from the data fields}
procedure TfrmLQTEarthMap.RectToGUI;
  procedure DoubleToEdit(const AValue : Double; const AEdit : TEdit);
  var
    s : String;
  begin
    s := Format('%1.2f',[AValue]);
    if AEdit.Text <> s then
      AEdit.Text := s;
  end;

begin
  FGUIUpdateInProgress := True;
  try
    DoubleToEdit(FTestRect.Left, edtALeft);
    DoubleToEdit(FTestRect.Top, edtATop);
    DoubleToEdit(FTestRect.Right, edtARight);
    DoubleToEdit(FTestRect.Bottom, edtABottom);
  finally
    FGUIUpdateInProgress := False;
  end;
end;
{ LQTFirstItemEnumerationCallBack is called when the tree enumerates the items.
    The original call is initiated in pbEarthMapMouseMove
    The goal here is just to pick one item and not collection all items into a list
    and than dicard 99.9% of the work.
}
function TfrmLQTEarthMap.LQTFirstItemEnumerationCallBack(Sender: TObject;
  ANode: TLQTNode; ANodeItem: TObject; ANodeItemIndex: Integer;
  AUserData: Pointer): Boolean;
begin
  // Simply copy the Pointers from Item
  TTestPOI(AUserData^) := TTestPOI(ANodeItem);
  // If the Item is a country shape then return true, to force the
  // Tree to look for more items. We do so, because otherwise in the
  // mixed POI / Country Shape examples only the countries are shown
  // and not the underlying POIs
  Result := (TTestPOI(ANodeItem) is TTestShape);
end;

procedure TfrmLQTEarthMap.SetQuadTree(Value: TLazQuadTree);
begin
  FQuadTree := Value;
end;

procedure TfrmLQTEarthMap.TreeContentChanged;
begin
  UpdateItemCountInRectangle;
  SetupCachedWorldMap; // Paint the rectangles into the cached world map
  pbEarthMap.Invalidate; // redraw the map
  pbEarthMap.Repaint;
end;

procedure TfrmLQTEarthMap.CallMeAfterFormActivated;
begin
  TreeContentChanged;
  pnlEarthMapChangeBounds(Self);
  tbEstimatedItemsThresholdChange(Self);
end;

constructor TfrmLQTEarthMap.Create(AOwner: TComponent);
var
  fp : String;
  jpg : TJpegImage;
begin
  inherited Create(AOwner);
  // Setup the user adjustable Test-Rectangle.
  // It will wrap around the date line and cover mainly the Pacific Ocean
  // (BTW as an not native english speaking I like the two words
  // "Pacific Ocean" because all three used "c" are pronounced different :-))
  // between Hawaii and Japan/North Australia.
  FTestRect.SetFrom(135.0, 35.0, -145.0, -15.0);
  // Create the helper class for the User-Interaction with the rectangle
  FRectShifter := TMouseRectShifter.Create;
  // Find the map folder
  fp := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName));
  FMapTilePath := fp+EarthMapTileFolder;
  if not DirectoryExists(FMapTilePath) then
  begin
    fp := IncludeTrailingPathDelimiter(ResolveDots(ExtractFilePath(Application.ExeName)+ExampleDataPath));
    FMapTilePath := fp+EarthMapTileFolder;
  end;
  // Create the two background maps (one for lesser, one for higher resolution of
  // the users Screen and Program Window size
  FWorldmap := TBitmap.Create;
  FWorldmapLevel1 := TBitmap.Create;
  try
    jpg := TJpegImage.Create;
    try
      try
        jpg.LoadFromFile(FMapTilePath+'NE_0_0_0.jpg');
        FWorldmap.Assign(jpg);
        FWorldmapLevel1.SetSize(1440,720);
        jpg.LoadFromFile(FMapTilePath+'NE_1_0_0.jpg');
        FWorldmapLevel1.Canvas.Draw(0,0,jpg);
        jpg.LoadFromFile(FMapTilePath+'NE_1_1_0.jpg');
        FWorldmapLevel1.Canvas.Draw(720,0,jpg);
        jpg.LoadFromFile(FMapTilePath+'NE_1_0_1.jpg');
        FWorldmapLevel1.Canvas.Draw(0,360,jpg);
        jpg.LoadFromFile(FMapTilePath+'NE_1_1_1.jpg');
        FWorldmapLevel1.Canvas.Draw(720,360,jpg);
      except
        FWorldmapLevel1.Canvas.TextOut(10,50,'Error loading enhanced Worldmap!');
      end;
    finally
      jpg.Free;
    end;
  finally
  end;
  // In the case that the Map could not be loaded, let the user know
  if (FWorldmap.Width <> 720) or
     (FWorldmap.Height <> 360) then
  begin
    FWorldmap.SetSize(720,360);
    FWorldmap.Canvas.TextOut(10,50,'Error loading Worldmap!');
  end;
  // Create the cach bitmap for the complex display
  FWorldmapChached := TBitmap.Create;
  // Bring the coordinates of the Testrectangle into the edit fields
  RectToGUI;
end;

destructor TfrmLQTEarthMap.Destroy;
begin
  // Free everything we created
  FRectShifter.Free;
  FWorldmap.Free;
  FWorldmapLevel1.Free;
  FWorldmapChached.Free;
  inherited Destroy;
end;


{pnlEarthMapChangeBounds is called, when the Panel where the Paintbox is placed on
   is changed in size }
procedure TfrmLQTEarthMap.pnlEarthMapChangeBounds(Sender: TObject);
var
  w, h : Integer;
  ratio : Double;
  pbh, pbw : Integer;
  pbc : Integer;
begin
  // The goal is to fit the paint box into the panel, keeping it's
  // AspectRation of 1:2
  w := pnlEarthMap.Width;
  h := pnlEarthMap.Height;
  if (w < 2) or (h < 1) then
  begin
    w := 2;
    h := 1;
  end;
  ratio := w / h;
  if ratio > 2.0 then
  begin // wider than our background
    pbh := h;
    pbw := Round(h * 2.0);
    pbc := (w - pbw) div 2;
    pbEarthMap.SetBounds(pbc,0,pbw,pbh);
  end
  else
  begin // taller than our background
    pbw := w;
    pbh := Round(w / 2.0);
    pbc := (h - pbh) div 2;
    pbEarthMap.SetBounds(0,pbc,pbw,pbh);
  end;
  // Adjust the WorldRectangle to ease the transformation of coordinates
  FPaintBoxWorld.SetFrom(0,0,pbw,pbh);
  // Setup the background map and the HelperClass
  SetupCachedWorldMap;
  SetupRectShifter;
end;

procedure TfrmLQTEarthMap.tbEstimatedItemsThresholdChange(Sender: TObject);
begin
  case tbEstimatedItemsThreshold.Position of
    0 : FEstimatedItemsThreshold := 4;
    1 : FEstimatedItemsThreshold := 10;
    2 : FEstimatedItemsThreshold := 50;
    3 : FEstimatedItemsThreshold := 100;
    4 : FEstimatedItemsThreshold := 200;
    5 : FEstimatedItemsThreshold := 500;
    6 : FEstimatedItemsThreshold := 1000;
  else
    tbEstimatedItemsThreshold.Position := 3;
    FEstimatedItemsThreshold := 100;
  end;
  lblEstimatedItemsThreshold.Caption := IntToStr(FEstimatedItemsThreshold);
  UpdateItemCountInRectangle;
end;


end.

