{ A simple QuadTree-Explorer Frame used in the QuadTree demo
  Copyright (C) 2023,2024 Ekkehard Domning (www.domis.de)

  License: modified LGPL with linking exception (like RTL, FCL and LCL)

  See the file COPYING.modifiedLGPL.txt, included in the Lazarus distribution,
  for details about the license.

  Version Date        Change
  0.0.1   2024-08-23  First release

  This frame allows the manipulation and exploring of the content of the QuadTree.
  The textbase access display the content of one selected node, its location within
  the tree and the contained items.
  The map based access shows where the items are located, agglomerating items
  where they are so close together that they could not distinguished.
  A small MapViewer is implemented, to allow a zoom into smaller details of the
  World. The biggest useful magnification is approx 2 mi / 3 km per Pixel.
  The used MapTiles are generated from the Natural Earth Maps, see details in the unit.
  A real world example is given in the OSMMapView Example.
}

unit uFrameLqtExplorer;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, StdCtrls, ExtCtrls, Graphics,
  LCLType, ClipBrd, ComCtrls, Dialogs, TAGraph, TASeries, Types,
  ulazQuadTreeGeometry,
  uBasicTestCommons,
  ulazQuadTree, uSimpleNEMapTiler;

type
  { TframeLQTExplorer }

  TframeLQTExplorer = class(TFrame)
    btnChildNE: TButton;
    btnChildNE1: TButton;
    btnChildNW: TButton;
    btnChildNW1: TButton;
    btnChildSE: TButton;
    btnChildSE1: TButton;
    btnChildSW: TButton;
    btnChildSW1: TButton;
    btnSearchNode: TButton;
    btnOneLevelUp1: TButton;
    btnRecurseNodeToClipboard: TButton;
    btnOneLevelUp: TButton;
    btnPackCurrentNode: TButton;
    btnExpandCurrentNode: TButton;
    btnTreeClear: TButton;
    btnPackTree: TButton;
    btnCreateItem: TButton;
    btnExpandTree: TButton;
    btnDeleteNodeItem: TButton;
    btnTreeToClipboard: TButton;
    btnSaveToFile: TButton;
    btnLoadFromFile: TButton;
    btnNextNode: TButton;
    btnNextItem: TButton;
    cbShowOccupiedRect: TCheckBox;
    chartNodeList: TChart;
    chartNodeListLineSeries1: TLineSeries;
    edtFindItemCaption: TEdit;
    edtDescr: TEdit;
    edtLatNode: TEdit;
    edtLonItem: TEdit;
    edtLatItem: TEdit;
    edtLonNode: TEdit;
    gbFindNode: TGroupBox;
    gbNodeLevel: TGroupBox;
    gbTree: TGroupBox;
    gbNewItem: TGroupBox;
    gbItemSearch: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    lbNodeLevel: TListBox;
    lbNodeItems: TListBox;
    memoNodeInfo: TMemo;
    OpenDialog1: TOpenDialog;
    pnlCurrentNodeMap: TPanel;
    pbCurrentNodeMap: TPaintBox;
    pcDisplay: TPageControl;
    pnlNavigation1: TPanel;
    pnlQuadTree: TPanel;
    pnlNodeItemControl: TPanel;
    pnlNodeItems: TPanel;
    pnlCenter: TPanel;
    pnlNavigation: TPanel;
    rgCurrentNodeReadCreate: TRadioGroup;
    SaveDialog1: TSaveDialog;
    Splitter1: TSplitter;
    Splitter2: TSplitter;
    Splitter3: TSplitter;
    sbExplorerMap: TStatusBar;
    tsTreeChart: TTabSheet;
    tsCurrentNodeText: TTabSheet;
    tsCurrentNodeMap: TTabSheet;
    procedure btnChildNEClick(Sender: TObject);
    procedure btnChildNWClick(Sender: TObject);
    procedure btnChildSEClick(Sender: TObject);
    procedure btnChildSWClick(Sender: TObject);
    procedure btnCreateItemClick(Sender: TObject);
    procedure btnDeleteNodeItemClick(Sender: TObject);
    procedure btnExpandCurrentNodeClick(Sender: TObject);
    procedure btnExpandTreeClick(Sender: TObject);
    procedure btnLoadFromFileClick(Sender: TObject);
    procedure btnNextItemClick(Sender: TObject);
    procedure btnNextNodeClick(Sender: TObject);
    procedure btnOneLevelUpClick(Sender: TObject);
    procedure btnPackCurrentNodeClick(Sender: TObject);
    procedure btnPackTreeClick(Sender: TObject);
    procedure btnRecurseNodeToClipboardClick(Sender: TObject);
    procedure btnSaveToFileClick(Sender: TObject);
    procedure btnSearchNodeClick(Sender: TObject);
    procedure btnTreeClearClick(Sender: TObject);
    procedure btnTreeToClipboardClick(Sender: TObject);
    procedure cbShowOccupiedRectChange(Sender: TObject);
    procedure edtLonLatChange(Sender: TObject);
    procedure lbNodeItemsClick(Sender: TObject);
    procedure lbNodeLevelSelectionChange(Sender: TObject; User: boolean);
    procedure pbCurrentNodeMapMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbCurrentNodeMapMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure pnlCurrentNodeMapChangeBounds(Sender: TObject);
    procedure pbCurrentNodeMapPaint(Sender: TObject);
    procedure rgCurrentNodeReadCreateClick(Sender: TObject);
  private
    FQuadTree : TLazQuadTree; // A refernce to the used QuadTree
    FQuadTreeChangedEvent : TNotifyEvent; // An Event variable to allow the notification of changes
    FCurrentNode : TLQTNode; // The current node, which is displayes
    FCurrentItem : TObject;
    FLatItem, FLonItem : Double; // Latitude and Longitude of an item to be created
    FLatNode, FLonNode : Double; // Latitude and Longitude of an node to be searched
    FPaintBoxWorld : TLQTRect; // The World Rectangle of the paintbox
    FNodeMap : TBitmap; // The loaded Node map from the file (720x360 px)
    procedure SetCurrentNode(Value : TLQTNode);
    procedure SetQuadTree(Value : TLazQuadTree);
    function IsInCreateNodeMode : Boolean;
    procedure UpdateDirectionButtons;
    function ClickableNodeFromMouseXY(const X, Y : Integer) : TLQTNode;
    procedure ChartFromTree;
  public
    property QuadTree : TLazQuadTree read FQuadTree write SetQuadTree;
    property QuadTreeChangedEvent : TNotifyEvent read FQuadTreeChangedEvent write FQuadTreeChangedEvent;
    property CurrentNode : TLQTNode read FCurrentNode write SetCurrentNode;
    procedure UpdateGUIFromCurrentNode;
    procedure TreeContentChanged;
    procedure CallMeAfterFormActivated;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;override;
  end;

const
  // The Parameters for the Agglomeration, could be changed to see what happens
  AgglomerateMinItems = 8;
  AgglomerateLevel    = 3;

implementation

{$R *.lfm}

{ TframeLQTExplorer }

procedure TframeLQTExplorer.btnOneLevelUpClick(Sender: TObject);
begin
  if not Assigned(FCurrentNode) then Exit;
  if Assigned(FCurrentNode.ParentNode) then // If the current node has a parent
    CurrentNode := FCurrentNode.ParentNode; // use the parent
end;

procedure TframeLQTExplorer.btnPackCurrentNodeClick(Sender: TObject);
begin
  if not Assigned(FCurrentNode) then Exit;
  FCurrentNode.PackNode; // Pack node and all children below
  CurrentNode := FCurrentNode;
  if Assigned(FQuadTreeChangedEvent) then
    FQuadTreeChangedEvent(Self);
end;

procedure TframeLQTExplorer.btnPackTreeClick(Sender: TObject);
begin
  if not Assigned(FQuadTree) then Exit;
  FQuadTree.RootNode.PackNode;  // Pack the whole Tree
  CurrentNode := FQuadTree.RootNode;
  ChartFromTree;
  if Assigned(FQuadTreeChangedEvent) then
    FQuadTreeChangedEvent(Self);
end;

procedure TframeLQTExplorer.btnRecurseNodeToClipboardClick(Sender: TObject);
var
  sl : TStringList;
  lOldCursor : TCursor;
begin
  if not Assigned(FCurrentNode) then Exit;
  sl := TStringList.Create;
  try
    lOldCursor := Screen.Cursor;  // Save the current mouse pointer
    Screen.Cursor := crHourGlass; // set it to the wainting one
    try
      FCurrentNode.NodeAsStrings(sl); // get the readable content
      ClipBoard.AsText := sl.Text; // Drop it into the clipboard
    finally
      Screen.Cursor := lOldCursor; // reinstall the previous mouse pointer
    end;
  finally
    sl.Free;
  end;
end;

procedure TframeLQTExplorer.btnSaveToFileClick(Sender: TObject);
var
  lOldCursor : TCursor;
begin
  if not Assigned(FQuadTree) then Exit;
  if SaveDialog1.Execute() then
  begin
    lOldCursor := Screen.Cursor;  // Save the current mouse pointer
    Screen.Cursor := crHourGlass; // set it to the wainting one
    try
      FQuadTree.WriteTreeToFile(SaveDialog1.FileName); // Write the Tree
    finally
      Screen.Cursor := lOldCursor; // reinstall the previous mouse pointer
    end;
  end;
end;

procedure TframeLQTExplorer.btnSearchNodeClick(Sender: TObject);
var
  n : TLQTNode;
  pt : TLQTPoint;
begin
  if not Assigned(FQuadTree) then Exit;
  pt.X := FLonNode;
  pt.Y := FLatNode;
  if not FQuadTree.WorldRect.ContainsPoint(pt,FQuadTree.WorldRect,FQuadTree.WorldModel) then
  begin
    Application.MessageBox('Coords out of World-rectangle!','Range error',MB_OK or MB_ICONSTOP);
    Exit;
  end;
  n := FQuadTree.FindCoordNode(FLonNode,FLatNode);
  if Assigned(n) then
    CurrentNode := n
  else if FCurrentNode <> FQuadTree.RootNode then
    CurrentNode := FQuadTree.RootNode
end;

procedure TframeLQTExplorer.btnTreeClearClick(Sender: TObject);
begin
  if not Assigned(FQuadTree) then Exit;

  if Application.MessageBox('All nodes in the QuadTree will be deleted! Delete?','Clear Quad-Tree',MB_YESNO) = IDYES then
  begin
    FQuadTree.Clear; // This will clear the tree, free all nodes and items
    CurrentNode := FQuadTree.RootNode;
    ChartFromTree;
    if Assigned(FQuadTreeChangedEvent) then
      FQuadTreeChangedEvent(Self);
  end;
end;

procedure TframeLQTExplorer.btnTreeToClipboardClick(Sender: TObject);
var
  sl : TStringList;
  lOldCursor : TCursor;
begin
  if not Assigned(FQuadTree) then Exit;
  sl := TStringList.Create;
  try
    lOldCursor := Screen.Cursor;  // Save the current mouse pointer
    Screen.Cursor := crHourGlass; // set it to the wainting one
    try
      FQuadTree.TreeAsStrings(sl); // Copy the Trees readable content
      ClipBoard.AsText := sl.Text; // into the clipboard
    finally
      Screen.Cursor := lOldCursor; // reinstall the previous mouse pointer
    end;
  finally
    sl.Free;
  end;
end;

procedure TframeLQTExplorer.cbShowOccupiedRectChange(Sender: TObject);
begin
  pbCurrentNodeMap.Invalidate;
  pbCurrentNodeMap.Repaint;
end;

procedure TframeLQTExplorer.edtLonLatChange(Sender: TObject);
var
  edt : TEdit;
  s : String;
  fs : TFormatSettings;
  d : Double;
begin
  if not (Sender is TEdit) then Exit;
  edt := TEdit(Sender);
  s := Trim(edt.Text);
  while Pos(',',s) > 0 do
    s[Pos(',',s)] := '.';
  fs.ThousandSeparator:= #0;
  fs.DecimalSeparator:='.';
  d := StrToFloatDef(s,-1e20,fs);
  if d <= -1e20 then
    edt.Color := clRed
  else if edt.Color <> clDefault then
    edt.Color := clDefault;
  if edt.Name = 'edtLonItem' then
    FLonItem := d
  else if edt.Name = 'edtLatItem' then
    FLatItem := d
  else if edt.Name = 'edtLonNode' then
    FLonNode := d
  else if edt.Name = 'edtLatNode' then
    FLatNode := d
  else
    Exit;
end;

procedure TframeLQTExplorer.lbNodeItemsClick(Sender: TObject);
begin
  // Depending of the Items selection, enable the delete Items Button
  btnDeleteNodeItem.Enabled := (lbNodeItems.ItemIndex >= 0);
end;

procedure TframeLQTExplorer.lbNodeLevelSelectionChange(Sender: TObject;
  User: boolean);
var
  n : TLQTNode;
begin
  if (not User) or (lbNodeLevel.ItemIndex < 0) then Exit;
  // In the Node-Level Listbox the parent and grandparent node until the
  // root node of the selected node are displayed.
  // If the user clicks to one entry, we pick the node from the objects
  // property of the items
  n := TLQTNode(lbNodeLevel.Items.Objects[lbNodeLevel.ItemIndex]);
  // and select the node if not the current one.
  if n <> FCurrentNode then
    CurrentNode := n;
end;

procedure TframeLQTExplorer.pbCurrentNodeMapMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  n : TLQTNode;
begin
  // If the user click with the mouse we try to find an agglomeration and
  // if found zoom in
  n := ClickableNodeFromMouseXY(X,Y);
  if Assigned(n) then
    CurrentNode := n;
end;

procedure TframeLQTExplorer.pbCurrentNodeMapMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  n : TLQTNode;
  newcursor : TCursor;
  ptx, pty : Double;
begin
  // If the user moves the mouse we try to find an agglomeration and
  // set the pointer to the HandPoint one to indicate that we found something
  newcursor := crDefault;
  n := ClickableNodeFromMouseXY(X,Y);
  if Assigned(n) then
    newcursor := crHandPoint;
  if Screen.Cursor <> newcursor then
    Screen.Cursor := newcursor;

  if Assigned(FCurrentNode) then
  begin
    TransformPoint(X,Y,
                   FPaintBoxWorld, FCurrentNode.NodeRect,
                   ptx,pty);
    sbExplorerMap.Panels[1].Text := Format('%1.6f',[pty]);
    sbExplorerMap.Panels[3].Text := Format('%1.6f',[ptx]);
  end;
end;

{pnlEarthMapChangeBounds is called, when the Panel where the Paintbox is placed on
   is changed in size }
procedure TframeLQTExplorer.pnlCurrentNodeMapChangeBounds(Sender: TObject);
var
  w, h : Integer;
  ratio : Double;
  pbh, pbw : Integer;
  pbc : Integer;
begin
  // The goal is to fit the paint box into the panel, keeping it's
  // AspectRation of 1:2
  w := pnlCurrentNodeMap.Width;
  h := pnlCurrentNodeMap.Height;
  if (w < 2) or (h < 1) then
  begin
    w := 2;
    h := 1;
  end;
  ratio := w / h;
  if ratio > 2.0 then
  begin // wider than our background
    pbh := h;
    pbw := Round(h * 2.0);
    pbc := (w - pbw) div 2;
    pbCurrentNodeMap.SetBounds(pbc,0,pbw,pbh);
  end
  else
  begin // taller than our background
    pbw := w;
    pbh := Round(w / 2.0);
    pbc := (h - pbh) div 2;
    pbCurrentNodeMap.SetBounds(0,pbc,pbw,pbh);
  end;
  FPaintBoxWorld.SetFrom(0,0,pbw,pbh);
end;

{ pbCurrentNodeMapPaint the central MapPaint event, several actions are performed here}
procedure TframeLQTExplorer.pbCurrentNodeMapPaint(Sender: TObject);
var
  r : TRect;
  lAgglomerationArray: TLQTAgglomerationArray = Nil;
  lItemList: TList = Nil;
  i,j : Integer;
  s : String;
  d : Integer;
  cnt : Integer;
  ptx, pty : Double;
  xi,yi : Integer;
  extent : TSize;
  poi : TTestPOI;
  shp : TTestShape;
  lqtrect : TLQTRect;
begin
  r.Left := 0;
  r.Top := 0;
  r.Right := pbCurrentNodeMap.Width;
  r.Bottom := pbCurrentNodeMap.Height;
  // Draw the current NodeMap as the background
  // This works every time, even if no Tree or Node ist in use
  pbCurrentNodeMap.Canvas.StretchDraw(r,SimpleNEMapTiler.CurrentTile);

  if Assigned(FCurrentNode) then
  begin
    if cbShowOccupiedRect.Checked then
    begin
      lqtrect := FCurrentNode.OccupiedRect;
      lqtrect.Transform(FCurrentNode.NodeRect,FPaintBoxWorld);
      pbCurrentNodeMap.Canvas.Pen.Width := 1;
      pbCurrentNodeMap.Canvas.Pen.Color := clYellow;
      pbCurrentNodeMap.Canvas.Brush.Color := clWhite;
      pbCurrentNodeMap.Canvas.Brush.Style := bsClear;
      r.Left := Round(lqtrect.Left);
      r.Top := Round(lqtrect.Top);
      r.Right := Round(lqtrect.Right);
      r.Bottom := Round(lqtrect.Bottom);
      pbCurrentNodeMap.Canvas.Rectangle(r);
    end;

    // A list to take our items
    lItemList := TList.Create;
    try
      // While in the Map display the enumeration of items is used, here we use
      // the AsList function.
      // In general AsList are easier to use, but waste memory.
      // Enumeration needs separate callback function, which are harder to understand
      // and to debug.
      // AgglomerateNodeAsList dive into the Tree beginning from the current
      // node. The dive stoppes if the maximum dive level (in this case AgglomerateLevel) has been reached
      // or the items count is bigger than given (in this case AgglomerateMinItems = 8).
      // The found items are listed, the agglomeration are stored in the dynamic array
      FQuadTree.AgglomerateNodeAsList(FCurrentNode,AgglomerateLevel,AgglomerateMinItems,lAgglomerationArray, lItemList);
      pbCurrentNodeMap.Canvas.Pen.Color := clNavy;
      pbCurrentNodeMap.Canvas.Brush.Color := clWhite;
      pbCurrentNodeMap.Canvas.Brush.Style := bsClear;
      // ... and font
      pbCurrentNodeMap.Canvas.Font.Color := clBlack;
      pbCurrentNodeMap.Canvas.Font.Name := 'Arial';
      pbCurrentNodeMap.Canvas.Font.Size := 6;
      pbCurrentNodeMap.Canvas.Font.Style := [];

      // Paint all found agglomerations, with the found number in a circle
      d := 10;
      for i := 0 to High(lAgglomerationArray) do
      begin
        TransformPoint(lAgglomerationArray[i].Point.X,
                       lAgglomerationArray[i].Point.Y,
                       FCurrentNode.NodeRect, FPaintBoxWorld,
                       ptx,pty);
        xi := Round(ptx);
        yi := Round(pty);
        pbCurrentNodeMap.Canvas.Ellipse(xi - d, yi - d,
                                        xi + d, yi + d);
        cnt := Round(lAgglomerationArray[i].Count);
        if cnt > 9999 then
          s := '10K+'
        else
          s := IntToStr(cnt);
        extent := pbCurrentNodeMap.Canvas.TextExtent(s);
        // Write
        pbCurrentNodeMap.Canvas.TextOut(xi - (extent.CX div 2), yi - (extent.CY div 2), s);
      end;
      pbCurrentNodeMap.Canvas.Pen.Color := clRed;
      pbCurrentNodeMap.Canvas.Brush.Style := bsSolid;
      pbCurrentNodeMap.Canvas.Brush.Color := clRed;
      pbCurrentNodeMap.Canvas.Font.Size := 7;
      // Iterate to the items list and paint all items.
      // Take care that some items are CountryShapes.
      // You may notice, that some shapes are painted, even if they are fully outside
      // the display. E.G. the US-Border is paint if zoomed to colorado and all
      // parts of the (mainland) Borders are outside.
      d := 2;
      for i := 0 to lItemList.Count-1 do
      begin
        poi := TTestPOI(lItemList.Items[i]);
        if not Assigned(poi) then Continue; // skip unnasigned items (will never catch)
        if poi is TTestShape then
        begin  // if we found a shape
          // Check wether the item belongs to the displayed node.
          if FCurrentNode.NodeItemIndex(poi) >= 0 then
            pbCurrentNodeMap.Canvas.Pen.Color := clRed // Yes, paint it red
          else
            pbCurrentNodeMap.Canvas.Pen.Color := clMaroon; // Else, dark red
          pbCurrentNodeMap.Canvas.Brush.Style := bsClear;
          // cast to shape
          shp := TTestShape(poi);
          // Paint the polyline
          for j := 0 to shp.CoordsCount-1 do
          begin
            TransformPoint(shp.CoordsLons[j],
                           shp.CoordsLats[j],
                           FCurrentNode.NodeRect, FPaintBoxWorld,
                           ptx,pty);
            if j = 0 then
              pbCurrentNodeMap.Canvas.MoveTo(Round(ptx),Round(pty))
            else
              pbCurrentNodeMap.Canvas.LineTo(Round(ptx),Round(pty));
          end;
          { // You can paint the bounding box if you like
          lqtrect.Assign(shp.ShapeBounds);
          lqtrect.Transform(FCurrentNode.NodeRect, FPaintBoxWorld);
          r.Left := Round(lqtrect.Left);
          r.Top := Round(lqtrect.Top);
          r.Right := Round(lqtrect.Right);
          r.Bottom := Round(lqtrect.Bottom);
          pbCurrentNodeMap.Canvas.Brush.Style := bsClear;
          pbCurrentNodeMap.Canvas.Rectangle(r);
          }
        end
        else
        begin // Paint the POI
          TransformPoint(poi.Lon,
                         poi.Lat,
                         FCurrentNode.NodeRect, FPaintBoxWorld,
                         ptx,pty);
          xi := Round(ptx);
          yi := Round(pty);
          pbCurrentNodeMap.Canvas.Brush.Style := bsSolid;
          pbCurrentNodeMap.Canvas.Ellipse(xi - d, yi - d,
                                          xi + d, yi + d);
          pbCurrentNodeMap.Canvas.Brush.Style := bsClear;
          s := poi.Caption;
          extent := pbCurrentNodeMap.Canvas.TextExtent(s);
          // Write the caption
          // You will see overlaping Captions on the Map
          pbCurrentNodeMap.Canvas.TextOut(xi + 4, yi - (extent.CY div 2), s);
        end;
      end;
    finally
      lItemList.Free; // don't forget to free the list
    end;
  end;
end;

procedure TframeLQTExplorer.rgCurrentNodeReadCreateClick(Sender: TObject);
begin
  UpdateDirectionButtons; // The buttons work "inverse" when switching between read and create nodes
end;

procedure TframeLQTExplorer.btnChildNWClick(Sender: TObject);
var
  cn : TLQTNode;
  canCreate : Boolean;
begin
  if not Assigned(FCurrentNode) then Exit;

  canCreate := IsInCreateNodeMode;
  if canCreate then
  begin  // In create mode we create not existing child nodes
    FCurrentNode.CreateOrGetChildNode(nclNW);
    UpdateGUIFromCurrentNode;
    if Assigned(FQuadTreeChangedEvent) then
      FQuadTreeChangedEvent(Self);
  end
  else
  begin
    cn := FCurrentNode.ChildNodeNW; // in read mode, we step into the existing node
    if Assigned(cn) then
      CurrentNode := cn;
  end;
end;

procedure TframeLQTExplorer.btnChildSEClick(Sender: TObject);
var
  cn : TLQTNode;
  canCreate : Boolean;
begin
  if not Assigned(FCurrentNode) then Exit;

  canCreate := IsInCreateNodeMode;
  if canCreate then
  begin
    FCurrentNode.CreateOrGetChildNode(nclSE);
    UpdateGUIFromCurrentNode;
    if Assigned(FQuadTreeChangedEvent) then
      FQuadTreeChangedEvent(Self);
  end
  else
  begin
    cn := FCurrentNode.ChildNodeSE;
    if Assigned(cn) then
      CurrentNode := cn;
  end;
end;

procedure TframeLQTExplorer.btnChildSWClick(Sender: TObject);
var
  cn : TLQTNode;
  canCreate : Boolean;
begin
  if not Assigned(FCurrentNode) then Exit;

  canCreate := IsInCreateNodeMode;
  if canCreate then
  begin
    FCurrentNode.CreateOrGetChildNode(nclSW);
    UpdateGUIFromCurrentNode;
    if Assigned(FQuadTreeChangedEvent) then
      FQuadTreeChangedEvent(Self);
  end
  else
  begin
    cn := FCurrentNode.ChildNodeSW;
    if Assigned(cn) then
      CurrentNode := cn;
  end;
end;

procedure TframeLQTExplorer.btnCreateItemClick(Sender: TObject);
var
  pt : TLQTPoint;
  poi : TTestPOI;
begin
  pt.X := FLonItem;
  pt.Y := FLatItem;
  if not FQuadTree.WorldRect.ContainsPoint(pt,FQuadTree.WorldRect,FQuadTree.WorldModel) then
  begin
    Application.MessageBox('Coords out of World-rectangle!','Range error',MB_OK or MB_ICONSTOP);
    Exit;
  end;
  poi := TTestPOI.Create(FLatItem,FLonItem,edtDescr.Text); // Create an item
  try
    FQuadTree.InsertItem(poi,True,False,FLonItem,FLatItem); // Insert the item into the Tree. The tree may become reconstructed during the insertion process
    poi := Nil;  // insertion was successful
  finally
    if Assigned(poi) then
      poi.Free; // under all circumstances, if the insertion failed, free the not inserted item.
  end;
  CurrentNode := FCurrentNode;
end;

procedure TframeLQTExplorer.btnDeleteNodeItemClick(Sender: TObject);
var
  ndx : Integer;
  o : TObject;
begin
  if not Assigned(FCurrentNode) then Exit;
  ndx := lbNodeItems.ItemIndex;  // Get the Index of the item in the list
  if ndx < 0 then Exit; // No item selected (should never happen)
  o := FCurrentNode.AssignedNodeItems[ndx]; // get the node item from the node
  if Assigned(o) then
  begin // if assigned, should be always the case
    o := FCurrentNode.ExtractNodeItem(o); // extract the item from the node
    if Assigned(o) then // if assigned (in this case this should be always the case)
    begin // this indicates, that the node has been removed. Since the item is stored in tree
      o.Free; // we have to free it here, to prevent memory leaks!
      CurrentNode := FCurrentNode; // Update the GUI
      // since Items are not displayed in the main form, the update call is not necesarry
      if Assigned(FQuadTreeChangedEvent) then
        FQuadTreeChangedEvent(Self);
    end;
  end;
end;

procedure TframeLQTExplorer.btnExpandCurrentNodeClick(Sender: TObject);
begin
  if not Assigned(FCurrentNode) then Exit;
  FCurrentNode.ExpandNode; // Expand the node. This means, all point items will be stored in leaf nodes
  CurrentNode := FCurrentNode;
  if Assigned(FQuadTreeChangedEvent) then
    FQuadTreeChangedEvent(Self);
end;

procedure TframeLQTExplorer.btnExpandTreeClick(Sender: TObject);
begin
  if not Assigned(FQuadTree) then Exit;
  FQuadTree.RootNode.ExpandNode; // Expand the tree. This means, all point items will be stored in leaf nodes
  CurrentNode := FQuadTree.RootNode;
  ChartFromTree;
  if Assigned(FQuadTreeChangedEvent) then
    FQuadTreeChangedEvent(Self);
end;

procedure TframeLQTExplorer.btnLoadFromFileClick(Sender: TObject);
var
  lOldCursor : TCursor;
begin
  if not Assigned(FQuadTree) then Exit;
  if OpenDialog1.Execute() then
  begin
    lOldCursor := Screen.Cursor;  // Save the current mouse pointer
    Screen.Cursor := crHourGlass; // set it to the wainting one
    try
      FQuadTree.ReadTreeFromFile(OpenDialog1.FileName); // Do the read
      QuadTree := FQuadTree;
      ChartFromTree;
      if Assigned(FQuadTreeChangedEvent) then
        FQuadTreeChangedEvent(Self);
    finally
      Screen.Cursor := lOldCursor; // reinstall the previous mouse pointer
    end;
  end;
end;

procedure TframeLQTExplorer.btnNextItemClick(Sender: TObject);
var
  rc : Boolean;
  ni : TObject;
  poi : TTestPOI absolute ni;
  nn : TLQTNode;
  ndx : Integer;
  loops : Integer;
  lOldCursor : TCursor;
begin
  if not Assigned(FQuadTree) then Exit;
  lOldCursor := Screen.Cursor;  // Save the current mouse pointer
  Screen.Cursor := crHourGlass; // set it to the wainting one
  try
    loops := 2; // Counter of loops (bypassing the root node)
    ni := FCurrentItem;
    nn := FCurrentNode;
    repeat
      // NextItem starts the search at the current Node/Item,
      // but the Item we are looking for may lay before the current node/item.
      // This means, we should tolerate one "not found" indication, since
      // this means, we only bypassed the root node.
      rc := FQuadTree.NextItem(ni,nn,ni,nn);
      if not rc then
      begin
        Dec(loops); // Dec the loop counter
        if loops > 0 then // Retry the search if in the second part
          rc := FQuadTree.NextItem(ni,nn,ni,nn);
        if not rc then // if failed
        begin
          gbItemSearch.Caption := 'Search Item (not found)';
          CurrentNode := FQuadTree.RootNode ;
          Exit;
        end;
      end;
      // an empty search field means, find anything
      // other values are taken im UpperCase and tested for beeing a part of
      if (Length(edtFindItemCaption.Text) <= 0) or
         (Pos(UpperCase(edtFindItemCaption.Text),UpperCase(poi.Caption)) > 0) then
      begin  // we found something
        if nn <> FCurrentNode then
          CurrentNode := nn;
        FCurrentItem := ni;
        ndx := FCurrentNode.NodeItemIndex(ni);
        gbItemSearch.Caption := Format('Search Item (ItemIndex: %d)',[ndx]);
        pbCurrentNodeMap.Invalidate;
        pbCurrentNodeMap.Repaint;
        Break; // Search done
      end;
    until False;
  finally
    Screen.Cursor := lOldCursor; // reinstall the previous mouse pointer
  end;
end;

procedure TframeLQTExplorer.btnNextNodeClick(Sender: TObject);
var
  n : TLQTNode;
begin
  if not Assigned(FQuadTree) then Exit;
  n := FQuadTree.NextNode(FCurrentNode);
  if not Assigned(n) then
    n := FQuadTree.NextNode(FCurrentNode);
  if FCurrentNode <> n then
  begin
    CurrentNode := n;
    pbCurrentNodeMap.Invalidate;
    pbCurrentNodeMap.Repaint;
  end;
end;

procedure TframeLQTExplorer.btnChildNEClick(Sender: TObject);
var
  cn : TLQTNode;
  canCreate : Boolean;
begin
  canCreate := IsInCreateNodeMode;
  if canCreate then
  begin
    FCurrentNode.CreateOrGetChildNode(nclNE);
    UpdateGUIFromCurrentNode;
    if Assigned(FQuadTreeChangedEvent) then
      FQuadTreeChangedEvent(Self);
  end
  else
  begin
    cn := FCurrentNode.ChildNodeNE;
    if Assigned(cn) then
      CurrentNode := cn;
  end;
end;

procedure TframeLQTExplorer.SetCurrentNode(Value: TLQTNode);
begin
  if Assigned(Value) and (Value.Owner <> FQuadTree) then
    ELQTException.Create('Only Nodes from this QuadTree could be used!');
  FCurrentNode := Value;
  FCurrentItem := Nil;
  if Assigned(FCurrentNode) then
    SimpleNEMapTiler.LoadTile(FCurrentNode.NodeLevel,FCurrentNode.Left,FCurrentNode.Top);
  UpdateGUIFromCurrentNode;
end;

procedure TframeLQTExplorer.SetQuadTree(Value: TLazQuadTree);
begin
  FQuadTree := Value;
  if Assigned(FQuadTree) then
    CurrentNode := FQuadTree.RootNode
  else
    CurrentNode := Nil;
end;

function TframeLQTExplorer.IsInCreateNodeMode: Boolean;
begin
  Result := (rgCurrentNodeReadCreate.ItemIndex = 1);
end;
{ UpdateDirectionButtons ensure that the buttons are enabled to possible directions }
procedure TframeLQTExplorer.UpdateDirectionButtons;
var
  canCreate : Boolean;
begin
  try
    if not Assigned(FCurrentNode) then
    begin
      btnChildNW.Enabled := False;
      btnChildNE.Enabled := False;
      btnChildSW.Enabled := False;
      btnChildSE.Enabled := False;
      Exit;
    end;
    canCreate := IsInCreateNodeMode;
    if canCreate then
    begin
      btnChildNW.Enabled := not Assigned(FCurrentNode.ChildNodeNW);
      btnChildNE.Enabled := not Assigned(FCurrentNode.ChildNodeNE);
      btnChildSW.Enabled := not Assigned(FCurrentNode.ChildNodeSW);
      btnChildSE.Enabled := not Assigned(FCurrentNode.ChildNodeSE);
    end
    else
    begin
      btnChildNW.Enabled := Assigned(FCurrentNode.ChildNodeNW);
      btnChildNE.Enabled := Assigned(FCurrentNode.ChildNodeNE);
      btnChildSW.Enabled := Assigned(FCurrentNode.ChildNodeSW);
      btnChildSE.Enabled := Assigned(FCurrentNode.ChildNodeSE);
    end;
  finally
    // Ensure that the secondary buttons follow the primary
    btnChildNW1.Enabled := btnChildNW.Enabled;
    btnChildNE1.Enabled := btnChildNE.Enabled;
    btnChildSW1.Enabled := btnChildSW.Enabled;
    btnChildSE1.Enabled := btnChildSE.Enabled;
  end;
end;

{ ClickableNodeFromMouseXY
  This method is a bit a tricky short cut, which work good here, but may not a good solution
  in other cases.
  One could store the Agglomerated regions in a global array, and than iterate through
  this array to find the possible one, where the mouse is hovering.
  But since the Agglomeration is performed on Nodes, we simply emulate the behaviour here.
  Find the Node where the mouse is over (fast Tree function FindCoordNode!) and
  than step the levels up until the node three levels below the current displayed
  node is found. If this node contains >= AgglomerateMinItems = 8 items, an Agglomeration is found.
}
function TframeLQTExplorer.ClickableNodeFromMouseXY(const X, Y: Integer
  ): TLQTNode;
var
  ptx, pty : Double;
  nodecenterx, nodecentery : Double;
  n : TLQTNode;
begin
  Result := Nil;
  if Assigned(FCurrentNode) then
  begin
    TransformPoint(X,Y,
                   FPaintBoxWorld, FCurrentNode.NodeRect,
                   ptx,pty);
    n := FCurrentNode.FindCoordNode(ptx,pty); // Find the node where the mouse is hovering
    while Assigned(n) and ((n.NodeLevel - FCurrentNode.NodeLevel) > AgglomerateLevel) do
      n := n.ParentNode; // step one level up, until 3 levels below the current one is reached
    // If such a node exists, an the node contains more items than displayable
    // a Agglomeration is found.
    if Assigned(n) and (n.TotalNodeItemsCount >= AgglomerateMinItems) then
    begin
      nodecenterx := (n.NodeRect.Right + n.NodeRect.Left) / 2.0;
      nodecentery := (n.NodeRect.Top + n.NodeRect.Bottom) / 2.0;
      TransformPoint(nodecenterx,nodecentery,
                     FCurrentNode.NodeRect, FPaintBoxWorld,
                     ptx,pty);
      if (Abs(ptx-X) <= 8) and
         (Abs(pty-Y) <= 8) then
        Result := n;
    end;
  end;
end;

procedure TframeLQTExplorer.ChartFromTree;
var
  list : TList;
  values : array of Double;
  cnt : Integer;
  i : Integer;
  n : TLQTNode;
begin
  if not Assigned(FQuadTree) then Exit;
  list := TList.Create;
  try
    FQuadTree.ListTreeNodes(list);
    cnt := list.Count;
    SetLength(values,cnt);
    for i := 0 to cnt-1 do
    begin
      n := TLQTNode(list.Items[i]);
      values[i] := n.NodeLevel;
    end;
    chartNodeListLineSeries1.Clear;
    chartNodeListLineSeries1.AddArray(values);
  finally
    list.Free;
  end;
end;

procedure TframeLQTExplorer.UpdateGUIFromCurrentNode;
// here we fill all the boxes with the content and enable/disable the buttons
// depending of the assignment and content of the Tree and the CurrentItem
var
  sl : TStringList;
  n : TLQTNode;
  s : String;
  ic : TLQTNodeChildLocation;
  orient : String;
  i : Integer;
  poi : TTestPOI;
  ndx : Integer;
begin
  btnOneLevelUp.Enabled := Assigned(FCurrentNode) and Assigned(FCurrentNode.ParentNode); // No "upper" Node to the RootNode exists
  btnOneLevelUp1.Enabled := btnOneLevelUp.Enabled;
  btnPackCurrentNode.Enabled := Assigned(FCurrentNode);
  btnExpandCurrentNode.Enabled := Assigned(FCurrentNode);
  btnRecurseNodeToClipboard.Enabled := Assigned(FCurrentNode);

  btnPackTree.Enabled := Assigned(FQuadTree);
  btnCreateItem.Enabled := Assigned(FQuadTree);
  btnTreeClear.Enabled := Assigned(FQuadTree);
  btnTreeToClipboard.Enabled := Assigned(FQuadTree);
  btnLoadFromFile.Enabled := Assigned(FQuadTree);
  btnSaveToFile.Enabled := Assigned(FQuadTree);

  btnSearchNode.Enabled := Assigned(FQuadTree);
  btnNextItem.Enabled := Assigned(FQuadTree);

  UpdateDirectionButtons;
  if Assigned(FCurrentNode) then
  begin
    // Get the content as a readable text from the node
    s := FCurrentNode.NodeAsString;
    // and add the information of the children (if some)
    for ic := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    begin
      n := FCurrentNode.ChildNodes[ic];
      if Assigned(n) then
        s := s + #13#10 + Format(' ChildNode[%s]: TotalItems %d',[LQTNodeChildLocationStrings[ic],n.TotalNodeItemsCount]);
    end;
    memoNodeInfo.Text := s;
  end
  else
    memoNodeInfo.Text := 'n/a';
  memoNodeInfo.SelStart := 0; // Scroll to Top
  memoNodeInfo.SelLength := 0;
  // Map update
  pbCurrentNodeMap.Invalidate;
  pbCurrentNodeMap.Repaint;
  sl := TStringList.Create;
  try
    // Fill the listbox with the items (if any)
    if Assigned(FCurrentNode) then
    begin
      // iterate through the items and drop a text into the stringlist
      for i := 0 to FCurrentNode.NodeItemsLength -1 do
      begin
        poi := TTestPOI(FCurrentNode.NodeItems[i]);
        if Assigned(poi) then
          sl.Add(Format('[%d] %1.6f %1.6f %s',[i, poi.Lat, poi.Lon, poi.Caption]));
      end;
    end;
    if lbNodeItems.Items.Text <> sl.Text then
    begin // if the items changed, update the listbox
      ndx := lbNodeItems.ItemIndex; // save the current index
      lbNodeItems.Items.Assign(sl);
      if ndx > lbNodeItems.Items.Count-1 then  // if the old index is beyond the scope
        ndx := lbNodeItems.Items.Count-1; // set to the last (or none)
      lbNodeItems.ItemIndex := ndx; // use the new index as selection
    end;
    btnDeleteNodeItem.Enabled := (lbNodeItems.ItemIndex >= 0); // if an item is selected, enable the delete button
    // Now fill the node level box, with all parents until the root is reached
    sl.Clear;
    n := FCurrentNode; // begin with the current node
    while Assigned(n) do // do, as long a node is in use
    begin
      // except for the root node, get the location of the current node within it's parent
      if n.GetNodesParentChildLocation(ic) then
        orient := LQTNodeChildLocationStrings[ic]
      else
        orient := 'n/a';
      // Prepare some information
      s := Format('TotalItems %d, %s, (Left/X: %1.6f, Top/Y %1.6f), (Right/X: %1.6f, Bottom/Y %1.6f)',
                   [n.TotalNodeItemsCount,
                   orient,
                   n.Left,
                   n.Top,
                   n.Right,
                   n.Bottom]);
      sl.Insert(0,s);
      sl.Objects[0] := n;
      n := n.ParentNode; // do next with the parent of the current node, will be Nil if the current is the root node
    end;
    lbNodeLevel.Items.Assign(sl); //Assign what we collected
    lbNodeLevel.ItemIndex := sl.Count-1; // select always the deepest node level, if one exists
  finally
    sl.Free;
  end;
end;

procedure TframeLQTExplorer.TreeContentChanged;
begin
  QuadTree := FQuadTree;
  ChartFromTree;
end;

procedure TframeLQTExplorer.CallMeAfterFormActivated;
begin
  pcDisplay.ActivePageIndex:=0;
  pnlCurrentNodeMapChangeBounds(Nil);
  edtLonLatChange(edtLonNode);
  edtLonLatChange(edtLatNode);
end;

constructor TframeLQTExplorer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FNodeMap := TBitmap.Create;
  CurrentNode := Nil; // This will update the GUI to a startup configuration
end;

destructor TframeLQTExplorer.Destroy;
begin
  FNodeMap.Free;
  inherited Destroy;
end;

end.

