{ A general demo for the Quad-Tree for Objects
  Copyright (C) 2023,2024 Ekkehard Domning (www.domis.de)

  The three tab-sheets contains:
  - An Earth-Map with an resizeable rectangle for selections and some item related
    display options: uFrameLqtEarthmap.pas
  - An QuadTree-Explorer with a text based display and editor of the content and
    a map based view only display: uFrameLqtExplorer.pas
  - An setup page for selecting some test cases and an about-page

  License: The same modified LGPL as the Free Pascal RTL
           See the file COPYING.modifiedLGPL for more details

  Author: Ekkehard Domning (ekkehard@domning.eu)

  CAUTION: The form uses frames, to encapsulate parts of the program into different
           units. Sometimes, after changes in the GUI of the frames, the program starts
           with heavy exceptions. In this case find the frame in the object inspector and
           delete it. Then insert the frame again (second right icon in the "Standard"-tab
           of the components-bar). This should fix the problem.
           To ease this process, never rename inserted frames, instead use the name
           that is assigned by the IDE.
           Also, never assign any events in the inserted frame. Events should only assigned
           in the unit, or (rarely used) manually in the code.

  Version Date        Change
  0.0.1   2024-08-04  First release
}

unit uLQTBasicTestMain;

{$mode objfpc}{$H+}

interface

{$DEFINE USE_EPIKTIMER}

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ComCtrls,
  ExtCtrls, LCLType,
  // unextfloat,
  IniFiles,lazFileUtils, // For loading the test cases
  uLoadCSV, // A form to show the loading progress
  ulazQuadTreeGeometry, ulazQuadTree, // The tree and the geometry sources
  uBasicTestCommons, // Some sample global definitions, including the POI-Definition
  uSimpleNEMapTiler, // A simple map tile for the Natural Earth map
  uFrameLqtExplorer, uFrameLqtEarthmap, // The two sample frames
  uPerformanceTest; // The PerformanceThread

type

  { TForm1 }

  TForm1 = class(TForm)
    btnLoadLargeSet: TButton;
    btnMediumSet: TButton;
    btnCountryShapes: TButton;
    btnStartStopPerformanceTest: TButton;
    cbClearPriorLoading: TCheckBox;
    frameLQTExplorer1: TframeLQTExplorer;
    frmLQTEarthMap1: TfrmLQTEarthMap;
    gbTestCases: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    lblPOIImportResult: TLabel;
    memoPerformanceTest: TMemo;
    pcMain: TPageControl;
    Panel1: TPanel;
    Panel2: TPanel;
    TabSheet1: TTabSheet;
    tsPerformanceTest: TTabSheet;
    tsSetup: TTabSheet;
    tsExplorer: TTabSheet;
    Timer1: TTimer;
    tsEarthMap: TTabSheet;
    procedure btnCountryShapesClick(Sender: TObject);
    procedure btnLoadLargeSetClick(Sender: TObject);
    procedure btnMediumSetClick(Sender: TObject);
    procedure btnStartStopPerformanceTestClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    FFormFullyLoaded : Boolean;
    FTestQuadTree : TLazQuadTree;
    FErroredLoadedFileLine : Integer;
    FCSVErrorMessage : String;
    FMediumSetLoaded : Boolean;
    FLargeSetLoaded : Boolean;
    FShapesLoaded : Boolean;
    FPerformanceTestThread : TPerformanceTestThread;
    // Three callbacks/events to provide needed Date to the QuadTree
    function OnTestQuadTreeGetItemCoord(Sender : TObject; AItem : TObject; var AX, AY : Double) : Boolean;
    function OnTestQuadTreeGetItemRect(Sender : TObject; AItem : TObject; var AItemRect : TLQTRect) : Boolean;
    function OnTestQuadTreeGetItemCaption(Sender : TObject; AItem : TObject; var AItemCaption : String) : Boolean;
    function OnLoadCSVLineIntoQuadTreeEvent(Sender : TObject; ALine : String) : Boolean;
    procedure OnPerformanceTestThreadStatusUpdate(Sender : TPerformanceTestThread; AStatusMessage : String; ACurrentTest : Integer);
    procedure LoadDefaultPOIs; // Loads the default POIs on startup
    procedure LoadLargeDataset;
    procedure LoadMediumDataset;
    procedure LoadCountryShapes;
    procedure LoadPOIs(const AFileName : String; const AClearQuadTree : Boolean = False); // Loads the POI from the POS.csv file
    procedure OnLQTExplorerTreeChanged(Sender : TObject);
    procedure StopPerformanceTest;
  public

  end;


// function _nextafter(x,y:cdouble):cdouble;cdecl;external 'msvcrt.dll' ;

const
  EarthMapTileFolder = 'EarthMapTiles\';

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }
procedure TForm1.Timer1Timer(Sender: TObject);
begin
  if FormLoadCSV.Visible then
  begin // During the loading process it is possible to update the display
    frameLQTExplorer1.TreeContentChanged;
    frmLQTEarthMap1.TreeContentChanged;
  end;

  if Assigned(FPerformanceTestThread) and
    FPerformanceTestThread.Terminated then
  begin
    memoPerformanceTest.Lines.Add('Stopping performance test.');
    btnStartStopPerformanceTest.Caption := 'Start';
    StopPerformanceTest;
    memoPerformanceTest.Lines.Add('Performance test stopped.');
  end;
end;

{ This event is called if the tree needs the point coordinate for an inserted item }
function TForm1.OnTestQuadTreeGetItemCoord(Sender: TObject; AItem: TObject;
  var AX, AY: Double): Boolean;
begin
  Result := False;
  if Assigned(AItem) then
  begin
    if (AItem is TTestPOI) then
    begin // both POIs and country shapes has a single point
      AX := TTestPOI(AItem).Lon;
      AY := TTestPOI(AItem).Lat;
      Result := True;
    end;
  end;
end;

{ This event is called if the tree needs the rectangle (bounding box) for an inserted item }
function TForm1.OnTestQuadTreeGetItemRect(Sender : TObject; AItem : TObject;
                                               var AItemRect : TLQTRect) : Boolean;
begin
  Result := False; // preset with false
  if Assigned(AItem) then
  begin
    if (AItem is TTestShape) then
    begin // Only country shapes has a bounding box
      AItemRect.Assign(TTestShape(AItem).ShapeBounds);
      Result := True;
    end;
  end;
end;

{ The event is called if the tree need the description of an item
  this is the case on all "get as string" methods of the tree}
function TForm1.OnTestQuadTreeGetItemCaption(Sender: TObject;
  AItem: TObject; var AItemCaption: String): Boolean;
begin
  Result := False;
  if Assigned(AItem) then
  begin
    if (AItem is TTestPOI) then
    begin
      AItemCaption := TTestPOI(AItem).Caption;
      Result := True;
    end;
  end;
end;

{ This event is called if the Modal-Form says we should load another item into the tree.
  To keep the program responsive, the loading is continued for 25ms then other
  Messages are processed }
function TForm1.OnLoadCSVLineIntoQuadTreeEvent(Sender: TObject; ALine: String
  ): Boolean;
var
  lpoi : TTestPOI;
  lErrorMessage : String = '';
  n : TLQTNode;
begin
  Result := False;
  lpoi := TTestPOI.Create(0.0,0.0,'');
  try
    if lpoi.SetPOIFromCSVLine(ALine,lErrorMessage) then
    begin
      n := FTestQuadTree.InsertItem(lpoi,True,False,lpoi.Lon,lpoi.Lat);
      if not Assigned(n) then
        // The POI could no be inserted, terminate loading
        FErroredLoadedFileLine := FormLoadCSV.LoadedItemCount
      else
        // The poi is inserted, nil the local reference
        lpoi := Nil;
      Result := True;
    end
    else
    begin
      FErroredLoadedFileLine := FormLoadCSV.LoadedItemCount;
      FCSVErrorMessage := lErrorMessage;
    end;
  finally
    if Assigned(lpoi) then // if assigned, the insertion into the tree failed
      FreeAndNil(lpoi); // free the item here
  end;
end;

procedure TForm1.LoadDefaultPOIs;
begin
  LoadMediumDataset
end;

procedure TForm1.LoadLargeDataset;
var
  fn, poipath : String;
begin
  poipath := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName));
  fn := poipath+'POI.csv';
  if not FileExists(fn) then
  begin
    poipath := IncludeTrailingPathDelimiter(ResolveDots(ExtractFilePath(Application.ExeName)+ExampleDataPath));
    fn := poipath+'POI.csv';
  end;
  LoadPOIs(fn);
  FLargeSetLoaded := True;
end;

procedure TForm1.LoadMediumDataset;
var
  fn, poipath : String;
begin
  poipath := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName));
  fn := poipath+'cities.csv';
  if not FileExists(fn) then
  begin
    poipath := IncludeTrailingPathDelimiter(ResolveDots(ExtractFilePath(Application.ExeName)+ExampleDataPath));
    fn := poipath+'cities.csv';
  end;
  LoadPOIs(fn);
  FMediumSetLoaded := True;
end;
{ The loading of the country shapes does not take that long, so we just display
  a waiting cursor }
procedure TForm1.LoadCountryShapes;
var
  fn, shppath : String;
  sl : TStringList;
  i : Integer;
  ini : TMemIniFile;
  keys : TStringList;
  shp : TTestShape;
  j : Integer;
  cnt : Integer = 0;
  shpstr : String;
  oldcursor : TCursor;
begin
  lblPOIImportResult.Caption :=  'Loading country shapes...';
  shppath := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName));
  fn := shppath+'CountryShapes.ini';
  if not FileExists(fn) then
  begin
    shppath := IncludeTrailingPathDelimiter(ResolveDots(ExtractFilePath(Application.ExeName)+ExampleDataPath));
    fn := shppath+'CountryShapes.ini';
  end;
  oldcursor := Screen.Cursor;
  try
    Screen.Cursor := crHourGlass;
    ini := TMemIniFile.Create(fn);
    try
      sl := Nil;
      keys := Nil;
      try
        sl := TStringList.Create;
        keys := TStringList.Create;
        ini.ReadSections(sl);
        for i := 0 to sl.Count-1 do
        begin
          ini.ReadSection(sl[i],keys);
          for j := 0 to keys.Count-1 do
          begin
            shpstr := ini.ReadString(sl[i],keys[j],'');
            if Length(shpstr) < 1 then Continue;
            // Create the shape class
            shp := TTestShape.Create(sl[i]);
            try
              // read the data from the line
              shp.ReadShapesFromString(shpstr);
              if shp.CoordsCount > 0 then
              begin // if there are any data in this line read
                // Insert the item into the tree
                if not Assigned(FTestQuadTree.InsertItem(shp)) then
                begin // the return value is nil, if the shape could not be inserted
                  // in this case display a message
                  lblPOIImportResult.Caption := Format('Error in section "%s" in key "%s"',[sl[i],keys[j]]);
                  // the class is freed below
                end
                else
                begin
                  // if sucessfully inserted, the tree takes care, so nil the local variable
                  shp := Nil;
                  // and increment the counter of loaded items
                  Inc(cnt);
                end;
              end;
            finally
              // if the shape class is assigned, the object is not inserted into the tree
              if Assigned(shp) then
                // so free the object here
                FreeAndNil(shp);
            end;
          end;
        end;
      finally
        if Assigned(sl) then
          sl.Free;
        if Assigned(keys) then
          keys.Free;
      end;
    finally
      ini.Free;
    end;
  finally
    Screen.Cursor := oldcursor;
  end;
  lblPOIImportResult.Caption :=  Format('Loaded %d shapes',[cnt]);
  FShapesLoaded := True;
end;

{ LoadPOIs will load the "Points of Interest". Since this could take a while, during
  loading a modal window ist displayed, which allows the termination of the loading,
  but also the display of a progress bar, to entertain the waiting time.
}
procedure TForm1.LoadPOIs(const AFileName: String; const AClearQuadTree: Boolean);
var
  sl : TStringList;
begin
  if AClearQuadTree then
    FTestQuadTree.Clear;
  sl := TStringList.Create;
  try
    try
      sl.LoadFromFile(AFileName);
    except
      on e : Exception do
      begin
        lblPOIImportResult.Caption := 'Error loading POI data. "'+e.Message+'" "'+AFileName+'"';
        Exit;
      end;
    end;
    FErroredLoadedFileLine := -1;
    FormLoadCSV.ClearListOnClose := True; // The internal list should be clear when we're finished
    FormLoadCSV.CloseOnLoadingDone := True; // The dialog should close when the job is done
    FormLoadCSV.OnLoadCSVLineIntoQuadTree := @OnLoadCSVLineIntoQuadTreeEvent; // The linie by line processing Event
    FormLoadCSV.LoadList := sl; // drop the data into the dialog
    FormLoadCSV.ShowModal; // show modal (this will prevent any user interaction within this form, but the timer is running!!)
    // After loading is done
    frameLQTExplorer1.TreeContentChanged;
    frmLQTEarthMap1.TreeContentChanged;
    if FErroredLoadedFileLine >= 0 then // Give the user a message what the result of the loading was
      lblPOIImportResult.Caption := Format('%d lines imported into %d objects. Error in Line %d, Msg "%s"',[sl.Count,FormLoadCSV.LoadedItemCount,FErroredLoadedFileLine, FCSVErrorMessage])
    else
      lblPOIImportResult.Caption := Format('%d lines imported into %d objects',[sl.Count,FormLoadCSV.LoadedItemCount]);
  finally
    if Assigned(sl) then
      sl.Free;
  end;
end;
procedure TForm1.FormCreate(Sender: TObject);
var
  fp : String;
begin
  // Creation of the Quadtree, the Tree will own the POI-objects
  FTestQuadTree := TLazQuadTree.Create(EarthWorld,lqtwmCylindrical,True);
  // Callbacks for the Tree to get the information from the POI Items
  FTestQuadTree.OnGetItemCoord := @OnTestQuadTreeGetItemCoord;
  FTestQuadTree.OnGetItemRect := @OnTestQuadTreeGetItemRect;
  FTestQuadTree.OnGetItemCaption := @OnTestQuadTreeGetItemCaption;
  // Callbacks to read and write the POI to and read from the stream
  FTestQuadTree.OnReadStreamItem := @TTestItemStreamHelper.OnTestQuadTreeReadStreamItem;
  FTestQuadTree.OnWriteStreamItem := @TTestItemStreamHelper.OnTestQuadTreeWriteStreamItem;


  // Compose the name of the path to the data folder
  // If the exe is used stand alone, the data is assumed to be "below" the exe-path
  // In the debug case, the path is assumed to be related to the source code.
  fp := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName));
  fp := fp+EarthMapTileFolder;
  if not DirectoryExists(fp) then
  begin
    fp := IncludeTrailingPathDelimiter(ResolveDots(ExtractFilePath(Application.ExeName)+ExampleDataPath));
    fp := fp+EarthMapTileFolder;
  end;
  SimpleNEMapTiler.TilePath := fp;

  // The explorer frame can change the QuadTree, but needs to notify the main
  // program if the tree has been changed by the user
  frameLQTExplorer1.QuadTree := FTestQuadTree;
  frameLQTExplorer1.QuadTreeChangedEvent := @OnLQTExplorerTreeChanged;
  frmLQTEarthMap1.QuadTree := FTestQuadTree;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FTestQuadTree.Free;
  StopPerformanceTest; // Don't forget to free the running thread
end;

procedure TForm1.FormActivate(Sender: TObject);
begin
  // Is only called once after the first activation of the main form
  OnActivate := Nil;
  pcMain.ActivePageIndex := 0;
  FFormFullyLoaded := True;
  // Update the frames to display an empty tree
  frmLQTEarthMap1.CallMeAfterFormActivated;
  frameLQTExplorer1.CallMeAfterFormActivated;
  // start loading the default test case
  LoadDefaultPOIs;
end;

procedure TForm1.btnLoadLargeSetClick(Sender: TObject);
begin
  if FLargeSetLoaded and (not cbClearPriorLoading.Checked) then
  begin
    Application.MessageBox('The large size dataset is already loaded. Check the "Clear Tree-Data ..." tick-box above to reload!',
                           'Loading not possible',MB_OK or MB_ICONHAND);
    Exit;
  end;
  if cbClearPriorLoading.Checked then
  begin
    FTestQuadTree.Clear;
    FMediumSetLoaded := False;
    FLargeSetLoaded := False;
    FShapesLoaded := False;
  end;
  LoadLargeDataSet;
end;

procedure TForm1.btnCountryShapesClick(Sender: TObject);
begin
  if FShapesLoaded and (not cbClearPriorLoading.Checked) then
  begin
    Application.MessageBox('The country shapes are already loaded. Check the "Clear Tree-Data ..." tick-box above to reload!',
                           'Loading not possible',MB_OK or MB_ICONHAND);
    Exit;
  end;
  if cbClearPriorLoading.Checked then
  begin
    FTestQuadTree.Clear;
    FMediumSetLoaded := False;
    FLargeSetLoaded := False;
    FShapesLoaded := False;
  end;
  LoadCountryShapes;
  frameLQTExplorer1.TreeContentChanged;
  frmLQTEarthMap1.TreeContentChanged;
end;

procedure TForm1.btnMediumSetClick(Sender: TObject);
begin
  if FMediumSetLoaded and (not cbClearPriorLoading.Checked) then
  begin
    Application.MessageBox('The medium size dataset is already loaded. Check the "Clear Tree-Data ..." tick-box above to reload!',
                           'Loading not possible',MB_OK or MB_ICONHAND);
    Exit;
  end;
  if cbClearPriorLoading.Checked then
  begin
    FTestQuadTree.Clear;
    FMediumSetLoaded := False;
    FLargeSetLoaded := False;
    FShapesLoaded := False;
  end;
  LoadMediumDataSet;
end;
procedure TForm1.OnPerformanceTestThreadStatusUpdate(Sender : TPerformanceTestThread; AStatusMessage : String; ACurrentTest : Integer);
begin
  memoPerformanceTest.Lines.Add(Format('[%d] %s', [ACurrentTest,AStatusMessage]));
end;

procedure TForm1.btnStartStopPerformanceTestClick(Sender: TObject);
begin
  if not Assigned(FPerformanceTestThread) then
  begin
    memoPerformanceTest.Text := 'Starting performance test.';
    btnStartStopPerformanceTest.Caption := 'Stop';
    FPerformanceTestThread := TPerformanceTestThread.Create(False, @OnPerformanceTestThreadStatusUpdate);
  end
  else
  begin
    memoPerformanceTest.Lines.Add('Stopping performance test.');
    btnStartStopPerformanceTest.Caption := 'Start';
    StopPerformanceTest;
    memoPerformanceTest.Lines.Add('Performance test stopped.');
  end;
end;

procedure TForm1.OnLQTExplorerTreeChanged(Sender: TObject);
begin
  // Forward the changed tree notification to the earth map frame
  frmLQTEarthMap1.TreeContentChanged;
  if FTestQuadTree.RootNode.TotalNodeItemsCount = 0 then
  begin
    FMediumSetLoaded := False;
    FLargeSetLoaded := False;
    FShapesLoaded := False;
  end;
end;

procedure TForm1.StopPerformanceTest;
begin
  if Assigned(FPerformanceTestThread) then
  begin
    if not FPerformanceTestThread.Terminated then
      FPerformanceTestThread.Terminate;
    if FPerformanceTestThread.Suspended then
      FPerformanceTestThread.Suspended := False;
    FPerformanceTestThread.WaitFor;
    FPerformanceTestThread.Free;
    FPerformanceTestThread := Nil
  end;
end;

end.

