{ Some performance tests for the QuadTree demo
  Copyright (C) 2023,2024 Ekkehard Domning (www.domis.de)

  License: modified LGPL with linking exception (like RTL, FCL and LCL)

  See the file COPYING.modifiedLGPL.txt, included in the Lazarus distribution,
  for details about the license.

  Version Date        Change
  0.0.1   2024-08-23  First release
}
unit uPerformanceTest;

{$mode ObjFPC}{$H+}

interface

// You can use the EPikTimer or the default Millsecond timer from the System
{$DEFINE UseEpikTimer}

uses
  Classes, SysUtils, Forms,
  {$IFDEF UseEpikTimer}
  epiktimer, // http://wiki.lazarus.freepascal.org/EpikTimer
  {$ENDIF}
  lazFileUtils, // ResolveDots
  ulazQuadTreeGeometry, // The underlying rectangkle geometry
  ulazQuadTree, // The QuadTree
  uBasicTestCommons; // The Eart and POI definition

type

  { TMeasureTimer }

  TMeasureTimer = class(Tobject)
  private
    {$IFDEF UseEpikTimer}
      FEpikTimer : TEpikTimer;
    {$ELSE}
      Ft0 : Int64;
      Ft1 : Int64;
      procedure WaitForBeginning;
    {$ENDIF}

    function GetDurationMS : Double;
  public
    property DurationMS : Double read GetDurationMS;
    procedure Start;
    procedure Stop;
    constructor Create;
    destructor Destroy;override;
  end;

  { TPerformanceTestThread }

  TPerformanceTestThread = class; // Forward

  TPerformanceTestThreadStatusUpdateEvent =
    procedure(Sender : TPerformanceTestThread; AStatusMessage : String; ACurrentTest : Integer) of object;

  { The TPerformanceTestThread will perform several tests an measure the time consumption
    to get an idea on how fast the algorithm is working. This may also helps on further
    development to quantify improvements.
    The main test are
    - importing data into the tree (from csv) *1)
    - saving to file (.lqtf) *1)
    - loading from file (.lqtf) *1)
    - estimating and counting items in the tree *2)
    - enumerating/listing items in the tree *2)

    *1) this is highly dependend of the used file source
    *2) this is performed with the identical query also on an unsorted list,
        and a pair of sorted lists to generate a ratio of performance,
        independend of the target machine.
  }
  TPerformanceTestThread = class(TThread)
  private
    FTempTreeFileName : String;
    FMeasureTimer : TMeasureTimer;
    FStatusUpdateEvent : TPerformanceTestThreadStatusUpdateEvent;
    FTestQuadTree : TLazQuadTree;
    FCurrentStatusMessage : String;
    FCurrentTest : Integer;
    FUnsortedItems : TList;
    FItemsSortedByLat : TList;
    FItemsSortedByLon : TList;

    function OnTestQuadTreeGetItemCoord(Sender : TObject; AItem : TObject; var AX, AY : Double) : Boolean;
    function OnTestQuadTreeGetItemRect(Sender : TObject; AItem : TObject; var AItemRect : TLQTRect) : Boolean;
    function OnTestQuadTreeGetItemCaption(Sender : TObject; AItem : TObject; var AItemCaption : String) : Boolean;
    procedure DoSynchedNotification;
    procedure SynchedNotification(ACurrentStatusMessage : String);
    procedure ImportCSVIntoTree;
    procedure SaveTreeToFile;
    procedure LoadTreeFromFile;
    procedure SetupAndSortPOILists;
    procedure EstimateAndCountItems;
  protected
    procedure Execute;override;
  public
    property Terminated;
    property CurrentTest : Integer read FCurrentTest;
    property StatusUpdateEvent : TPerformanceTestThreadStatusUpdateEvent read FStatusUpdateEvent;
    constructor Create(CreateSuspended: Boolean;
                       const AStatusUpdateEvent : TPerformanceTestThreadStatusUpdateEvent;
                       const StackSize: SizeUInt = DefaultStackSize);
    destructor Destroy;override;
  end;

implementation

const
  { EstimatedItemThreshold, if TotalNodeItemsCount is equal or less this value
     then the estimation is terminated for this node and a part of the value is returned.
     Higher Value = faster calculation and greater deviation
  }
  EstimatedItemThreshold          = 500;
  { EstimatedItemDeviationThreshold, the minimum estimated items where the
    percentage of deviation is measured, assuming that the estimated value
    in a real application is less than this, a counting is subsequent counting
    is performed.}
  EstimatedItemDeviationThreshold = 100;
  { RandSeedEstimateAndCount will fix that all tests have the same "random"-Picking  }
  RandSeedEstimateAndCount = $BADC0DED;
  RandSeedItemsShuffle     = $ABAD1DEA;

  PoiLatLeftFoundFlag    = $00000001;
  PoiLonTopFoundFlag     = $00000002;
  PoiLatRightFoundFlag   = $00000004;
  PoiLonBottomFoundFlag  = $00000008;
  PoiPositionValid = PoiLatLeftFoundFlag or PoiLonTopFoundFlag or
                     PoiLatRightFoundFlag or PoiLonBottomFoundFlag;


{ TMeasureTimer }
{$IFNDEF UseEpikTimer}
procedure TMeasureTimer.WaitForBeginning;
begin
  Ft0 := GetTickCount64;
  repeat
    Ft1 := GetTickCount64;
  until Ft0 <> Ft1;
  Ft0 := Ft1;
end;
{$ENDIF}

function TMeasureTimer.GetDurationMS: Double;
begin
  {$IFDEF UseEpikTimer}
     Result := FEpikTimer.Elapsed * 1000;
  {$ELSE}
     Result := Ft1-Ft0;
  {$ENDIF}
end;

procedure TMeasureTimer.Start;
begin
  {$IFDEF UseEpikTimer}
     FEpikTimer.Clear;
     FEpikTimer.Start;
  {$ELSE}
    WaitForBeginning;
  {$ENDIF}
end;

procedure TMeasureTimer.Stop;
begin
  {$IFDEF UseEpikTimer}
     FEpikTimer.Stop;
  {$ELSE}
    Ft1 := GetTickCount64;
  {$ENDIF}
end;

constructor TMeasureTimer.Create;
begin
  {$IFDEF UseEpikTimer}
    FEpikTimer := TEpikTimer.Create(Nil);
  {$ENDIF}
end;

destructor TMeasureTimer.Destroy;
begin
  {$IFDEF UseEpikTimer}
    if Assigned(FEpikTimer) then
      FEpikTimer.Free;
  {$ENDIF}
  inherited;
end;

{ TPerformanceTestThread }

{ This event is called if the tree needs the point coordinate for an inserted item }
function TPerformanceTestThread.OnTestQuadTreeGetItemCoord(Sender: TObject; AItem: TObject;
  var AX, AY: Double): Boolean;
begin
  Result := False;
  if Assigned(AItem) then
  begin
    if (AItem is TTestPOI) then
    begin // both POIs and country shapes has a single point
      AX := TTestPOI(AItem).Lon;
      AY := TTestPOI(AItem).Lat;
      Result := True;
    end;
  end;
end;

{ This event is called if the tree needs the rectangle (bounding box) for an inserted item }
function TPerformanceTestThread.OnTestQuadTreeGetItemRect(Sender : TObject; AItem : TObject;
                                               var AItemRect : TLQTRect) : Boolean;
begin
  Result := False; // preset with false
  if Assigned(AItem) then
  begin
    if (AItem is TTestShape) then
    begin // Only country shapes has a bounding box
      AItemRect.Assign(TTestShape(AItem).ShapeBounds);
      Result := True;
    end;
  end;
end;

{ The event is called if the tree need the description of an item
  this is the case on all "get as string" methods of the tree}
function TPerformanceTestThread.OnTestQuadTreeGetItemCaption(Sender: TObject;
  AItem: TObject; var AItemCaption: String): Boolean;
begin
  Result := False;
  if Assigned(AItem) then
  begin
    if (AItem is TTestPOI) then
    begin
      AItemCaption := TTestPOI(AItem).Caption;
      Result := True;
    end;
  end;
end;

procedure TPerformanceTestThread.DoSynchedNotification;
begin
  if Assigned(FStatusUpdateEvent) then
    FStatusUpdateEvent(Self,FCurrentStatusMessage,FCurrentTest);
end;

procedure TPerformanceTestThread.SynchedNotification(
  ACurrentStatusMessage: String);
begin
  FCurrentStatusMessage := ACurrentStatusMessage;
  Synchronize(@DoSynchedNotification);
end;

procedure TPerformanceTestThread.ImportCSVIntoTree;
var
  lpoi : TTestPOI;
  lErrorMessage : String = '';
  n : TLQTNode;
  fn, poipath : String;
  sl : TStringList;
  i : Integer;
  errline, cntpoi : Integer;
begin
  SynchedNotification('Import CSV into the tree');

  poipath := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName));
  fn := poipath+'POI.csv';
  if not FileExists(fn) then
  begin
    poipath := IncludeTrailingPathDelimiter(ResolveDots(ExtractFilePath(Application.ExeName)+ExampleDataPath));
    fn := poipath+'POI.csv';
  end;

  errline := -1;
  cntpoi := 0;
  FTestQuadTree.Clear;
  sl := TStringList.Create;
  try
    FMeasureTimer.Start;
    try
      sl.LoadFromFile(fn);
    except
      on e : Exception do
      begin
        SynchedNotification('Error loading POI data. "'+e.Message+'" "'+fn+'"');
        Exit;
      end;
    end;
    FMeasureTimer.Stop;
    SynchedNotification(Format('Loading the file in %1.2f ms ("%s")',[FMeasureTimer.DurationMS,fn]));
    FMeasureTimer.Start;
    for i := 0 to sl.Count-1 do
    begin
      if Terminated then Exit;
      lpoi := TTestPOI.Create(0.0,0.0,'');
      try
        if lpoi.SetPOIFromCSVLine(sl[i],lErrorMessage) then
        begin
          n := FTestQuadTree.InsertItem(lpoi,True,False,lpoi.Lon,lpoi.Lat);
          if not Assigned(n) then
            // The POI could no be inserted, terminate loading
            errline := i
          else
          begin
            // The poi is inserted, nil the local reference
            lpoi := Nil;
            Inc(cntpoi);
          end;
        end
        else
          errline := i;
      finally
        if Assigned(lpoi) then // if assigned, the insertion into the tree failed
          FreeAndNil(lpoi); // free the item here
      end;
    end;
    FMeasureTimer.Stop;
    if errline >= 0 then // Give the user a message what the result of the loading was
      SynchedNotification(Format('%d lines imported into %d objects in %1.2f ms. Error in Line %d, Msg "%s"',[sl.Count,cntpoi,errline, FMeasureTimer.DurationMS]))
    else
      SynchedNotification(Format('%d lines imported into %d objects in %1.2f ms',[sl.Count,cntpoi,FMeasureTimer.DurationMS]));
  finally
    if Assigned(sl) then
      sl.Free;
  end;
end;

{ Here I would like to separate the pure file writing and reading
  from the processing of the content,
  to get compareable values to the import of the CSV-File.
  This requires to get one protected method into visibility.
  So I derive a class, override the method, calling inherited.
  Since this method is used only to read the first few bytes from the
  file, which identify the format. It is not necesarry to alter
  the class of the TestTree!
}

type
  TUnprotectedQuadTree = class(TLazQuadTree)
  public
    procedure ReadFileStreamStart(const AStream : TStream);override;
  end;
procedure TUnprotectedQuadTree.ReadFileStreamStart(const AStream : TStream);
begin
  inherited;
end;


procedure TPerformanceTestThread.SaveTreeToFile;
begin
  FTempTreeFileName := GetTempFileName;
  FTempTreeFileName := FTempTreeFileName + QuadTreeFileExtension;
  SynchedNotification(Format('QuadTree saving File "%s"',[FTempTreeFileName]));
  FMeasureTimer.Start;
  FTestQuadTree.WriteTreeToFile(FTempTreeFileName);
  FMeasureTimer.Stop;
  SynchedNotification(Format('QuadTree saved in %1.2f ms into file "%s"',[FMeasureTimer.DurationMS, FTempTreeFileName]))
end;

procedure TPerformanceTestThread.LoadTreeFromFile;
var
  fs : TFileStream;
  ms : TMemoryStream;
  upt : TUnprotectedQuadTree;
begin
  SynchedNotification(Format('QuadTree File reading into Memory "%s"',[FTempTreeFileName]));
  FMeasureTimer.Start;
  fs := Nil;
  ms := Nil;
  try
    ms := TMemoryStream.Create;
    fs := TFileStream.Create(FTempTreeFileName,fmOpenRead or fmShareDenyWrite);
    fs.Position := 0;
    ms.CopyFrom(fs, fs.Size);
    FMeasureTimer.Stop;
    SynchedNotification(Format('QuadTree File read in %1.2f ms into Memory "%s"',[FMeasureTimer.DurationMS, FTempTreeFileName]));
    ms.Position := 0;
    // Now read the StreamStart. Will work always, since we wrote the file a few milliseconds earlier.
    // But in the case an exception is raised
    upt := TUnprotectedQuadTree.Create(FTestQuadTree.RootNode.NodeRect,FTestQuadTree.WorldModel);
    try
      upt.ReadFileStreamStart(ms);
    finally
      upt.Free;
    end;
    // Now load the data from the memory stream
    FMeasureTimer.Start;
    FTestQuadTree.ReadTreeFromStream(ms);
    FMeasureTimer.Stop;
    SynchedNotification(Format('QuadTree read in %1.2f ms from memory',[FMeasureTimer.DurationMS]));
  finally
    if Assigned(fs) then
      fs.Free;
    if Assigned(ms) then
      ms.Free;
  end;
end;

function ListSortComparePoiByLat(Item1: Pointer;Item2: Pointer):Integer;
var
  poi1 : TTestPOI absolute Item1;
  poi2 : TTestPOI absolute Item2;
begin
  if poi1.Lat < poi2.Lat then
    Result := -1
  else if poi1.Lat > poi2.Lat then
    Result := 1
  else
    Result := 0;
end;
function ListSortComparePoiByLon(Item1: Pointer;Item2: Pointer):Integer;
var
  poi1 : TTestPOI absolute Item1;
  poi2 : TTestPOI absolute Item2;
begin
  if poi1.Lon < poi2.Lon then
    Result := -1
  else if poi1.Lon > poi2.Lon then
    Result := 1
  else
    Result := 0;
end;

procedure TPerformanceTestThread.SetupAndSortPOILists;
var
  i : Integer;
  p : Pointer;
  ni : TObject;
  nn : TLQTNode;
  ndx0, ndx1 : Integer;
  cnt : Integer;
begin
  FUnsortedItems.Clear;
  SynchedNotification('Collecting Items from the QuadTree');
  // Build the list from the tree
  nn := Nil;
  ni := Nil;
  repeat
    if not FTestQuadTree.NextItem(ni,nn,ni,nn) then Break;
    FUnsortedItems.Add(ni);
  until False;
  // shuffle the items
  SynchedNotification('Shuffle Items');
  RandSeed := RandSeedItemsShuffle;
  cnt := FUnsortedItems.Count;
  for i := 0 to (cnt div 2)-1 do
  begin
    ndx0 := Random(cnt);
    ndx1 := Random(cnt);
    p := FUnsortedItems.Items[ndx0];
    FUnsortedItems.Items[ndx0] := FUnsortedItems.Items[ndx1];
    FUnsortedItems.Items[ndx1] := p;
  end;

  // Start the sorting
  SynchedNotification('Sorting Items');
  FMeasureTimer.Start;

  FItemsSortedByLat.Clear;
  FItemsSortedByLon.Clear;
  for i := 0 to FUnsortedItems.Count-1 do
  begin
    p := FUnsortedItems.Items[i];
    FItemsSortedByLat.Add(p);
    FItemsSortedByLon.Add(p);
  end;
  FItemsSortedByLat.Sort(@ListSortComparePoiByLat);
  FItemsSortedByLon.Sort(@ListSortComparePoiByLon);
  FMeasureTimer.Stop;
  SynchedNotification(Format('Sorted Items in %1.1f ms',[FMeasureTimer.DurationMS]));
end;

procedure TPerformanceTestThread.EstimateAndCountItems;
  { CountListItemsInUnsortedLists runs to all items, tests whether they are inside
    the rectangle and returns the count. If AList is passed, the items are dropped into the list }
  function CountListItemsInUnsortedLists(const ARect : TLQTRect; const AList : TList = Nil) : Integer;
  var
    poi : TTestPOI;
    i,j : Integer;
    cnt : Integer;
    lNormRects : TLQTRectArray; // The normalized Rectangles
    lNormWorld : TLQTRect;
  begin
    cnt := 0;
    NormalizeRectangle(ARect,FTestQuadTree.RootNode.NodeRect,FTestQuadTree.WorldModel,
                       lNormRects,lNormWorld);
    for i := 0 to High(lNormRects) do
    begin
      if lNormWorld.WorldXAxisLeftPlus <> FTestQuadTree.WorldXAxisLeftPlus then
      begin
        lNormRects[i].Left:= -lNormRects[i].Left;
        lNormRects[i].Right:= -lNormRects[i].Right;
      end;
      if lNormWorld.WorldYAxisUpPlus <> FTestQuadTree.WorldYAxisUpPlus then
      begin
        lNormRects[i].Top:= -lNormRects[i].Top;
        lNormRects[i].Bottom:= -lNormRects[i].Bottom;
      end;
      for j := 0 to FUnsortedItems.Count-1 do
      begin
        poi := TTestPOI(FUnsortedItems.Items[j]);
        if lNormRects[i].ContainsPoint(poi.Lon, poi.Lat) then
        begin
          Inc(cnt);
          if Assigned(AList) then
            AList.Add(poi);
        end;
      end;
    end;
    Result := cnt;
  end;

  { BinarySearchLatLon finds the index of the ASearchValue in the sorted Lat/Lon list
     Parameters
      ASearchValue the value to look for
      ASearchLat if True the Latitude is used, if False the Longitude
     Output
       AIndex the index that is the SearchValue (if Result = True) or the
       next smaller Value (if Result = False) in the list.
       AIndex may be -1 if the value in Index 0 is greater than the ASearchValue
     Return
       True if the Index specifies the Item (direct hit)
       False if the index specifies the next smaller item im the list
     Caution
       if the result is false, the AIndex might be -1, indication that
       the item in index 0 is greater than the ASearchValue
  }
  function BinarySearchLatLon(const ASearchValue : Double; const ASearchLat : Boolean;
                              out AIndex : Integer) : Boolean;
  var
    ndxlo, ndxhi : Integer;
    poi : TTestPOI;
    d : Double;
  begin
    ndxlo := 0;
    ndxhi := FUnsortedItems.Count-1;
    while (ndxlo <= ndxhi) do
    begin
      AIndex := ndxlo + (ndxhi - ndxlo) div 2;
      // Check if x is present at mid
      if ASearchLat then
      begin
        poi := TTestPOI(FItemsSortedByLat.Items[AIndex]);
        d := poi.Lat;
      end
      else
      begin
        poi := TTestPOI(FItemsSortedByLon.Items[AIndex]);
        d := poi.Lon;
      end;
      if d = ASearchValue then
      begin
        AIndex := AIndex;
        Exit(True);
      end;
      // If x greater, ignore left half
      if (d < ASearchValue) then
        ndxlo := AIndex + 1
      // If x is smaller, ignore right half
      else
        ndxhi := AIndex - 1;
    end;
    // If we reach here, then element was not present
    if d > ASearchValue then
      Dec(AIndex);
    Result := False;
  end;

  { CountListItemsInSortedLists is very simmilar to CountListItemsInUnsortedLists.
    It runs to all items, tests whether they are inside
    the rectangle and returns the count. If AList is passed, the items are dropped into the list }
  function CountListItemsInSortedLists(const ARect : TLQTRect; const AList : TList = Nil) : Integer;
  var
    poi : TTestPOI;
    i,j : Integer;
    lNormRects : TLQTRectArray; // The normalized Rectangles
    lNormWorld : TLQTRect;
    ndxlatlo, ndxlathi : Integer;
    ndxlonlo, ndxlonhi : Integer;
    cnt : Integer;
  begin
    Result := 0;
    NormalizeRectangle(ARect,FTestQuadTree.RootNode.NodeRect,FTestQuadTree.WorldModel,
                       lNormRects,lNormWorld);
    for i := 0 to High(lNormRects) do
    begin
      // Reset the Tag field. The field will be used to speed up the search
      for j := 0 to FUnsortedItems.Count-1 do
      begin
        poi := TTestPOI(FUnsortedItems.Items[j]);
        poi.Tag := 0;
      end;

      if lNormWorld.WorldXAxisLeftPlus <> FTestQuadTree.WorldXAxisLeftPlus then
      begin
        lNormRects[i].Left:= -lNormRects[i].Left;
        lNormRects[i].Right:= -lNormRects[i].Right;
      end;
      if lNormWorld.WorldYAxisUpPlus <> FTestQuadTree.WorldYAxisUpPlus then
      begin
        lNormRects[i].Top:= -lNormRects[i].Top;
        lNormRects[i].Bottom:= -lNormRects[i].Bottom;
      end;


      // Find all Items in the Latitude range
      // First find the lowest and the highest index of the items
      if not BinarySearchLatLon(lNormRects[i].Bottom,True,ndxlatlo) then
        Inc(ndxlatlo);
      BinarySearchLatLon(lNormRects[i].Top,True,ndxlathi);

      // Now, mark all Items in the Latitude
      for j := ndxlatlo to ndxlathi do
      begin
        poi := TTestPOI(FItemsSortedByLat.Items[j]);
        poi.Tag := poi.Tag or PoiLatLeftFoundFlag or PoiLatRightFoundFlag;
      end;

      // Same with Longitude
      // First the lowest and highest index
      if not BinarySearchLatLon(lNormRects[i].Left,False,ndxlonlo) then
        Inc(ndxlonlo);
      BinarySearchLatLon(lNormRects[i].Right,False,ndxlonhi);

      cnt := 0;
      // Iterate to all found items
      // To avoid a secondary loop to all item, we combine the search, counting and adding here
      for j := ndxlonlo to ndxlonhi do
      begin
        poi := TTestPOI(FItemsSortedByLon.Items[j]);
        poi.Tag := poi.Tag or PoiLonTopFoundFlag or PoiLonBottomFoundFlag;
        if (poi.Tag and PoiPositionValid) = PoiPositionValid then
        begin // Item fits in Lat and on
          Inc(cnt); // Count
          if Assigned(AList) then
            AList.Add(poi); // Add to list if wanted
        end;
      end;
      Result := Result + cnt;
    end;
  end;

const
  TestArrLen = 1000;
var
  r : TLQTRect;
  w, h : Double;
  l, b : Double;
  d : Double;
  i, j, k : Integer;
  cntd : Double;
  cnt0, cnt2 : Integer;
  loops : Integer;
  list : TList;
  cntds : array[0..TestArrLen] of Double;
  cntdndx : Integer;
  cnts : array[0..TestArrLen] of Integer;
  cntndx : Integer;
  cntalts : array[0..TestArrLen] of Integer;
  cntaltndx : Integer;
  ndx : Integer;
  delta, avgdelta, maxdelta : Double;
begin
  SynchedNotification('Several retrieving methods. Result in calls/second.');

  cntdndx := 0;
  cntndx := 0;
  cntaltndx := 0;

  w := FTestQuadTree.RootNode.NodeRect.Width;
  h := FTestQuadTree.RootNode.NodeRect.Height;
  l := FTestQuadTree.RootNode.NodeRect.Left;
  b := FTestQuadTree.RootNode.NodeRect.Bottom;
  for i := 0 to 6 do
  begin
    loops := 0;
    // To get compareable results, we use "random"-Values for the choosen rectangles and
    // ensure that each test, will work with the same "random"-Values.
    // This is not Threadsafe, but we have only one thread running and the main program ist
    // not using this random functions.
    RandSeed := RandSeedEstimateAndCount;

    // Our tests are running for 250ms
    FMeasureTimer.Start;
    repeat
      // Setup an rectangle
      d := Random * w;
      r.Left := d + l;
      d := Random * w;
      r.Right := d + l;
      d := Random * h;
      r.Top := d + b;
      d := Random * h;
      r.Bottom := d + b;
      if r.Top < r.Bottom then
      begin // Swap the top and bottom (no wrap around allowed
        d := r.Top;
        r.Top := r.Bottom;
        r.Bottom := d;
      end;
      // perform the tests
      case i of
        0 : begin
              cnt0 := FTestQuadTree.CountAssignedItemsInRect(r);
              if cntndx < Length(cnts) then
              begin
                cnts[cntndx] := cnt0;
                Inc(cntndx);
              end;
            end;
        1 : begin
              cntd := FTestQuadTree.EstimatedCountOfAssignedItemsInRect(r,EstimatedItemThreshold);
              if cntdndx < Length(cntds) then
              begin
                cntds[cntdndx] := cntd;
                Inc(cntdndx);
              end;
            end;
        2 : begin
              CountListItemsInUnsortedLists(r);
            end;
        3 : begin
              cnt2 := CountListItemsInSortedLists(r);
              if cntaltndx < Length(cntalts) then
              begin
                cntalts[cntaltndx] := cnt2;
                Inc(cntaltndx);
              end;
            end;
        4 : begin
              list := TList.Create;
              try
                FTestQuadTree.ListItemsInRect(r,list);
              finally
                list.Free;
              end;
            end;
        5 : begin
              list := TList.Create;
              try
                CountListItemsInUnsortedLists(r,list);
              finally
                list.Free;
              end;
            end;
        6 : begin
              list := TList.Create;
              try
                CountListItemsInSortedLists(r,list);
              finally
                list.Free;
              end;
            end;
      end;
      Inc(loops);
    until FMeasureTimer.DurationMS > 250;
    FMeasureTimer.Stop;
    d := loops / FMeasureTimer.DurationMS * 1000;
    case i of
      0 : SynchedNotification(Format('QuadTree count Items %1.2f per second',[d]));
      1 : begin
            // cnt0 := 0;
            avgdelta := 0.0;
            maxdelta := 0.0;
            if cntdndx < cntndx then
              ndx := cntdndx
            else
              ndx := cntndx;
            for j := 0 to ndx-1 do
            begin
              // Validation of the circumstance that count is not zero, but estimation is zero
              // if (cnts[j] > 0) and (cntds[j] <= 0) then
              //  Inc(cnt0);
              if cntds[j] > EstimatedItemDeviationThreshold then
                delta := Abs(((cnts[j]/(cnts[j] + cntds[j]))*2)-1)
              else
                delta := 0.0;
              if maxdelta < delta then
                maxdelta := delta;
              avgdelta := avgdelta + delta;
            end;
            avgdelta := avgdelta / ndx;
            SynchedNotification(Format('QuadTree estimated Items %1.2f per second (Deviation (cnt > %d) Avg. %1.1f %%, Max. %1.1f %%)',
                                        [d,EstimatedItemDeviationThreshold, avgdelta * 100, maxdelta * 100]));
          end;
      2 : SynchedNotification(Format('UnsortedItems count Items %1.2f per second',[d]));
      3 : SynchedNotification(Format('SortedItems count Items %1.2f per second',[d]));
      4 : SynchedNotification(Format('QuadTree List Items %1.2f per second',[d]));
      5 : SynchedNotification(Format('UnsortedItems List Items %1.2f per second',[d]));
      6 : SynchedNotification(Format('SortedItems List Items %1.2f per second',[d]));
    end;
  end;
  // for two of the tests check, wether the results are identical, to verify that
  // the algorithms are working same.
  k := cntndx;
  if k > cntaltndx then
    k := cntaltndx;
  for j := 0 to k-1 do
  begin
    if cnts[j] <> cntalts[j] then
      SynchedNotification(Format('Test %d, differs: %d %d',[j, cnts[j], cntalts[j]]));
  end;
end;


procedure TPerformanceTestThread.Execute;
begin
  try
    try
      FCurrentTest := 0;
      if Terminated then Exit;
      SynchedNotification('Welcome to the performance tests!');
      {$IFDEF UseEpikTimer}
        SynchedNotification('Using HighPerformance timer!');
      {$ELSE}
        SynchedNotification('Using default system timer!');
      {$ENDIF}
      if Terminated then Exit;
      FCurrentTest := 1;
      ImportCSVIntoTree;
      if Terminated then Exit;
      FCurrentTest := 2;
      SaveTreeToFile;
      if Terminated then Exit;
      FCurrentTest := 3;
      LoadTreeFromFile;
      if Terminated then Exit;
      FCurrentTest := 4;
      SetupAndSortPOILists;
      EstimateAndCountItems;
      if Terminated then Exit;

      SynchedNotification('Performance tests done!');
    except
      on e : Exception do
      begin
        if Terminated then Exit;
        SynchedNotification(Format('Exception in Test. Terminated! Message: "%s"',[e.Message]));
      end;
    end;
    if Terminated then Exit;
  finally
    if not Terminated then
      Terminate;
  end;
end;

constructor TPerformanceTestThread.Create(CreateSuspended: Boolean;
  const AStatusUpdateEvent: TPerformanceTestThreadStatusUpdateEvent;
  const StackSize: SizeUInt);
begin
  inherited Create(True,StackSize);
  FMeasureTimer := TMeasureTimer.Create;
  FStatusUpdateEvent := AStatusUpdateEvent;
  FTestQuadTree := TLazQuadTree.Create(EarthWorld,lqtwmCylindrical,True);
  FTestQuadTree.OnGetItemCoord := @OnTestQuadTreeGetItemCoord;
  FTestQuadTree.OnGetItemRect := @OnTestQuadTreeGetItemRect;
  FTestQuadTree.OnGetItemCaption := @OnTestQuadTreeGetItemCaption;
  FTestQuadTree.OnReadStreamItem := @TTestItemStreamHelper.OnTestQuadTreeReadStreamItem;
  FTestQuadTree.OnWriteStreamItem := @TTestItemStreamHelper.OnTestQuadTreeWriteStreamItem;
  FUnsortedItems := TList.Create;
  FItemsSortedByLat := TList.Create;
  FItemsSortedByLon := TList.Create;
  Suspended := CreateSuspended;
end;

destructor TPerformanceTestThread.Destroy;
begin
  if not Terminated then Terminate;
  if Suspended then
    Suspended := False;
  WaitFor;
  if Assigned(FTestQuadTree) then
    FTestQuadTree.Free;
  if Assigned(FMeasureTimer) then
    FMeasureTimer.Free;
  if (Length(FTempTreeFileName) > 0) and
     FileExists(FTempTreeFileName) then
  begin
    try
      DeleteFile(FTempTreeFileName);
    except
    end;
  end;
  if Assigned(FUnsortedItems) then
    FUnsortedItems.Free;
  if Assigned(FItemsSortedByLat) then
    FItemsSortedByLat.Free;
  if Assigned(FItemsSortedByLon) then
    FItemsSortedByLon.Free;
  inherited Destroy;
end;

end.

