{ A simple Map-Tiler for the Natural Earth map tiles in the QuadTree demo
  Copyright (C) 2023,2024 Ekkehard Domning (www.domis.de)

  License: modified LGPL with linking exception (like RTL, FCL and LCL)

  See the file COPYING.modifiedLGPL.txt, included in the Lazarus distribution,
  for details about the license.

  Version Date        Change
  0.0.1   2024-08-23  First release
}

unit uSimpleNEMapTiler;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, Graphics;
type

  { TSimpleNEMapTiler, a class which deals with the provide Natural Earth Map tiles
     The class can load the tile from file, but also can extract non existing
     tiles from the pixels from the parent tile.
     As deliviered the map set consist out of 341 tiles.
     The tile names are "NE_z_x_y.jpg", where z is the level and
     x, y the coordinates in the level.
     Level 0 contains   1 tile   e.g NE_0_0_0.jpg
     Level 1 contains   4 tiles  e.g NE_1_1_1.jpg
     Level 2 contains  16 tiles  e.g NE_2_3_3.jpg
     Level 3 contains  64 tiles  e.g NE_3_7_7.jpg
     Level 4 contains 256 tiles  e.g NE_4_15_15.jpg
     The tile in level 0 are 360 deg width and 180 deg high
     In level 1 the tiles are 180 deg x 90 deg.
     In level 2 the tiles are 90 deg x 45 deg.
     In level 3 the tiles are 45 deg x 22.5 deg.
     In level 4 the tiles are 22.5 deg x 11.25 deg.
     Every tile has 720 x 360 pixels.
     They are stored in high compressed jpeg-format, so the full set is less than 3MB.
     All tiles are derived from the file found on page
     https://www.naturalearthdata.com/downloads/10m-raster-data/10m-cross-blend-hypso/
     Cross Blended Hypso with Shaded Relief and Water, medium size
     https://www.naturalearthdata.com/http//www.naturalearthdata.com/download/10m/raster/HYP_LR_SR_W.zip
  }
  TSimpleNEMapTiler = class(TObject)
  private
    FTilePath : String;
    FCurrentTile : TBitmap;
    FCurrentTileFileName : String;
    function TileIDFromLevelLonLat(const ALevel : Integer; const ALon, ALat : Double;
                                     out ATileIDX, ATileIDY : Integer) : Boolean;
  public
    { The path of the tiles }
    property TilePath : String read FTilePath write FTilePath;
    { The current loaded tile }
    property CurrentTile : TBitmap read FCurrentTile;
    { The filename of the current tile, even if the non existing file is interpolated! }
    property CurrentTileFileName : String read FCurrentTileFileName;
    { Load the tile for the given level, Longitude and Latitude }
    procedure LoadTile(const ALevel : Integer; const ALon, ALat : Double);
    constructor Create;
    destructor Destroy;override;
  end;

var
  SimpleNEMapTiler : TSimpleNEMapTiler = Nil;
const
  MaxLevel = 13; // At Level 13 an interpolated tile contains only one pixel from the level 4 tile
  TileWidth = 720; // The tiles are 720 px * 360 px
  TileHeight = TileWidth div 2;

implementation

{ TSimpleNEMapTiler }

function TSimpleNEMapTiler.TileIDFromLevelLonLat(const ALevel: Integer;
  const ALon, ALat: Double; out ATileIDX, ATileIDY: Integer): Boolean;
var
  tilewidth : Double;
  i : Integer;
begin
  Result := False;
  if (ALevel > MaxLevel) or
     (ALon < -180.0) or (ALon >= 180.0) or
     (ALat <= -90.0) or (ALat > 90.0) then Exit;
  tilewidth := 360.0;
  for i := 0 to ALevel-1 do
    tilewidth := tilewidth / 2.0;
  ATileIDX := Round((ALon + 180.0) / tilewidth);
  ATileIDY := Round((90.0 - ALat) / (tilewidth / 2.0));
  Result := True;
end;

procedure TSimpleNEMapTiler.LoadTile(const ALevel: Integer; const ALon,
  ALat: Double);
var
  lZ : Integer;
  lTileIDX, lTileIDY: Integer;
  lIDX, lIDY : Integer;
  fnstart, fn, fp : String;
  jpg : TJpegImage = Nil;
  w,h : Integer;
  lDeltaZoom : Integer;
  bmp : TBitmap = Nil;
  rs, rt : TRect;
  xfrac, yfrac : Double;
  px, py : Integer;
begin
  if TileIDFromLevelLonLat(ALevel,ALon,ALat,lTileIDX, lTileIDY) then
  begin
    lIDX := lTileIDX;
    lIDY := lTileIDY;
    lZ := ALevel;
    fn := Format('NE_%d_%d_%d.jpg',[ALevel,lTileIDX, lTileIDY]);
    if fn = FCurrentTileFileName then Exit; // Current tile already loaded
    fp := IncludeTrailingPathDelimiter(FTilePath) + fn;
    if FileExists(fp) then
    begin  // File exists
      jpg := TJpegImage.Create;
      try
        jpg.LoadFromFile(fp);
        FCurrentTile.Assign(jpg);
        FCurrentTileFileName := fn;
      finally
        jpg.Free;
      end;
      Exit; //Done
    end
    else
    begin
      // The following code is taken from the lazMapViewer, but was developed by me
      // for a simmilar purpose, to create an interpolated preview, while the
      // full scale tile is (probably slowly) loaded from the internet.
      //
      // Here we will not download anything, we will interpolate an existing tile for higher levels.
      // If the searched tile does not exists, the content
      // is the part of the containing tile that covers the location of the wanted tile
      // Every decrement of Zoom reduces the tile area by 4 (half of x and y direction)
      // So incrementing Z and dividing X and Y in the Id will lead us to the containing tile
      // The fraction of the division points to the location of the preview
      // e.g 0.5 = right or lower half of the tile, when divided by 2
      fnstart := fn; // Save the original filename we searched for
      lDeltaZoom := 1;
      w := TileWidth;
      h := TileHeight;
      repeat
        w := w shr 1;
        h := h shr 1;
        Dec(lZ);
        lDeltaZoom := lDeltaZoom shl 1;
        xfrac := lTileIDX / lDeltaZoom; // xfrac, yfrac contains the tile number
        yfrac := lTileIDY / lDeltaZoom;
        lIDX := Trunc(xfrac);
        lIDY := Trunc(yfrac);
        fn := Format('NE_%d_%d_%d.jpg',[lZ,lIDX, lIDY]);
        fp := IncludeTrailingPathDelimiter(FTilePath) + fn;
        if FileExists(fp) then
        begin // We found a tile on the disk that contains the preview
          xfrac := xfrac - lIDX; //xfrac and yfrac calculated for the position in the tile from the cache
          yfrac := yfrac - lIDY;
          px := Trunc(xfrac * TileWidth); //x and y are the percentage of the tile width
          py := Trunc(yfrac * TileHeight);
          rs := Rect(px, py, px+w, py+h);
          jpg := TJpegImage.Create;
          try
            jpg.LoadFromFile(fp);
            bmp := TBitmap.Create; // The bitmap wil hold the copied part
            try
              bmp.SetSize(w,h); // it has the small sub tile size
              rt := Rect(0,0,w,h); // The target rectangle is the small size
              bmp.Canvas.CopyRect(rt,jpg.Canvas,rs); // Copy the small rectangle from the jpg to the bitmap
              rt := Rect(0,0,TileWidth,TileHeight);
              FCurrentTile.Canvas.StretchDraw(rt,bmp); // draw the small bitmap streched to the used tile
              FCurrentTileFileName := fnstart; // Assign the original file name
              Exit; // Done
            finally
              bmp.Free;
            end;
          finally
            jpg.Free;
          end;
        end;
      until (w <= 1) or (h <= 1) or (lZ <= 0); // if the size is only one pixel than we are done
    end;
  end;
  // Here we go, if no file found, so we use a gray flat instead.
  FCurrentTile.Canvas.Pen.Style := psClear;
  FCurrentTile.Canvas.Brush.Style := bsSolid;
  FCurrentTile.Canvas.Brush.Color := clDkGray;
  FCurrentTile.Canvas.Rectangle(0,0,TileWidth,TileHeight);
  FCurrentTileFileName := 'maxedout';
end;

constructor TSimpleNEMapTiler.Create;
begin
  FCurrentTile := TBitmap.Create;
  FCurrentTile.SetSize(TileWidth,TileHeight);
end;

destructor TSimpleNEMapTiler.Destroy;
begin
  if Assigned(FCurrentTile) then
    FCurrentTile.Free;
  inherited Destroy;
end;

initialization
  SimpleNEMapTiler := TSimpleNEMapTiler.Create;

finalization
  if Assigned(SimpleNEMapTiler) then
    SimpleNEMapTiler.Free;
end.

