{
ufrmLQTRectOverlapTest.pas

A test-bed for the overlapping of two rectangles with different world and axis settings

License: The same modified LGPL as the Free Pascal RTL
         See the file COPYING.modifiedLGPL for more details

Author: Ekkehard Domning (ekkehard@domning.eu)

}

unit ufrmLQTRectOverlapTest;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ComCtrls, ExtCtrls,
  StdCtrls, LCLType,
  Math {for definition of Infinity},
  lazFileUtils,
  uLazQuadTreeGeometry, uMouseRectShifter;

type
  { TTestCaseRec
      contains the data for each Test Case.
      A description, that goes to the drop down list and
      a world defition and the two rectangles}
  TTestCaseRec = record
    Desc : String;
    WorldModel : TLQTWorldModel;
    RectA : TLQTRect;
    RectB : TLQTRect;
  end;

  { TFrameLQTRectOverlapTest }

  TFrameLQTRectOverlapTest = class(TFrame)
    cbShowOverlappingAreas: TCheckBox;
    cbTestCases: TComboBox;
    cbShowUnion: TCheckBox;
    cbShowUnionWithPoint: TCheckBox;
    edtABottom: TEdit;
    edtALeft: TEdit;
    edtARight: TEdit;
    edtATop: TEdit;
    edtBBottom: TEdit;
    edtBLeft: TEdit;
    edtBRight: TEdit;
    edtBTop: TEdit;
    gbRectOverlapSettings: TGroupBox;
    gbTestCases: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    memoOverlapRect: TMemo;
    pnlSettings: TPanel;
    pnlOutput: TPanel;
    pbRectangleOverlap: TPaintBox;
    pcRectOverlapRectSettings: TPageControl;
    pnlPaintBox: TPanel;
    rgWorldModel: TRadioGroup;
    sbRectangleOverlap: TStatusBar;
    Splitter1: TSplitter;
    tsRectASettings: TTabSheet;
    tsRectBSettings: TTabSheet;
    procedure cbShowOverlappingAreasChange(Sender: TObject);
    procedure cbShowUnionChange(Sender: TObject);
    procedure cbTestCasesChange(Sender: TObject);
    procedure pbRectangleOverlapMouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure pbRectangleOverlapMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure pbRectangleOverlapMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbRectangleOverlapPaint(Sender: TObject);
    procedure pcRectOverlapRectSettingsChange(Sender: TObject);
    procedure pnlPaintBoxChangeBounds(Sender: TObject);
    procedure rgWorldModelClick(Sender: TObject);
    procedure edtRectangleCoordsChange(Sender: TObject);
  private
    FGUIUpdateInProgress : Boolean; // A flag to prevent unwanted loops during internal update
    // The two bitmaps for the cylindrical and toroidal world models
    FCylindricBmp : TBitmap;
    FToroidalBmp : TBitmap;
    FWorldRect : TLQTRect;     // The world rectangle
    FWorldModel : TLQTWorldModel; // The world model
    FRectA : TLQTRect;  // the two rectangles
    FRectB : TLQTRect;
    FRectShifter : TMouseRectShifter; // A HelperClass for the mouse interaction for resizing the rectangles
    FShifterXInverseMode : Boolean; // Two Flags to ease the user interface for the flat and cylindrical world
    FShifterYInverseMode : Boolean;
    FMouseMapCoords : TLQTPoint;
    procedure SetupRectShifter;  // Does the setup of the HelperClass
    procedure RectToGUI; // Fills the 8 Edits with the Rectangle Coords
    procedure RectangleValuesToMemo; // Writes the current State to the memo
    procedure RectangleToStatusBar;
    procedure OverlappingToLines(const ALines : TStrings); // Copies the overlapping rectangle values to the strings
    procedure UnionToLine(const ALines : TStrings); // Copies the union rectangle values to the strings
    procedure SetupTestCases; // Loads the test cases into the drop down list
    function IsEqualToTestCaseIndex : Integer; // returns the index of the test case, if the current setup meets one of them
    procedure AdjustTestCaseDropDownListToSetup; // select the correct entry of the drop down list to the current setting
  public
    property CylindricBmp : TBitmap read FCylindricBmp;
    property ToroidalBmp : TBitmap read FToroidalBmp;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy;override;
  end;

const
  sbndxMouseWorldX   = 1;
  sbndxMouseWorldY   = 3;
  sbndxPointInRect   = 4;
  sbndxOverlapResult = 6;
  sbndxWorldModel    = 8;
  sbndxWorldCoords   = 9;
  sbndxWorldAxisX    = 10;
  sbndxWorldAxisY    = 11;

  // Do not change the constants, there are several of hidden dependencies in the code (sorry).
  WorldQuarterWidth = 180;
  WorldQuarterHeight = 90;
  // Flat world, like an ordinary bitmap, zero is top-left
  FlatWorld : TLQTRect = (
    Left :     0.0;
    Top :      0.0;
    Right :  WorldQuarterWidth*2.0;
    Bottom : WorldQuarterWidth
  );
  // Cylindrical world, like a normal map zero is center
  CylindricalWorld : TLQTRect = (
    Left :   -WorldQuarterWidth;
    Top :     WorldQuarterHeight;
    Right :   WorldQuarterWidth;
    Bottom : -WorldQuarterHeight
  );
  // Torodial world, like a sheet of paper, zero is bottom-left
  TorodialWorld : TLQTRect = (
    Left :     0.0;
    Top :    WorldQuarterHeight*2.0;
    Right :  WorldQuarterWidth*2.0;
    Bottom :   0.0
  );

  CylindricalImgFileName = 'Cylindrical1080x720.jpg';
  ToroidalImgFileName = 'Toroidal1080x720.jpg';

  // Do not change this values too
  WorldPxScale = 2;
  WorldWidthPx  = WorldQuarterWidth*2*WorldPxScale;
  WorldHeightPx = WorldQuarterHeight*2*WorldPxScale;
  WorldOutSidePx  = WorldQuarterWidth;
  WorldTotalWidthPx = WorldOutSidePx + WorldWidthPx + WorldOutSidePx;
  WorldTotalHeightPx = WorldOutSidePx + WorldHeightPx + WorldOutSidePx;


  // Colors may be changed with out side effects
  clSteelBlue  = TColor($C08040); // The color of the blue rectangles displayed in the flat world
  clRectA      = clFuchsia;
  clRectB      = clLime;
  clOverlap    = clYellow;
  clOverlapDim = clOlive;
  clUnion      = clRed;
  clUnionMouse = clMaroon;

  // Width of the drawing pen for the two rectangles, might be changed withoout side effects.
  pen1Width = 5;
  pen2Width = 3;
  penUnionWidth = 2;
  penUnionMouseWidth = 1;


  // The several test cases
  // They show, with increasing complexity, the different overlapping scenarios of the
  // different worlds
  TestCases : array[0..13] of TTestCaseRec = (
    (
    // Case 0: Two non overlapping rectangles.
    //  Value: qroNone
    //  See, that the two rectangles share a common border, but did not overlap.
    //  This reflects the design goal of not including the right and bottom border
    //  to the area of the rectangle, which allows the seamless tiling of an area
    //  with non overlapping rectangles.
    // Caution: Do not change the world model of this case!!
      Desc : 'Flat, None (Border sharing)';
      WorldModel : lqtwmFlat;
      RectA : (Left: 100; Top : 50;
               Right: 200; Bottom : 150);
      RectB : (Left: 200; Top : 50;
               Right: 300; Bottom : 150)
    ),(
    // Case 1: RectB is located fully inside RectA
    // Value: qroInside
    //  In this case all four corners of the rectangle B are
    //  inside the RectA.
    //  To be precise, the result is gathered from the areas, thus
    //  the right and bottom side may shared, having the corresponding
    //  corners outside of the rectangles :-)
      Desc : 'Flat, B inside A';
      WorldModel : lqtwmFlat;
      RectA : (Left: 100; Top : 50;
               Right: 200; Bottom : 150);
      RectB : (Left: 120; Top : 70;
               Right: 180; Bottom : 130)
   ),(
   // Case 2: Both rectangles are eqal
   // Value: qroEqual
     Desc : 'Flat, A equal B';
     WorldModel : lqtwmFlat;
     RectA : (Left: 100; Top : 50;
              Right: 200; Bottom : 150);
     RectB : (Left: 100; Top : 50;
              Right: 200; Bottom : 150)
   ),(
   // Case 3: RectB covers A fully
   // Value: qroFull
   //   This is the opposite situation of Case 1
   //   It is essential to distiguish between this two cases
   //   since in one case the search could be terminated,
   //   while in the other the zoom must be increased
     Desc : 'Flat, B covers A';
     WorldModel : lqtwmFlat;
     RectA : (Left: 120; Top : 50;
              Right: 180; Bottom : 150);
     RectB : (Left: 100; Top : 30;
              Right: 200; Bottom : 170)
    ),(
    // In the remaining cases the two rectangles covering each other only partly
    // Value: qroPartly

    // Case 4: Two corners of B are inside A
      Desc : 'Flat, partly (2 corners)';
      WorldModel : lqtwmFlat;
      RectA : (Left: 100; Top : 50;
               Right: 200; Bottom : 150);
      RectB : (Left: 150; Top : 70;
               Right: 250; Bottom : 130)
    ),(
    // Case 5: One corner of B are inside A
      Desc : 'Flat, partly (1 corner)';
      WorldModel : lqtwmFlat;
      RectA : (Left: 100; Top : 50;
               Right: 200; Bottom : 150);
      RectB : (Left: 150; Top : 70;
               Right: 250; Bottom : 170)
    ),(
    // Case 6: No corner of B are inside A, but still a party overlap!
      Desc : 'Flat, partly (0 corners)';
      WorldModel : lqtwmFlat;
      RectA : (Left: 100; Top : 40;
               Right: 280; Bottom : 140);
      RectB : (Left: 140; Top : 10;
               Right: 240; Bottom : 170)
    ),(
    // Case 7: In the cylindrical world it is possible to have
    //   two separated overlapping areas. This is caused if one
    //   area is looping around an overlaps the other area on the
    //   right and on the left side.
      Desc : 'Cyl., partly (2 overlaps)';
      WorldModel : lqtwmCylindrical;
      RectA : (Left: 10; Top : 53;
               Right: 0; Bottom : -50);
      RectB : (Left: -42; Top : 80;
               Right: 30; Bottom : -80)
    ),(
    // Case 8: Same as Case 7, but one overlapping rectangles is crossing the
    //   right or left border of the world. So three different areas are
    //   marked on the map.
    //   Keep in mind, that there are still only 2 overlapping areas!
      Desc : 'Cyl., partly (2 overlaps, +1)';
      WorldModel : lqtwmCylindrical;
      RectA : (Left: -74; Top : 50;
               Right: -118; Bottom : -50);
      RectB : (Left: 151; Top : 80;
               Right: -42; Bottom : -80)
    ),(
    // Case 9: On the toroidal world the strange behaviour of the cylindrical
    //   world becomes more complicated, since the overlapping could be done
    //   in the vertical axis (Y) too.
    //   So it is possible to create 4 independend overlapping areas.
      Desc : 'Toroid., partly (4 overlaps)';
      WorldModel : lqtwmToroidal;
      RectA : (Left: 60; Top : 140;
               Right: 300; Bottom : 40);
      RectB : (Left: 270; Top : 80;
               Right: 90; Bottom : 100)
    ),(
    // Case 10: This is the toroidal equivalent to the Case 8, the right or left
    //   overlapping areas are wrapping horizontal around the world
    //   forming a total of 6 different marked areas.
    //   Keep in mind, that there are still only 4 overlapping areas!
      Desc : 'Toroid., partly (4 overlaps, +2H)';
      WorldModel : lqtwmToroidal;
      RectA : (Left: 330; Top : 140;
               Right: 300; Bottom : 40);
      RectB : (Left: 270; Top : 80;
               Right: 90; Bottom : 100)
    ),(
    // Case 11: Same as Case 10, but vertical wrapping around.
    //   Keep in mind, that there are still only 4 overlapping areas!
      Desc : 'Toroid., partly (4 overlaps, +2V)';
      WorldModel : lqtwmToroidal;
      RectA : (Left: 60; Top : 20;
               Right: 300; Bottom : 40);
      RectB : (Left: 270; Top : 80;
               Right: 90; Bottom : 100)
    ),(
    // Case 12: A combination of case 10 and 11, forming a total of 9 independend
    //   areas.
    //   Keep in mind, that there are still only 4 overlapping areas!
      Desc : 'Toroid., partly (4 overlaps, +5)';
      WorldModel : lqtwmToroidal;
      RectA : (Left: 330; Top : 20;
               Right: 300; Bottom : 40);
      RectB : (Left: 270; Top : 80;
               Right: 90; Bottom : 100)
    ),(
    // Case 13: Two identical rectangles at the same position
    //
    //
      Desc : 'Toroid., equal (4 overlaps)';
      WorldModel : lqtwmToroidal;
      RectA : (Left: 270; Top : 80;
               Right: 90; Bottom : 100);
      RectB : (Left: 270; Top : 80;
               Right: 90; Bottom : 100);
    )
  );

implementation

{$R *.lfm}

{ TFrameLQTRectOverlapTest }

{ pbRectangleOverlapMouseMove
  The movement of the mouse are traced.
  The coordinates of the mouse are passed to the helper class.
  Depending of the down-state of the mouse button, the area, side or corner of
  a rectangle are shiftet around.
  If the part of the rectangle could be wrapping around, this is performed,
  if not the sides could change the place, giving a natural shift expirience.
}
procedure TFrameLQTRectOverlapTest.pbRectangleOverlapMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
  // The following "PerformXXX" procedures, do the actial adjusting of the
  // side of the active rectangle.
  procedure PerformRightSide(const ANewX : Double; var ARect : TLQTRect);
  begin
    if (FWorldModel <> lqtwmFlat) then
    begin
      ARect.Right := ANewX;
      Exit;
    end;
    if FShifterXInverseMode then
    begin
      if (ANewX > ARect.Right) then
        FShifterXInverseMode := False;
    end
    else
    begin
      if (ANewX < ARect.Left) then
        FShifterXInverseMode := True;
    end;
    if FShifterXInverseMode then
      ARect.Left := ANewX
    else
      ARect.Right := ANewX;
  end;
  procedure PerformLeftSide(const ANewX : Double; var ARect : TLQTRect);
  begin
    if (FWorldModel <> lqtwmFlat) then
    begin
      ARect.Left := ANewX;
      Exit;
    end;
    if FShifterXInverseMode then
    begin
      if (ANewX < ARect.Left) then
        FShifterXInverseMode := False;
    end
    else
    begin
      if (ANewX > ARect.Right) then
        FShifterXInverseMode := True;
    end;
    if FShifterXInverseMode then
      ARect.Right := ANewX
    else
      ARect.Left := ANewX;
  end;
  procedure PerformTopSide(const ANewY : Double; var ARect : TLQTRect);
   begin
     if (FWorldModel = lqtwmToroidal) then
     begin
       ARect.Top := ANewY;
       Exit;
     end;
     if FShifterYInverseMode then
     begin
       if (ANewY < ARect.Top) then
         FShifterYInverseMode := False;
     end
     else
     begin
       if (ANewY > ARect.Bottom) then
         FShifterYInverseMode := True;
     end;
     if FShifterYInverseMode then
       ARect.Bottom := ANewY
     else
       ARect.Top := ANewY;
   end;
   procedure PerformBottomSide(const ANewY : Double; var ARect : TLQTRect);
   begin
     if (FWorldModel = lqtwmToroidal) then
     begin
       ARect.Bottom := ANewY;
       Exit;
     end;
     if FShifterYInverseMode then
     begin
       if (ANewY > ARect.Bottom) then
         FShifterYInverseMode := False;
     end
     else
     begin
       if (ANewY < ARect.Top) then
         FShifterYInverseMode := True;
     end;
     if FShifterYInverseMode then
       ARect.Top := ANewY
     else
       ARect.Bottom := ANewY;
   end;
var
  lscale : Double;
  cofs : Integer;
  w, h : Double;
  lx,ly, lx0, ly0 : Double;
  lFlatWorld, lUsedWorld : TLQTRect;
  lRect0 : TLQTRect;
  lHitItem : TMouseHitItem;
  pt : TLQTPoint;
  s : String;
begin
  // First we need to calculate the current offset and scale of the
  // displayed map, since the PaintBox are changing the size acording
  // to the size of the window
  w := pbRectangleOverlap.Width;
  lscale := (w / WorldTotalWidthPx) * WorldPxScale;
  cofs := Round((w - (w / 1.5))/2.0);
  lx := (X-cofs)/lscale;
  ly := (Y-cofs)/lscale;
  lFlatWorld.Assign(FlatWorld);
  case FWorldModel of
    lqtwmCylindrical :
      lUsedWorld := CylindricalWorld;
    lqtwmToroidal :
      lUsedWorld := TorodialWorld;
  else
    // lqtwmFlat
    lUsedWorld := FlatWorld;
  end;
  // The normated mouse coords are translated into the
  // coords of the current world
  TransformPoint(lx,ly,
                 lFlatWorld,lUsedWorld,
                 lx,ly);
  // Display the currend coords in the statusbar
  NormatePointToWorld(lx,ly,
                      lUsedWorld,FWorldModel,
                      lx0,ly0);
  sbRectangleOverlap.Panels[sbndxMouseWorldX].Text := Format('%1.1f',[lx0]);
  sbRectangleOverlap.Panels[sbndxMouseWorldY].Text := Format('%1.1f',[ly0]);
  FMouseMapCoords.X := lx0;
  FMouseMapCoords.Y := ly0;

  // Pass the parameters to the HelperClass.
  FRectShifter.OnMouseMove(Sender,Shift,X,Y,lHitItem);
  if Assigned(lHitItem) then
  begin
    // lHitItem will be assigned if one item is selected.
    pbRectangleOverlap.Cursor := lHitItem.Cursor;  // Set the cursor to the hit items kind
    if lHitItem.MouseDownFlag then // if the mouse is down
    begin
      // Select the actual rectangle A or B
      if ((lHitItem.Tag and $FF00) = $0100) then
        lRect0.Assign(FRectA)
      else
        lRect0.Assign(FRectB);
      // Transform it to the Flat world
      lRect0.Transform(lUsedWorld, lFlatWorld);
      // The HitItem is in Flat World orientation
      lx := (lHitItem.OrgX+lHitItem.MouseDeltaX-cofs)/lscale;
      ly := (lHitItem.OrgY+lHitItem.MouseDeltaY-cofs)/lscale;
      // Normate to world, including wrap arounds
      NormatePointToWorld(lx,ly,lFlatWorld,FWorldModel,lx,ly);
      // case of the several actions
      case (lHitItem.Tag and $F) of
        tagRectArea :
          begin
            // Special operation for the movement of the full rectangle.
            // Just locate the rectangle at the new position ...
            w := lRect0.Width(lFlatWorld,FWorldModel);
            h := lRect0.Height(lFlatWorld,FWorldModel);
            lRect0.Left := lx;
            lRect0.Top := ly;
            lRect0.Right := lRect0.Left + w;
            lRect0.Bottom := lRect0.Top + h;
            // ... but disallow shifting beyond the borders
            if FWorldModel = lqtwmFlat then
            begin
              // In the flat world model right and left are fixed borders
              if lRect0.Right >= lFlatWorld.Right then
                lRect0.Left := lFlatWorld.Right - w;
            end;
            if (FWorldModel <> lqtwmToroidal) then
            begin
              // In the flat and cylindrical world model top and bottom are fixed borders
              lRect0.Bottom := lRect0.Top + h;
              if lRect0.Bottom >= lFlatWorld.Bottom then
                lRect0.Top := lFlatWorld.Bottom - h;
            end;
          end;
        tagRectLeft :
          begin
            PerformLeftSide(lx,lRect0);
          end;
        tagRectTop :
          begin
            PerformTopSide(ly,lRect0);
          end;
        tagRectRight :
          begin
            PerformRightSide(lx,lRect0);
          end;
        tagRectBottom :
          begin
            PerformBottomSide(ly,lRect0);
          end;
        tagRectTopLeft :
          begin
            PerformTopSide(ly,lRect0);
            PerformLeftSide(lx,lRect0);
          end;
        tagRectTopRight :
          begin
            PerformTopSide(ly,lRect0);
            PerformRightSide(lx,lRect0);
          end;
        tagRectBottomRight :
          begin
            PerformBottomSide(ly,lRect0);
            PerformRightSide(lx,lRect0);
          end;
        tagRectBottomLeft :
          begin
            PerformBottomSide(ly,lRect0);
            PerformLeftSide(lx,lRect0);
          end;
      end;
      // Now return adjust the changed coords and transform to the world coordinates
      lRect0.NormateToWorld(lFlatWorld, FWorldModel);
      lRect0.Transform(lFlatWorld, lUsedWorld);
      // And return into the active rectangle A or B
      if ((lHitItem.Tag and $FF00) = $0100) then
        FRectA.Assign(lRect0)
      else
        FRectB.Assign(lRect0);
      RectToGUI; // Write the changed values into the edit fields
      AdjustTestCaseDropDownListToSetup; // set the drop down list, if we hit a special case
      pbRectangleOverlap.Invalidate; // redraw the map
      pbRectangleOverlap.Repaint;
    end;
  end
  else
  begin
    pbRectangleOverlap.Cursor := crDefault;  // no hit, set the default cursor
  end;
  if cbShowUnionWithPoint.Checked then
  begin
    pbRectangleOverlap.Invalidate; // redraw the map
    pbRectangleOverlap.Repaint;
  end;

  pt.X := lx;
  pt.Y := ly;
  s := '';
  if PointInRectangle(FRectA, pt, lUsedWorld, FWorldModel) then
    s := 'A';
  if PointInRectangle(FRectB, pt, lUsedWorld, FWorldModel) then
    s := s + 'B';
  sbRectangleOverlap.Panels[sbndxPointInRect].Text := s;
end;

procedure TFrameLQTRectOverlapTest.pbRectangleOverlapMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FRectShifter.OnMouseUp(Sender,Button,Shift,X,Y); // The shifting ends
  SetupRectShifter;  // Setup the HelperClass for the current setting
end;

procedure TFrameLQTRectOverlapTest.pbRectangleOverlapMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  // On MouseDown the helper stores the last position
  FRectShifter.OnMouseDown(Sender,Button,Shift,X,Y);
  // Reset the inverters
  FShifterXInverseMode := False;
  FShifterYInverseMode := False;
end;

procedure TFrameLQTRectOverlapTest.cbShowOverlappingAreasChange(Sender: TObject);
begin
  // If the user changes the visibility of the overlapping areas, simply redraw the map
  pbRectangleOverlap.Invalidate;
  pbRectangleOverlap.RePaint;
end;

procedure TFrameLQTRectOverlapTest.cbShowUnionChange(Sender: TObject);
begin
  // If the user changes the visibility of the union, simply redraw the map
  pbRectangleOverlap.Invalidate;
  pbRectangleOverlap.RePaint;
end;

{ cbTestCasesChange loads a new testcase into the data, but only if it differs from the
  current setting to avoid flickering loops }
procedure TFrameLQTRectOverlapTest.cbTestCasesChange(Sender: TObject);
var
  ndx, ndx0 : Integer;
begin
  ndx := cbTestCases.ItemIndex;
  if ndx >= 0 then
  begin
    // Check if the current setup is equal to the selected one from the DropDownList
    ndx0 := IsEqualToTestCaseIndex;
    if ndx = ndx0 then Exit; // Is equal, so nothing to do

    FWorldModel := TestCases[ndx].WorldModel;
    case FWorldModel of
      lqtwmCylindrical :
        begin
          rgWorldModel.ItemIndex := 1;
          FWorldRect.Assign(CylindricalWorld);
        end;
      lqtwmToroidal :
        begin
          rgWorldModel.ItemIndex := 2;
          FWorldRect.Assign(TorodialWorld);
        end;
     else
       rgWorldModel.ItemIndex := 0;
       FWorldRect.Assign(FlatWorld);
     end;
    FRectA.Assign(TestCases[ndx].RectA);
    FRectB.Assign(TestCases[ndx].RectB);

    RectToGUI;
    AdjustTestCaseDropDownListToSetup;
    SetupRectShifter;
  end;
end;

{ pbRectangleOverlapPaint is the central and long section of the display of this
  demonstration. It performs all needed calculations and makes the most usage of the
  library unit}
procedure TFrameLQTRectOverlapTest.pbRectangleOverlapPaint(Sender: TObject);
  // PaintRectangle will draw one of the two rectangles A or B
  // It takes care if the rectangle is wrapping around.
  // For a better understanding the "out of the map" areas are duplicated with
  // thin lines to give a better impression what the loop around does.
  procedure PaintRectangle(const ALQTRect : TLQTRect;
                           const ALQTWorldRect : TLQTRect;
                           const ALQTWorldModel : TLQTWorldModel;
                           const AColor : TColor;
                           const AScale : Double;
                           const AOfs : Integer;
                           const ALineWidth : Integer
                           );
  var
    lLQTRect,lLQTRect0 : TLQTRect;
    lNormalizedRectangles : TLQTRectArray; // The normalized Rectangles
    lNormalizedWorldRect : TLQTRect;
    i,j,k : Integer;
    r,r0 : TRect;
    lCanvasWorld : TLQTRect;
    ww, wh : Integer;
  begin
    pbRectangleOverlap.Canvas.Pen.Width := ALineWidth;

    // Prepare the canvas
    pbRectangleOverlap.Canvas.Pen.Style := psSolid;
    pbRectangleOverlap.Canvas.Pen.Color := AColor;
    pbRectangleOverlap.Canvas.Brush.Style := bsClear;
    // The flat world is our canvas world
    lCanvasWorld.Assign(FlatWorld);
    try
      // Calculate the actual width and height in pixel
      ww := Round(lCanvasWorld.Width * AScale);
      wh := Round(lCanvasWorld.Height * AScale);
      lLQTRect.Assign(ALQTRect);
      lLQTRect.NormateToWorld(ALQTWorldRect,ALQTWorldModel);
      // NormalizeRectangle will split wrapping around rectangles
      // into several ones, laying within the map.
      // The lNormalizedWorldRect contains an upright version
      // of the used world, so that X increases to the right and Y downwards.
      // lNormalizedRectangles contains the rectangles (1, 2 or 4)
      NormalizeRectangle(lLQTRect,
                         ALQTWorldRect, ALQTWorldModel,
                         lNormalizedRectangles,
                         lNormalizedWorldRect);
      for i := 0 to High(lNormalizedRectangles) do
      begin
        // Now we transform the rectangles to our drawing surface
        lLQTRect0.Assign(lNormalizedRectangles[i]);
        lLQTRect0.Transform(lNormalizedWorldRect,lCanvasWorld);
        // do the stretch and ofset
        r.Left := Round((lLQTRect0.Left * AScale) + AOfs);
        r.Top := Round((lLQTRect0.Top * AScale) + AOfs);
        r.Right := Round((lLQTRect0.Right * AScale) + AOfs);
        r.Bottom := Round((lLQTRect0.Bottom * AScale) + AOfs);
        // nd finally paint to the canvas
        pbRectangleOverlap.Canvas.Rectangle(r);
      end;

      // If we are on a cylindrical ar toroidal world, we paint a bit of the out side
      if ALQTWorldModel <> lqtwmFlat then
      begin
        // This is fairly the same a above, but avoids the "normalization".
        // Instead, if the left side of a rectangle is bigger than the right side
        // the value is decreased by the width.
        // The top and bottom are adjusted in the same manner.
        pbRectangleOverlap.Canvas.Pen.Width := 1;  // Make the pen thinner
        lLQTRect0.Assign(lLQTRect);
        lLQTRect0.Transform(ALQTWorldRect,lCanvasWorld);
        r.Left := Round((lLQTRect0.Left*AScale) + AOfs);
        r.Top := Round((lLQTRect0.Top*AScale) + AOfs);
        r.Right := Round((lLQTRect0.Right*AScale) + AOfs);
        r.Bottom := Round((lLQTRect0.Bottom*AScale) + AOfs);
        if r.Left > r.Right then
          r.Left := r.Left - ww;
        if r.Bottom < r.Top then
         r.Bottom := r.Bottom + wh;
        // Duplicate the rectangles to the dimmed doubled picture areas (only on cylindrical or toroidal worlds)
        // Since the color of the rectangle is the same, we simply ignore what we painted
        // so far, it will not harm if we overwrite existing lines.
        for j := -1 to 1 do  // Y = (above, center, below)
        begin
          for k := -1 to 1 do // X = (left, center, right)
          begin
            if (ALQTWorldModel = lqtwmCylindrical) and (j <> 0) then Continue; // In  the cylindrical world only right and left needed
            r0 := r; // Take the last painted rectangle
            // Simply subtract or add the width and height of the world to the coords of the rectangle.
            r0.Right := r0.Right + (ww*k);
            r0.Left := r0.Left + (ww*k);
            r0.Top := r0.Top + (wh*j);
            r0.Bottom := r0.Bottom + (wh*j);
            pbRectangleOverlap.Canvas.Rectangle(r0);  // Paint the thinner rectangle
          end;
        end;
      end;
    except
    end;
  end;

  // PaintOverlapping does the painting of the overlapping rectangles
  // The method is fairly the same as for the Rectangle A and B
  // Instead the overlapping rectangles are calculated inside the
  // procedure
  procedure PaintOverlapping(const ALQTWorldRect : TLQTRect;
                             const ALQTWorldModel : TLQTWorldModel;
                             const AScale : Double;
                             const AOfs : Integer);
  var
    ovl : TLQTRectOverlapKind;
    CRectangles: TLQTRectArray;
    i,j,k,l : Integer;
    lLQTRect, lLQTRect0 : TLQTRect;
    r,r0 : TRect;
    lCanvasWorld : TLQTRect;
    lNormalizedRectangles : TLQTRectArray; // The normalized Rectangles
    lNormalizedWorldRect : TLQTRect;
    ww, wh : Integer;
  begin
    // Here we calculate the overlapping rectangles. As described above, this will
    // be alyways 0, 1, 2 or 4
    ovl := RectanglesOverlapEx(FRectA, FRectB, FWorldRect, FWorldModel, CRectangles);
    if ovl = qroNone then Exit;
    lCanvasWorld.Assign(FlatWorld);
    // Calculate the actual width and height in pixel
    ww := Round(lCanvasWorld.Width * AScale);
    wh := Round(lCanvasWorld.Height * AScale);
    pbRectangleOverlap.Canvas.Brush.Style := bsSolid;
    // Now we loop to all of the overlapping rectangles (1, 2 or 4)
    for i := 0 to High(CRectangles) do
    begin
      // Since the rectangle could wrap around, we have to "normalize" the overlapping
      // rectangle
      lLQTRect.Assign(CRectangles[i]);
      NormalizeRectangle(lLQTRect,
                         ALQTWorldRect,ALQTWorldModel,
                         lNormalizedRectangles,
                         lNormalizedWorldRect);
      for j := 0 to High(lNormalizedRectangles) do
      begin
        // For all the normalizes rectangles (1, 2 or 4)
        // We calculate the area where on the canvas the rectangle
        // has to be painted.
        // But we will not paint yet!
        lLQTRect0.Assign(lNormalizedRectangles[j]);
        lLQTRect0.Transform(lNormalizedWorldRect,lCanvasWorld);
        r.Left := Round((lLQTRect0.Left*AScale) + AOfs);
        r.Top := Round((lLQTRect0.Top*AScale) + AOfs);
        r.Right := Round((lLQTRect0.Right*AScale) + AOfs);
        r.Bottom := Round((lLQTRect0.Bottom*AScale) + AOfs);

        // If we are on a cylindrical ar toroidal world, we paint a bit of the out side
        if ALQTWorldModel <> lqtwmFlat then
        begin
          pbRectangleOverlap.Canvas.Pen.Style := psSolid;
          pbRectangleOverlap.Canvas.Pen.Width := 1;
          pbRectangleOverlap.Canvas.Pen.Color := clOverlapDim;
          pbRectangleOverlap.Canvas.Brush.Color := clOverlapDim;
          // Duplicate the rectangles to the dimmed doubled picture areas (only on cylindrical or toroidal worlds)
          for l := -1 to 1 do  // Y = (above, center, below)
          begin
            for k := -1 to 1 do // X = (left, center, right)
            begin
              if (ALQTWorldModel = lqtwmCylindrical) and (l <> 0) then Continue; // In  the cylindrical world only right and left needed
              r0 := r; // Take the last painted rectangle
              // Simply subtract or add the width and height of the world to the coords of the rectangle.
              r0.Right := r0.Right + (ww*k);
              r0.Left := r0.Left + (ww*k);
              r0.Top := r0.Top + (wh*l);
              r0.Bottom := r0.Bottom + (wh*l);
              // Make the pen thinner
              pbRectangleOverlap.Canvas.Rectangle(r0);
            end;
          end;
        end;
        // So the reason for the late painting of the original rectangle
        // is that we must avoid the overwriting with the different color.
        pbRectangleOverlap.Canvas.Pen.Style := psClear;
        pbRectangleOverlap.Canvas.Brush.Color := clOverlap;
        pbRectangleOverlap.Canvas.Rectangle(r);
      end;
    end;
  end;

  procedure PerformUnion(const AScale : Double;
                         const AOfs : Integer);
  var
    lRectUnion : TLQTRect;
  begin
    lRectUnion.Assign(FRectA);
    // Draw the different worlds
    case FWorldModel of
      lqtwmCylindrical : // Cylindrical
        begin
          lRectUnion.Union(FRectB,CylindricalWorld,lqtwmCylindrical);
          PaintRectangle(lRectUnion,CylindricalWorld,lqtwmCylindrical,clUnion,AScale,AOfs,penUnionWidth);
        end;
      lqtwmToroidal : // Toroidal
        begin
          lRectUnion.Union(FRectB,TorodialWorld,lqtwmToroidal);
          PaintRectangle(lRectUnion,TorodialWorld,lqtwmToroidal,clUnion,AScale,AOfs,penUnionWidth);
        end;
    else
      // Flat
      lRectUnion.Union(FRectB,FlatWorld,lqtwmFlat);
      PaintRectangle(lRectUnion,FlatWorld,lqtwmFlat,clUnion,AScale,AOfs,penUnionWidth);
    end;
  end;

  procedure PerformUnionWithPoint(const AScale : Double;
                                  const AOfs : Integer;
                                  const APoint : TLQTPoint);
  var
    lRectUnion : TLQTRect;
  begin
    if pcRectOverlapRectSettings.TabIndex = 0 then
      lRectUnion.Assign(FRectA)
    else
      lRectUnion.Assign(FRectB);
    // Draw the different worlds
    case FWorldModel of
      lqtwmCylindrical : // Cylindrical
        begin
          lRectUnion.Union(APoint,CylindricalWorld,lqtwmCylindrical);
          PaintRectangle(lRectUnion,CylindricalWorld,lqtwmCylindrical,clUnionMouse,AScale,AOfs,penUnionMouseWidth);
        end;
      lqtwmToroidal : // Toroidal
        begin
          lRectUnion.Union(APoint,TorodialWorld,lqtwmToroidal);
          PaintRectangle(lRectUnion,TorodialWorld,lqtwmToroidal,clUnionMouse,AScale,AOfs,penUnionMouseWidth);
        end;
    else
      // Flat
      lRectUnion.Union(APoint,FlatWorld,lqtwmFlat);
      PaintRectangle(lRectUnion,FlatWorld,lqtwmFlat,clUnionMouse,AScale,AOfs,penUnionMouseWidth);
    end;
  end;


var
  w,h : Integer;
  r : TRect;
  i,j,k : Integer;
  x,y : Integer;
  cbw : Integer;
  lscale : Double;
  cofs : Integer;
  clRect1, clRect2 : TColor;
  lRect1, lRect2 : TLQTRect;
begin
  // We calculate the scale and offset of map on the canvas
  // We assume a ratio between height and width of 1:2
  w := pbRectangleOverlap.Width;
  h := pbRectangleOverlap.Height;
  r.Left := 0;
  r.Top := 0;
  r.Right := w;
  r.Bottom := h;
  lscale := (w / WorldTotalWidthPx) * WorldPxScale;
  cbw := Round(w / 1.5 / 12.0);
  cofs := Round((w - (w / 1.5))/2.0);
  // Depending which rectangle (A, B) has been selected (Button of the pagecontrol)
  // store the setting for RectA and B into local variable 1 and 2,
  // this puts the selected rectangle in front and in thicker lines
  if pcRectOverlapRectSettings.TabIndex = 0 then
  begin
    lRect1.Assign(FRectA);
    lRect2.Assign(FRectB);
    clRect1 := clRectA;
    clRect2 := clRectB;
  end
  else
  begin
    lRect1.Assign(FRectB);
    lRect2.Assign(FRectA);
    clRect1 := clRectB;
    clRect2 := clRectA;
  end;
  // Draw the different worlds
  case FWorldModel of
    lqtwmCylindrical : // Cylindrical
      begin
        pbRectangleOverlap.Canvas.StretchDraw(r,FCylindricBmp); // The background map
        // The overlapping rectangle
        if cbShowOverlappingAreas.Checked then
          PaintOverlapping(CylindricalWorld,lqtwmCylindrical,lscale,cofs);
        // The Union Rectangle
        if cbShowUnion.Checked then
          PerformUnion(lscale,cofs);
        // Union with Mouse
        if cbShowUnionWithPoint.Checked then
          PerformUnionWithPoint(lscale,cofs,FMouseMapCoords);

        // The two rectangles, first the one, that should be stayed in the background
        PaintRectangle(lRect2,CylindricalWorld,lqtwmCylindrical,clRect2,lscale,cofs,pen2Width);
        // then the selected one in front of everything
        PaintRectangle(lRect1,CylindricalWorld,lqtwmCylindrical,clRect1,lscale,cofs,pen1Width);
      end;
    lqtwmToroidal : // Toroidal
      begin
        pbRectangleOverlap.Canvas.StretchDraw(r,FToroidalBmp);
        if cbShowOverlappingAreas.Checked then
          PaintOverlapping(TorodialWorld,lqtwmToroidal,lscale,cofs);
        if cbShowUnion.Checked then
          PerformUnion(lscale,cofs);
        // Union with Mouse
        if cbShowUnionWithPoint.Checked then
          PerformUnionWithPoint(lscale,cofs,FMouseMapCoords);

        PaintRectangle(lRect2,TorodialWorld,lqtwmToroidal,clRect2,lscale,cofs,pen2Width);
        PaintRectangle(lRect1,TorodialWorld,lqtwmToroidal,clRect1,lscale,cofs,pen1Width);
      end;
  else
    // Flat
    // Instead of a predefined bitmap, we draw a checker board for the flat world
    pbRectangleOverlap.Canvas.Pen.Width := 1;
    pbRectangleOverlap.Canvas.Pen.Style := psSolid;
    pbRectangleOverlap.Canvas.Pen.Color := clDkGray;
    pbRectangleOverlap.Canvas.Brush.Style := bsSolid;
    pbRectangleOverlap.Canvas.Brush.Color := clBlack;
    pbRectangleOverlap.Canvas.Rectangle(0,0,w,h);
    for k := 0 to 1 do // For the two colors of the checker board
    begin
      if k = 0 then
        pbRectangleOverlap.Canvas.Brush.Color := clSteelBlue
      else
        pbRectangleOverlap.Canvas.Brush.Color := clDkGray;
      for i := 0 to 5 do // Iterate through the 6 fields of same colored fields in a row
      begin
        for j := 0 to 5 do  // Iterate through the 6 rows of the board
        begin
          x := (i * 2) + ((j+k) mod 2); // the X of the board
          y := j; // The Y
          x := (x * cbw) + cofs;
          y := (y * cbw) + cofs;
          pbRectangleOverlap.Canvas.Rectangle(x,y,x+cbw,y+cbw);
        end;
      end;
    end;
    // Paint the overlapping parts of the rectangles if desired
    if cbShowOverlappingAreas.Checked then
      PaintOverlapping(FlatWorld,lqtwmFlat,lscale,cofs);
    // The Union Rectangle
    if cbShowUnion.Checked then
      PerformUnion(lscale,cofs);
    // Union with Mouse
    if cbShowUnionWithPoint.Checked then
      PerformUnionWithPoint(lscale,cofs,FMouseMapCoords);
    // Paint the two rectangles, the not selected first, the selected last to bring it in the front
    PaintRectangle(lRect2,FlatWorld,lqtwmFlat,clRect2,lscale,cofs,pen2Width);
    PaintRectangle(lRect1,FlatWorld,lqtwmFlat,clRect1,lscale,cofs,pen1Width);
  end;
end;

procedure TFrameLQTRectOverlapTest.pcRectOverlapRectSettingsChange(
  Sender: TObject);
begin
  // The selection of the rectangle changed
  // Setup the HelperClass (to have a higher Z-Order for the selected rectangle).
  SetupRectShifter;
  // and repaint
  pbRectangleOverlap.Invalidate;
  pbRectangleOverlap.RePaint;
end;

procedure TFrameLQTRectOverlapTest.pnlPaintBoxChangeBounds(Sender: TObject);
var
  w, h : Integer;
  ratio : Double;
  pbh, pbw : Integer;
  pbc : Integer;
begin
  // The panel on which the paintbox is located changed its size
  // now we resize the paintbox, so that it fits in a 1:2 scale on this
  // panel.
  w := pnlPaintBox.Width;
  h := pnlPaintBox.Height;
  if (w < 2) or (h < 1) then
  begin
    w := 2;
    h := 1;
  end;
  ratio := w / h;
  if ratio > 1.5 then
  begin // wider than our background
    pbh := h;
    pbw := Round(h * 1.5);
    pbc := (w - pbw) div 2;
    pbRectangleOverlap.SetBounds(pbc,0,pbw,pbh);
  end
  else
  begin // taller than our background
    pbw := w;
    pbh := Round(w / 1.5);
    pbc := (h - pbh) div 2;
    pbRectangleOverlap.SetBounds(0,pbc,pbw,pbh);
  end;
  // Update the helper class for the mouse shifting
  SetupRectShifter;
end;

procedure TFrameLQTRectOverlapTest.rgWorldModelClick(Sender: TObject);
var
  lLastWorld: TLQTRect;
begin
  // The user change the world model
  // this causes a lot of changes
  lLastWorld := FWorldRect; // save the current world
  // Setup the new world
  case rgWorldModel.ItemIndex of
    1 : begin
          FWorldRect.Assign(CylindricalWorld);
          FWorldModel := lqtwmCylindrical;
        end;
    2 : begin
          FWorldRect.Assign(TorodialWorld);
          FWorldModel := lqtwmToroidal;
        end;
  else
    FWorldRect.Assign(FlatWorld);
    FWorldModel := lqtwmFlat;
  end;
  // Transform and normate the Rectangle coords from the old to the new world
  // This causes in general rectangles to be kept on the same position on the
  // display. If reducing the worlds complexity, the rectangle will change there
  // position and shape.
  FRectA.Transform(lLastWorld,FWorldRect);
  FRectA.NormateToWorld(FWorldRect,FWorldModel);
  FRectB.Transform(lLastWorld,FWorldRect);
  FRectB.NormateToWorld(FWorldRect,FWorldModel);
  // Write the Values back to the edit fields
  RectToGUI;
  AdjustTestCaseDropDownListToSetup; // check if we have a predefined test case
  SetupRectShifter; // setup the Helper Class
  // Repaint the map
  pbRectangleOverlap.Invalidate;
  pbRectangleOverlap.RePaint;
end;

{ edtRectangleCoordsChange is called if the user change the text in one of the
    coords edit fields.
    If the value could not be converted, the field becomes marked with red color.}
procedure TFrameLQTRectOverlapTest.edtRectangleCoordsChange(Sender: TObject);
var
  edt : TEdit;
  s : String;
  fs : TFormatSettings;
  d : Double;
begin
  if not (Sender is TEdit) then Exit;
  edt := TEdit(Sender);
  s := Trim(edt.Text);
  while Pos(',',s) > 0 do
    s[Pos(',',s)] := '.';
  fs.DecimalSeparator:='.';
  d := StrToFloatDef(s,Infinity,fs);
  if d = Infinity then
  begin
    edt.Color := clRed;
    Exit;
  end;
  if edt.Color <> clDefault then
    edt.Color := clDefault;
  if edt.Name = 'edtALeft' then
    FRectA.Left := d
  else if edt.Name = 'edtATop' then
    FRectA.Top := d
  else if edt.Name = 'edtARight' then
    FRectA.Right := d
  else if edt.Name = 'edtABottom' then
    FRectA.Bottom := d
  else if edt.Name = 'edtBLeft' then
    FRectB.Left := d
  else if edt.Name = 'edtBTop' then
    FRectB.Top := d
  else if edt.Name = 'edtBRight' then
    FRectB.Right := d
  else if edt.Name = 'edtBBottom' then
    FRectB.Bottom := d
  else
    Exit;
  // Reflect the change of the value to the rest of the GUI
  AdjustTestCaseDropDownListToSetup;
  RectangleValuesToMemo;
  RectangleToStatusBar;
  pbRectangleOverlap.Invalidate;
  pbRectangleOverlap.Repaint;
end;

{ SetupRectShifter will fill the Helper Class for the mouse movement with the
    values, so that the mouse pointer will show sensitive items }
procedure TFrameLQTRectOverlapTest.SetupRectShifter;
const
  InflateSensitiveArea = 2; // this increases the pixels of the sensitive items

  // AddRectangle does exactly this, placing all elements for one of the
  //   rectangkles A or B into the helper class
  //   The method is fairly the same as the painting does.
  //   But instead of "normalization" the rectangles are duplicated around
  //   giving most of the work to windows instead of doing the cropping by our self
  procedure AddRectangle(const ALQTRect : TLQTRect;
                           const ALQTWorldRect : TLQTRect;
                           const ALQTWorldModel : TLQTWorldModel;
                           const AScale : Double;
                           const AOfs : Integer;
                           const ATargetRectNdx : Integer = 0
                           );
  var
    lLQTRect : TLQTRect;

    j,k : Integer;
    r,r0 : TRect;
    lCanvasWorld : TLQTRect;
    z : Integer;
    ww,wh : Integer;
  begin
    // The flat world is our canvas world
    lCanvasWorld.Assign(FlatWorld);
    try
      // Calculate the actual width and height in pixel
      ww := Round(lCanvasWorld.Width * AScale);
      wh := Round(lCanvasWorld.Height * AScale);
      lLQTRect.Assign(ALQTRect);
      lLQTRect.NormateToWorld(ALQTWorldRect,ALQTWorldModel);
      lLQTRect.Transform(ALQTWorldRect,lCanvasWorld);
      r.Left := Round((lLQTRect.Left*AScale) + AOfs);
      r.Top := Round((lLQTRect.Top*AScale) + AOfs);
      r.Right := Round((lLQTRect.Right*AScale) + AOfs);
      r.Bottom := Round((lLQTRect.Bottom*AScale) + AOfs);
      if r.Left > r.Right then
        r.Left := r.Left - ww;
      if r.Bottom < r.Top then
       r.Bottom := r.Bottom + wh;

      if (ATargetRectNdx <> 0) then
      begin
        // The Z order is calculated by the active page index and the parameter for the saving.
        // If ActivePageIndex = 0 the Rectangle A is in front of B
        // If ActivePageIndex = 1 the Rectangle B is in front of A
        // (((0*2)-1) * 1) + 5 = -1 + 5 = 4
        // (((0*2)-1) * 2) + 5 = -2 + 5 = 3
        // (((1*2)-1) * 1) + 5 =  1 + 5 = 6
        // (((1*2)-1) * 2) + 5 =  2 + 5 = 7
        z := ((((pcRectOverlapRectSettings.ActivePageIndex*2)-1) * ATargetRectNdx) + 5) * 10;
        if ATargetRectNdx = 1 then
          FRectShifter.AddRectangle(r,InflateSensitiveArea,z,$0100)
        else
          FRectShifter.AddRectangle(r,InflateSensitiveArea,z,$0200);
        if ALQTWorldModel = lqtwmFlat then Exit;

        // Duplicate the rectangles to the dimmed doubled picture areas (only on cylindrical or toroidal worlds)
        for j := -1 to 1 do  // Y = (above, center, below)
        begin
          for k := -1 to 1 do // X = (left, center, right)
          begin
            if (j = 0) and (k = 0) then Continue; // The center is allready painted
            if (ALQTWorldModel = lqtwmCylindrical) and (j <> 0) then Continue; // In  the cylindrical world only right and left needed
            r0 := r; // Take the last painted rectangle
            // Simply subtract or add the width and height of the world to the coords of the rectangle.
            r0.Right := r0.Right + (ww*k);
            r0.Left := r0.Left + (ww*k);
            r0.Top := r0.Top + (wh*j);
            r0.Bottom := r0.Bottom + (wh*j);
            if ATargetRectNdx = 1 then
              FRectShifter.AddRectangle(r0,InflateSensitiveArea,z,$0100)
            else
              FRectShifter.AddRectangle(r0,InflateSensitiveArea,z,$0200);
          end;
        end;
      end;

    except
    end;
  end;

var
  w : Integer;
  lscale : Double;
  cofs : Integer;
begin
  FRectShifter.Clear;

  w := pbRectangleOverlap.Width;
  // h := pbRectangleOverlap.Height; // H is nit needed, since the ratio is fixed to 1:2
  lscale := (w / WorldTotalWidthPx) * WorldPxScale;
  cofs := Round((w - (w / 1.5))/2.0);

  case rgWorldModel.ItemIndex of
    1 : // Cylindrical
      begin
        AddRectangle(FRectA,CylindricalWorld,lqtwmCylindrical,lscale,cofs,1);
        AddRectangle(FRectB,CylindricalWorld,lqtwmCylindrical,lscale,cofs,2);
      end;
    2 : // Toroidal
      begin
        AddRectangle(FRectA,TorodialWorld,lqtwmToroidal,lscale,cofs,1);
        AddRectangle(FRectB,TorodialWorld,lqtwmToroidal,lscale,cofs,2);
      end;
  else
    // Flat
    AddRectangle(FRectA,FlatWorld,lqtwmFlat,lscale,cofs, 1);
    AddRectangle(FRectB,FlatWorld,lqtwmFlat,lscale,cofs, 2);
  end;
end;

{ RectToGUI fills the Edit fields with the values from the data fields}
procedure TFrameLQTRectOverlapTest.RectToGUI;
  procedure DoubleToEdit(const AValue : Double; const AEdit : TEdit);
  var
    s : String;
  begin
    s := Format('%1.2f',[AValue]);
    if AEdit.Text <> s then
      AEdit.Text := s;
  end;

begin
  FGUIUpdateInProgress := True;
  try
    DoubleToEdit(FRectA.Left, edtALeft);
    DoubleToEdit(FRectA.Top, edtATop);
    DoubleToEdit(FRectA.Right, edtARight);
    DoubleToEdit(FRectA.Bottom, edtABottom);
    DoubleToEdit(FRectB.Left, edtBLeft);
    DoubleToEdit(FRectB.Top, edtBTop);
    DoubleToEdit(FRectB.Right, edtBRight);
    DoubleToEdit(FRectB.Bottom, edtBBottom);
    RectangleValuesToMemo;
    RectangleToStatusBar;
  finally
    FGUIUpdateInProgress := False;
  end;
end;

{RectangleValuesToMemo fill some values together with the overlapping status an the
  coords of the overlapping rectangles to the memo. This allowes the user to
  copy those information }
procedure TFrameLQTRectOverlapTest.RectangleValuesToMemo;
begin
  memoOverlapRect.Lines.BeginUpdate;
  try
    memoOverlapRect.Clear;
    memoOverlapRect.Lines.Add('World: (model: '+LQTWorldModelString[FWorldModel]+ '): '+
                              FWorldRect.CoordsAsString('L %1.2f, T %1.2f, R %1.2f, B %1.2f'));
    memoOverlapRect.Lines.Add(FRectA.CoordsAsString('A: L %1.2f, T %1.2f, R %1.2f, B %1.2f'));
    memoOverlapRect.Lines.Add(FRectB.CoordsAsString('B: L %1.2f, T %1.2f, R %1.2f, B %1.2f'));
    // Calculate the overlapping and drop the data into the memo
    OverlappingToLines(memoOverlapRect.Lines);
    UnionToLine(memoOverlapRect.Lines);
    // Plave the input cursor at the beginning of the text
    memoOverlapRect.SelStart := 0;
    memoOverlapRect.SelLength := 0;
  finally
    memoOverlapRect.Lines.EndUpdate;
  end;
end;

procedure TFrameLQTRectOverlapTest.RectangleToStatusBar;
var
  ovl : TLQTRectOverlapKind;
  CRectangles: TLQTRectArray;
  s : String;
begin
  ovl := RectanglesOverlapEx(FRectA, FRectB, FWorldRect, FWorldModel, CRectangles);
  sbRectangleOverlap.Panels[sbndxOverlapResult].Text := RectOverlapKindString[ovl];
  sbRectangleOverlap.Panels[sbndxWorldModel].Text := LQTWorldModelString[FWorldModel];
  sbRectangleOverlap.Panels[sbndxWorldCoords].Text := FWorldRect.CoordsAsString('L %1.2f, T %1.2f, R %1.2f, B %1.2f');
  if FWorldRect.WorldXAxisLeftPlus then
    s := '←'
  else
    s := '→';
  sbRectangleOverlap.Panels[sbndxWorldAxisX].Text := s;
  if FWorldRect.WorldYAxisUpPlus then
    s := '↑'
  else
    s := '↓';
  sbRectangleOverlap.Panels[sbndxWorldAxisY].Text := s;
end;

{ OverlappingToLines calculates the overlapping and drops the result into the lines }
procedure TFrameLQTRectOverlapTest.OverlappingToLines(const ALines: TStrings);
var
  ovl : TLQTRectOverlapKind;
  CRectangles: TLQTRectArray;
  lLQTRect : TLQTRect;
  i : Integer;
begin
  ovl := RectanglesOverlapEx(FRectA, FRectB, FWorldRect, FWorldModel, CRectangles);
  ALines.Add('Overlapping: '+RectOverlapKindString[ovl]);
  if ovl = qroNone then Exit;
  ALines.Add('Overlapping rect count: '+IntToStr(Length(CRectangles)));
  for i := 0 to High(CRectangles) do
  begin
    lLQTRect.Assign(CRectangles[i]);
    ALines.Add('  R'+IntToStr(i)+': '+lLQTRect.CoordsAsString('L %1.2f, T %1.2f, R %1.2f, B %1.2f'));
  end;
end;

procedure TFrameLQTRectOverlapTest.UnionToLine(const ALines: TStrings);
var
  lRectUnion : TLQTRect;
begin
  lRectUnion.Assign(FRectA);
  // Draw the different worlds
  case FWorldModel of
    lqtwmCylindrical : // Cylindrical
        lRectUnion.Union(FRectB,CylindricalWorld,lqtwmCylindrical);
    lqtwmToroidal : // Toroidal
        lRectUnion.Union(FRectB,TorodialWorld,lqtwmToroidal);
  else
    // Flat
    lRectUnion.Union(FRectB,FlatWorld,lqtwmFlat);
  end;
  ALines.Add('Union of Rect A and B:');
  ALines.Add(lRectUnion.CoordsAsString('L %1.2f, T %1.2f, R %1.2f, B %1.2f'));

end;

{ SetupTestCases copies the Description from the TestCases into the drop down list }
procedure TFrameLQTRectOverlapTest.SetupTestCases;
var
  i : Integer;
begin
  cbTestCases.Clear;
  for i := 0 to High(TestCases) do
    cbTestCases.Items.Add(TestCases[i].Desc);
end;

{ IsEqualToTestCaseIndex returns the index of the testcase if the current used
  values are identical to one of them }
function TFrameLQTRectOverlapTest.IsEqualToTestCaseIndex: Integer;
var
  i : Integer;
begin
  Result := -1;
  for i := 0 to High(TestCases) do
  begin
    if (FWorldModel = TestCases[i].WorldModel) and
       FRectA.IsEqual(TestCases[i].RectA) and
       FRectB.IsEqual(TestCases[i].RectB) then
    begin
      Result := i;
      Exit
    end;
  end;
end;

procedure TFrameLQTRectOverlapTest.AdjustTestCaseDropDownListToSetup;
var
  ndx : Integer;
begin
  ndx := IsEqualToTestCaseIndex;
  cbTestCases.ItemIndex := ndx;
end;
{ Create of the Frame }
constructor TFrameLQTRectOverlapTest.Create(AOwner: TComponent);
var
  exepath : String;
  mappath : String;
  jpg : TJpegImage;
begin
  inherited Create(AOwner);
  // Load the two bitmaps, if not loadable replace them by a colored dummy
  exepath := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName));
  mappath := IncludeTrailingPathDelimiter(ResolveDots(exepath+'..\..\Maps'));
  FCylindricBmp := TBitmap.Create;
  try
    jpg := TJpegImage.Create;
    try
      jpg.LoadFromFile(mappath+CylindricalImgFileName);
      FCylindricBmp.Assign(jpg);
    finally
      jpg.Free;
    end;
  except
    FCylindricBmp.SetSize(WorldTotalWidthPx,WorldTotalHeightPx);
    FCylindricBmp.Canvas.Brush.Style := bsSolid;
    FCylindricBmp.Canvas.Brush.Color := clBlack;
    FCylindricBmp.Canvas.Rectangle(0,0,WorldTotalWidthPx,WorldTotalHeightPx);
    FCylindricBmp.Canvas.Brush.Color := clSteelBlue;
    FCylindricBmp.Canvas.Rectangle(WorldOutSidePx,WorldOutSidePx,WorldOutSidePx+WorldWidthPx,WorldOutSidePx+WorldHeightPx);
    FCylindricBmp.Canvas.Brush.Color := clLtGray;
    FCylindricBmp.Canvas.Rectangle(0,WorldOutSidePx,WorldOutSidePx,WorldOutSidePx+WorldHeightPx);
    FCylindricBmp.Canvas.Rectangle(900,WorldOutSidePx,WorldTotalWidthPx,WorldOutSidePx+WorldHeightPx);
  end;
  FToroidalBmp := TBitmap.Create;
  try
    jpg := TJpegImage.Create;
    try
      jpg.LoadFromFile(mappath+ToroidalImgFileName);
      FToroidalBmp.Assign(jpg);
    finally
      jpg.Free;
    end;
  except
    FToroidalBmp.SetSize(WorldTotalWidthPx,WorldTotalHeightPx);
    FToroidalBmp.Canvas.Brush.Color := clLtGRay;
    FToroidalBmp.Canvas.Rectangle(0,0,WorldTotalWidthPx,WorldTotalHeightPx);
    FToroidalBmp.Canvas.Brush.Color := clSteelBlue;
    FToroidalBmp.Canvas.Rectangle(WorldOutSidePx,WorldOutSidePx,WorldOutSidePx+WorldWidthPx,WorldOutSidePx+WorldHeightPx);
  end;
  // Setup the data from the first test case
  // Assume a flat world
  FRectA.Assign(TestCases[0].RectA);
  FRectB.Assign(TestCases[0].RectB);
  FWorldRect.Assign(FlatWorld);
  FWorldModel := lqtwmFlat;
  FRectShifter := TMouseRectShifter.Create;
  RectToGUI;
  RectangleValuesToMemo;
  RectangleToStatusBar;
  SetupTestCases;
  AdjustTestCaseDropDownListToSetup;
end;

destructor TFrameLQTRectOverlapTest.Destroy;
begin
  // Free created items
  FRectShifter.Free;
  FCylindricBmp.Free;
  FToroidalBmp.Free;
  inherited Destroy;
end;

end.

