{ uLQTRectOverlapThread.pas
  A threaded Testcase for the rectangle overlapping
  functions defined in the uLazQuadTreeGeometry.pas unit

License: The same modified LGPL as the Free Pascal RTL
         See the file COPYING.modifiedLGPL for more details

Author: Ekkehard Domning (ekkehard@domning.eu)

  As explained in detail in the used unit, two rectangles could overlap in several
  ways. This unit validates the implementation of the overlapping function, by
  generating all possible constellations of rectangles.

}
unit uLQTRectOverlapThread;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, uLazQuadTreeGeometry;
const
  // Size of the test bed is choosen to allow all possible overlapping scenarios.
  // - separated by a free field,
  // - sharing a common border
  // - partly overlapping, without sharing any common borders
  // - partly overlapping, with sharing common borders and / or corners
  // - fully covered with common border
  // - fully covered without common borders

  OverlapTestfieldHeight = 5;
  OverlapTestfieldWidth = 10;




  ExpectedTestCount = 6195772; //

  // Flags for the Testfield
  tfClear = $00;
  tfBase  = $01;
  tfTest  = $02;

type
  TTestRectangle = record
    Width, Height : Integer;
  end;

  { TRectangleOverlapTestThread }

  TRectangleOverlapTestThread = class(TThread)
  private
    FRectangles : array of TTestRectangle;
    FTestField : array[0..OverlapTestfieldWidth-1, 0..OverlapTestfieldHeight-1] of Byte;
    FTestCounter : Integer;
    FMaxOverlapRectangleCount : Integer;
    FRectA : TLQTRect;
    FRectB : TLQTRect;
    FOverlappingRectangles : TLQTRectArray;
    FHitTestError : Boolean;
    FHitExpected : TLQTRectOverlapKind;
    FHitCalculated : TLQTRectOverlapKind;
    procedure CreateRectangles;
  protected
    procedure Execute;override;
  public
    property Terminated;
    property HitTestError : Boolean read FHitTestError;
    property TestCounter : Integer read FTestCounter;  // Runs from 0 to 6195771
    property MaxOverlapRectangleCount : Integer read FMaxOverlapRectangleCount;
    property RectA : TLQTRect read FRectA ;
    property RectB : TLQTRect read FRectB ;
    property HitExpected : TLQTRectOverlapKind read FHitExpected;
    property HitCalculated : TLQTRectOverlapKind read FHitCalculated;
    constructor Create(CreateSuspended: Boolean;
                       const StackSize: SizeUInt = DefaultStackSize);
    destructor Destroy; override;
  end;

implementation

{ TRectangleOverlapTestThread }

procedure TRectangleOverlapTestThread.CreateRectangles;
var
  i,j : Integer;
  ndx : Integer;
begin
  SetLength(FRectangles,OverlapTestfieldHeight*OverlapTestfieldWidth);
  ndx := 0;
  for i := 1 to OverlapTestfieldHeight do
  begin
    for j := 1 to OverlapTestfieldWidth do
    begin
      FRectangles[ndx].Height := i;
      FRectangles[ndx].Width := j;
      Inc(ndx);
    end;
  end;
end;

procedure TRectangleOverlapTestThread.Execute;
var
  i,j : Integer; // Loop through the rectangles for Base and Test
  xb,yb : Integer; // Loop through the possible positions of Base-Rect
  xt,yt : Integer; // Loop throug the possible position of Test-Rect
  xx,yy, xr,yr : Integer; //Loop through the used rectangle fields
  hb,wb, ht,wt : Integer; //Height and Width of Base- and Test-Rectangle
  ri :Integer;
  hittest, evalhittest : TLQTRectOverlapKind;
  cntclr, cntbase, cnttest, cntboth : Integer; //Counter of used fields
  v : Byte;
  world : TLQTRect;
  overlaplat, overlaplon : Double;
  pt : TLQTPoint;
  b0 : Boolean;
  overlapequal : Boolean;
  wm :   TLQTWorldModel;
begin
  try
    world.SetFrom(-180.0,90.0,180.0,-90.0); // Caution World is assumed later to be symmetrical around 0,0!!
    CreateRectangles;
    for wm := Low(TLQTWorldModel) to High(TLQTWorldModel) do
    begin
      // For each rectangle in the list, used as "Base"
      for i := 0 to High(FRectangles) do
      begin
        if Terminated then Exit;
        // For each coordinate in the field
        hb := FRectangles[i].Height;
        wb := FRectangles[i].Width;
        // Limit the position to the heights of the rectangles
        for yb := 0 to OverlapTestfieldHeight-1 do
        begin
          // In the case of a flat or cylindrical world the upper and lower borders of the world
          // could not be traversed
          // A 1-height rectangle could be placed everywhere, a 5-height only on position [0]
          if ((wm = lqtwmFlat) or (wm = lqtwmCylindrical)) and
             (hb+yb > OverlapTestfieldHeight) then
                Break;
          // In the toroidial world rectangles with the full height could be placed only on the position 0
          if (wm = lqtwmToroidal) and
             (hb = OverlapTestfieldHeight) and (yb > 0) then Break;
          for xb := 0 to OverlapTestfieldWidth-1 do
          begin
            // In the flat world rectangles could not pass the right border
            if (wm = lqtwmFlat) and
               (wb+xb > OverlapTestfieldWidth) then
                  Break;
            // In the cylindric and the toroidial world
            // width could not be limited, since overlapping to the right is a feature.
            // An exception is a 10-width rectangle, wich is catched separately
            if (wb = OverlapTestfieldWidth) and (xb > 0) then Break;
            // Clear the Field
            FillChar(FTestField[0],SizeOf(FTestField),tfClear);
            // Calc the real world coordinates
            FRectA.Left := Round(((xb/OverlapTestfieldWidth)*360.0)-180);
            FRectA.Top := Round(-((yb/OverlapTestfieldHeight)*180)+90);
            FRectA.Right := Round((((xb+wb)/OverlapTestfieldWidth)*360.0)-180);
            // If cylindrical or toroidal the rectangle may wrap around from the right to the left side
            if FRectA.Right > world.Right then
              FRectA.Right := FRectA.Right - world.Width;
            FRectA.Bottom := Round(-(((yb+hb)/OverlapTestfieldHeight)*180)+90);
            // In the toroidal world the wrap around may also occour from the bottom to the top
            // The non torodial cases are sorted out before
            if FRectA.Bottom < world.Bottom then
              FRectA.Bottom := FRectA.Bottom + world.Height;
            // Place the Base-rectangle on the field
            for xx := 0 to wb-1 do
              for yy := 0 to hb-1 do
              begin
                xr := xx+xb;
                if xr >= OverlapTestfieldWidth then
                  xr := xr - OverlapTestfieldWidth;
                yr := yy+yb;
                if yr >= OverlapTestfieldHeight then
                  yr := yr - OverlapTestfieldHeight;
                FTestField[xr,yr] := tfBase;
              end;
            // Now the same for the Test Rectangle
            // For each rectangle in the list, used as "Test"
            for j := 0 to High(FRectangles) do
            begin
              // For each coordinate in the field
              ht := FRectangles[j].Height;
              wt := FRectangles[j].Width;
              // Limit the position to the heights of the rectangles
              // A 1-height rectangle could be placed everywhere, a 5-height only on position [0]
              for yt := 0 to OverlapTestfieldHeight-1 do
              begin
                if ((wm = lqtwmFlat) or (wm = lqtwmCylindrical)) and
                   (ht+yt > OverlapTestfieldHeight) then
                      Break;
                // In the toroidial world rectangles with the full height could be placed only on the position 0
                if (wm = lqtwmToroidal) and
                   (ht = OverlapTestfieldHeight) and (yt > 0) then Break;
                // Width could not be limited, since overlapping to the right is a feature.
                // An exception is a 10-width rectangle, wich is catche separately
                for xt := 0 to OverlapTestfieldWidth-1 do
                begin
                  if Terminated then Exit;
                  // In the flat world rectangles could not pass the right border
                  if (wm = lqtwmFlat) and
                     (wt+xt > OverlapTestfieldWidth) then
                        Break;
                  if (wt = OverlapTestfieldWidth) and (xt > 0) then Break;
                  Inc(FTestCounter);
                  // Calc the real world coordinates
                  FRectB.Left := Round(((xt/OverlapTestfieldWidth)*360.0)-180);
                  FRectB.Top := Round(-((yt/OverlapTestfieldHeight)*180)+90);
                  FRectB.Right := Round((((xt+wt)/OverlapTestfieldWidth)*360.0)-180);
                  if FRectB.Right > world.Right then
                    FRectB.Right := FRectB.Right - world.Width;
                  FRectB.Bottom := Round(-(((yt+ht)/OverlapTestfieldHeight)*180)+90);
                  if FRectB.Bottom < world.Bottom then
                    FRectB.Bottom := FRectB.Bottom + world.Height;
                  // Place the Test-rectangle on the field, by "or"-ing the fields
                  for xx := 0 to wt-1 do
                    for yy := 0 to ht-1 do
                    begin
                      xr := xx+xt;
                      if xr >= OverlapTestfieldWidth then
                        xr := xr - OverlapTestfieldWidth;
                      yr := yy+yt;
                      if yr >= OverlapTestfieldHeight then
                        yr := yr - OverlapTestfieldHeight;
                      FTestField[xr,yr] := FTestField[xr,yr] or tfTest;
                    end;

                  hittest := RectanglesOverlapEx(FRectA,FRectB,
                                    world, wm,
                                    FOverlappingRectangles);
                  if FMaxOverlapRectangleCount < Length(FOverlappingRectangles) then
                    FMaxOverlapRectangleCount := Length(FOverlappingRectangles);
                  cntclr := 0;
                  cntbase := 0;
                  cnttest := 0;
                  cntboth := 0;
                  overlapequal := True;
                  for yy := 0 to OverlapTestfieldHeight-1 do
                    for xx := 0 to OverlapTestfieldWidth-1 do
                    begin
                      v := FTestField[xx,yy];
                      if v = tfClear then
                        Inc(cntclr)
                      else
                      begin
                        if (v and (tfBase or tfTest)) = (tfBase or tfTest) then
                        begin
                          // An overlapping field is detected
                          Inc(cntboth);
                          overlaplat := Round(-((yy/OverlapTestfieldHeight)*180)+90);
                          overlaplon := Round(((xx/OverlapTestfieldWidth)*360.0)-180);
                          pt.X := overlaplon;
                          pt.Y := overlaplat;
                          // Now test the coordinates in the overlapping area
                          b0 := False;
                          for ri := 0 to High(FOverlappingRectangles) do
                          begin
                            b0 := PointInRectangle(FOverlappingRectangles[ri],
                                                   pt,world,wm);
                            if b0 then Break;
                          end;
                          if not b0 then
                            overlapequal := False;
                        end
                        else if (v and tfBase) <> 0 then
                          Inc(cntbase)
                        else if (v and tfTest) <> 0 then
                          Inc(cnttest);
                      end;
                    end;
                  if cntboth = 0 then
                    evalhittest := qroNone {no hit}
                  else if (cntbase = 0) and (cnttest = 0) then
                    evalhittest := qroEqual
                  else if (cnttest = 0) then
                    evalhittest := qroInside
                  else if (cntbase = 0) then
                    evalhittest := qroFull
                  else
                    evalhittest := qroPartly;
                  if  (not overlapequal) or (hittest <> evalhittest) then
                  begin
                    FHitExpected := evalhittest;
                    FHitCalculated := hittest;
                    FHitTestError := True;
                    // For debugging purpose a recall of the possible wrong coded detection function
                    hittest := RectanglesOverlapEx(FRectA,FRectB,
                                      world, wm,
                                      FOverlappingRectangles);
                    Exit;
                  end;
                  // clear the test rectangle
                  for yy := 0 to OverlapTestfieldHeight-1 do
                    for xx := 0 to OverlapTestfieldWidth-1 do
                    begin
                      v := FTestField[xx,yy];
                      v := (v and (not tfTest));
                      FTestField[xx,yy] := v;
                    end;
                end;
              end;
            end;
          end;
        end;
      end;
    end;
  finally
    Terminate;
  end;
end;

constructor TRectangleOverlapTestThread.Create(CreateSuspended: Boolean;
  const StackSize: SizeUInt);
begin
  inherited;
end;

destructor TRectangleOverlapTestThread.Destroy;
begin
  Suspended := False;
  Terminate;
  WaitFor();
  inherited Destroy;
end;

end.

