{ A Demo project for using the LazQuadTree in conjunction with the lazMapViewer
  Copyright (C) 2024 Ekkehard Domning (www.domis.de)

  License: modified LGPL with linking exception (like RTL, FCL and LCL)

  See the file COPYING.modifiedLGPL.txt, included in the Lazarus distribution,
  for details about the license.

  Version Date        Change
  0.0.1   2024-05-31  First release
  0.0.2   2024-08-31  Update of the new QuadTree-Unit and the Aggregation method
  0.0.3   2024-10-27  Added Copyright information for OpenStreetMap
}
unit uLazMapViwerQuadTreeDemo;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Windows, Graphics, Dialogs, ComCtrls, ExtCtrls,
  StdCtrls,
  LCLIntf, // For OpenURL
  lazFileUtils,
  // If one of this Units not found, download from
  // https://sourceforge.net/p/lazarus-ccr/svn/HEAD/tree/components/lazmapviewer/
  // There is no need to install those components into Lazarus
  // the components are loaded during runtime, not during design time.
  mvMapViewer, mvGpsObj, mvDrawingEngine,
  mvGeoMath, mvTypes, mvEngine,
  uLazQuadTreeGeometry, uLazQuadTree, uOSMQuadTree;
const
  // The sample file is located at the directory of the exe file
  SamplePOIFileName = 'POI.csv';
  // This defines how many group items will be displyed in one tile
  AggregationLevel = 3; // We will dive 3 Levels into the Tile = 64 Grouped Items per Tile
  ZoomInPerGroup = 2; // We zoom 2 Levels in
  //
  MinItemsPerGroup = 4; // If more than MinItemsPerGroup are in a specific area then those items are grouped

type
  { TGpsPointGroupMarker
     Is displayed as a circle with the amount of grouped items
   }
   TGpsPointGroupMarker = class(TGPSPoint)
   private
     FGroupedItemsCount : Integer;
   public
     property GroupedItemsCount : Integer read FGroupedItemsCount write FGroupedItemsCount;
   end;

  { TForm1 }

  TForm1 = class(TForm)
    cbShowPOICaptions: TCheckBox;
    ImageList1: TImageList;
    Label1: TLabel;
    Label2: TLabel;
    Label4: TLabel;
    lblOSMCopyright: TLabel;
    lblLatitude: TLabel;
    lblLongitude: TLabel;
    lblSampleImportResult: TLabel;
    Label3: TLabel;
    lblListItemsCount: TLabel;
    lblPOIDesc: TLabel;
    Panel1: TPanel;
    Panel2: TPanel;
    pnlMapView: TPanel;
    rgDefaultVsQuadTree: TRadioGroup;
    procedure cbShowPOICaptionsChange(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure lblOSMCopyrightClick(Sender: TObject);
    procedure MapViewBeforeDrawObjects(Sender: TObject);
    procedure MapViewChange(Sender: TObject);
    procedure MapViewDrawGpsPoint(Sender: TObject;
      ADrawer: TMvCustomDrawingEngine; APoint: TGpsPoint);
    procedure MapViewMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MapViewMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure MapViewMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pnlMapViewChangeBounds(Sender: TObject);
    procedure rgDefaultVsQuadTreeClick(Sender: TObject);
  private
    FMapView: TMapView;
    FMapPOIList: TGPSObjectList;
    // Track of the Mouse movement
    FMapViewMouseDownX : Integer;
    FMapViewMouseDownY : Integer;
    // Track of the change in the map display
    FLastMapArea : TRealArea;
    FLastMapCenter : TRealPoint;
    // The QuadTree for the POIs
    FTestQuadTree : TOSMQuadTree;
    // The extracted POI values from the file
    FPOICoords : array of TLQTPoint;
    FPOICaptions : TStringList;
    // Events / CallBacks for the specific implementation
    // First, the Events/Callbacks for the QuadTree to get the Coords of the POI
    function OnTestQuadTreeGetItemCoord(Sender : TObject; AItem : TObject; var AX, AY : Double) : Boolean;
    function OnTestQuadTreeGetItemCaptionEvent(Sender : TObject; AItem : TObject; var AItemCaption : String) : Boolean;
    // Second, the MapView Event for displaying the POIs
    // A Helper function to calculate the tile coords of the displayed map
    // will be very helpful for derived procets
    procedure CalculateMapPixelCoords(var ALeft,ATop,ARight,ABottom : Int64; var AOSMMapSize : Int64; var AOfsX, AOfsY : Integer);
    // Import the Sample POIs
    procedure SamplePOIImport;
    function AgglomerationItemEnumerationCB(Sender : TObject;
                                            ANode : TLQTNode;
                                            ANodeItem : TObject;
                                            ANodeItemIndex : Integer;
                                            AUserData : Pointer) : Boolean;
    function AgglomerationEnumerationCB(ANode : TLQTNode;
                                        const ATreeAgglomerationRec : TLQTAgglomerationRec;
                                        AUserData : Pointer) : Boolean;

  public

  end;


var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }

procedure TForm1.MapViewChange(Sender: TObject);
begin
  lblPOIDesc.Caption := '';
end;

procedure TForm1.FormCreate(Sender: TObject);
  procedure AddTraditionalMarker(const ALon, ALat : Double; ACaption : String);
  var
    gpsPt: TGpsPointOfInterest;
  begin
    gpsPt := TGpsPointOfInterest.Create(ALon,ALat);
    try
      gpsPt.Name := ACaption;
      gpsPt.ImageIndex := 0;
      FMapView.GPSItems.Add(gpsPt, 100);
      gpsPt := Nil;
    finally
      if Assigned(gpsPt) then
        gpsPt.Free;
    end;
  end;

begin
  FMapView := TMapView.Create(Self);
  with FMapView do
  begin
    Parent := pnlMapView;
    Align := alClient;
    Cyclic := True;
    DebugTiles := True;
    DefaultTrackColor := clBlue;
    DefaultTrackWidth := 3;
    Font.Color := clBlack;
    MapProvider := 'OpenStreetMap Mapnik';
    UseThreads := True;
    OnBeforeDrawObjects := @MapViewBeforeDrawObjects;
    OnChange := @MapViewChange;
    OnDrawGpsPoint := @MapViewDrawGpsPoint;
    OnMouseDown := @MapViewMouseDown;
    OnMouseMove := @MapViewMouseMove;
    OnMouseUp := @MapViewMouseUp;
    OnBeforeDrawObjects := @MapViewBeforeDrawObjects;
    Active := True;

    // An Example of a different item, that is used like "normal"
    POIImages := ImageList1;


    // Save the GPSLayer[0] to ease the access for the specific group
    FMapPOIList := GPSLayer[0];    // ... so schon. Wegen Index 0 werden die Items auf diesem Layer unter GPSItems (Index 5) gezeichnet.
  end;
  AddTraditionalMarker(0.0, 51.4825766,'Greenwich');
  AddTraditionalMarker(2.2945500,48.8582300,'Tour d´Eiffel, Paris');
  AddTraditionalMarker(-79.3884000,43.6439500,'CN Tower, Toronto');
  AddTraditionalMarker(-157.7739800,21.2716900,'Kahala Avenue, Honolulu');
  AddTraditionalMarker(114.1497900,22.2708100,'The Peak, Hong Kong');
  AddTraditionalMarker(13.377778,52.516389,'Brandenburger Tor, Berlin');
  AddTraditionalMarker(-58.3722400,-34.6084700,'Pirámide de Mayo, Buenos Aires');
  AddTraditionalMarker(151.2082800,-33.8707000,'Sydney Tower Skywalk, Sydney');
  AddTraditionalMarker(139.7021800,35.6787500,'Meiji Jingu Shrine, Tokyo');
  lblOSMCopyright.Parent := FMapView;

  // To simplify the conversion of the Coords, this format ist fixed here
  DefaultFormatSettings.DecimalSeparator:='.';
  DefaultFormatSettings.ThousandSeparator:=#0;
  FTestQuadTree := TOSMQuadTree.Create(False); // The Tree does not owns the items
  FTestQuadTree.MaxTreeLevel := FMapView.ZoomMax; //Equal Tree and Map-Zoom levels
  FTestQuadTree.OnGetItemCoord := @OnTestQuadTreeGetItemCoord;
  FTestQuadTree.OnGetItemCaption := @OnTestQuadTreeGetItemCaptionEvent;
  FLastMapArea.Init(0,0,0,0);
  // If the MapViewer-Component is updated, this Event could be assigned using the ObjectInspector
  FPOICaptions := TStringList.Create;
end;

procedure TForm1.FormActivate(Sender: TObject);
var
  oldCursor : TCursor;
begin
  OnActivate := Nil;
  pnlMapViewChangeBounds(Nil);
  oldCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  try
    pnlMapView.Repaint; // Paint the Background direct
    // Import the Sample POIs on activating, takes a while, so show wainting mouse pointer
    SamplePOIImport;
  finally
    Screen.Cursor := oldCursor;
    pnlMapView.Repaint; // Force displaying the Items
  end;
end;

procedure TForm1.cbShowPOICaptionsChange(Sender: TObject);
begin
  if (rgDefaultVsQuadTree.ItemIndex = 1) and
     cbShowPOICaptions.Checked then
  begin
    if Application.MessageBox('Warning! Showing all Captions may overload the capacity of the MapViewer! Continue?',
                              'Warning! Overload possible!',MB_ICONWARNING or mb_YesNo) = idNo then
    begin
      cbShowPOICaptions.Checked := False;
      Exit;
    end;
  end;
  FMapView.Invalidate;
  FMapView.Repaint;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  // Destroy created objects
  FPOICaptions.Free;
  FTestQuadTree.Free;
end;

procedure TForm1.lblOSMCopyrightClick(Sender: TObject);
begin
  OpenURL('https://www.openstreetmap.org/copyright');
end;

{ AgglomerationItemEnumerationCB this callback-function will be called by the
   Agglomeration Method of the QuadTree, which is called in the
   MapViewBeforeDrawObjects method below in this unit.
   The job here is to create the GPSPints, which represent one
   POI of the original POI.csv file}
function TForm1.AgglomerationItemEnumerationCB(Sender : TObject;
                                               ANode : TLQTNode;
                                               ANodeItem : TObject;
                                               ANodeItemIndex : Integer;
                                               AUserData : Pointer) : Boolean;
var
  llon, llat : Double;
  gpsPt : TGpsPoint;
  s : String;
begin
  s := '';
  llat := 0.0;
  llon := 0.0;
  // Ask the Item for it's coordinates (we reuse the Event-Method here)
  if OnTestQuadTreeGetItemCoord(Self,TObject(ANodeItem),llon,llat) then
  begin
    // Create a GpsPoint and add the POI to the list of the MapView
    gpsPt := TGpsPoint.Create(llon,llat);
    try
    // Ask the Item for it's Caption (we reuse the Event-Method here)
    if OnTestQuadTreeGetItemCaptionEvent(Self,TObject(ANodeItem),s) then
      gpsPt.Name := s;
    FMapPOIList.Add(gpsPt, 0); // Add the created item to the map
    gpsPt := Nil;
    except
      if Assigned(gpsPt) then
        gpsPt.Free;
    end;
  end;
  Result := True;
end;
{ AgglomerationEnumerationCB this callback-function will be called by the
   Agglomeration Method of the QuadTree, which is called in the
   MapViewBeforeDrawObjects method below in this unit.
   The job here is to create the GroupMarkers, which will represent
   areas in the map, who are to overcrowded. }
function TForm1.AgglomerationEnumerationCB(ANode : TLQTNode;
                                           const ATreeAgglomerationRec : TLQTAgglomerationRec;
                                           AUserData : Pointer) : Boolean;
var
  lItemGroupMarker : TGpsPointGroupMarker;
begin
  lItemGroupMarker := TGpsPointGroupMarker.Create(ATreeAgglomerationRec.Point.X,ATreeAgglomerationRec.Point.Y);
  lItemGroupMarker.GroupedItemsCount := Round(ATreeAgglomerationRec.Count);
  FMapPOIList.Add(lItemGroupMarker, 0);  // Add the group indicator to the map
  Result := True;
end;

procedure TForm1.MapViewBeforeDrawObjects(Sender: TObject);
var
  maparea : TRealArea;
  x, y : Integer;
  ofsx, ofsy : Integer;
  i : Integer;
  rpt0 : TRealPoint;
  lTileCntX, lTileCntY : Integer;
  lMapWidthPx : Int64;
  lLeft,lTop,lRight,lBottom : Int64;
  lTileX, lTileY : Int64;
  n : TOSMQuadNode;
  lNodeList : TList;
  cnt : Integer;
  found : Boolean;
begin
  // If the User wants to see ungrouped items than we dont have anything to do here
  if (rgDefaultVsQuadTree.ItemIndex <> 0) then
  begin
    // Update the Number of GpsItems in the list to the GUI
    lblListItemsCount.Caption := IntToStr(FMapPOIList.Count);  // Display the number of items in the Map
    Exit;
  end;
  // Check if the Map has changed or moved in the mean time.
  // If not, exit.
  maparea := FMapView.GetVisibleArea;
  rpt0 := FMapView.Center;
  if FLastMapArea.Equal(maparea) and
    (FLastMapCenter.Lat = rpt0.Lat) and
    (FLastMapCenter.Lon = rpt0.Lon)  then Exit;
  // Update the current position of the map
  FLastMapCenter.InitXY(rpt0.Lon,rpt0.Lat);
  FLastMapArea.Init(maparea.TopLeft,maparea.BottomRight);

  FMapView.BeginUpdateObjects;
  try

    // Keep the compiler calm
    ofsx := 0;
    ofsy := 0;
    lLeft := 0;
    lTop := 0;
    lRight := 0;
    lBottom := 0;
    lMapWidthPx := 0;

    // Clear all Marker-Items in the MapView
    // This could be done smarter, by keeping unchanged items
    // but for the ease of demonstration, we restart everytime from scratch.
    FMapPOIList.ClearAll; // If compiling fails here, you missed one update,
    // download the current component or use the next line
    // FMapPOIList.Clear(0); // Takes surprisingly long!!

    // Calculate the tiles area
    // The method returns the coordinates in pixel for the current map, adjusted
    // to the full OSM-Tiles. OfsX and OfsY are the pixel shift between the
    // display and the tile upperleft edge.
    CalculateMapPixelCoords(lLeft,lTop,lRight,lBottom,lMapWidthPx,ofsx,ofsy);
    // Calculate the number of tiles in the x and y direction
    lTileCntX := (lRight-lLeft) div TILE_SIZE;
    lTileCntY := (lBottom-lTop) div TILE_SIZE;

    // Now we run to the displayed tiles
    // (doubles caused by loop around display are not pop up here!)
    // and collect the correspondend TreeNodes
    lNodeList := TList.Create;
    try
      for y := 0 to lTileCntY-1 do
      begin
        lTileY := lTop + (y * TILE_SIZE); // Top of the Tile
        for x := 0 to lTileCntX do
        begin
          lTileX := lLeft + (x * TILE_SIZE); // Left of the Tile
          // Caution, due to a wrap around at the Date-Border (-180/180 Degree)
          // the sign changes.
          // So negative X values are Left from -180 Degree!
          // X values >= the MapWidth are right from 180 Degree, so wrap back
          if lTileX < 0 then
            lTileX := lTileX + lMapWidthPx
          else if lTileX >= lMapWidthPx then
            lTileX := lTileX - lMapWidthPx;
          // Find the Node who is the related one
          n := FTestQuadTree.FindOSMTileNode(FMapView.Zoom,lTileX,lTileY);
          // Caution: In the area where only few items available the
          // node is not at the deepest possible level, thus the found node
          // may already be in the list, so we only insert nodes, we did not found
          // so far.
          if lNodeList.IndexOf(n) < 0 then
          begin // The insertion node is not in the list.
            // The following stuff is a bit tricky, you can comment everything until
            // lNodeList.Add(n);
            // and see what happen.
            // You will find that sometimes Group markers of Group markers are inserted
            //    So we like to               We want to avoid
            //    have grouped markers        something like this
            //
            //     10     20                     10     20
            //                                      100
            //     30     40                     30     40

            // To avoid this, we have to follow the rule that children (descendants)
            // have a higher rank than parents (ascendants).
            // So we have to remove Nodes in the list, where the insertion node is an descendant
            cnt := lNodeList.Count;
            for i := cnt-1 downto 0 do
            begin
              if n.IsDescendantOf(TOSMQuadNode(lNodeList.Items[i])) then
                lNodeList.Delete(i);
            end;
            // Also the reverse case has to be checked.
            // Is one node in the list an descendant of the node that is about to be inserted?
            found := False; // Preset to "Insert Node"
            cnt := lNodeList.Count;
            for i := cnt-1 downto 0 do
            begin
              if TOSMQuadNode(lNodeList.Items[i]).IsDescendantOf(n) then
              begin
                found := True; // We found an descendant of the insertion node, omit insertion
                Break;
              end;
            end;
            if not Found then // Only if no relation found, we will insert this node
              lNodeList.Add(n);
          end;
        end;
      end;
      // From here it will be fairly simple
      // Now we iterate through all Nodes in the list...
      cnt := lNodeList.Count;
      for i := 0 to cnt-1 do
      begin
        n := TOSMQuadNode(lNodeList.Items[i]);
        // ... and Agglomerate the content
        // The creation of the Items for the MavView is made in the CallBack functions above.
        n.AgglomerateNodeAsEnumeration(AggregationLevel,MinItemsPerGroup,
                                       @AgglomerationItemEnumerationCB,Nil,
                                       @AgglomerationEnumerationCB, Nil);
      end;
    finally
      lNodeList.Free;
    end;
  finally
    FMapView.EndUpdateObjects;
  end;
  // Update the Number of GpsItems in the list to the GUI
  cnt := FMapPOIList.Count;
  lblListItemsCount.Caption := IntToStr(cnt);  //
end;

{ MapViewDrawGpsPoint draws the GpsPoints of the MapViewer
   The drawing separates the display of the GroupItems and the single POIs }
procedure TForm1.MapViewDrawGpsPoint(Sender: TObject;
  ADrawer: TMvCustomDrawingEngine; APoint: TGpsPoint);
var
  P: TPoint;
  Pt: TPoint;
  PtCyc: TPointArray;
  s : String;
  d: Integer;
  clr: TColor;
  extent: TSize;
  lMapView: TMapView;
  cnt : Integer;
  fn: String;
  fs: Integer;
  fc: TColor;
begin
  lMapView := TMapView(Sender);
  // Screen coordinates of the GPS point
  P := lMapView.LatLonToScreen(APoint.RealPoint);
  PtCyc := FMapView.CyclicPointsOf(P);
  // Presetting
  clr := clRed;
  d := 5;
  ADrawer.PenColor := clr;
  ADrawer.PenWidth:= 3;
  for Pt in PtCyc do
  begin
    // If the Point is our GroupMarker, then display a circle with the amount of
    // grouped Items in in, if 10.000 and more then display "10K+" instead
    if APoint is TGpsPointGroupMarker then
    begin
      ADrawer.PenColor := clNavy;
      ADrawer.BrushStyle := bsSolid;
      ADrawer.BrushColor := clWhite;
      // Store font data
      fc := ADrawer.FontColor;
      fn := ADrawer.FontName;
      fs := ADrawer.FontSize;
      // Change font
      ADrawer.FontColor := clBlack;
      ADrawer.FontName := 'Arial';
      ADrawer.FontSize := 6;
      ADrawer.FontStyle := [];

      d := 10;
      ADrawer.Ellipse(Pt.X - d, Pt.Y - d,
                      Pt.X + d, Pt.Y + d);
      cnt := TGpsPointGroupMarker(APoint).GroupedItemsCount;
      if cnt > 9999 then
        s := '10K+'
      else
        s := IntToStr(cnt);
      extent := ADrawer.TextExtent(s);
      // ... and font
      ADrawer.BrushStyle := bsClear;

      // Write the POI text
      ADrawer.TextOut(Pt.X - (extent.CX div 2), Pt.Y - (extent.CY div 2), s);

      // Restore font data
      ADrawer.FontColor := fc;
      ADrawer.FontName := fn;
      ADrawer.FontSize := fs;

    end
    else // Al other objects ends in a "red cross" (optional with Caption)
    begin
      ADrawer.Line(Pt.X - d, Pt.Y, Pt.X + d, Pt.Y);
      ADrawer.Line(Pt.X, Pt.Y - d, Pt.X , Pt.Y + d);
{ Displaying the Captions of the single POI will looks ugly due to overlapping
  something to do better in the future. Here just left for testing
}
      if cbShowPOICaptions.Checked then
      begin
        // Prepare text output: background color...
        extent := ADrawer.TextExtent(APoint.Name);
        ADrawer.BrushStyle := bsClear;
        // ... and font
        ADrawer.FontColor := FMapView.Font.Color;
        ADrawer.FontName := FMapView.Font.Name;
        ADrawer.FontSize := 7; // MapView.Font.Size;
        ADrawer.FontStyle := FMapView.Font.Style;


        // Write the POI text
        ADrawer.TextOut(Pt.X + 2, Pt.Y + 2, APoint.Name);
      end;
    end;
  end;
end;

procedure TForm1.MapViewMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  // Keep track of the mouse
  FMapViewMouseDownX := X;
  FMapViewMouseDownY := Y;
end;

// MouseMove finds the Map-GpsItems under the mouse pointer and display their
// captions in a label
procedure TForm1.MapViewMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
const
  DELTA = 3;
var
  rArea: TRealArea;
  rPoint : TRealPoint;
  gpsList: TGpsObjList;
  sl : TStringList;
  i: Integer;
  pt : TPoint;
  foundgroupmarker : Boolean;
begin
  pt.X := X;
  pt.Y := Y;
  rPoint := FMapView.ScreenToLatLon(pt);
  lblLongitude.Caption := Format('%1.6f°',[rPoint.Lon]);
  lblLatitude.Caption := Format('%1.6f°',[rPoint.Lat]);

  foundgroupmarker := False;
  pt.X := X-DELTA;
  pt.Y := Y-DELTA;
  rArea.TopLeft := FMapView.ScreenToLatLon(pt);
  pt.X := X+DELTA;
  pt.Y := Y+DELTA;
  rArea.BottomRight := FMapView.ScreenToLatLon(pt);
  gpsList := FMapPOIList.GetObjectsInArea(rArea);  // Return all items (which belongs to this "layer") on the screen in one list
  try
    if gpsList.Count > 0 then
    begin
      sl := TStringList.Create;
      try
        for i:=0 to gpsList.Count-1 do
        begin
          if ((gpsList[i] is TGpsPointGroupMarker)) then
            foundgroupmarker := True
          else if (gpsList[i] is TGpsPoint) then
            sl.Add(TGpsPoint(gpsList[i]).Name);
        end;
        lblPOIDesc.Caption := sl.Text;
      finally
        sl.Free;
      end;
    end
    else
      lblPOIDesc.Caption := '';
  finally
    gpsList.Free;
  end;
  if foundgroupmarker then
     // if Groupmarker was found, change the pointer to indicate that the item could be clicked
    FMapView.Cursor := crHandPoint
  else
    FMapView.Cursor := crDefault;
end;

// MouseUp will perform the "ungroup"-Action, but only if the
// Mouse was'nt moved to far.
// Clicking on the GroupItems will set the center of the Map to the group and
// increase the zoom so that the desired area is using the area of the group.
procedure TForm1.MapViewMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
const
  DELTA = 12;
var
  rArea: TRealArea;
  gpsList: TGpsObjList;
  L: TStrings;
  i: Integer;
  pt : TPoint;
begin
  if (Abs(FMapViewMouseDownX-X) > 2) or
     (Abs(FMapViewMouseDownY-Y) > 2) then Exit; // Mouse was moved, no Group action
  // calculate the Area of the possible Group Item
  pt.X := X-DELTA;
  pt.Y := Y-DELTA;
  rArea.TopLeft := FMapView.ScreenToLatLon(pt);
  pt.X := X+DELTA;
  pt.Y := Y+DELTA;
  rArea.BottomRight := FMapView.ScreenToLatLon(pt);
  // Query for GPSItems in the MapView
  gpsList := FMapPOIList.GetObjectsInArea(rArea);
  try
    if gpsList.Count > 0 then
    begin
      L := TStringList.Create;
      try
        for i:=0 to gpsList.Count-1 do
        begin
          if gpsList[i] is TGpsPointGroupMarker then
          begin  // Found a GroupItem, perform the zoom and exit
            FMapView.Center := TGpsPointGroupMarker(gpsList[i]).RealPoint;
            FMapView.Zoom := FMapView.Zoom + ZoomInPerGroup;
            Exit;
          end;
        end;
      finally
        L.Free;
      end;
    end;
  finally
    gpsList.Free;
  end;
end;

procedure TForm1.pnlMapViewChangeBounds(Sender: TObject);
var
  h : Integer;
begin
  h := lblOSMCopyright.Height;
  lblOSMCopyright.Top := pnlMapView.Height - h;
end;

procedure TForm1.rgDefaultVsQuadTreeClick(Sender: TObject);
var
  i : Integer;
  gpsPt : TGpsPoint;
  oldCursor : TCursor;
begin
  oldCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  try
    if rgDefaultVsQuadTree.ItemIndex = 1 then
    begin
      cbShowPOICaptions.Checked := False;
      // If the user wants to see the default behaviour of the mapview, simply
      // drop all items into the MapViewers GpsItem list
      FMapPOIList.BeginUpdate;
      try
        FMapPOIList.ClearAll; // If compiling fails here, you missed one update,
        // download the current component or use the next line
        // FMapPOIList.Clear(0); // Takes surprisingly long!!

        for i := 0 to High(FPOICoords) do
        begin
          // Create a GpsPoint and add the POI to the list of the MapView
          gpsPt := TGpsPoint.Create(FPOICoords[i].X, FPOICoords[i].Y);
          gpsPt.Name := FPOICaptions[i];
          FMapPOIList.Add(gpsPt, 0);
        end;
      finally
        FMapPOIList.EndUpdate;
      end;
    end
    else // if the grouped behaviour should be shown, simply empty the list
    begin
      FLastMapArea.Init(0,0,0,0); // Force Repaint
      FMapView.Active := False;
      FMapPOIList.BeginUpdate;  // was: FMapView.GPSItems.BeginUpdate;
      try
        FMapPOIList.ClearAll; // If compiling fails here, you missed one update,
        // download the current component or use the next line
        // FMapPOIList.Clear(0); // Takes surprisingly long!!
      finally
        FMapPOIList.EndUpdate;
        FMapView.Active := True;
      end;
    end;
    FMapView.Invalidate;
    FMapView.Repaint;
  finally
    Screen.Cursor := oldCursor;
  end;
end;

{ OnTestQuadTreeGetItemCoord is called, when the QuadTree needs the coordinates of an Item }
function TForm1.OnTestQuadTreeGetItemCoord(Sender: TObject; AItem: TObject;
  var AX, AY: Double): Boolean;
var
  ndx : NativeInt;
begin
  Result := False;
  ndx := NativeInt(AItem)-1;
  if (ndx < 0) or
     (ndx >= Length(FPOICoords)) then Exit;
  AX := FPOICoords[ndx].X;
  AY := FPOICoords[ndx].Y;
  Result := True;
end;
{ OnTestQuadTreeGetItemCaptionEvent is called when the QuadTree needs the caption of an Item }
function TForm1.OnTestQuadTreeGetItemCaptionEvent(Sender: TObject;
  AItem: TObject; var AItemCaption: String): Boolean;
var
  ndx : NativeInt;
begin
  Result := False;
  ndx := NativeInt(AItem)-1;
  if (ndx < 0) or
     (ndx >= FPOICaptions.Count) then Exit;
  AItemCaption := FPOICaptions[ndx];
  Result := True;
end;

{SamplePOIImport Imports the POIs from the sample file
The file is comma separated values in three columns, first longitude,
second latitude (both full degree plus decimal fraction), third Caption
Example:
13.3777900,52.5162700,"[DE-10117] Brandenburger Tor [D-Berlin]"
}
procedure TForm1.SamplePOIImport;
var
  sl, line : TStringList;
  i : Integer;
  llat, llon : Double;
  lineerr, cnt0, cnt1 : Integer;
  fn : String;
  ndx : Integer;
begin
  if FMapPOIList.Count > 0 then
  begin
    FMapPOIList.ClearAll; // If compiling fails here, you missed one update,
    // download the current component or use the next line
    // FMapPOIList.Clear(0); // Takes surprisingly long!!
  end;
  FLastMapArea.Init(0,0,0,0); // Force Repaint
  FTestQuadTree.Clear; // Clear the Tree
  // Clear the imported data
  SetLength(FPOICoords,0);
  FPOICaptions.Clear;
  sl := Nil;
  line := Nil;
  try
    sl := TStringList.Create;
    line := TStringList.Create;
    line.Delimiter:=',';
    line.StrictDelimiter := True;
    line.QuoteChar:='"';
    fn := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName))+SamplePOIFileName;
    if not FileExists(fn) then
      fn := ResolveDots(IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName))+'..\..\..\LQTExampleData\'+SamplePOIFileName);

    // Load the file
    sl.LoadFromFile(fn);
    cnt0 := sl.Count;
    SetLength(FPOICoords,cnt0);
    cnt1 := 0;
    lineerr := -1;
    ndx := 0;
    for i := 0 to sl.Count-1 do
    begin
      line.DelimitedText := sl[i]; // Split one line
      if line.Count < 3 then Continue; // To few information
      // Convert the data
      llon := StrToFloatDef(line[0],-9999);
      llat := StrToFloatDef(line[1],-9999);
      if (llat < OSMMinLatitude) or (llat > OSMMaxLatitude) or
         (llon < OSMMinLongitude) or (llon > OSMMaxLongitude) then
      begin // Track error
        if lineerr < 0 then
          lineerr := i;
        Continue;
      end;
      // Assign the imported data
      FPOICoords[ndx].X := llon;
      FPOICoords[ndx].Y := llat;
      FPOICaptions.Add(line[2]);
      Inc(ndx);
      // Insert the data in the tree
      // We "misuse" the Item : TObject parameter by casting it to an Index to the
      // FPOICoords and FPOICaptions POI-Data
      // Since 0 would be interpreted as Nil, we increment the used index by 1
      // reducing this in the actual usage
      if not Assigned(FTestQuadTree.InsertItem(TObject(NativeInt(i+1)),
                      True, False, llon,llat)) then  // Point is Valid, Area is unvalid, Coords
      begin
        if lineerr < 0 then
          lineerr := i;
      end
      else
        Inc(cnt1);
    end;
    if lineerr >= 0 then
      lblSampleImportResult.Caption := Format('%d lines imported into %d objects. Error in Line %d',[cnt0,cnt1,lineerr])
    else
      lblSampleImportResult.Caption := Format('%d lines imported into %d objects',[cnt0,cnt1]);
  finally
    if Assigned(line) then
      line.Free;
    if Assigned(sl) then
      sl.Free;
    SetLength(FPOICoords, FPOICaptions.Count);
  end;
end;
{ OnMapViewBeforeDrawObjects
    This Event is new and has to be implemented in the
    TMapView.Class
    It will be called in the TMapView.Paint method
    (and there in the method local procedure FullRedraw).
    This point allows any modification in the GPSItems-List of the
    MapViewer without any further side effects, since after
    returning from the event, this list is drawn.

    The implementation here ist the central link between the
    QuadTree and the MapView.
    The displayed map (and some overlapping area) are fractioned into
    OSM-Tiles plus smaller parts of it (defined on the very top of
    this file). If not altered, each 256px Tile is divided into
    4 by 4 32px width rectangles.
    For each of those rectangle the QuadTree is asked how many objects
    are stored in this are. This estimation(!) is very fast, since no actual item
    coordinate is queried, instead the internal counter are fetched.
    If the number is bigger than (if not altered) 4 then the items are grouped
    into an special GroupMarker.
    If equal or less then the items are displayed as generic GpsPoints.
    This limits the total displayed items on a full screen map:
    (1920 x 1080px => 60 x 33 fields => 1980 items.
}

{ CalculateMapPixelCoords
  calculates the coordinates of the displayed map.
  This coordinates are adjusted to the full tiles, that are necesarry
  to cover the displayed map.
  This means:
  On a small zoom level (approx. < 5) than the full earth is returned.
  If the map is in the Cyclic mode, depending of the center position,
  the left-Coordinate may be negative, indicating that the map extends to the
  eastern Hemisphere. If the right coordinate is larger than the MapSize, this
  indicates that the map extends to the western hemishere.
  Example: If the procedure return
  -256,0,256,512
  than full earth is shown in Zoom-Level 1, the date border is in the middle, subsequently
  the eastern hemisphere is displayed left, the western hemisphere is displayed right.
}
procedure TForm1.CalculateMapPixelCoords(var ALeft, ATop, ARight,
  ABottom: Int64; var AOSMMapSize: Int64; var AOfsX, AOfsY: Integer);
var
  pt0 : TPoint;
  osmapsz : Int64;
  wpx,hpx : Int64;
  minx, maxx, miny, maxy : Int64;
  mcl : Double;
  d : Double;
begin
  wpx := FMapView.Width;
  hpx := FMapView.Height;
  osmapsz := round(mvGeoMath.ZoomFactor(FMapView.Zoom)) * TILE_SIZE;
  AOSMMapSize := osmapsz;
  mcl := FMapView.Center.Lon;
  // The pixelcoordinates of the MapView Center on the screen are always on the map, never outside
  pt0 := DegreesToPixelsOSM(FMapView.Zoom,FMapView.Center.Lat,mcl);
  // calculate the offset to the upper-left corner of the OSM-Tile that contains the center
  AOfsX := (pt0.X mod TILE_SIZE);
  AOfsY := (pt0.Y mod TILE_SIZE);

  // Calculate the number of tiles to the left border of the MapView
  d := ((wpx / 2)-AOfsX) / TILE_SIZE;
  minx := Trunc(d);
  if Frac(d) > 0 then // if a fraction exists, add the full tile
    Inc(minx);
  // calculate the pixel coord of the tile. This eleminates the offset for later use
  minx := minx * TILE_SIZE;
  minx := pt0.X - AOfsX - minx;

  // Same for the distance to the right border
  d := ((wpx / 2)+AOfsX) / TILE_SIZE;
  maxx := Trunc(d);
  if Frac(d) > 0 then
    Inc(maxx);
  maxx := maxx * TILE_SIZE;
  maxx := pt0.X - AOfsX + maxx;

  // Catch the extra behaviour of the map
  // if the map is not cyclic (means no crossing of the DateBorder could happen
  if not FMapView.Cyclic then
  begin
    // Limit the left and the right side
    if minx < 0 then
      minx := 0;
    if maxx > osmapsz then
      maxx := osmapsz;
  end;
  // In all cases, if the window is wider than the OSM-Map, than took the Map width
  if (maxx-minx) > osmapsz then
  begin
    minx := 0;
    maxx := osmapsz;
  end;

  // Samme procedure with the vertical size
  d := ((hpx / 2)-AOfsY) / TILE_SIZE;
  miny := Trunc(d);
  if Frac(d) > 0 then
    Inc(miny);
  miny := miny * TILE_SIZE;
  miny := pt0.Y - AOfsY - miny;

  d := ((hpx / 2)+AOfsY) / TILE_SIZE;
  maxy := Trunc(d);
  if Frac(d) > 0 then
    Inc(maxy);
  maxy := maxy * TILE_SIZE;
  maxy := pt0.Y - AOfsY + maxy;

  // And also limit to the maximum extent
  if (maxy-miny) > osmapsz then
  begin
    miny := 0;
    maxy := osmapsz;
  end;
  if miny < 0 then
    miny := 0;
  if maxy > osmapsz then
    maxy := osmapsz;

  // Assign the return values
  ALeft := minx;
  ARight := maxx;
  ATop := miny;
  ABottom := maxy;
end;


end.

