{ A general Quad-Tree for Objects (and arbitrary data)
  Copyright (C) 2023,2024 Ekkehard Domning (www.domis.de)

  License: modified LGPL with linking exception (like RTL, FCL and LCL)

  See the file COPYING.modifiedLGPL.txt, included in the Lazarus distribution,
  for details about the license.

  Version Date        Change
  0.0.1   2023-04-27  First release
  0.0.2   2024-05-25  Bugfixes, changes in visibility of virtual methods
  0.0.3   2024-05-29  Caching for NodeItem-Positions and Area
  0.0.4   2024-08-05  Using new ulazQuadTreeGeometry.pas for all rectangles
  0.0.5   2024-08-24  Bugfixes for area items and cleanups, removing code duplicity
  0.0.6   2024-08-30  Write and Read from Stream
  0.0.7   2024-08-31  Speedup CoordIsInNodeRect
  0.0.8   2024-09-14  OccupiedRect contains all items in one node.
  0.0.9   2024-09-24  Renaming Declarations for streamlining
}
unit uLazQuadTree;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils,
  uLazQuadTreeGeometry;

const
  { DefaultMaxTreeLevel: Each level quaters the area of the level above.
    Subsequently the one Node in Level 25 contains, if used as an earth map,
    roughly one squaremeter (at the equator).
    Keep in mind that the amount of Items that *could* be stored increases very fast.
    The default is about 1280 TerraItems, far beyond what can be actually processed.
    Used as initial value of the MaxQuadTreeLevel property of the QuadTree. }
  DefaultMaxTreeLevel = 25;
  { DefaultMaxNodeItemCount: If a node contains more items, it will be split in subnodes.
    Used as initial value of the MaxQuadNodeItemCount property of the QuadTree.
    if the items are close together, this will be continued until MaxLazQuadTreeLevels is reached}
  DefaultMaxNodeItemCount = 4;

  { Some remarks to the Read and Write from Streams.
    To give as much flexibility in the usage of the Tree, Nodes and NodeItems,
    the Stream Methods are using virtual submethods, which could be overwritten
    by descendands to extend or replace the capabilities of the basic classes.
    In general the written parts are defined as follow
    Bytes  Description
    [0..3] ClassToken a Value that identify the following data
    [4..7] n, Length of the following Data in Bytes
    if n > 0
    [8..8+n] Data of the class

    The QuadTree and the LQTNode write the first 4 Bytes in the *Data* section a
    Tree- or NodeDataVersion, to allow the adaption of changed datasets during the
    progress of the development and the backward compatibility

    The *File* starts always with a 4-byte QuadTreeFileStartCode, to abort loading
    of non data files as early as possible.

    Keeping this in mind, it is possible to have a QuadTree(-Descendant) as a NodeItem in a
    QuadNode of a tree and this item could be read and written with the whole bunch around.
  }
  { QuadTreeFileExtension defines the default file extension of a QuadTree-File }
  QuadTreeFileExtension = '.lqtf';
  { QuadTreeFileStartCode the first few bytes in the File(!) indicating a QuadTree following }
  QuadTreeFileStartCode = $ED6C5154; //ED+'lQT'
  { DefaultQuadTreeToken indicates the trunk with a QuadTree definition the last two bytes could be used
    to identify different QuadTree-Class-Descendants }
  DefaultQuadTreeToken  = $51540000; //'QT'+0000
  { DefaultLQTNodeToken indicates the trunk with a QuadNode definition the last two bytes could be used
    to identify different QuadNode-Class-Descendants }
  DefaultLQTNodeToken  = $514E0000; //'QN'+0000
  { DefaultQuadTreeStreamDataVersion a const that allows modification within the data of a QuadTree and
      the backward compatibilty of writen Streams
    The 4 bytes are used as follow
      AABBCCDD
      ^ ^ ^ ^------ User Tree Minor changes
      | | +---------User Tree Major changes
      | +-----------Unit Tree Minor changes
      +-------------Unit Tree Major changes
    The code in this unit will raise an exception on read from stream if the two Unit
    Bytes are not reflected in the code.
    The two UserBytes are user implementation dependend and ignored.
  }
  DefaultQuadTreeStreamDataVersion = $00000000;
  { DefaultLQTNodeStreamDataVersion a const that allows modification within the data of a QuadNode and
      the backward compatibilty of writen Streams
    The 4 bytes are used as follow
      AABBCCDD
      ^ ^ ^ ^------ User Node Minor changes
      | | +---------User Node Major changes
      | +-----------Unit Node Minor changes
      +-------------Unit Node Major changes
    The code in this unit will raise an exception on read from stream if the two Unit
    Bytes are not reflected in the code.
    The two UserBytes are user implementation dependend and ignored.
  }
  DefaultLQTNodeStreamDataVersion = $00000000;

  // lqtStreamDataVersionXXXMasks can be used to "and" with the received Mask Value
  lqtStreamDataVersionUnitMajorMask = $FF000000;
  lqtStreamDataVersionUnitMinorMask = $00FF0000;
  lqtStreamDataVersionUserMajorMask = $0000FF00;
  lqtStreamDataVersionUserMinorMask = $000000FF;
  // lqtStreamDataVersionXXXShr can be used to shift the "anded" Value to the right
  lqtStreamDataVersionUnitMajorShr  = 24;
  lqtStreamDataVersionUnitMinorShr  = 16;
  lqtStreamDataVersionUserMajorShr  =  8;
  lqtStreamDataVersionUserMinorShr  =  0;

  // StreamReadOptions Flags
  // The following Flags could be "or"ed together to enable the correspondend Option
  { srofLazyLengthProcessing
     If the length of a Trunk is longer than read, the remainder is skipped without error}
  srofLazyLengthProcessing = $00000001;

type
  ELQTException = class(Exception);
  ELQTStreamException = class(ELQTException);

  { TLQTNodeChildLocation the Index of the four ChildNodes}
  TLQTNodeChildLocation = (nclNW,nclNE,nclSE,nclSW); //Clockwise Left, Top, Right, Bottom

  // Forward declaration
  TLazQuadTree = class;

  { TLQTNodeItemsRec is used by the Node to store the Items and chache there positions }
  TLQTNodeItemsRec = record
    Item : TObject;
    PointValid : Boolean;
    RectValid : Boolean;
    Point : TLQTPoint;
    Rect : TLQTRect;
  end;

  // Forward declaration
  TLQTNode = class;

  { TLQTNodeEnumerationCallBack
     This event/callback is used in the enumeration functions.
     It will be called for each found Node.
    Parameters
      Sender is the tree to which the node belongs
      ANode contains the found node.
      AUserData is passed unchanged from the function call and
       may be used by the callback-function to identify data, avoiding the usage of
      global variable.}
  TLQTNodeEnumerationCallBack = function (Sender : TObject;
                                                    ANode : TLQTNode;
                                                    AUserData : Pointer) : Boolean of object;
  { TLQTItemEnumerationCallBack
     This event is used in the enumeration function.
     It will be called for each found Item.
    Parameters
     Sender is the tree to which the node belongs
     ANode contains the found node.
     ANodeItem contains the Item.
     ANodeItemIndex contains the index of the ANodeItem in the NodeItems-Array.
     AUserData is passed unchanged from the function call and
      may be used by the callback-function to identify data, avoiding the usage of
      global variable.}
  TLQTItemEnumerationCallBack = function (Sender : TObject;
                                          ANode : TLQTNode;
                                          ANodeItem : TObject;
                                          ANodeItemIndex : Integer;
                                          AUserData : Pointer) : Boolean of object;


  { TLQTAgglomerationRec is used by the Agglomeration method
      Point contains the center coordinate of the agglomeration (= center of the node rectangle)
      Count the estimated number of items within the agglomeration. }
  TLQTAgglomerationRec = record
    Point : TLQTPoint;
    Count : Double;
    Node : TLQTNode;
  end;
  { PLQTAgglomerationArray, an pointer to an array of TLazQuadTreeAgglomerationRec }
  PLQTAgglomerationArray = ^TLQTAgglomerationArray;
  { TLQTAgglomerationArray an array of TLazQuadTreeAgglomerationRec }
  TLQTAgglomerationArray = array of TLQTAgglomerationRec;

  { TLQTAgglomerationEnumerationCallBack is used in the Agglomeration Enumeration
     and is called for every build agglomeration
    Parameters
      ANode contains the operated node.
      ATreeAgglomerationRec contains the agglomerated data for this node
      AUserData is passed unchanged from the function call and
       may be used by the callback-function to identify data, avoiding the usage of
       global variable.}
  TLQTAgglomerationEnumerationCallBack = function (ANode : TLQTNode;
                                                   const ATreeAgglomerationRec : TLQTAgglomerationRec;
                                                   AUserData : Pointer) : Boolean of object;


  { TLQTNode
    A node in the Tree.
    The node is assumed to be the root node, if the ParentNode is not assigned.
    The node can contain data items. These data items are generic TObjects.
    There actual Coordinates are fetched via an Event in the Tree
  }
  TLQTNode = class(TObject)
  private
    FNodeToken : Cardinal;
    FNodeLevel : Integer;
    FNodeRect : TLQTRect;
    FOwner : TLazQuadTree;
    FParentNode : TLQTNode;
    FChildNodes : array[TLQTNodeChildLocation] of TLQTNode;
    FNodeItems : array of TLQTNodeItemsRec;
    FNodeTouched : Boolean;
    FOccupiedRect : TLQTRect;
    procedure SetTotalNodeItemsCount(Value : Integer);
  protected
    { DoFetchItemCoords performs the Notification to get the Item Coords, if not already cached
      The result ist true if at least on of the two coords (point or Rect) the valid }
    function DoFetchItemCoords(const AItemIndex : Integer) : Boolean;virtual;
    function GetNodeItemsLength : Integer;virtual;
    procedure SetNodeItemsLength(Value : Integer);virtual;
    function GetNodeItems(Index : Integer) : TObject;virtual;
    procedure SetNodeItems(Index : Integer; Value : TObject);virtual;
    function GetAssignedNodeItemsCount : Integer;virtual;
    function GetAssignedNodeItems(Index : Integer) : TObject;virtual;
    function GetChildNodes(Index : TLQTNodeChildLocation) : TLQTNode;virtual;
    function GetChildRects(Index : TLQTNodeChildLocation) : TLQTRect;virtual;
    { MoveItemToParentNode moves the specified item to the parent item
      The function result is true, if the item has been move, false otherwise.}
    function MoveItemToParentNode(const AItemIndex : Integer; out ATargetItemIndex : Integer) : Boolean;virtual;
    { MoveItemToChildNode moves the specified item to a fitting childnode
      The result is true if the movement was succesful, false if not. In the latter case
      the item was not moved.}
    function MoveItemToChildNode(const AItemIndex : Integer; out ATargetChildNode : TLQTNode; out ATargetItemIndex : Integer) : Boolean;virtual;
    { NodeItemToListEnumerationCallBack is used by the some ListItems methods
           to accumlate the found items into a list.
           Caution: The function assumes that the parameter UserData is a pointer to a TList Descendand!
           The method is used only internal but is declared protected and virtual for
           future expansions.
           If the parameter ANodeItem is Nil, nothing is inserted into the list
           The result true if the AUserData is assigned, else false}
    function NodeItemToListEnumerationCallBack(Sender : TObject;
                                               ANode : TLQTNode;
                                               ANodeItem : TObject;
                                               ANodeItemIndex : Integer;
                                               AUserData : Pointer) : Boolean;virtual;
    { NodeAgglomerationEnumerationCallBack is used by the AgglomerateNodeAsList method
       to store the agglomeration data in the data array
       Caution: The function assumes that the parameter UserData is a Pointer from the type
         PLQTAgglomerationArray.
       The method is used only internal but is declared protected and virtual for
       future expansions.
       If the parameter ANodeItem is Nil, nothing is inserted into the list
       The result true if the AUserData is assigned, else false}
    function NodeAgglomerationToArrayEnumerationCallBack(ANode : TLQTNode;
                                                  const ATreeAgglomerationRec : TLQTAgglomerationRec;
                                                  AUserData : Pointer) : Boolean;
    { the following Methods could be overwritten by derived classes to add extended
      data }
    { WriteNodeHeader writes the Header to the stream, containing the StartCode and
      the Version Code. Could be overwritten to write a modified StreamDataVersion }
    procedure WriteStreamNodeHeader(const AStream : TStream;
                              const ATreeStreamDataVersion : Cardinal;
                              const ANodeStreamDataVersion : Cardinal);virtual;
    { WriteNodeProperties writes the property fields to the stream }
    procedure WriteStreamNodeProperties(const AStream : TStream;
                                  const ATreeStreamDataVersion : Cardinal;
                                  const ANodeStreamDataVersion : Cardinal);virtual;
    { WriteNodeChildren writes the property fields to the stream }
    procedure WriteStreamNodeChildren(const AStream : TStream;
                                      const ATreeStreamDataVersion : Cardinal;
                                      const ANodeStreamDataVersion : Cardinal);virtual;
    { WriteNodeItems writes the items to the stream }
    procedure WriteStreamNodeItems(const AStream : TStream;
                            const ATreeStreamDataVersion : Cardinal;
                            const ANodeStreamDataVersion : Cardinal);virtual;
    { WriteNodeItem write one items to the stream }
    procedure WriteStreamNodeItem(const AStream : TStream;
                           const AQuadTreeStreamDataVersion : Cardinal;
                           const ANodeStreamDataVersion : Cardinal;
                           const AItemIndex : Integer);virtual;

    { ReadNodeHeader reads the header and the Version.
        raises an exception if the Header Code is wrong or the version incompatible }
    procedure ReadStreamNodeHeader(const AStream : TStream;
                               const ATreeStreamDataVersion : Cardinal;
                               out ANodeStreamDataVersion : Cardinal);virtual;
    { ReadNodeProperties reads the property fields from the stream }
    procedure ReadStreamNodeProperties(const AStream : TStream;
                                 const ATreeStreamDataVersion : Cardinal;
                                 const ANodeStreamDataVersion : Cardinal);virtual;
    { ReadNodeProperties reads the children fields from the stream }
    procedure ReadStreamNodeChildren(const AStream : TStream;
                                     const ATreeStreamDataVersion : Cardinal;
                                     const ANodeStreamDataVersion : Cardinal);virtual;

    { ReadNodeItems reads the items from the stream }
    procedure ReadStreamNodeItems(const AStream : TStream;
                            const ATreeStreamDataVersion : Cardinal;
                            const ANodeStreamDataVersion : Cardinal);virtual;
    { ReadNodeItems reads the items from the stream }
    procedure ReadStreamNodeItem(const AStream : TStream;
                           const ATreeStreamDataVersion : Cardinal;
                           const ANodeStreamDataVersion : Cardinal;
                           const AItemIndex : Integer);virtual;
    { UpdateOccupiedRectLocal Updates the UpdateOccupiedRect,
        by including the UpdateOccupiedRect
        from the Children and Items contained by this node.
        It returns true, if the bounding box changed.
        The function will *not* recurse downwards, just working local in this node }
    function UpdateOccupiedRectLocal : Boolean;virtual;
    { UpdateOccupiedRect calls UpdateOccupiedRectLocal and if something changed,
       it calls the parent node to update the chain upwards
       It returns true, if the bounding box changed. }
    function UpdateOccupiedRect : Boolean;virtual;

    { GetOccupiedRect returns the FOccupiedRect field value.
       The method could be overwritten to implement a different way to obtain the
       occupied area of the item in the tree}
    function GetOccupiedRect : TLQTRect;virtual;
  public
    FTotalNodeItemsCount : Integer;

    { Owner: The Tree this node belongs to }
    property Owner : TLazQuadTree read FOwner;
    { ParentNode: The node, where this node is one of the four ChildNodes.
      If Nil, this node is the RootNode of the Tree }
    property ParentNode : TLQTNode read FParentNode;
    { ChildNodes: Indexed array of the four ChildNode. Each Node covers one quadrant of the
      Rect of this node }
    property ChildNodes[Index : TLQTNodeChildLocation] : TLQTNode read GetChildNodes;
    property ChildNodeNE : TLQTNode index nclNE read GetChildNodes;
    property ChildNodeSE : TLQTNode index nclSE read GetChildNodes;
    property ChildNodeSW : TLQTNode index nclSW read GetChildNodes;
    property ChildNodeNW : TLQTNode index nclNW read GetChildNodes;
    { ChildRects: Contains the definition of the for quadrants (=Rects) of the ChildNodes.
      The virtual getter function GetChildRects might be overwritten in derived classes to
      implement a different distribution to adapt to different projections.
    }
    property ChildRects[Index : TLQTNodeChildLocation] : TLQTRect read GetChildRects;
    { NodeLevel: The depth of this node. 0 for the RootNode }
    property NodeLevel : Integer read FNodeLevel;
    { NodeRect: The Rect covered by this node.
      Care mus be taken, by interpreting the values of the Rect.
      Depending of the configuration of the WorldRect of the related tree, the value
      Top is greater or lesser than Bottom.
      The NodeRect is guaranteed never passing the Borders of the Tree's WorldRect.
    }
    property NodeRect : TLQTRect read FNodeRect;
    property Left : Double read FNodeRect.Left;
    property Top : Double read FNodeRect.Top;
    property Right : Double read FNodeRect.Right;
    property Bottom : Double read FNodeRect.Bottom;
    { OccupiedRect comprise all items of this node and all subnodes.
       If TotalNodeItems = 0, the the result is an empty rectangle }
    property OccupiedRect : TLQTRect read GetOccupiedRect;

    { TotalNodeItemsCount: Contains the maintained Number of assigned Items of
      this Node and all Child-nodes to the bottom }
    property TotalNodeItemsCount : Integer read FTotalNodeItemsCount write SetTotalNodeItemsCount;
    { NodeItemsLength contains the Length of the NodeItems array.
      Caution: When writing a smaller value than the current length,
       the ItemsArePacked and if still need to be shortended, some Items are freed,
       if the related Tree owns the objects! }
    property NodeItemsLength : Integer read GetNodeItemsLength write SetNodeItemsLength;
    { NodeItems gives a raw access to the NodeItems array.
       Caution: overwriting existing Items will not free them, this has to be done manually if needed!
       Caution: placing items into the wrong node is not checked!}
    property NodeItems[Index : Integer] : TObject read GetNodeItems write SetNodeItems;
    { AssignedNodeItemsCount gives the number of assigned items in this node. }
    property AssignedNodeItemsCount : Integer read GetAssignedNodeItemsCount;
    { AssignedNodeItems gives access to the assigned items in the raw NodeItems array }
    property AssignedNodeItems[Index : Integer] : TObject read GetAssignedNodeItems;
    { DefaultLQTNodeToken a value that is used during the Stream read and write methods
      identifying the class of the QuadNode }
    property NodeToken : Cardinal read FNodeToken;
    { CountNodes recurses through the Childnodes and return the sum of all nodes below.
      The minimum result is one, since this node is assigned. }
    function CountNodes : Integer;
    { CreateOrGetChildNode creates a childnode unless one exists in this location.
       The result is the new created or the existing ChildNode from this position.
       CAUTION: If a new childnode is created, the new node tooks all belonging items
        from his parent node (this node), thus the number of items in this node may
        be reduced down to zero. A call of AssignedNodeItemsCount might be necesarry!}
    function CreateOrGetChildNode(const AChildLocation : TLQTNodeChildLocation) : TLQTNode;
    { DeleteChildNode deletes the childnode (if existing) and all subchildnodes.
      The function moves all NodeItems from the deleted ChildNodes to this Node. }
    procedure DeleteChildNode(const AChildLocation : TLQTNodeChildLocation);
    { CoordIsInNodeRect returns true if the Point AX, AY is located within the NodeRect. }
    function CoordIsInNodeRect(const AX, AY : Double) : Boolean;virtual;
    { FindCoordNode iterates through the part of the tree that is covered by this node.
        returns the deepest (leaf)node which contains the coordinate.
        The result is nil, if the coordinate is not located within this part of the tree.}
    function FindCoordNode(const AX, AY : Double) : TLQTNode;virtual;

    { NodeRectOverlapTest returns the Test result of the comparism of ARect with the NodeRect. }
    function NodeRectOverlapTest(const ARect : TLQTRect) : TLQTRectOverlapKind;virtual;
    { OccupiedRectOverlapTest returns the Test result of the comparism of ARect with the OccupiedRect. }
    function OccupiedRectOverlapTest(const ARect : TLQTRect) : TLQTRectOverlapKind;virtual;

    { NodeRectOverlap returns true, if ARect overlap at least partly the NodeRect. }
    function NodeRectOverlap(const ARect : TLQTRect) : Boolean;virtual;
    { AddNodeItem insert the AItem into the node items array.
      The optional parameters could be used to speed up the processing, since
        a callback to query the data is suppressed.
        APointValid, ARectValid, APointX, APointY
        ARectLeft, ARectTop, ARectRight, ARectBottom
    }
    function AddNodeItem(const AItem : TObject;
                         const APointValid : Boolean = False;
                         const ARectValid : Boolean = False;
                         const APointX : Double = 0.0;
                         const APointY : Double = 0.0;
                         const ARectLeft : Double = 0.0;
                         const ARectTop : Double = 0.0;
                         const ARectRight : Double = 0.0;
                         const ARectBottom : Double = 0.0
      ) : Integer;virtual;
    { ExtractNodeItem removes an Item identified by its ItemIndex from the internal storage,
        the current field in the Items-array is set to Nil
       Parameter: AItemIndex, contains the index of the raw data storage
        Internal counters are updated.
        the result contains the removed item, Nil if no item has been removed.
        The item has to be freed by the user, if the Tree Owns the Items }
    function ExtractNodeItem(const AItemIndex : Integer) : TObject;
    { ExtractNodeItem removes an Item identified by its Adress from the internal storage,
        the current field in the Items-array is set to Nil
       Parameter: ANodeItem, contains the (pointer of the) item which has to be removed.
        Internal counters are updated.
        the result contains the removed item, Nil if the item is not found in this node
        The item has to be freed by the user, if the Tree Owns the Items }
    function ExtractNodeItem(const ANodeItem : TObject) : TObject;
    { SplitNode split the node into child nodes if necesarry. Recurses down with splitting, if necesarry.
      Returns true, if at least one node is splitted. }
    function SplitNode : Boolean;virtual;
    { PackNode tries to remove unused ChildNodes. Recurses down with packing, if useful.
      Return true if at least one node was combined }
    function PackNode : Boolean;virtual;
    { ExpandNode place all point items into leaf nodes, creating those ChildNodes, if necesarry.
      Recurses down with expanding, if needed.
      Area items will remain here or, if fitting into a smaller child node, moved
      Return true if at least one node was expanded }
    function ExpandNode : Boolean;virtual;
    { NodeHasZeroDataItems  recurses through all subnodes and counts all populated DataItems.
      If there are none, the result is true. }
    function NodeHasZeroDataItems : Boolean;virtual;
    { HasChildNodes returns true if at least one of the four possible ChildNodes exists }
    function HasChildNodes : Boolean;
    { IsDescendantOf returns True if this node is an descendand of the passed node,
        else the Result is false}
    function IsDescendantOf(ANode : TLQTNode) : Boolean;
    { NodeItemIndex returns the Index of the passed Item, -1 if not found}
    function NodeItemIndex(const AItem : TObject) : Integer;
    { PackNodeItems removes unused and duplicate items.
       If APackToZero is true, the NodeItems array are set to the assigned items count}
    function PackNodeItems(const APackToZero : Boolean = False) : Boolean;virtual;
    { GetNodesParentChildLocation determine the ChildLocation of this node in its Parent's
        ChildNode array. The Value is valid if the return value of the function is true.
        The return value false means, that the Node is the Root node and no parent node is
        assignd.}
    function GetNodesParentChildLocation(var AChildLocation : TLQTNodeChildLocation) : Boolean;
    { AgglomerateNodeAsEnumeration build a summary of the Items within this node and his children.
        The usage is e.g. to build clickable items on a map, which represents a bunch of items
        below, to prevent that thousands of items will be drawn, while a "10K"-sign would
        do the job.
        Another usage would be to colorize parts of an image recording to the "density" of the
        Agglomerations.
       Parameters
        AAgglomerationDepth determines how many node-leves below should be taken into account.
         Due to the math (every level contains 4 times nodes than the parents level) a value
         of 1 will return up to 4 agglomeration records, 2 will return 16 and 3 = 64.
         If one Node cover an square of e.g. 256 pixel (as one OSM-Tile does), a level of
         3 will place up to 64 ( = 16 * 16) agglomeration records on this tile.
        AMinimumItemNumberInAgglomeration will control the minimum ammount in one agglomeration.
         The value should not be less than MaxQuadNodeItemCount. A good value depends on the
         Data structure and the tile size. If one agglomeration contains a square of 16 x 16
         pixel, the value could be 16 to place (in average) 4 x 4 items in this area.
        AItemEnumFunc,
        AAgglomerationEnumFunc both callback function must be provided to allow a sucessful
         operation. Each function will be called once for every item or build agglomeration.
        AItemUserData,
        AAgglomerationUserData are passed unmodified to the callback functions to allow the management
          of the passed data.
    }
    function AgglomerateNodeAsEnumeration(const AAgglomerationDepth : Integer;
                              const AMinimumItemNumberInAgglomeration : Integer;
                              const AItemEnumFunc : TLQTItemEnumerationCallBack;
                              const AItemUserData : Pointer;
                              const AAgglomerationEnumFunc : TLQTAgglomerationEnumerationCallBack;
                              const AAgglomerationUserData : Pointer) : Boolean;virtual;
    { AgglomerateNodeAsList build a summary of the Items within this node and his children.
        The usage is to build clickable items on a map, which represents a bunch of items
        below, to prevent that thousands of items will be drawn, while a "10K"-sign would
        do the job.
       Parameters
        AAgglomerationDepth determines how many node-leves below should be taken into account.
         Due to the math (every level contains 4 times nodes than the parents level) a value
         of 1 will return up to 4 agglomeration records, 2 will return 16 and 3 = 64.
         If one Node cover an square of e.g. 256 pixel (as one OSM-Tile does), a level of
         3 will place up to 64 ( = 16 * 16) agglomeration records on this tile.
        AMinimumItemNumberInAgglomeration will control the minimum ammount in one agglomeration.
         The value should not be less than MaxQuadNodeItemCount. A good value depends on the
         Data structure and the tile size. If one agglomeration contains a square of 16 x 16
         pixel, the value could be 16 to place (in average) 4 x 4 items in this area
        AAgglomerationArray will be populated with the agglomeration records
        AItemList must be created prior and freed after the call by the using code.
         The list will contain the items outside of the agglomerations }
    procedure AgglomerateNodeAsList(const AAgglomerationDepth : Integer;
                              const AMinimumItemNumberInAgglomeration : Integer;
                              var AAgglomerationArray : TLQTAgglomerationArray;
                              const AItemList : TList);
    { RebuildOccupiedRect iterate through all Nodes and Items and refreshes the Nodes FOccupiedRect.
        This method might be useful after a manually reconfiguration of the Tree. }
    procedure RebuildOccupiedRect;
    { NodeAsString returns some information of the current node in a string }
    function NodeAsString : String;virtual;
    { NodeAsStrings returns the content of the Node and its children in
      readable form for debugging purpose. }
    procedure NodeAsStrings(const AStrings : TStrings);virtual;
    { Clear recursivly delete all Childnodes. If the related Tree OwnsObjects,
      all containing Items are freed }
    procedure Clear;

    { WriteNodeToStream writes this Node and all Children into the stream }
    procedure WriteNodeToStream(const AStream : TStream; const AQuadTreeStreamDataVersion : Cardinal);virtual;
    { ReadNodeFromStream read the content of this node and create and read all Child-Nodes }
    procedure ReadNodeFromStream(const AStream : TStream;
                                 const AQuadTreeStreamDataVersion : Cardinal;
                                 const AStreamReadOptions : Cardinal = 0);virtual;

    { Create the Node within the owning Tree, a Parent and an covered Rect.
      Caution: It is not allowed that Parent- and Child-node have different Owners.
      Caution: It is assumed that the NodeRect contains a part of the Rect of the Parent,
      in a way, that the Rects of the four children covers exactly the Rect of the parent node.
      Saying this, Create will be used in very rare cases outside this unit. }
    constructor Create(const AOwner : TLazQuadTree; const AParentNode : TLQTNode; const ALQTRect : TLQTRect);
    { Destroy frees the Node. Clear is called prior inherited destroy, causing all
      Items are freed, if the owning tree owns the items, and all child nodes are freed. }
    destructor Destroy;override;
  end;

  { TLQTGetItemCoordEvent
    An event which has to return an X and a Y Coordinate for the given ItemIndex.
    Return True if the return values AX and AY are valid, False if not. }
  TLQTGetItemCoordEvent = function (Sender : TObject; AItem : TObject; var AX, AY : Double) : Boolean of object;
  { TLQTGetItemRectEvent
    An event which may return an Rect which contains the entire item. If the item has no
    Rect, the Event should return false.
    Return True if the return value AItemRect is valid, False if not. }
  TLQTGetItemRectEvent = function (Sender : TObject; AItem : TObject; var AItemRect : TLQTRect) : Boolean of object;
  { TLQTGetItemCaptionEvent
    An event which can return a Caption (title, descriptive string) of the item.
    If the item has not such a Caption than the result shlould return false
    Return True if the Caption is valid, False if not. }
  TLQTGetItemCaptionEvent = function (Sender : TObject; AItem : TObject; var ACaption : String) : Boolean of object;
  { TLQTReadStreamItemEvent
    The event is used within the tree stream reading method.
    An event which has to read an Object from the given Stream.
    Depending from the Tree.OwnsObjects property the returned object has to be
    a real created object (True) or arbitrary data (False).
    The first DWord of the Stream contains a UserValue <> 0 that describes
    the following data. Most common is this an UserID where the User code
    knows what Object to create.
    The second DWord of the Stream contains the length of the remaining data.
    If this value is 0, no bytes are following.
  }
  TLQTReadStreamItemEvent = function (AOwner : TLazQuadTree;
                                              ANode : TLQTNode;
                                              AItemIndex : Integer;
                                              AStream : TStream) : TObject of object;
  { TLQTWriteStreamItemEvent
    The event is used within the tree stream reading method.
    An event which has to write an Object to the given Stream.
    The data is user dependend, but has to follow the general rule of writing
    data into the .lqtf-File.
    The first DWord that is written to the Stream mus be an UserID which allows
    the User code to reconstruct the data on the read procedure.
    The second DWord will contain the dta length of the remaining data.
    To ease the write process this value is calculated by the caller, but
    the space must reserved, by writing an DWord value 0
  }
  TLQTWriteStreamItemEvent = procedure(AOwner : TLazQuadTree;
                                              ANode : TLQTNode;
                                              AItem : TObject;
                                              AItemIndex : Integer;
                                              AStream : TStream) of object;
  { TLQTCreateNodeEvent
     An Event which has to return a new created QuadNode. This event could be used to insert derived
     nodes into the QuadTree.
    Parameter
     AOwner the tree where the node belongs to.
     AParentNode the parent node (if Assigned, otherwise RootNode)
     ALQTRect the NodeRectangle (see ANodeToken) for an important remark.
     ANodeToken if the parameter is 0 then the function has to create an
      node by the provided information in the passed Owner and ParentNode.
      The function may decide to return Nil, in this case the caller will create
      a default Node.
      If the parameter is not 0 then the value is derived from
      DefaultLQTNodeToken = $81780000, with a different value at the lower two bytes.
      In this case the passed value comes from a Stream which is about to be read and
      the value identify the class which has to be created.
      In this case the function has to create a class, which is able to read the
      data from the stream. If this is not possible return Nil
     Result the creeated class.
      If the ANodeToken is 0 then Nil will be fine, else the stream reading
      process ends with an Exception.
  }
  TLQTCreateNodeEvent = function (AOwner : TLazQuadTree;
                                  AParentNode : TLQTNode;
                                  ALQTRect : TLQTRect;
                                  ANodeToken : Cardinal = 0) : TLQTNode of object;
  { TLQTGetNodeChildRectEvent
    This event is called whenever a queried ChildRect Size is variable, e.g. not fixed
    by other childrens positions. It allows to modify the suggested ChildRect.
    The purpose is to shift the borders between the newly created children so that a
    uniform distribution of the items between the nodes are possible. If using non point
    items (e.g. Shapes of countries, tracks) a good selection of borders, help to avoid
    inefficent long list of Items at a high level. On a globe, the 0°-Longitude crosses a
    long list of countries, which by default all will be placed on the root node.
    By moving the division to -25° this could be reduced to one country.
    Caution: Asking the AParentNode for other ChildRects may cause a recursive call of this event!
    Asking the ParentNode for its NodeRect is always possible since this Rect is fixed.
    The parameters AChangeAllowedTop, ... , ...Left indicates, which of the field of the
    AChildRect could be modified. Modification of the not allowed fields are ignored.
  }
  TLQTGetNodeChildRectEvent = procedure (AOwner : TLazQuadTree;
                                               AParentNode : TLQTNode;
                                               AChildLocation : TLQTNodeChildLocation;
                                               AChangeAllowedTop, AChangeAllowedRight, AChangeAllowedBottom, AChangeAllowedLeft : Boolean;
                                               var AChildRect : TLQTRect) of object;
  { TLazQuadTree
    A QuadTree for Lazarus.
    Organizes arbitrary Items who are ordered on a plane by X and Y coordinates for
    a fast access for a given Rect.
    The TLazQuadTree could be used to store tenthousands of items and access them very fast,
    because items that are not within the desired Rect are sorted out quickly.
    To allow the processing of any kind of objects or data, an event is used
    where the user has to provide the location or Rect of a specific item in the tree.
    The tree could be used for a flat world with borders on all four edges or
    a cylindric world, where the right and left edge are connected. Additionally the
    orientation of the Y-Axis could be adjusted, to cover cartesian-coordinate-systems
    (like Maps or mathematical function diagrams, where the Y value increases when going upwards) or
    computer-coordinate-system (like Bitmaps or Editors, where the Y increases when going downwards).
    To aggregate data on smaller magnification (e.g. showing bigger parts of the plane)
    a sum of TotalItems in the (Sub-)Nodes is maintained. This allowes, instead of a
    search of a bunch of unusefull Items, the display of a Placeholder/Sum of the
    items which are hidden.
  }
  TLazQuadTree = class(TObject)
  private
    FTreeToken : Cardinal;
    FRootNode : TLQTNode;
    FOwnsObjects: Boolean;
    FUpdateCount : Integer;
    FWorldModel : TLQTWorldModel;
    FMaxTreeLevel : Integer;
    FMaxNodeItemCount : Integer;
    FGetItemCoordEvent : TLQTGetItemCoordEvent;
    FGetItemRectEvent : TLQTGetItemRectEvent;
    FGetItemCaptionEvent : TLQTGetItemCaptionEvent;

    FCreateNodeEvent : TLQTCreateNodeEvent;
    FReadStreamItemEvent : TLQTReadStreamItemEvent;
    FWriteStreamItemEvent : TLQTWriteStreamItemEvent;

    function GetWorldXAxisLeftPlus : Boolean;
    function GetWorldYAxisUpPlus : Boolean;
  protected
    function GetWorldRect: TLQTRect;virtual;
    {InternalCreateQuadNode creates a new node.
       If FCreateNodeEvent is assigned, it is called to allow
       Users to create Nodes of a derived class.
       If this failed, a Node of the type TLQTNode is created and returned.
    }
    function InternalCreateNode(const AParentNode : TLQTNode;
                                    const ALQTRect : TLQTRect;
                                    const ANodeToken : Cardinal = 0) : TLQTNode;virtual;
    // The following "Internal" methods, implements the recursive variants of
    // the non-"Internal" methods in the public section.
    // They expect nonwrapped rectangles as ALQTRect
    // E.g. the public EnumerateNodesInRect will call the private InternalEnumerateNodesInRect
    // with the RootNode as the parameter
    // Caution: All functions expect an unwrapped rectangle
    function InternalEnumerateNodesInRect(
                                    const ANode : TLQTNode;
                                    const AUnwrappedRect : TLQTRect;
                                    const AEnumCallBackFunc : TLQTNodeEnumerationCallBack;
                                    const AUserData : Pointer) : Boolean;virtual;

    function InternalEnumerateItemsInRect(
                                    const ANode : TLQTNode;
                                    const AUnwrappedRect : TLQTRect;
                                    const AEnumCallBackFunc : TLQTItemEnumerationCallBack;
                                    const AUserData : Pointer) : Boolean;virtual;

    function InternalEnumerateItemsInParentNodes(
                                             const ANode : TLQTNode;
                                             const AUnwrappedRect : TLQTRect;
                                             const AEnumCallBackFunc : TLQTItemEnumerationCallBack;
                                             const AUserData : Pointer) : Boolean;virtual;

    function InternalCountNodesInRect(const ANode : TLQTNode;
                                                 const AUnwrappedRect : TLQTRect) : Integer;virtual;
    function InternalCountAssignedNodeItemsInRect(const ANode : TLQTNode;
                                                 const AUnwrappedRect : TLQTRect) : Integer;virtual;
    function InternalEstimatedCountOfAssignedItemsInRect(const ANode : TLQTNode;
                                                 const AUnwrappedRect : TLQTRect;
                                                 const AItemThreshold : Integer) : Double;virtual;
    { DoGetItemCoord: Fetches the coordinates for the given AItem.
      The coordinates (AX, AY) are valid if the return value is true.
      The function calls the assigned GetItemCoordEvent.
      Derived class may overwrite the method and implement a different approach to get the coordinates. }
    function DoGetItemCoord(AItem : TObject; var AX, AY : Double) : Boolean;virtual;

    { DoGetItemRect: Fetches the covered Rect for the given AItem.
      The AItemRect are valid if the return value is true.
      The function calls the assigned GetItemRectEvent.
      Derived class may overwrite the method and implement a different approach to get the Rect. }
    function DoGetItemRect(AItem : TObject; var AItemRect : TLQTRect) : Boolean;virtual;

    { DoGetItemCaption: Fetches the Caption of the given AItem.
      If the AItemCaption is valid, the result is True.
      The function call the assigned GetItemCaptionEvent.}
    function DoGetItemCaption(AItem : TObject; var AItemCaption : String) : Boolean;virtual;

    { TreeItemToListEnumerationCallBack is used by the several ListItems methods
       to accumlate the found items into a list.
       Caution: The function assumes that the parameter UserData is a pointer to a TList Descendand!
       The method is used only internal but is declared protected and virtual for
       future expansions.
       If the parameter ANodeItem is Nil, nothing is inserted into the list
       The result true if the AUserData is assigned, else false}
    function TreeItemToListEnumerationCallBack(Sender : TObject;
                                               ANode : TLQTNode;
                                               ANodeItem : TObject;
                                               ANodeItemIndex : Integer;
                                               AUserData : Pointer) : Boolean;virtual;
    { TreeNodeToListEnumerationCallBack is used by the ListNode methods to accumulate
      found Nodes into a list.
      Caution: The function assumes that the parameter UserData is a pointer to a TList Descendand!
      The method is used only internal but is declared protected and virtual for
      future expansions.
      If the parameter ANodeItem is Nil, nothing is inserted into the list
      The result true if the AUserData is assigned, else false}
    function TreeNodeToListEnumerationCallBack(Sender : TObject;
                                               ANode : TLQTNode;
                                               AUserData : Pointer) : Boolean;virtual;
    { the following Methods could be overwritten by derived classes to add extended
      data }
    { WriteFileStreamStart will write the QuadTreeFileStartCode to the passed Stream}
    procedure WriteFileStreamStart(const AStream : TStream);virtual;
    { WriteStreamHeader writes the Header to the stream, containing the StartCode and
      the Version Code. Could be overwritten to write a modified StreamDataVersion }
    procedure WriteStreamTreeHeader(const AStream : TStream; const AQuadTreeStreamDataVersion : Cardinal);virtual;
    { WriteTreeProperties writes the property fields to the stream }
    procedure WriteStreamTreeProperties(const AStream : TStream; const AQuadTreeStreamDataVersion : Cardinal);virtual;
    { WriteRootNode write the RootNode }
    procedure WriteStreamTreeRootNode(const AStream : TStream; const AQuadTreeStreamDataVersion : Cardinal);virtual;
    { ReadFileStreamStart read the QuadTreeFileStartCode from passed Stream.
        if the value is different from the const, an exception is raised}
    procedure ReadFileStreamStart(const AStream : TStream);virtual;
    { ReadStreamHeader reads the header and the Version.
        raises an exception if the Header Code is wrong or the version incompatible }
    procedure ReadStreamTreeHeader(const AStream : TStream; out AQuadTreeStreamDataVersion : Cardinal);virtual;
    { ReadStreamTreeProperties reads the property fields from the stream }
    procedure ReadStreamTreeProperties(const AStream : TStream; const AQuadTreeStreamDataVersion : Cardinal);virtual;
    { ReadRootNode calls the StreamRed method of the root node }
    procedure ReadStreamTreeRootNode(const AStream : TStream; const AQuadTreeStreamDataVersion : Cardinal);virtual;

    { GetOccupiedRect returns the OccupiedRect of the RootNode.
       The method could be overwritten to implement a differen way to obtain the
       occuopied area of the items in the tree}
    function GetOccupiedRect : TLQTRect;virtual;
    { GetActualNodeRect returns the NodeRect from the passed ANode.
       If the ANode is not assigned the RootNode is used
       This Method is prepared to provide forward compatibility for derived trees.
       The method could be overwritten to modify the returned value to an "actual" value.
       The reason for this subtle distinguish is, that QuadTrees could be used
       to implement polygon representation for fast collision detection. In this
       scenario the full tree needs to be translated and rotated, since the polygon
       (imagine a car / heavy cruizer in a complex polygon-world) is moving and
       turning. It is very ineffective to rebuild the full tree after such an (small)
       change. So it is much better to store the roation and translation value in the tree
       and calculate only the rotated and translated result of the several NodeRect, on request.
       The question if two of such complex polygon items collides could be
       very fast denied by comparing such "actual"-Bonding boxes.
       For example Tree with a world of (-1, -1, 1, 1) wich is rotated by 45° would
       return (-1.414..., -1.414..., 1.414..., 1.414...),
       This approach helps to deal with different capable trees.
       }
    function GetActualNodeRect(const ANode : TLQTNode) : TLQTRect;virtual;
    { GetActualNodeRectAsTetragonCorners calls the GetActualNodeRect (definition below)
       procedure and return the indexed corner from the Tetragon.
       This function could be overwritten return the rotated rectangle, instead of the
       bounding box of the rotated rectangle.
       Following the example from GetActualNodeRect above, the four corners would be
         (0,-1.414...), (1.414...,0), (0, 1.414...),(-1.414...,0)
       }
    function GetActualNodeRectAsTetragonCorners(ANode : TLQTNode; Index : Integer) : TLQTPoint;virtual;
    { GetActualOccupiedRect does the same as GetActualNodeRect does, but using the FOccupiedRect}
    function GetActualOccupiedRect(const ANode : TLQTNode) : TLQTRect;virtual;
    { GetActualOccupiedRectAsTetragonCorners does the same as
      GetActualNodeRectAsTetragonCorners does, but using the FOccupiedRect}
    function GetActualOccupiedRectAsTetragonCorners(ANode : TLQTNode; Index : Integer) : TLQTPoint;virtual;
  public
    { WorldModel contains the extent of the Worls covered by this tree }
    property WorldModel : TLQTWorldModel read FWorldModel;
    property WorldRect : TLQTRect read GetWorldRect;
    { WorldXAxisLeftPlus is true if the values on the X-Axis increment traveling to the left, false increment to the right }
    property WorldXAxisLeftPlus : Boolean read GetWorldXAxisLeftPlus;
    { WorldYAxisUpPlus is true if the values on the Y-Axis increment traveling to the top, false increment to the bottom }
    property WorldYAxisUpPlus : Boolean read GetWorldYAxisUpPlus;
    { OwnsObjects: True if the Tree is the owner of the added NodeItems, so that they will be freed
        on destroying the tree. }
    property OwnsObjects: Boolean read FOwnsObjects write FOwnsObjects;
    { RootNode: The RootNode where the tree is build below }
    property RootNode : TLQTNode read FRootNode;
    { MaxNodeItemCount: This is the amount of NodeItems one Node could hold. If the amount of NodeItems
        increases over this amount, the Node is split and the items are distributed to the new children.
        This value is ignored for Nodes at the deepest level, since this nodes are not allowed to split.
        A too small value will create a lot of nodes, with no Items in.
        A too big value will reduce the performance, since the items are processed like
        an unordered list. The value has a big performance impact depending of the desired
        functionality. The default value is 4}
    property MaxNodeItemCount : Integer read FMaxNodeItemCount write FMaxNodeItemCount;
    { MaxTreeLevel: This is the maximum depth the tree could have. By default the value is 25.
      Keep in mind that the amount of Items that *could* be stored increases very fast.
      The default is about 1280 TerraItems, and a value of 32 exceeds the range of the used Integer variable.
      The value should be choosen, so that the median distance of the used items,
      will distribute them into different nodes.
      A too small value may lead into the situation, where to many items are collected
      in one node at the maximum depth, reducing the performance back to an unordered list.
      A too big value instead, creates a tree with only few, but very long branches,
      causing a lot of running around in empty nodes. }
    property MaxTreeLevel : Integer read FMaxTreeLevel write FMaxTreeLevel default DefaultMaxTreeLevel;
    { DefaultQuadTreeToken is used to identify the Tree class in the Stream read and write methods}
    property TreeToken : Cardinal read FTreeToken;
    { OccupiedRect the rectangle which contains all items of the tree.
       If the tree is empty = (RootNode.TotalItemsCount = 0), the result is the NullRectangle.
       Caution: The rectangle might be empty or not an area,
        dependend of the extension(s) and location(s) of the item(s)}
    property OccupiedRect : TLQTRect read GetOccupiedRect;
    { ActualNodeRect in this implementation the same value as WorldRect.
       See GetActualNodeRect for more details. }
    property ActualNodeRect[Index : TLQTNode] : TLQTRect read GetActualNodeRect;
    { ActualNodeRectAsTetragonCorners in this implementation the four corners of the
       ActualNodeRect, see GetActualNodeRectAsTetragonCorners for more details. }
    property ActualNodeRectAsTetragonCorners[ANode : TLQTNode; Index : Integer] : TLQTPoint read GetActualNodeRectAsTetragonCorners;
    { ActualOccupiedRect in this implementation the same value as WorldRect.
       See GetActualOccupiedRect for more details. }
    property ActualOccupiedRect[Index : TLQTNode] : TLQTRect read GetActualOccupiedRect;
    { ActualOccupiedRectAsTetragonCorners in this implementation the four corners of the
       ActualOccupiedRect, see GetActualOccupiedRectAsTetragonCorners for more details. }
    property ActualOccupiedRectAsTetragonCorners[ANode : TLQTNode; Index : Integer] : TLQTPoint read GetActualOccupiedRectAsTetragonCorners;

    { OnGetItemCoord: The Event that is called to get the location of a specific NodeItem in the tree.
        If instantiate this class this Event *must* be assigned, since the Tree
        needs the get the coordinates of the Items to get the tree organized.
        A derived class of the tree may overwrite the DoGetItemCoord method to
        collect the coordinates of the items direct. }
    property OnGetItemCoord : TLQTGetItemCoordEvent read FGetItemCoordEvent write FGetItemCoordEvent;
    { OnGetItemRect: The Event that is called to get the covered Rect of a specific NodeItem in the tree.
        This Event may be not assigned, if only point-like NodeItems are used.
        If NodeItems are covering an Rect, this event must be assigned. }
    property OnGetItemRect : TLQTGetItemRectEvent read  FGetItemRectEvent write FGetItemRectEvent;
    { OnGetItemCaption: The Event is called to get a descriptive caption for a specific item in the tree.
        This event may be not assigned if items have no caption or the caption is not used.}
    property OnGetItemCaption : TLQTGetItemCaptionEvent read FGetItemCaptionEvent write FGetItemCaptionEvent;
    { OnCreateNode: This Event could be assigned to create descendants
        classes of the TLQTNode to be used by the tree.
        Under nomal circumstances it remains unassigned. }
    property OnCreateNode : TLQTCreateNodeEvent read FCreateNodeEvent write FCreateNodeEvent;
    { OnReadStreamItem: This Event must be assigned, when the ReadStreamTree method is used.
       The event will be called for every Item that needs to be read from the stream. }
    property OnReadStreamItem : TLQTReadStreamItemEvent read FReadStreamItemEvent write FReadStreamItemEvent;
    { OnWriteStreamItem: This Event must be assigned, when the WriteStreamTree method is used.
       The event will be called for every Item that needs to be written to the stream. }
    property OnWriteStreamItem : TLQTWriteStreamItemEvent read FWriteStreamItemEvent write FWriteStreamItemEvent;

    { InsertItem insert the Item into the tree.
      The optional parameters could be used to speed up the processing
        APointValid, ARectValid, APointX, APointY
        ARectLeft, ARectTop, ARectRight, ARectBottom
      The function returns the node where the item was inserted.
      If the item is already in the tree, the containing node is returned.
      If the node could not be inserted, the result is Nil!
      The tree might be reorganized if necesarry and possible, thus some or all
      Items in the tree may be moved in different nodes or indices than before this call. }
    function InsertItem(const AItem : TObject;
                        const APointValid : Boolean = False;
                        const ARectValid : Boolean = False;
                        const APointX : Double = 0.0;
                        const APointY : Double = 0.0;
                        const ARectLeft : Double = 0.0;
                        const ARectTop : Double = 0.0;
                        const ARectRight : Double = 0.0;
                        const ARectBottom : Double = 0.0
                       ) : TLQTNode;virtual;
    { ExtractItem removes an Item from the Tree.
      The function returns the removed Item.
      On any other case the result is Nil!
      The tree might be reorganized if necesarry and possible, thus Items may be moved
      to different nodes than they located before this call. }
    function ExtractItem(const AItem : TObject) : TObject;virtual;
    { MoveItemAfterCoordChange moves the given Item to a new Node, reflecting an
      updated internal position.
      This may cause the relocation of the Item and a restructure of the tree. }
    function MoveItemAfterCoordChange(const AItem : TObject) : TLQTNode;virtual;
    { GetActualNodeTetragon returns the same values as the ActualBoundingTetragonEdges
       but effective in one call.
       If ANode is Nil, the FRootNode will be used to obtain the world's actual BoundingBox}
    procedure GetActualNodeTetragon(const ANode : TLQTNode; out AActualBoundingtTetragon : TLQTPointArr);
    { FindRectNode returns the deepest (leaf)node which contains (or is equal to) the ALQTRect. }
    function FindRectNode(const ALQTRect : TLQTRect) : TLQTNode;virtual;
    { FindCoordNode returns the deepest (leaf)node which contains the coordinate.
        As long as the coordinate is within the world, the result will be assigned.}
    function FindCoordNode(const AX, AY : Double) : TLQTNode;virtual;
    { FindItemNode returns the Node where the Item is located.
      The Tree tries to locate the Item by its coords, if this fails, the full tree
      is searched. So it is very time consuming looking for Items that are not in the tree.
      If the item is not found the result is Nil. }
    function FindItemNode(const AItem : TObject) : TLQTNode;virtual;
    { EnumerateNodesInNodeRect enumerates all Nodes in the Tree who are at least partly covered by
      the passed ALQTRect. For each found node, the passed callback function is called.
      Since the RootNode contains the whole world, the callback is called always at least once.
      See the TLQTNodeEnumerationCallBack defintion for more details.
      AUserData is a generic Pointer, containing any data what the caler likes.
      The value is unchanged passed to the callback. This allows the usage of local
      variables as a context to be used in the callback, avoiding the usage ob global
      variables. }
    function EnumerateNodesInRect(const ALQTRect : TLQTRect;
                                  const AEnumCallBackFunc : TLQTNodeEnumerationCallBack;
                                  const AUserData : Pointer = Nil) : Boolean;virtual;
    { ListNodesInNodeRect lists all Nodes in the Tree who are at least partly covered by
      the passed ALQTRect.
      Since the RootNode contains the whole world, the List will contain always
      at least one Node.
      The AList must be created prior and freed after the call by the using code. }
    procedure ListNodesInRect(const ALQTRect : TLQTRect;
                                  const AList : TList);virtual;
    { CountNodesInRect returns the number of Nodes in the passed Rect.
      The function will return the precise count of nodes in the given Rect.
      This function has to traverse the covered parts of the tree, while nodes
      outside the ALQTRect are completely ignored. }
    function CountNodesInRect(const ALQTRect : TLQTRect) : Integer;virtual;
    { EnumerateItemsInRect enumerates all Nodes in the Tree, who are at least partly covered by
      the passed ALQTRect. For each found node, the passed callback function is called.
      Since the RootNode contains the whole world, the callback is called always at least once.
      See the definition of TLQTItemEnumerationCallBack for more details.
      AUserData is a generic Pointer, containing any data what the caler likes.
      The value is unchanged passed to the callback. This allows the usage of local
      variables as a context to be used in the callback, avoiding the usage ob global
      variables. }
    function EnumerateItemsInRect(const ALQTRect : TLQTRect;
                                    const AEnumCallBackFunc : TLQTItemEnumerationCallBack;
                                    const AUserData : Pointer = Nil) : Boolean;virtual;
    { ListItemsInRect lists all Items in the Tree who are inside the passed ALQTRect.
      The AList must be created prior and freed after the call by the using code.}
    procedure ListItemsInRect(const ALQTRect : TLQTRect;
                                  const AList : TList);virtual;
    { CountAssignedItemsInRect returns the exact number of (assigned) items in the passed Rect.
        This method returns the exact number of items in the Rect.
        It's still a fast method, since for all fully covered rectangles the
        TotalNodeItemsCount is taken. For all partly covered rectangles
        all items in the nodes are queried for their exact position.
        Thus, the method may takes her time. If only an estimation
        is wanted use the faster EstimatedCountOfAssignedItemsInRect instead. }
    function CountAssignedItemsInRect(const ALQTRect : TLQTRect) : Integer;virtual;
    { EstimatedCountOfAssignedItemsInRect returns the estimated amount of Items in the passed Rect.
        This estimation is faster than counting, since no items are queried for there positions.
        Only the TotalNodeItemsCount is used for the estimation and the fraction of the passed Rect.
       Parameters
        ALQTRect, the rectangle i which the item count should be estimated
        AItemThreshold, if TotalNodeItemsCount is equal or less then the estimation is
          terminated for this node and a part of the value is returned.
       Remarks
        The main use of this method is to decide quickly, whether too many items
        in an Rect exists, so that only a placeholder should be displayed,
        instead of an bunch of overlapping items.
        A result greater than 0 will not guaranty that items are really exists in the
        passed Rect. If a node is covering a large Rect, but all items sit in a small region,
        than a mismatch between the reality and the estimation occour.
        A larger AItemThreshold will increase the execution speed, by reducing the
        accuracy. Depending of the data values between 100 and 500 gives an accuracy
        between 10%-20%. }
    function EstimatedCountOfAssignedItemsInRect(const ALQTRect : TLQTRect; const AItemThreshold : Integer) : Double;virtual;
    { AgglomerateNodeAsEnumeration build a summary of the Items within this node and his children.
        The usage is e.g. to build clickable items on a map, which represents a bunch of items
        below, to prevent that thousands of items will be drawn, while a "10K"-sign would
        do the job.
        Another usage would be to colorize parts of an image recording to the "density" of the
        Agglomerations.
       Parameters
        ANode, the Node where the agglomeration starts.
        AAgglomerationDepth determines how many node-leves below should be taken into account.
         Due to the math (every level contains 4 times nodes than the parents level) a value
         of 1 will return up to 4 agglomeration records, 2 will return 16 and 3 = 64.
         If one Node cover an square of e.g. 256 pixel (as one OSM-Tile does), a level of
         3 will place up to 64 ( = 16 * 16) agglomeration records on this tile.
        AMinimumItemNumberInAgglomeration will control the minimum ammount in one agglomeration.
         The value should not be less than MaxQuadNodeItemCount. A good value depends on the
         Data structure and the tile size. If one agglomeration contains a square of 16 x 16
         pixel, the value could be 16 to place (in average) 4 x 4 items in this area.
        AItemEnumFunc,
        AAgglomerationEnumFunc both callback function must be provided to allow a sucessful
         operation. Each function will be called once for every item or build agglomeration.
        AItemUserData,
        AAgglomerationUserData are passed unmodified to the callback functions to allow the management
         of the passed data.
    }
    function AgglomerateNodeAsEnumeration(
                              const ANode : TLQTNode;
                              const AAgglomerationDepth : Integer;
                              const AMinimumItemNumberInAgglomeration : Integer;
                              const AItemEnumFunc : TLQTItemEnumerationCallBack;
                              const AItemUserData : Pointer;
                              const AAgglomerationEnumFunc : TLQTAgglomerationEnumerationCallBack;
                              const AAgglomerationUserData : Pointer) : Boolean;virtual;

    { AgglomerateNodeAsList (see the method with the same name of the TreeNode for a detailed explanation)
        The method build a summary of the Items related within the passed node.
       Parameters (see the method with the same name of the TreeNode for a detailed explanation)
        ANode, the Node where the agglomeration starts.
       Remarks: The method does exactly the same a the method of the node (actually it calls
         this method), but adds all related items of the parent and grand-parent nodes using
         the ListItemsInParentNodes method. }
    procedure AgglomerateNodeAsList(const ANode : TLQTNode;
                              const AAgglomerationDepth: Integer;
                              const AMinimumItemNumberInAgglomeration: Integer;
                              var AAgglomerationArray: TLQTAgglomerationArray;
                              const AItemList: TList);
    { TreeAsStrings returns the content of the Tree in a readable form for debugging purpose. }
    procedure TreeAsStrings(const AStrings : TStrings);virtual;
    { ListTreeNodes recurses through the Nodes of the Tree and stores them in the passed list.
       The AList must be created prior and freed after the call by the using code.}
    procedure ListTreeNodes(const AList : TList);virtual;
    { ListSubNodes recurses through the ANode and all SubNodes and stores them in the passed list.
       if ANode is Nil, the RootNode is used and the full tree listed.
        The AList must be created prior and freed after the call by the using code.}
    procedure ListSubNodes(const ANode : TLQTNode;
                           const AList : TList);virtual;
    { NodesTreeCount recurses through all Nodes of the Tree return the sum of all nodes.
      The minimum result is one if the tree is empty, since the root node must always exists.
      If the result is zero, the tree is malformed}
    function CountTreeNodes : Integer;
    { CountNodes returns the number of SubNodes of the passed ANode (inclusive).
        If ANode is Nil, the result is 0, if ANode has no children, the result is 1}
    function CountNodes(const ANode : TLQTNode) : Integer; virtual;

    { EnumerateItemsInSubNodes recurses through the ANode and all SubNodes and
        calls the callback function once for every found Item.
        See the definition of TLQTItemEnumerationCallBack for details.
        AUserData is a generic Pointer, containing any data what the caller likes. }
    function EnumerateItemsInSubNodes(const ANode : TLQTNode;
                           const AEnumCallBackFunc : TLQTItemEnumerationCallBack;
                           const AUserData : Pointer = Nil) : Boolean;virtual;
    { ListItemsInSubNodes recurses through the ANode and all SubNodes and
        stores the found Items in the passed list.
        The AList must be created prior and freed after the call by the using code.}
    procedure ListItemsInSubNodes(const ANode : TLQTNode;
                           const AList : TList);virtual;

    { EnumerateItemsInParentNodes recurses upwards through the Parent of ANode and Grandparents
        until the root node is reached and stores the Items located in the NodeArea
        in the passed list.
        This seems stupid, since all the items should reside in the smallest possible
        rectangle, and those will be collected with the Enumerate-/ListItemsInSubNodes method.
        But for items which covers an area the smallest rectangle might be a generations
        above. E.g. the shape of the USA-Mainland will reside much higher than
        the location of the Liberty-Statue in New York, but covers her fully.
        The AList must be created prior and freed after the call by the using code.}
    function EnumerateItemsInParentNodes(const ANode : TLQTNode;
                           const AEnumCallBackFunc : TLQTItemEnumerationCallBack;
                           const AUserData : Pointer = Nil) : Boolean;virtual;
    { ListItemsInParentNodes recurses upwards through the Parent of ANode and Grandparents
        until the root node is reached and stores the Items located in the NodeArea
        in the passed list.
        This seems stupid, since all the items should reside in the smallest possible
        rectangle, and those will be collected with the Enumerate-/ListItemsInSubNodes method.
        But for items which covers an area the smallest rectangle might be a generations
        above. E.g. the shape of the USA-Mainland will reside much higher than
        the location of the Liberty-Statue in New York, but covers her fully.
        The AList must be created prior and freed after the call by the using code.}
    procedure ListItemsInParentNodes(const ANode : TLQTNode;
                           const AItemList : TList);virtual;

    { NextNode will return the next node in the tree, behind the given node.
       If the passed node is empty, the RootNode is returned. If StartNode is the last node,
       than Nil is returned. }
    function NextNode(const AStartNode : TLQTNode = Nil) : TLQTNode;
    { NextItem will return the next Item behind the given one.
       Parameters
        output
         ANextItem     the next found item, nil if nothing found
         ANextItemNode the node where the next out item is found, nil if nothing
        input
         AStartItem the Item where the search starts. If Nil, the first*) Item is returned.
         AStartItemNode the Node where the AStartItem resides. If Nil, the RootNode is used.
        *) If AStartItem is Nil, but AStartItemNode assigned, then the first Item in this Node,
           or if no Item exists, the next item is returned.
           If AStartItem is assigned, but AStartNode is Nil, then the item is searched
           in the whole tree, which is very not efficent!
       Result true if something found, false if not.
       A loop:
       var
         item : TObject = Nil;
         node : TLQTNode = Nil;
       ...
         while NextItem(item, node, item, node) do
         begin
           // do something here
         end;
       ...
       will iterate through all items.
    }
    function NextItem(out ANextItem : TObject; out ANextItemNode : TLQTNode;
                      const AStartItem : TObject = Nil; const AStartItemNode : TLQTNode = Nil) : Boolean;
    { RecountTotalNodeItems iterate through all Nodes and Items and refreshes the Nodes TotalNodeItemsCount.
        This method might be useful after a manually reconfiguration of the Tree. }
    procedure RecountTotalNodeItems; virtual;
    { RebuildOccupiedRect iterate through all Nodes and Items and refreshes the Nodes FOccupiedRect.
        This method might be useful after a manually reconfiguration of the Tree. }
    procedure RebuildOccupiedRect; virtual;
    { Clear, empties the QuadTree, frees all Nodes and if OwnsObjects is true Items are freed too }
    procedure Clear; virtual;
    { BeginUpdate initiate a BeginUpdate..EndUpdate sequences. If inside this sequence splitting and
        packing nodes and packing node items are suppressed. BeginUpdate..EndUpdate calls could be nested.
        Caution: BeginUpdate..EndUpdate slows down the adding of many items to the tree, because all new
                 Items are stored in the root node!}
    procedure BeginUpdate;
    { EndUpdate finalize a BeginUpdate..EndUpdate sequences. If the last EndUpdate call from a BeginUpdate
        sequence is reached, the RootNode is "Splitted" and "Packed" (including cleaning up the Items) and
        this is recursed through the whole tree.}
    procedure EndUpdate;
    { WriteQuadTreeToStream writes the Tree including all Nodes to the stream }
    procedure WriteTreeToStream(const AStream : TStream);virtual;
    { ReadNodeFromStream read the content of the Tree including all Nodes from the stream }
    procedure ReadTreeFromStream(const AStream : TStream;
                                     const AStreamReadOptions : Cardinal = 0);virtual;
    { WriteQuadTreeToFile writes the Tree including all Nodes to the stream }
    procedure WriteTreeToFile(const AFileName : String);virtual;
    { ReadQuadTreeFromFile read the content of the Tree including all Nodes from the stream }
    procedure ReadTreeFromFile(const AFileName : String;
                                   const AStreamReadOptions : Cardinal = 0);virtual;

    { Create: Creates the Tree.
        AWorldSize defines the used world. if AXAxisIsCylindrical is True (this is the default),
         the Right-value should be the negative of the Left-value, to allow a seamless wrap around of
         the NodeItems.
        AFreeObjects defines whether the Tree owns the objects (and frees them, when needed) or
         if the NodeItems are held somewhere else and the NodeItems are only references.
        AXAxisIsCylindrical defines whether the World is flat, like a piece of paper (=False) or
         wraping around like a cylinder (=True, the default).
        AYAxisDirection defines whether Top is larger than Bottom (= qydUpPlus, the default) or
         Top is smaller than Bottom (=qydUpMinus). The first case is used for maps or cartesian
         projections, the second for Bitmaps or Editors
        AQuadTreeCreateQuadNodeEvent could be passed to allow the creation of derived classes of
         Nodes, which than extend or replace certain functionallities than the default Node-class.
         Usually this parameter is not used (=Nil)
    }
    constructor Create(const AWorldRect : TLQTRect;
                       const AWorldModel : TLQTWorldModel;
                       const AFreeObjects: Boolean = False;
                       const AQuadTreeCreateQuadNodeEvent : TLQTCreateNodeEvent = Nil);
    { Destroy frees the Tree. It frees the RootNode and subsequently all data in the tree.}
    destructor Destroy;override;
  end;

  { TLQTStreamWriteHelper an internal helper for the writing of the LengthField}
  TLQTStreamWriteHelper = class
  private
    FStream : TStream;
    FLengthFieldPosition : Int64;
  public
    property LengthFieldPosition : Int64 read FLengthFieldPosition;
    procedure WriteLengthField;
    constructor Create(const AStream : TStream);
    destructor Destroy;override;
  end;

  function ReadIntegerFromStream(const AStream : TStream) : Integer;
  procedure WriteIntegerToStream(const AStream : TStream; const AInteger : Integer);
  function ReadDoubleFromStream(const AStream : TStream) : Double;
  procedure WriteDoubleToStream(const AStream : TStream; const ADouble : Double);
  procedure WriteStringToStream(const AStream : TStream; const AString : String);
  function ReadStringFromStream(const AStream : TStream) : String;



const
  LQTNodeChildLocationStrings : array[TLQTNodeChildLocation] of String = (
    'NorthEast','SouthEast', 'SouthWest', 'NorthWest');

implementation

function ReadIntegerFromStream(const AStream : TStream) : Integer;inline;
begin
  Result := 0; // Keep the compiler calm
  AStream.ReadBuffer(Result,SizeOf(Result));
end;
procedure WriteIntegerToStream(const AStream : TStream; const AInteger : Integer);inline;
begin
  AStream.WriteBuffer(AInteger,SizeOf(AInteger));
end;
function ReadDoubleFromStream(const AStream : TStream) : Double;inline;
begin
  Result := 0; // Keep the compiler calm
  AStream.ReadBuffer(Result,SizeOf(Result));
end;
procedure WriteDoubleToStream(const AStream : TStream; const ADouble : Double);inline;
begin
  AStream.WriteBuffer(ADouble,SizeOf(ADouble));
end;
function ReadStringFromStream(const AStream : TStream) : String;
var
  len : Integer;
begin
  Result := '';
  len := ReadIntegerFromStream(AStream);
  if len > 0 then
  begin
    SetLength(Result,len);
    AStream.ReadBuffer(Result[1],len);
  end;
end;
procedure WriteStringToStream(const AStream : TStream; const AString : String);
var
  len : Integer;
begin
  len := Length(AString);
  WriteIntegerToStream(AStream, len); // First write the length = number of bytes
  if len > 0 then // If the string is not empty
    AStream.WriteBuffer(AString[1],len); // write the string buffer
end;

{ TLQTNode }

function TLQTNode.GetNodeItemsLength: Integer;
begin
  Result := Length(FNodeItems);
end;

procedure TLQTNode.SetNodeItemsLength(Value: Integer);
var
  len, i : Integer;
  o : TObject;
begin
  // If the length of the Item-array should be shortended
  // extract the Items in this part of the array and free them if the Tree own them.
  len := Length(FNodeItems);
  if Assigned(FOwner) and FOwner.OwnsObjects and
    (Value < len) then
  begin
    PackNodeItems(True);
    len := Length(FNodeItems);
    if (Value < len) then
    begin
      for i := Value to len-1 do
      begin
        o := ExtractNodeItem(i);
        if Assigned(o) then
          o.Free;
      end;
    end;
  end;
  SetLength(FNodeItems,Value);
end;

function TLQTNode.GetNodeItems(Index: Integer): TObject;
begin
  Result := FNodeItems[Index].Item;
end;

procedure TLQTNode.SetNodeItems(Index: Integer; Value: TObject);
begin
  FNodeItems[Index].Item := Value;
end;

function TLQTNode.GetAssignedNodeItemsCount: Integer;
var
  i : Integer;
begin
  Result := 0;
  for i := 0 to High(FNodeItems) do
    if Assigned(FNodeItems[i].Item) then
      Inc(Result);
end;

function TLQTNode.GetAssignedNodeItems(Index: Integer): TObject;
var
  i : Integer;
  ndx0 : Integer;
begin
  Result := Nil;
  ndx0 := 0;
  for i := 0 to High(FNodeItems) do
  begin
    if not Assigned(FNodeItems[i].Item) then Continue;
    if ndx0 = Index then
    begin
      Result := FNodeItems[i].Item;
      Exit;
    end;
    Inc(ndx0);
  end;
end;

function TLQTNode.GetChildNodes(Index: TLQTNodeChildLocation
  ): TLQTNode;
begin
  Result := FChildNodes[Index];
end;

function TLQTNode.GetChildRects(Index: TLQTNodeChildLocation
  ): TLQTRect;
var
  lLLA : TLQTRect;
  w2, h2 : Double;
  lw2, th2 : Double;
  ydir : Double;
begin
  if Assigned(FChildNodes[Index]) then
    Result := FChildNodes[Index].NodeRect
  else
  begin
    if Fowner.WorldYAxisUpPlus then
      ydir := 1.0
    else
      ydir := -1.0;
    lLLA := FNodeRect;
    w2 := Abs(lLLA.Left - lLLA.Right) / 2.0;
    h2 := Abs(lLLA.Top - lLLA.Bottom) / 2.0;
    lw2 := lLLA.Left + w2;
    th2 := lLLA.Top - (h2*ydir);
    case Index of
      nclNE :
        begin
          lLLA.Left   := lw2;
          lLLA.Bottom := th2;
        end;
      nclSE :
        begin
          lLLA.Left := lw2;
          lLLA.Top  := th2;
        end;
      nclSW :
        begin
          lLLA.Right := lw2;
          lLLA.Top   := th2;
        end;
      nclNW :
        begin
          lLLA.Right  := lw2;
          lLLA.Bottom := th2;
        end;
    end;
    Result := lLLA;
  end;
end;

function TLQTNode.MoveItemToParentNode(const AItemIndex: Integer; out
  ATargetItemIndex: Integer): Boolean;
var
  len : Integer;
  i : Integer;
  ir : TLQTNodeItemsRec;
begin
  Result := False;
  ATargetItemIndex := -1;
  if not Assigned(FParentNode) then Exit;
  if (AItemIndex < 0) or (AItemIndex >= Length(FNodeItems)) then Exit;
  // To prevent a lot of calculations in the TotalNodeItemsCount value,
  // which remains constant by this manipulation,
  // the extract and add functions are not used.
  // Instead the Items are copied by direct access to the arrays

  ir := FNodeItems[AItemIndex]; // Copy the data
  if not Assigned(ir.Item) then Exit; // Nothing to do, Exit
  // Now look for a free place in the parents items array
  len := FParentNode.NodeItemsLength;
  for i := len-1 downto 0 do
  begin
    if not Assigned(FParentNode.FNodeItems[i].Item) then
    begin
      ATargetItemIndex := i; // found one
      Break;
    end;
  end;
  if ATargetItemIndex < 0 then
  begin // no free space found
    ATargetItemIndex := len; // took the previous length as the target index
    FParentNode.NodeItemsLength := ATargetItemIndex + FOwner.MaxNodeItemCount; // Increase the length of the array
  end;
  FParentNode.FNodeItems[ATargetItemIndex] := ir;
  FNodeItems[AItemIndex].Item := Nil;
  Dec(FTotalNodeItemsCount); // Reduce the total item count here, the sum at the parents node remains untouched!
  // The Parents OccupiedRect will not change, but this one
  UpdateOccupiedRect;
  Result := True;
end;

function TLQTNode.MoveItemToChildNode(const AItemIndex: Integer; out
  ATargetChildNode: TLQTNode; out ATargetItemIndex: Integer): Boolean;
var
  len : Integer;
  i : Integer;
  ir : TLQTNodeItemsRec;
  cn : TLQTNode;
  li : TLQTNodeChildLocation;
  ovl : TLQTRectOverlapKind;
  moveitem : Boolean;
begin
  Result := False;
  ATargetChildNode := Self;
  ATargetItemIndex := -1;
  if (AItemIndex < 0) or (AItemIndex >= Length(FNodeItems)) then Exit; // Index out of range
  if not HasChildNodes then Exit; // No childnodes
  ir := FNodeItems[AItemIndex]; // Copy the data fo ease the handling and debugging
  if not Assigned(ir.Item) then Exit; // No Item at this place, nothing to do, Exit
  if (not ir.RectValid) and (not ir.PointValid) then Exit; // No coords valid, Exit
  // Now we iterate through the child nodes
  moveitem := False; // Preset with "nothing found"
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[li];
    if not Assigned(cn) then Continue; // No child at this position, continue
    if ir.RectValid then // if the item covers a rectangle
    begin // check if the rectangle is equal or inside the new node's rectangle
      ovl := cn.NodeRectOverlapTest(ir.Rect);
      moveitem := (ovl = qroInside) or // (B is located inside A) or (A and B are equal)
                  (ovl = qroEqual);
    end
    else if ir.PointValid then // If it is an point
      moveitem := cn.CoordIsInNodeRect(ir.Point.X, ir.Point.Y); // check if the point is in the new node's rectangle
    if moveitem then Break; // We found a node, where the item could be moved in
  end;
  if not moveitem then Exit; // No childnode found where the item match, Exit

  // To prevent a lot of calculations in the TotalNodeItemsCount value,
  // which remains constant by this manipulation,
  // the extract and add-nodeitem functions are not used.
  // Instead the Item are copied by direct access to the arrays

  // Now look for a free place in the child node items array
  len := cn.NodeItemsLength;
  for i := len-1 downto 0 do
  begin
    if not Assigned(cn.FNodeItems[i].Item) then
    begin // found a free place
      ATargetItemIndex := i; // remember the index
      Break;
    end;
  end;
  if ATargetItemIndex < 0 then
  begin // no free space found
    ATargetItemIndex := len; // took the previous length as the target index
    cn.NodeItemsLength := len + FOwner.MaxNodeItemCount; // Increase the length of the array
  end;
  cn.FNodeItems[ATargetItemIndex] := ir; // copy the data
  Inc(cn.FTotalNodeItemsCount); // Update the totalitems count of the child node
  FNodeItems[AItemIndex].Item := Nil; // clear the item here

  // The OccupiedRect of this will not change, but the BoundingBox of the ChildNode
  cn.UpdateOccupiedRectLocal;
  ATargetChildNode := cn;
  Result := True;
end;

function TLQTNode.NodeItemToListEnumerationCallBack(Sender: TObject;
  ANode: TLQTNode; ANodeItem: TObject; ANodeItemIndex: Integer;
  AUserData: Pointer): Boolean;
begin
  Result := Assigned(AUserData);
  if Assigned(ANodeItem) then
    TList(AUserData).Add(ANodeItem);
end;

function TLQTNode.NodeAgglomerationToArrayEnumerationCallBack(
  ANode: TLQTNode;
  const ATreeAgglomerationRec: TLQTAgglomerationRec; AUserData: Pointer
  ): Boolean;
var
  paa : PLQTAgglomerationArray absolute AUserData;
  ndx : Integer;
begin
  Result := True;
  ndx := Length(paa^);
  SetLength(paa^,ndx+1);
  paa^[ndx] := ATreeAgglomerationRec;
end;

procedure TLQTNode.WriteStreamNodeHeader(const AStream: TStream;
  const ATreeStreamDataVersion: Cardinal; const ANodeStreamDataVersion: Cardinal
  );
begin
  AStream.WriteDWord(DefaultLQTNodeStreamDataVersion); // 0 DefaultLQTNodeStreamDataVersion (Option for updates)
end;

procedure TLQTNode.WriteStreamNodeProperties(const AStream: TStream;
  const ATreeStreamDataVersion: Cardinal; const ANodeStreamDataVersion: Cardinal
  );
var
  c : Cardinal;
begin
  WriteIntegerToStream(AStream,FNodeLevel);     // 2: FNodeLevel
  c := SizeOf(FNodeRect);
  AStream.WriteDWord(c);                 // 3: SizeOf FNodeRect
  AStream.WriteBuffer(FNodeRect,c);      // 4: FNodeRect
end;

procedure TLQTNode.WriteStreamNodeChildren(const AStream: TStream;
  const ATreeStreamDataVersion: Cardinal;
  const ANodeStreamDataVersion: Cardinal);
var
  li : TLQTNodeChildLocation;
begin
  // 5: 4 Children, each Word 0 / 1; if 1 then ChildNode Data
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    if Assigned(FChildNodes[li]) then
      FChildNodes[li].WriteNodeToStream(AStream, ATreeStreamDataVersion) // ChildNode
    else
    begin
      AStream.WriteDWord(0); // Marker for no child node exists (= Nil)
      AStream.WriteDWord(0); // Length of Data = 0
    end;
  end;
end;

procedure TLQTNode.WriteStreamNodeItems(const AStream: TStream;
  const ATreeStreamDataVersion: Cardinal; const ANodeStreamDataVersion: Cardinal
  );
var
  c : Cardinal;
  i : Integer;
begin
  c := Length(FNodeItems);
  AStream.WriteDWord(c);    // 6: Length of FNodeItems
  // 7: Items
  for i := 0 to High(FNodeItems) do
    WriteStreamNodeItem(AStream,ATreeStreamDataVersion,ANodeStreamDataVersion,i);
end;

procedure TLQTNode.WriteStreamNodeItem(const AStream: TStream;
  const AQuadTreeStreamDataVersion: Cardinal;
  const ANodeStreamDataVersion: Cardinal; const AItemIndex: Integer);
var
  posi0, posi1 : Int64;
  len : Cardinal;
begin
  if (not Assigned(FOwner.OnWriteStreamItem)) or
     (not Assigned(FNodeItems[AItemIndex].Item)) then
  begin
    // 8: No Item
    AStream.WriteDWord(0); // ItemToken 0 for no Item following (room for future expansion :-)
    AStream.WriteDWord(0); // Item Length is 0
  end
  else
  begin
    // 8: User Data
    posi0 := AStream.Position; // Remember the start position in the stream
    FOwner.OnWriteStreamItem(FOwner,Self,FNodeItems[AItemIndex].Item,AItemIndex,AStream);
    posi1 := AStream.Position; // Get the end position in the stream after the user writing
    len := (posi1 - posi0); // calculate the number of written by the user
    if len < 2 * SizeOf(Cardinal) then // smaller than token plus len, raise Exception
      raise ELQTStreamException.CreateFmt('Minimum DataSize for NodeItems in WriteStreamNodeItem not reached! Position %d',
                                          [posi0]);
    AStream.Position := posi0 + SizeOf(Cardinal); // Seek to the length field
    len := len - 2 * SizeOf(Cardinal); // calculate the user data length
    AStream.WriteDWord(len); // Write the field
    AStream.Position := posi1; // seek back to the end
  end;
end;

procedure TLQTNode.ReadStreamNodeHeader(const AStream: TStream;
  const ATreeStreamDataVersion: Cardinal; out ANodeStreamDataVersion: Cardinal);
begin
  ANodeStreamDataVersion := AStream.ReadDWord; // 0 DefaultQuadTreeStreamDataVersion (Option for updates)
  if (ANodeStreamDataVersion and lqtStreamDataVersionUnitMajorMask) <> (DefaultLQTNodeStreamDataVersion and lqtStreamDataVersionUnitMajorMask) then
    raise ELQTStreamException.CreateFmt('Streamdata node-version %d not supported! Position %d',[ANodeStreamDataVersion, AStream.Position-SizeOf(Cardinal)]);
end;

procedure TLQTNode.ReadStreamNodeProperties(const AStream: TStream;
  const ATreeStreamDataVersion: Cardinal; const ANodeStreamDataVersion: Cardinal
  );
var
  c : Cardinal;
begin
  FNodeLevel := ReadIntegerFromStream(AStream);      // 2: FNodeLevel
  c := AStream.ReadDWord;                // 3: SizeOf FNodeRect
  if c <> SizeOf(FNodeRect) then
    raise ELQTStreamException.CreateFmt('SizeOf NodeRect %d differ from expected %d. Position: %d',
                                         [c, SizeOf(FNodeRect), AStream.Position-Sizeof(Cardinal)]);
  AStream.ReadBuffer(FNodeRect,c);       // 4: FNodeRect
end;

procedure TLQTNode.ReadStreamNodeChildren(const AStream: TStream;
  const ATreeStreamDataVersion: Cardinal;
  const ANodeStreamDataVersion: Cardinal);

var
  c : Cardinal;
  li : TLQTNodeChildLocation;
begin
  // 5: The four Children, each one Word 0 / or a NodeToken then create/read ChildNode Data
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    c := AStream.ReadDWord; // Read the NodeToken
    if c = 0 then // Token is nil
    begin
      c := AStream.ReadDWord; // Read the dummy length
      Continue; // Next Node
    end;
    FChildNodes[li] := FOwner.InternalCreateNode(Self,FNodeRect,c);
    AStream.Position := AStream.Position-SizeOf(Cardinal); // Rewind before the NodeToken
    FChildNodes[li].ReadNodeFromStream(AStream, ATreeStreamDataVersion);
    if FChildNodes[li].NodeLevel <> FNodeLevel+1 then
      raise ELQTStreamException.CreateFmt('Wrong NodeLevel %d expected %d! Position: %d',
                                          [FChildNodes[li].NodeLevel, FNodeLevel+1, AStream.Position-Sizeof(Cardinal)]);
  end;
end;

procedure TLQTNode.ReadStreamNodeItems(const AStream: TStream;
  const ATreeStreamDataVersion: Cardinal; const ANodeStreamDataVersion: Cardinal
  );
var
  c : Cardinal;
  i : Integer;
begin
  c := AStream.ReadDWord;    // 6: Length of FNodeItems
  SetLength(FNodeItems,c);
  // 7: Items, each Word 0 (= no Item)
  for i := 0 to High(FNodeItems) do
    ReadStreamNodeItem(AStream,ATreeStreamDataVersion,ANodeStreamDataVersion,i);
end;

procedure TLQTNode.ReadStreamNodeItem(const AStream: TStream;
  const ATreeStreamDataVersion: Cardinal;
  const ANodeStreamDataVersion: Cardinal; const AItemIndex: Integer);
var
  posi0, posi1 : Int64;
  len, delta : Cardinal;
  c : Cardinal;
begin
  posi0 := AStream.Position; // Remember the start position in the stream
  c := AStream.ReadDWord; // The ItemsToken
  len := AStream.ReadDWord; // The Items Data
  if c = 0 then  // Nil token
  begin
    FNodeItems[AItemIndex].Item := Nil;
    if len > 0 then
      AStream.Position := AStream.Position+len;
    Exit;
  end;
  if (not Assigned(FOwner.OnWriteStreamItem)) then
    raise ELQTStreamException.CreateFmt('UserData in the Stream found, but not UserEvent Assigned! Position: %d',[AStream.Position]);
  // 8: User Data
  AStream.Position := posi0; // Seek Back to the beginning
  FNodeItems[AItemIndex].Item := FOwner.OnReadStreamItem(FOwner,Self,AItemIndex,AStream);
  posi1 := AStream.Position; // Get the end position in the stream after the user writing
  delta := (posi1 - posi0); // calculate the number of written by the user
  if delta < (2 * SizeOf(Cardinal)) then
    raise ELQTStreamException.CreateFmt('Less than 2 DWord read from Stream by user! Position: %d',
                                        [AStream.Position-Sizeof(Cardinal)]);
  delta := delta - (2 * SizeOf(Cardinal));
  if len <> delta then // smaller than token plus len, raise Exception
    raise ELQTStreamException.CreateFmt('DataSize %d for NodeItem differs from expectation! Position %d',[len, delta, AStream.Position]);
end;

function TLQTNode.UpdateOccupiedRectLocal: Boolean;
var
  il : TLQTNodeChildLocation;
  cn : TLQTNode;
  r : TLQTRect;
  i : Integer;
  rbb : TLQTRect;
begin
  rbb.Clear;
  for il := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[il];
    if Assigned(cn) then
    begin
      r := cn.OccupiedRect;
      if r.IsNotArea then Continue;
      if rbb.IsNotArea then
        rbb.Assign(r)
      else
        rbb.Union(r,FOwner.WorldRect,FOwner.WorldModel);
    end;
  end;
  for i := 0 to High(FNodeItems) do
  begin
    if not Assigned(FNodeItems[i].Item) then Continue;
    if not DoFetchItemCoords(i) then
       Continue; //if this fail, skip this item and try the next
    if FNodeItems[i].RectValid then
    begin
      if rbb.IsNotArea then
        rbb.Assign(FNodeItems[i].Rect)
      else
        rbb.Union(FNodeItems[i].Rect,FOwner.WorldRect,FOwner.WorldModel)
    end
    else if FNodeItems[i].PointValid then
    begin
      if rbb.IsNotArea then
        rbb.SetFromPoint(FNodeItems[i].Point,FOwner.WorldRect)
      else
        rbb.Union(FNodeItems[i].Point,FOwner.WorldRect,FOwner.WorldModel);
    end;
  end;
  Result := not rbb.IsEqual(FOccupiedRect);
  if Result then
    FOccupiedRect.Assign(rbb);
end;

function TLQTNode.UpdateOccupiedRect : Boolean;
begin
  Result := UpdateOccupiedRectLocal;
  if Result and Assigned(FParentNode) then
    FParentNode.UpdateOccupiedRect;
end;

function TLQTNode.GetOccupiedRect: TLQTRect;
begin
  Result := FOccupiedRect;
end;

function TLQTNode.CountNodes: Integer;
var
  cn : TLQTNode;
  li : TLQTNodeChildLocation;
begin
  Result := 1; // Count this node
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[li];
    if not Assigned(cn) then Continue; // No child at this position, continue
    Result := Result + cn.CountNodes; // Add the count result of each child node
  end;
end;

procedure TLQTNode.SetTotalNodeItemsCount(Value: Integer);
var
  delta : Integer;
begin
  if Value = FTotalNodeItemsCount then Exit;
  delta := Value - FTotalNodeItemsCount;
  FTotalNodeItemsCount := Value;
  if Assigned(FParentNode) then
    FParentNode.TotalNodeItemsCount := FParentNode.TotalNodeItemsCount+delta;
end;

function TLQTNode.DoFetchItemCoords(const AItemIndex : Integer) : Boolean;
begin
  if (not FNodeItems[AItemIndex].PointValid) and (not FNodeItems[AItemIndex].RectValid) then
  begin
    FNodeItems[AItemIndex].RectValid := FOwner.DoGetItemRect(FNodeItems[AItemIndex].Item,FNodeItems[AItemIndex].Rect); // Try to get the items Rect
    FNodeItems[AItemIndex].PointValid := FOwner.DoGetItemCoord(FNodeItems[AItemIndex].Item,FNodeItems[AItemIndex].Point.X,FNodeItems[AItemIndex].Point.Y); // try to get the coords from the item
  end;
  Result := FNodeItems[AItemIndex].PointValid or FNodeItems[AItemIndex].RectValid;
end;

function TLQTNode.CreateOrGetChildNode(
  const AChildLocation: TLQTNodeChildLocation): TLQTNode;
var
  lLLA : TLQTRect;
  i : Integer;
  ndx : Integer;
  cn : TLQTNode;
begin
  Result := FChildNodes[AChildLocation];
  if not Assigned(Result) then
  begin
    lLLA := ChildRects[AChildLocation];
    Result := FOwner.InternalCreateNode(Self,lLLA);
    FChildNodes[AChildLocation] := Result;
    if not Assigned(Result) then Exit;
    // If a new node is created, move the fitting items to the new created node
    for i := 0 to High(FNodeItems) do
    begin
      if not Assigned(FNodeItems[i].Item) then Continue; // No Item at this position
      if not DoFetchItemCoords(i) then
        Continue; //if this fail, skip this item and try the next
      // Move the item, if possible. Result ignored
      MoveItemToChildNode(i,cn,ndx);
    end;
  end;
end;

procedure TLQTNode.DeleteChildNode(
  const AChildLocation: TLQTNodeChildLocation);
var
  cn : TLQTNode;
  li : TLQTNodeChildLocation;
  assignedcnt, len, ndx, availcnt : Integer;
  i : Integer;
begin
  cn := FChildNodes[AChildLocation];
  if not Assigned(cn) then Exit;
  // Compress the current NodeItems array, to avoid gaps
  PackNodeItems;
  // recurse and delete all four grandchild nodes. They will accumulate all the user data
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    cn.DeleteChildNode(li);
  // Append the accumulated child data to our data
  assignedcnt := cn.AssignedNodeItemsCount;
  if assignedcnt > 0 then
  begin // There are Child data to append
    len := NodeItemsLength;
    availcnt := len-AssignedNodeItemsCount;
    if availcnt < assignedcnt then
      NodeItemsLength := len+(assignedcnt-availcnt); //Increase the array
    len := cn.NodeItemsLength; //The length of the child data
    for i := 0 to len-1 do // run through all fields in the NodeItems array
    begin
      if not Assigned(cn.FNodeItems[i].Item) then Continue;
      if not cn.MoveItemToParentNode(i,ndx) then; // what to do here?
    end;
    // TotalNodeItemsCount remains on the same value!
  end;
  FChildNodes[AChildLocation] := Nil; //Nil the relation
  cn.Free; //Free the child object
end;

function TLQTNode.CoordIsInNodeRect(const AX, AY: Double): Boolean;
// We can assume here, that the NodeRect is never passing the worlds borders.
// This allows to shorten the very complicated calculations
// with signs and wrapping worlds :-)
var
  w2, h2 : Double;
  cx, cy : Double;
begin
  // the center of the node is the average of the borders, regardless of the signs
  cx := (FNodeRect.Right+FNodeRect.Left) / 2.0;
  cy := (FNodeRect.Bottom+FNodeRect.Top) / 2.0;
  // The width and the height is the Abs() of the delta between the sides, regardless of the signs
  w2 := Abs(FNodeRect.Right-FNodeRect.Left) / 2.0;
  h2 := Abs(FNodeRect.Top-FNodeRect.Bottom) / 2.0;
  // Inside is if the distance in X and Y are smaller than half the width / height
  // or Coords laying on the left or top border
  Result := ((Abs(cx-AX) < w2) or (AX = FNodeRect.Left)) and
           ((Abs(cy-AY) < h2) or (AY = FNodeRect.Top));
end;


function TLQTNode.FindCoordNode(const AX, AY: Double): TLQTNode;
var
  cn : TLQTNode;
  li : TLQTNodeChildLocation;
begin
  Result := Nil;
  if not CoordIsInNodeRect(AX,AY) then Exit; // Wrong (sub-)part of the tree, exit
  // From here we are in the correct part of the Tree. This node might be the right one.
  // But we iterate through the four children (if existing) and recurse down, since
  // a child with a smaller rectangle may exist.
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[li]; // Get the child
    if Assigned(cn) then
    begin // child exists
      Result := cn.FindCoordNode(AX,AY); // Recurse down
      if Assigned(Result) then Break; // If the function return a "better" node, search is done
    end;
  end;
  if not Assigned(Result) then // if nothing better found...
    Result := Self; // ... this node is the best node
end;

function TLQTNode.NodeRectOverlapTest(const ARect: TLQTRect
  ): TLQTRectOverlapKind;
var
  wr : TLQTRect;
  wm : TLQTWorldModel;
  cr : TLQTRectArray;
begin
  if Assigned(FOwner) then
  begin
    wr.Assign(FOwner.WorldRect);
    wm := FOwner.WorldModel;
  end
  else
  begin
    wr.Assign(LQTRectInfinite);
    wm := lqtwmFlat;
  end;
  Result := RectanglesOverlapEx(FNodeRect, ARect,
                                wr,wm,
                                cr);
end;

function TLQTNode.OccupiedRectOverlapTest(const ARect: TLQTRect
  ): TLQTRectOverlapKind;
var
  wr : TLQTRect;
  wm : TLQTWorldModel;
  cr : TLQTRectArray;
begin
  if Assigned(FOwner) then
  begin
    wr.Assign(FOwner.WorldRect);
    wm := FOwner.WorldModel;
  end
  else
  begin
    wr.Assign(LQTRectInfinite);
    wm := lqtwmFlat;
  end;
  Result := RectanglesOverlapEx(FOccupiedRect, ARect,
                                wr,wm,
                                cr);
end;

function TLQTNode.NodeRectOverlap(const ARect: TLQTRect): Boolean;
begin
  Result := NodeRectOverlapTest(ARect) <> qroNone;
end;

function TLQTNode.AddNodeItem(const AItem: TObject;
                                  const APointValid : Boolean = False;
                                  const ARectValid : Boolean = False;
                                  const APointX : Double = 0.0;
                                  const APointY : Double = 0.0;
                                  const ARectLeft : Double = 0.0;
                                  const ARectTop : Double = 0.0;
                                  const ARectRight : Double = 0.0;
                                  const ARectBottom : Double = 0.0
                                  ) : Integer;
var
  i : Integer;
  len : Integer;
begin
  Result := -1;
  if not Assigned(AItem) then Exit;
  len := NodeItemsLength;
  if len > 0 then
  begin
    // In the case that the Owner is in the BeginUpdate...EndUpdate mode
    // or not the Owner of the items
    // ignore the duplicate Items, they will be sorted out later
    if (not Assigned(FOwner)) or
       (not FOwner.OwnsObjects) or
       (FOwner.FUpdateCount <= 0) then
    begin
      // Avoid duplicate entries
      for i := 0 to len-1 do
      begin
        if FNodeItems[i].Item = AItem then
        begin
        // Entry is in, exit
          Result := i;
          Exit;
        end;
      end;
    end;
    // Try to find an unused space
    for i := len-1 downto 0 do
    begin
      if not Assigned(FNodeItems[i].Item) then
      begin // Found a free space
        FNodeItems[i].Item := AItem;  // use it and exit
        FNodeItems[i].PointValid := APointValid;
        FNodeItems[i].RectValid := ARectValid;
        FNodeItems[i].Point.X := APointX;
        FNodeItems[i].Point.Y := APointY;
        FNodeItems[i].Rect.Left := ARectLeft;
        FNodeItems[i].Rect.Top := ARectTop;
        FNodeItems[i].Rect.Right := ARectRight;
        FNodeItems[i].Rect.Bottom := ARectBottom;
        TotalNodeItemsCount := FTotalNodeItemsCount+1; //Update the number of items in this and all parent nodes
        UpdateOccupiedRect;
        Result := i;
        Exit;
      end;
    end;
  end;
  // Append the data at the end of the array
  // If the Owner is in the  BeginUpdate...EndUpdate mode reserve more spce at the end to speedup the insertion
  // of many Items
  if (Assigned(FOwner) and (FOwner.FUpdateCount > 0)) then
    NodeItemsLength := len+(FOwner.MaxNodeItemCount*10)
  else
    NodeItemsLength := len+FOwner.MaxNodeItemCount;
  FNodeItems[len].Item := AItem;
  FNodeItems[len].PointValid := APointValid;
  FNodeItems[len].RectValid := ARectValid;
  FNodeItems[len].Point.X := APointX;
  FNodeItems[len].Point.Y := APointY;
  FNodeItems[len].Rect.Left := ARectLeft;
  FNodeItems[len].Rect.Top := ARectTop;
  FNodeItems[len].Rect.Right := ARectRight;
  FNodeItems[len].Rect.Bottom := ARectBottom;
  TotalNodeItemsCount := TotalNodeItemsCount+1; //Update the number of items in this and all parent nodes
  UpdateOccupiedRect;
  Result := len; // High(FNodeItems);
end;

function TLQTNode.ExtractNodeItem(const AItemIndex: Integer): TObject;
begin
  Result := FNodeItems[AItemIndex].Item;
  FNodeItems[AItemIndex].Item := Nil;
  if Assigned(Result) then
  begin
    TotalNodeItemsCount := FTotalNodeItemsCount-1; //Update the number of items in this and all parent nodes
    UpdateOccupiedRect;
  end;
end;

function TLQTNode.ExtractNodeItem(const ANodeItem: TObject): TObject;
var
  i : Integer;
begin
  Result := Nil;
  if not Assigned(ANodeItem) then Exit; // Nothing to do
  for i := 0 to NodeItemsLength-1 do // iterate to the raw array
  begin
    if FNodeItems[i].Item = ANodeItem then // if the item is found
    begin
      Result := ExtractNodeItem(i); // extract the item by it's index
      Exit;
    end;
  end;
end;

function TLQTNode.SplitNode: Boolean;
var
  li : TLQTNodeChildLocation;
  i : Integer;
  datalen, cnt : Integer;
  cn : TLQTNode;
  childcreated :  array[TLQTNodeChildLocation] of Boolean;
  lhascn : Boolean;
  n : Integer;
  ndx : Integer;
begin
  Result := False;
  if (not Assigned(FOwner)) then Exit; // Without Owner we have to quit
  if FOwner.FUpdateCount > 0 then
  begin
    FNodeTouched := True;
    Exit; // SplittingNodes are not allowed during BeginUpdate ... EndUpdate
  end;
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    childcreated[li] := False;

  if FNodeTouched then
  begin
    // Pack DataItems, they may contain unwanted duplicates or empty spaces
    PackNodeItems;
    FNodeTouched := False;
  end;
  cnt := AssignedNodeItemsCount;
  if (cnt <= FOwner.MaxNodeItemCount) then Exit; // Node is fine
  datalen := NodeItemsLength;
  lhascn := HasChildNodes;
  // The node must be split, if the array is overpopulated
  if (FNodeLevel < FOwner.MaxTreeLevel) and  // Do not split any node on the ground
     (
      ((cnt > 0) and lhascn) or // split if items here, but has already childnodes
      (cnt > FOwner.MaxNodeItemCount) // split if to much children here
     ) then
  begin
    // loop through all four sub quadrants
    for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    begin
      cn := FChildNodes[li];
      childcreated[li] := not Assigned(cn); // remember that we created this node here
      // Create all four childnodes
      if not Assigned(cn) then
        cn := CreateOrGetChildNode(li); // If the child does not exists, create one, this will move fitting items to the new node
    end;
    // A created node collects the items which better stay there,
    // but some items may not have been moved, because they have to stay here, or because the node exists allready before
    // Reread the data count to get the change
    n := AssignedNodeItemsCount;
    Result := n <> cnt;
    cnt := n;
    datalen := NodeItemsLength;
    // Loop through all NodeItems, since all may have to be distributed to the new children
    for i := 0 to datalen-1 do
    begin
      if not Assigned(FNodeItems[i].Item) then Continue;
      if not DoFetchItemCoords(i) then
        Continue; //if this fail, skip this item and try the next
      // Move the item, if possible. Result ignored
      if MoveItemToChildNode(i,cn,ndx) then
        Result := True; //Return the "something changed" Result
    end;
    // Pack NodeItems
    // Since some/all of the indexes are moved, they contains now Nil
    // we clear the unused array space
    PackNodeItems;
    // loop through all four sub quadrants and delete the empty ones, if we created them here
    for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    begin
      if not childcreated[li] then Continue;
      cn := FChildNodes[li]; // Get the child node
      if Assigned(cn) and (cn.AssignedNodeItemsCount = 0) and (not cn.HasChildNodes) then
        DeleteChildNode(li);
    end;
  end;

  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin  // loop through all four children
    cn := ChildNodes[li];
    if Assigned(cn) then
      if cn.SplitNode then //Recurse down, sum the "something changed flag"
        Result := True;
  end;

end;

function TLQTNode.PackNode: Boolean;
  // We can try to get rid of some of the child nodes
  procedure ProcessThisNode;
  var
    availcnt, cnt : Integer;
    li, limin : TLQTNodeChildLocation;
    cn : TLQTNode;
    childnodeitemscnt : array[TLQTNodeChildLocation] of Integer;
  begin
    // Now we will try to delete the node(s) with the smallest amount of items first
    // This gives us the chance to delete more nodes, than if we started with the biggest
    // In fact this could be done better, if we would create a permutation of the
    // 4 items and process only the one combination where the maximum of nodes would be
    // deleted. Something for the future :-)
    cnt := 0;
    for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    begin
      childnodeitemscnt[li] := 0;
      cn := FChildNodes[li];
      // Again, combining is only possible, if the childnode (cn) has no
      // children (since this means, that somewhere down a combining was not possible!!)
      if Assigned(cn) and (not cn.HasChildNodes) then
      begin
        childnodeitemscnt[li] := cn.AssignedNodeItemsCount;
        if childnodeitemscnt[li] = 0 then
        begin
          DeleteChildNode(li); // Delete the empty child
          Result := True;
        end;
        cnt := cnt + childnodeitemscnt[li];
      end;
    end;
    if cnt <= 0 then Exit; // Nothing to do here
    // childnodeitemscnt contains now the numbers of items for each childnode
    // (except those who have at least one child).
    // Now we pick the child with the fewest items first
    // Process the child and items and continue, until the node is full or all children processed
    repeat
      cnt := -1; // preset
      limin := Low(TLQTNodeChildLocation); // preset
      for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
      begin // Iterate to the item number array
        if (cnt < 0) and (childnodeitemscnt[li] > 0) then
        begin // if cnt is uninitialized the take the first node, with more than zero items in
          cnt := childnodeitemscnt[li];
          limin := li;
        end;
        if (cnt > 0) and (childnodeitemscnt[li] > 0) and (childnodeitemscnt[li] < cnt)  then
        begin // if cnt contains an amount > 0 then compare to the current item, take this if less item
          cnt := childnodeitemscnt[li];
          limin := li;
        end;
      end;
      // if cnt > 0 it means an node was found and limin contains the node index
      if cnt > 0 then
      begin // we found an node
        childnodeitemscnt[limin] := 0; // mark as done
        availcnt := FOwner.MaxNodeItemCount - AssignedNodeItemsCount; // calculate the free space
        if availcnt < cnt then Break; // we can not include the current smallest node, so we're done here
        DeleteChildNode(limin); // This deletion, will move the contained Items one step up, in our item array
        Result := True; // Something changed
      end;
    until cnt <= 0;
  end;

var
  li : TLQTNodeChildLocation;
  cn : TLQTNode;
  rc : Boolean;
  nodeprocessed : Boolean;
begin
  Result := False;
  if (not Assigned(FOwner)) then Exit; // Without an Owner packing is not possible
  if FOwner.FUpdateCount > 0 then
  begin
    FNodeTouched := True;
    Exit; // Packing Nodes are not allowed in this moment
  end;
  if FNodeTouched then
  begin
    // Pack DataItems, they may contain unwanted duplicates or empty spaces
    PackNodeItems;
    FNodeTouched := False;
  end;
  nodeprocessed  := False; // Preset
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[li];
    if Assigned(cn) then // The childnode exists
    begin
      rc := cn.PackNode;
      Result := Result or rc; // Recurse down with packing
      if rc then // If in this sub node something changed
      begin
        // Process this node, because items could be shifted upwards
        ProcessThisNode;
        nodeprocessed := True; // remember that we processed this node
      end;
    end;
  end;
  // if not processed before, do it here
  if not nodeprocessed then
    ProcessThisNode;
  Result := Result or PackNodeItems(True); // If the Result is still false, check if the Packing change something
end;

function TLQTNode.ExpandNode: Boolean;
var
  cnt, ndx : Integer;
  li : TLQTNodeChildLocation;
  cn, cn0 : TLQTNode;
  childcreated : array [TLQTNodeChildLocation] of Boolean;
  i : Integer;
begin
  Result := False;
  if (not Assigned(FOwner)) then Exit; // Without an Owner expanding is not possible
  if FOwner.FUpdateCount > 0 then
  begin
    FNodeTouched := True;
    Exit; // Expanding Nodes are not allowed in this moment
  end;
  FNodeTouched := False;
  SplitNode; // Split the node, removing overpopulation
  cnt := AssignedNodeItemsCount;
  if (cnt > 0) and HasChildNodes then // if there are items here and child nodes exists
  begin
    // Iterate through all child nodes, create not existing nodes, to ease the process
    for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    begin
      childcreated[li] := not Assigned(FChildNodes[li]); // Remember the creation
      cn := CreateOrGetChildNode(li);
      cnt := AssignedNodeItemsCount; // Recounting is necesarry, since the created node collects all belonging items from here
      if cnt > 0 then
      begin // if any items still here
        for i := 0 to cnt-1 do
          MoveItemToChildNode(i,cn0,ndx); // try to move (the target node and index, are ignored)
      end;
    end;
    // cleaup, delete the child nodes we created here, but still unused
    for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    begin
      if not childcreated[li] then Continue; // Node was not created here
      cn := FChildNodes[li];
      if Assigned(cn) and (cn.AssignedNodeItemsCount <= 0) and (not cn.HasChildNodes) then
        DeleteChildNode(li);  // Delete if not empty
    end;
  end;
  // Iterate through the nodes and expand them
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[li];
    if Assigned(cn) then
      cn.ExpandNode;
  end;
end;

function TLQTNode.NodeHasZeroDataItems: Boolean;
var
  li : TLQTNodeChildLocation;
  i : Integer;
  cn : TLQTNode;
begin
  Result := True; //Assume emptiness
  for i := 0 to NodeItemsLength-1 do
  begin
    if Assigned(FNodeItems[i].Item) then
    begin // It is not empty, further search will not find anything else
      Result := False;
      Exit;
    end;
  end;
  // here are no data stored, but maybe in the children
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := ChildNodes[li];
    if Assigned(cn) then
      Result := cn.NodeHasZeroDataItems; // check the child
    if not Result then Exit; // is not empty, then finish the search
  end;
end;

function TLQTNode.HasChildNodes: Boolean;
begin
  Result := Assigned(ChildNodeNE) or
            Assigned(ChildNodeSE) or
            Assigned(ChildNodeSW) or
            Assigned(ChildNodeNW);
end;

function TLQTNode.IsDescendantOf(ANode: TLQTNode): Boolean;
var
  n : TLQTNode;
begin
  n := Self;
  repeat
    Result := (n = ANode);
    if Result then Exit;
    n := n.FParentNode;
  until not Assigned(n);
end;

function TLQTNode.NodeItemIndex(const AItem: TObject): Integer;
var
  i : Integer;
begin
  Result := -1;
  for i := 0 to High(FNodeItems) do
    if FNodeItems[i].Item = AItem then
    begin
      Result := i;
      Exit;
    end;
end;

function TLQTNode.PackNodeItems(const APackToZero: Boolean): Boolean;
var
  i,j : Integer;
  cnt : Integer;
  len : Integer;
  newlen : Integer;
  nir : TLQTNodeItemsRec;
  tmpitems : array of TLQTNodeItemsRec = Nil;
  ndx : Integer;
  startlen : Integer;
begin
  Result := False;
  if (Assigned(FOwner)) and (FOwner.FUpdateCount > 0) then Exit; // Packing Nodes not allowed here
  startlen := Length(FNodeItems);

  if FOwner.OwnsObjects then
  begin
    // Remove duplicate Entries
    len := Length(FNodeItems);
    if len > 1 then
    begin
      // For all items, but not the last
      for i := 0 to High(FNodeItems)-1 do
      begin
        nir := FNodeItems[i]; // get the item
        if not Assigned(nir.Item) then Continue; //if not assigned, take the next
        for j := i+1 to High(FNodeItems) do // for all items behind the current one, including the last
        begin
          if nir.Item = FNodeItems[j].Item then // if equal then Nil the item
            FNodeItems[j].Item := Nil;
        end;
      end;
    end;
  end;

  // Count the unused data ndxs
  cnt := AssignedNodeItemsCount;
  // The easier case
  if cnt = 0 then
  begin  // All places are unused
    SetLength(FNodeItems,0); //Dont waste memory. But keeping a minimum, may speedier? FOwner.MaxNodeItemCount);
    Exit;
  end;
  len := Length(FNodeItems);
  if len = cnt then Exit; // No free places in the list
  // there are unused fields
  // Copy the assigned items into an temporary field;
  SetLength(tmpitems,cnt);
  ndx := 0;
  for i := 0 to len-1 do
  begin
    // Run through all places
    if Assigned(FNodeItems[i].Item) then
    begin
      tmpitems[ndx] := FNodeItems[i];
      FNodeItems[i].Item := Nil; // Clear the field!
      Inc(ndx);
    end;
    if ndx >= cnt then Break;
  end;
  // Move the items from the temporary field back
  for i := 0 to cnt-1 do
    FNodeItems[i] := tmpitems[i];
  // Set the array to the count of the used places length, keep the minimum length
  newlen := cnt;
  if (not APackToZero) and (newlen < FOwner.MaxNodeItemCount) then
    newlen := FOwner.MaxNodeItemCount; // Per default, keep the array at the default length
  SetLength(FNodeItems,newlen);
  Result := (startlen <> newlen);
end;

function TLQTNode.GetNodesParentChildLocation(
  var AChildLocation: TLQTNodeChildLocation): Boolean;
var
  ic : TLQTNodeChildLocation;
begin
  Result := False;
  if not Assigned(FParentNode) then Exit;
  for ic := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    if FParentNode.ChildNodes[ic] = Self then
    begin
      AChildLocation := ic;
      Result := True;
      Exit;
    end;
  end;
end;

function TLQTNode.AgglomerateNodeAsEnumeration(
  const AAgglomerationDepth: Integer;
  const AMinimumItemNumberInAgglomeration: Integer;
  const AItemEnumFunc: TLQTItemEnumerationCallBack;
  const AItemUserData: Pointer;
  const AAgglomerationEnumFunc: TLQTAgglomerationEnumerationCallBack;
  const AAgglomerationUserData: Pointer): Boolean;
var
  li : TLQTNodeChildLocation;
  cn : TLQTNode;
  cnt : Integer;
  i : Integer;
  o : TObject;
  ar : TLQTAgglomerationRec;
begin
  Result := True;
  if AAgglomerationDepth < 0 then Exit; // Nothing to do, exit
  if (FTotalNodeItemsCount <= AMinimumItemNumberInAgglomeration) or // Minimum items for one agglomeration reached
     (FNodeLevel >= FOwner.MaxTreeLevel) or  // or on the bottom of the tree
     (not HasChildNodes) then  // No childnodes
  begin
    Result := FOwner.EnumerateItemsInSubNodes(Self,AItemEnumFunc,AItemUserData); // Enumerate the items here
    if not Result then Exit;
  end
  else if AAgglomerationDepth = 0 then
  begin // if we are down at the bottom of the query, build teh agglomeration
    ar.Count := FTotalNodeItemsCount;
    ar.Point.X := (FNodeRect.Right + FNodeRect.Left) / 2.0;
    ar.Point.Y := (FNodeRect.Top + FNodeRect.Bottom) / 2.0;
    ar.Node := Self;
    Result := AAgglomerationEnumFunc(Self,ar,AAgglomerationUserData); // Enumerate the agglomerations here
    if not Result then Exit;
  end
  else // recurse downwards though the child nodes
  begin
    // Children on the way down should be enumerated.
    // Those children are usually area based and will otherwise not be covered by the
    // method.
    // Maybe in a later development this could be changed
    cnt := AssignedNodeItemsCount;
    if cnt > 0 then
    begin
      for i := 0 to High(FNodeItems) do
      begin
        o := FNodeItems[i].Item;
        if Assigned(o) then
        begin
          Result := AItemEnumFunc(FOwner,Self,o,i,AItemUserData);
          if not Result then Exit;
        end;
      end;
    end;
    // Now the recursion downwards
    for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    begin
      cn := FChildNodes[li];
      if not Assigned(cn) then Continue; // No child at this position, continue
      // A child exists, recurse downwards
      Result := cn.AgglomerateNodeAsEnumeration(AAgglomerationDepth-1,AMinimumItemNumberInAgglomeration,
                                                AItemEnumFunc,AItemUserData,
                                                AAgglomerationEnumFunc,AAgglomerationUserData);
      if not Result then Exit;
    end;
  end;
end;
procedure TLQTNode.AgglomerateNodeAsList(const AAgglomerationDepth: Integer;
  const AMinimumItemNumberInAgglomeration: Integer;
  var AAgglomerationArray: TLQTAgglomerationArray;
  const AItemList: TList);
begin
  AgglomerateNodeAsEnumeration(AAgglomerationDepth,AMinimumItemNumberInAgglomeration,
    @NodeItemToListEnumerationCallBack, AItemList,
    @NodeAgglomerationToArrayEnumerationCallBack, @AAgglomerationArray);
end;

procedure TLQTNode.RebuildOccupiedRect;
var
  il : TLQTNodeChildLocation;
  cn : TLQTNode;
begin
  FOccupiedRect.Clear; // Clear the box
  // Iterate and recurse to/through all children
  for il := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[il];
    if Assigned(cn) then
      cn.RebuildOccupiedRect;
  end;
  UpdateOccupiedRect;
end;

function TLQTNode.NodeAsString: String;
var
  i : Integer;
  s, orient : String;
  inds : String;
  cn : TLQTNode;
  ic : TLQTNodeChildLocation;
  cnt, cnt1 : Integer;
  childrencnt : Integer;
  lItemCaptionValid : Boolean;
  lItemCaption : String;
  nodecnt : Integer;
begin
  lItemCaption := '';
  inds := '';
  for i := 0 to FNodeLevel do
    inds := inds + ' ';
  ic := nclNE; // Keep the compiler calm
  if GetNodesParentChildLocation(ic) then
    orient := LQTNodeChildLocationStrings[ic]
  else
    orient := 'n/a';
  nodecnt := CountNodes;
  s := Format('Level: %d, TotalNodes %d, TotalItems %d, Rect("%s") (Left/X: %1.6f, Top/Y %1.6f), (Right/X: %1.6f, Bottom/Y %1.6f)',
               [FNodeLevel,
               nodecnt,
               FTotalNodeItemsCount,
               orient,
               FNodeRect.Left,
               FNodeRect.Top,
               FNodeRect.Right,
               FNodeRect.Bottom]);
  s := s + #13#10 + 'Occupied Rect ' + FOccupiedRect.CoordsAsString('(Left/X: %1.6f, Top/Y %1.6f), (Right/X: %1.6f, Bottom/Y %1.6f)');
  cnt1 := AssignedNodeItemsCount;
  cnt := NodeItemsLength;
  s := s + #13#10 + Format('+ DataItems Assigned %d of total %d',[cnt1, cnt]);
  for i := 0 to cnt-1 do
  begin
    s := s + #13#10 + Format('  [%d] = ',[i]);
    if Assigned(FNodeItems[i].Item) then
    begin
      DoFetchItemCoords(i);
      if FNodeItems[i].RectValid then
        s := s + Format('L: %1.6f, T: %1.6f, R: %1.6f, B: %1.6f',
                        [FNodeItems[i].Rect.Left, FNodeItems[i].Rect.Top, FNodeItems[i].Rect.Right, FNodeItems[i].Rect.Bottom])
      else if FNodeItems[i].PointValid then
        s := s + Format('X: %1.6f, Y: %1.6f',[FNodeItems[i].Point.X,FNodeItems[i].Point.Y])
      else
        s := s + 'Coords inaccessable!';
      lItemCaptionValid := FOwner.DoGetItemCaption(FNodeItems[i].Item,lItemCaption);
      if lItemCaptionValid then
        s := s + ' "'+lItemCaption+'"';
    end
    else
      s := s + '<unused>';
  end;
  childrencnt := 0;
  for ic := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[ic];
    if Assigned(cn) then
      Inc(childrencnt);
  end;
  s := s + #13#10 + 'Children '+IntToStr(childrencnt);
  Result := s;
end;

procedure TLQTNode.NodeAsStrings(const AStrings: TStrings);
var
  i : Integer;
  s : String;
  inds : String;
  cn : TLQTNode;
  ic : TLQTNodeChildLocation;
  cnt, cnt1 : Integer;
  childrencnt : Integer;
  lItemCaptionValid : Boolean;
  lItemCaption : String = '';
begin
  inds := '';
  for i := 0 to FNodeLevel do
    inds := inds + ' ';
  s := Format('Level: %d, TotalItems %d, Rect (Left/X: %1.6f, Top/Y %1.6f), (Right/X: %1.6f, Bottom/Y %1.6f)',
               [FNodeLevel,
               FTotalNodeItemsCount,
               FNodeRect.Left,
               FNodeRect.Top,
               FNodeRect.Right,
               FNodeRect.Bottom]);
  AStrings.Add(inds+s);
  s := 'Occupied Rect ' + FOccupiedRect.CoordsAsString('(Left/X: %1.6f, Top/Y %1.6f), (Right/X: %1.6f, Bottom/Y %1.6f)');
  AStrings.Add(inds+s);
  cnt1 := AssignedNodeItemsCount;
  cnt := NodeItemsLength;
  AStrings.Add(inds+Format('+ DataItems Assigned %d, Total %d',[cnt1, cnt]));
  for i := 0 to cnt-1 do
  begin
    s := Format('[%d] = ',[i]);
    if Assigned(FNodeItems[i].Item) then
    begin
      DoFetchItemCoords(i);
      if FNodeItems[i].RectValid then
        s := s + Format('L: %1.6f, T: %1.6f, R: %1.6f, B: %1.6f',
                        [FNodeItems[i].Rect.Left, FNodeItems[i].Rect.Top, FNodeItems[i].Rect.Right, FNodeItems[i].Rect.Bottom])
      else if FNodeItems[i].PointValid then
        s := s + Format('X: %1.6f, Y: %1.6f',[FNodeItems[i].Point.X,FNodeItems[i].Point.Y])
      else
        s := s + 'Coords inaccessable!';
      lItemCaptionValid := FOwner.DoGetItemCaption(FNodeItems[i].Item,lItemCaption);
      if lItemCaptionValid then
        s := s + ' "'+lItemCaption+'"';
    end
    else
      s := s + '<unused>';
    AStrings.Add(inds+' '+s);
  end;
  childrencnt := 0;
  for ic := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[ic];
    if Assigned(cn) then
      Inc(childrencnt);
  end;
  AStrings.Add(inds+'+ Children '+IntToStr(childrencnt));
  for ic := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[ic];
    if Assigned(cn) then
    begin
      s := inds+'ChildNode: '+LQTNodeChildLocationStrings[ic];
      AStrings.Add(s);
      cn.NodeAsStrings(AStrings);
    end;
  end;
end;

procedure TLQTNode.Clear;
var
  cn : TLQTNode;
  li : TLQTNodeChildLocation;
  i : Integer;
  o : TObject;
begin
  if Assigned(FOwner) and FOwner.OwnsObjects then
  begin
    for i := 0 to High(FNodeItems) do
    begin
      o := ExtractNodeItem(i);
      if Assigned(o) then
      begin
        o.Free;
        o := Nil;
      end;
    end;
  end;
  SetLength(FNodeItems,FOwner.MaxNodeItemCount);
  // recurse and free all four child nodes. They will free there children too
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := FChildNodes[li];
    if Assigned(cn) then
    begin
      cn.Clear;
      cn.Free;
      FChildNodes[li] := Nil;
    end;
  end;
  // Update the TotalNodeItemsCount
  TotalNodeItemsCount := 0;
  UpdateOccupiedRect;
end;

procedure TLQTNode.WriteNodeToStream(const AStream: TStream;
  const AQuadTreeStreamDataVersion: Cardinal);
var
  sh : TLQTStreamWriteHelper;
begin
  AStream.WriteDWord(FNodeToken);
  sh := TLQTStreamWriteHelper.Create(AStream);  // Writes the Token and a dummy len
  try
    WriteStreamNodeHeader(AStream,AQuadTreeStreamDataVersion,DefaultLQTNodeStreamDataVersion);
    WriteStreamNodeProperties(AStream,AQuadTreeStreamDataVersion,DefaultLQTNodeStreamDataVersion);
    WriteStreamNodeChildren(AStream,AQuadTreeStreamDataVersion,DefaultLQTNodeStreamDataVersion);
    WriteStreamNodeItems(AStream,AQuadTreeStreamDataVersion,DefaultLQTNodeStreamDataVersion);
  finally
    sh.Free; // Updates the written length
  end;
end;

procedure TLQTNode.ReadNodeFromStream(const AStream: TStream;
  const AQuadTreeStreamDataVersion: Cardinal;
  const AStreamReadOptions : Cardinal = 0);
var
  c : Cardinal;
  startposi : Int64;
  len : Cardinal;
  lQuadNodeStreamDataVersion : Cardinal;
  delta : Int64;
begin
  Clear; // Just for sanity
  c := AStream.ReadDWord; // NodeToken
  if c <> FNodeToken then
    raise ELQTStreamException.CreateFmt('Wrong DataToken %8.8x expected %8.8x!',
                                        [c, FNodeToken,AStream.Position-Sizeof(Cardinal)]);
  len := AStream.ReadDWord; // Data to read from here
  startposi := AStream.Position; // Memorize current position
  ReadStreamNodeHeader(AStream,AQuadTreeStreamDataVersion,lQuadNodeStreamDataVersion);
  ReadStreamNodeProperties(AStream,AQuadTreeStreamDataVersion,lQuadNodeStreamDataVersion);
  ReadStreamNodeChildren(AStream,AQuadTreeStreamDataVersion,lQuadNodeStreamDataVersion);
  ReadStreamNodeItems(AStream,AQuadTreeStreamDataVersion,lQuadNodeStreamDataVersion);
  delta := AStream.Position-startposi; // The delta between the two positions
  if delta <> len then // Different from what we expect
  begin
    // If we read fewer data than expected and the lazy mode is enabled, just seek forward
    if (delta < len) and ((AStreamReadOptions and srofLazyLengthProcessing) <> 0) then
      AStream.Position := startposi + len
    else
      raise ELQTStreamException.CreateFmt('Stream Data corrupted! Read Bytes %d Expected Bytes %d! Position: %d',
                                          [delta, len, AStream.Position]);
  end;
end;

constructor TLQTNode.Create(const AOwner: TLazQuadTree;
  const AParentNode: TLQTNode; const ALQTRect: TLQTRect);
begin
  inherited Create;
  FNodeToken := DefaultLQTNodeToken;
  FOwner := AOwner;
  FParentNode := AParentNode;
  FNodeRect := ALQTRect;
  // Calculate the NodeLevel. Store the result so that it is readable during lifetime
  FNodeLevel := 0;
  if Assigned(FParentNode) then
    FNodeLevel := FParentNode.NodeLevel+1;
end;

destructor TLQTNode.Destroy;
begin
  Clear;
  inherited;
end;

function TLazQuadTree.GetWorldXAxisLeftPlus: Boolean;
begin
  Result := FRootNode.FNodeRect.WorldXAxisLeftPlus;
end;

function TLazQuadTree.GetWorldRect: TLQTRect;
begin
  if Assigned(FRootNode) then
    Result := FRootNode.FNodeRect
  else
    Result := LQTRectEmpty;
end;

function TLazQuadTree.GetWorldYAxisUpPlus: Boolean;
begin
  Result := FRootNode.FNodeRect.WorldYAxisUpPlus ;
end;

{ TLazQuadTree }
function TLazQuadTree.InternalCreateNode(const AParentNode: TLQTNode;
  const ALQTRect: TLQTRect; const ANodeToken: Cardinal): TLQTNode;
begin
  Result := Nil;
  if Assigned(FCreateNodeEvent) then
    Result := FCreateNodeEvent(Self, AParentNode, ALQTRect, ANodeToken);
  if not Assigned(Result) then // The user did not create the node, try is by ourself
    // We create if the Token is 0 (=Nil) or the DefaultLQTNodeToken
    if (ANodeToken = 0) or (ANodeToken = DefaultLQTNodeToken) then
      Result := TLQTNode.Create(Self, AParentNode, ALQTRect);
end;

function TLazQuadTree.InternalEnumerateNodesInRect(
  const ANode: TLQTNode; const AUnwrappedRect: TLQTRect;
  const AEnumCallBackFunc: TLQTNodeEnumerationCallBack;
  const AUserData: Pointer): Boolean;
var
  cn : TLQTNode;
  hit : TLQTRectOverlapKind;
  i : TLQTNodeChildLocation;
  cRect : TLQTRect;
begin
  Result := True;
  if not Assigned(AEnumCallBackFunc) then Exit;
  if not Assigned(ANode) then Exit;
  hit := RectanglesOverlap(ANode.NodeRect,AUnwrappedRect,CRect,FRootNode.NodeRect);
  if hit = qroNone then Exit; //This branch contains no relevant data
  if ANode.NodeItemsLength > 0 then
    Result := AEnumCallBackFunc(Self,ANode,AUserData); //Call the users callback function with the found node
  if not Result then Exit; // Abortion as the user requests
  // Iterate through the four children (if existing)
  for i := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := ANode.ChildNodes[i]; // Get the child
    if Assigned(cn) then // Recurse throught the node
    begin
      Result := InternalEnumerateNodesInRect(cn,AUnwrappedRect,AEnumCallBackFunc,AUserData);
      if not Result then Exit; // Abortion as the user requests
    end;
  end;
end;

procedure TLazQuadTree.ListSubNodes(const ANode: TLQTNode;
  const AList: TList);
var
  n, cn : TLQTNode;
  li : TLQTNodeChildLocation;
begin
  if not Assigned(AList) then Exit;
  if not Assigned(ANode) then
    n := FRootNode
  else
    n := ANode;
  AList.Add(n);
  // Iterate through the four children (if existing)
  for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := n.ChildNodes[li]; // Get the child
    if Assigned(cn) then // Recurse throught the node
      ListSubNodes(cn,AList);
  end;
end;

function TLazQuadTree.CountTreeNodes: Integer;
begin
  Result := FRootNode.CountNodes;
end;

function TLazQuadTree.InternalEnumerateItemsInRect(const ANode: TLQTNode;
  const AUnwrappedRect: TLQTRect;
  const AEnumCallBackFunc: TLQTItemEnumerationCallBack;
  const AUserData: Pointer): Boolean;
var
  cn : TLQTNode;
  hit : TLQTRectOverlapKind;
  i : TLQTNodeChildLocation;
  ii : Integer;
  ni : TObject;
  lCRect : TLQTRect;
begin
  Result := True;
  if not Assigned(FGetItemCoordEvent) then Exit;
  if not Assigned(AEnumCallBackFunc) then Exit;
  if not Assigned(ANode) then Exit;
  hit := RectanglesOverlap(ANode.FOccupiedRect, AUnwrappedRect, lCRect,
                       FRootNode.NodeRect);
  if hit = qroNone then Exit; //This branch contains no relevant data
  // If the Nodes Rect covers the node or the node is fully inside the Rect...
  if (hit = qroEqual) or
     (hit = qroFull) then
  begin
    // ... than all Items in the SubNodes could be processed, without any further investigation
    // Remark: CountNodes returns 1 for the passed ANode
    Result := EnumerateItemsInSubNodes(ANode,AEnumCallBackFunc,AUserData);
    Exit;
  end;
  // The node is covered partly,
  // So some Items may be inside, some not
  // and it is up to the children to process further
  for ii := 0 to High(ANode.FNodeItems) do
  begin
    ni := ANode.FNodeItems[ii].Item;
    if Assigned(ni) then
    begin
      if not ANode.DoFetchItemCoords(ii) then
        Continue; //if this fail, skip this item and try the next
      if (ANode.FNodeItems[ii].RectValid and
          (RectanglesOverlap(AUnwrappedRect, ANode.FNodeItems[ii].Rect, lCRect, FRootNode.NodeRect) <> qroNone)
         ) or (
          ANode.FNodeItems[ii].PointValid and
          AUnwrappedRect.ContainsPoint(ANode.FNodeItems[ii].Point) // We can use the faster ContainsPoint here, since the FNodeRect is guaranteed to be flat!
         ) then
      begin
        Result := AEnumCallBackFunc(Self,ANode,ni,ii,AUserData);
        if not Result then Exit; // Abortion as the user requests
      end;
    end;
  end;
  // Iterate through the four children (if existing)
  for i := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := ANode.ChildNodes[i]; // Get the child
    if Assigned(cn) then // Recurse throught the node
    begin
      Result := InternalEnumerateItemsInRect(cn,AUnwrappedRect,AEnumCallBackFunc,AUserData);
      if not Result then Exit; // Abortion as the user requests
    end;
  end;
end;

function TLazQuadTree.InternalEnumerateItemsInParentNodes(
  const ANode: TLQTNode; const AUnwrappedRect: TLQTRect;
  const AEnumCallBackFunc: TLQTItemEnumerationCallBack;
  const AUserData: Pointer): Boolean;
var
  ii : Integer;
  ni : TObject;
  lCRect : TLQTRect;
begin
  Result := True; // The Result will always be true, except that the user returns different
  if not Assigned(ANode) then Exit; // The Root-Node was reached
  for ii := 0 to High(ANode.FNodeItems) do
  begin
    ni := ANode.FNodeItems[ii].Item;
    if Assigned(ni) then
    begin
      if not ANode.DoFetchItemCoords(ii) then
        Continue; //if this fail, skip this item and try the next
      if (ANode.FNodeItems[ii].RectValid and
          (RectanglesOverlap(AUnwrappedRect, ANode.FNodeItems[ii].Rect, lCRect, FRootNode.NodeRect) <> qroNone)
         ) or (
          ANode.FNodeItems[ii].PointValid and
          AUnwrappedRect.ContainsPoint(ANode.FNodeItems[ii].Point) // We can use the faster ContainsPoint here, since the FNodeRect is guaranteed to be flat!
         ) then
      begin
        Result := AEnumCallBackFunc(Self,ANode,ni,ii,AUserData);
        if not Result then Exit; // User tells us to terminate
      end;
    end;
  end;
  // Recurse upwards
  Result := InternalEnumerateItemsInParentNodes(ANode.ParentNode,AUnwrappedRect,AEnumCallBackFunc,AUserData);
end;

function TLazQuadTree.CountNodes(const ANode: TLQTNode): Integer;
begin
  Result := 0;
  if not Assigned(ANode) then
    Result := FRootNode.CountNodes
  else
    Result := ANode.CountNodes; // Count the passed Node
end;

function TLazQuadTree.EnumerateItemsInSubNodes(const ANode: TLQTNode;
  const AEnumCallBackFunc: TLQTItemEnumerationCallBack;
  const AUserData: Pointer): Boolean;
var
  cn : TLQTNode;
  i : TLQTNodeChildLocation;
  ii : Integer;
  ni : TObject;
begin
  Result := True;
  if not Assigned(AEnumCallBackFunc) then Exit;
  if not Assigned(ANode) then Exit;
  for ii := 0 to High(ANode.FNodeItems) do
  begin
    ni := ANode.FNodeItems[ii].Item;
    if Assigned(ni) then
    begin
      Result := AEnumCallBackFunc(Self,ANode,ni,ii,AUserData);
      if not Result then Exit;
    end;
  end;
  // Iterate through the four children (if existing)
  for i := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := ANode.ChildNodes[i]; // Get the child
    if Assigned(cn) then // Recurse throught the node
    begin
      Result := EnumerateItemsInSubNodes(cn,AEnumCallBackFunc,AUserData);
      if not Result then Exit; // Abortion as the user requests
    end;
  end;
end;
procedure TLazQuadTree.ListItemsInSubNodes(const ANode: TLQTNode;
  const AList: TList);
begin
  EnumerateItemsInSubNodes(ANode,@TreeItemToListEnumerationCallBack,AList);
end;

function TLazQuadTree.EnumerateItemsInParentNodes(const ANode: TLQTNode;
  const AEnumCallBackFunc: TLQTItemEnumerationCallBack;
  const AUserData: Pointer): Boolean;
begin
  Result := InternalEnumerateItemsInParentNodes(ANode,ANode.FNodeRect,AEnumCallBackFunc,AUserData);
end;

procedure TLazQuadTree.ListItemsInParentNodes(const ANode: TLQTNode;
  const AItemList: TList);
begin
  if not Assigned(ANode) then Exit;
  EnumerateItemsInParentNodes(ANode,@TreeItemToListEnumerationCallBack,AItemList);
end;

function TLazQuadTree.InternalCountNodesInRect(
  const ANode: TLQTNode; const AUnwrappedRect: TLQTRect): Integer;
var
  cn : TLQTNode;
  hit : TLQTRectOverlapKind;
  i : TLQTNodeChildLocation;
  lCRect : TLQTRect;
begin
  Result := 0;
  if not Assigned(ANode) then Exit;
  hit := RectanglesOverlap(AUnwrappedRect, ANode.FNodeRect, lCRect, FRootNode.NodeRect);
  if hit = qroNone then Exit; //This branch contains no relevant data
  // If the Nodes Rect covers the node or the node is fully inside the Rect...
  if (hit = qroEqual) or
     (hit = qroFull) then
  begin
    // ... than all subnodes could be inserted, without any further investigation
    // Remark: CountNodes returns 1 for the passed ANode
    Result := CountNodes(ANode);
    Exit;
  end;
  // The node is covered partly, so it is up to the children to decide
  Result := 1;
  // Iterate through the four children (if existing)
  for i := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := ANode.ChildNodes[i]; // Get the child
    if Assigned(cn) then // Recurse throught the node
      Result := Result + InternalCountNodesInRect(cn,AUnwrappedRect);
  end;
end;

function TLazQuadTree.InternalCountAssignedNodeItemsInRect(
  const ANode: TLQTNode; const AUnwrappedRect: TLQTRect): Integer;
var
  cn : TLQTNode;
  hit : TLQTRectOverlapKind;
  i : TLQTNodeChildLocation;
  ni : TObject;
  ii : Integer;
  lCRect : TLQTRect;

begin
  Result := 0;
  if not Assigned(ANode) then Exit;
  if ANode.TotalNodeItemsCount = 0 then Exit; // No Items here, nothing to do
  hit := RectanglesOverlap(ANode.FOccupiedRect, AUnwrappedRect, lCRect,
                           FRootNode.NodeRect);
  if hit = qroNone then Exit; //This branch contains no relevant data
  // If the Nodes Rect covers the node or the node is fully inside the Rect...
  if (hit = qroEqual) or
     (hit = qroFull) then
  begin
    // ... than the sum of all items of the subnodes could be inserted, without any further investigation
    Result := ANode.TotalNodeItemsCount;
    Exit;
  end;
  // The node is covered partly, so it is up to the children (if any) to decide.
  // But we may have items here too, because of no children, or big item extent
  for ii := 0 to High(ANode.FNodeItems) do // run through all item
  begin
    ni := ANode.FNodeItems[ii].Item;
    if Assigned(ni) then
    begin
      if not ANode.DoFetchItemCoords(ii) then
        Continue; //if this fail, skip this item and try the next
      if (ANode.FNodeItems[ii].RectValid and
          (RectanglesOverlap(AUnwrappedRect, ANode.FNodeItems[ii].Rect,lCRect,FRootNode.NodeRect) <> qroNone)
         ) or (
          ANode.FNodeItems[ii].PointValid and
          AUnwrappedRect.ContainsPoint(ANode.FNodeItems[ii].Point) // We can use the faster ContainsPoint here, since the FNodeRect is guaranteed to be flat!
         ) then
        Inc(Result); // Add to result if the item lays in the processed Rect
    end;
  end;

  // Iterate through the four children (if existing)
  for i := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := ANode.ChildNodes[i]; // Get the child
    if Assigned(cn) then // Recurse throught the node
      Result := Result + InternalCountAssignedNodeItemsInRect(cn,AUnwrappedRect);
  end;
end;

function TLazQuadTree.InternalEstimatedCountOfAssignedItemsInRect(
  const ANode: TLQTNode; const AUnwrappedRect: TLQTRect;
  const AItemThreshold: Integer): Double;
{ Generally the estimation is performed by calculation the
  fraction of the overlapping Rect of the node's Rect and
  multiplying this value with the TotalNodeItemsCount of the
  node, which is the accumulated count of all nodes in this part of the tree.
  So if a node is overlapped by the test Rect by 1/3 and has 15 items,
  then 5.0 is returned.
  The details are a bit more complex, because it is necesarry to dive deeper in the
  tree, to avoid a big quality range in the estimation.
  You may think about a small Rect in London around the Longitude "0°".
  It is not wise, to take all items in the NorthEast and NorthWest quadrants
  of the whole world, to estimate the items located in London.
  So recursing through the tree is done as deep as necessarry:
  - If a node is fully outside of the TestRect, this part of the Tree is ignored.
  - If a node is fully inside the TestRect, than all items below are counted, but
    no further investigation is necesarry.
  - For all parts of a node, where no children exists, the four quadrants are
    examined and if they fully in, the calculation is performed as described
    above (fraction of the Rect, multiplyed by TotalItemsCount).
  - To avoid double counting the calculation ist divided in parts where children exists and where not.
  And all Nodes, who are not covered by the test Rect are ignored as soon as possible.
  Even if the code is a bit complex, the speed is good, since wide Rects of the
  tree are excluded from searching very soon.
}

  {FractionOfNodeRectPartly: Calculates the fraction of a partly covered Rect. }
  function FractionOfNodeRectPartly (
             const ABaseRect : TLQTRect; // The Rect to proof (NodeRect or ChildRect)
             const AOverlappingRect : TLQTRect) : Double;
  var
    arn, ar : Double;
    w,h : Double;
  begin
    // calculate the area for all the rectioning parts
    w := AOverlappingRect.Width;
    h := AOverlappingRect.Height;
    ar := (w*h);
    // calculate the total area
    w := ABaseRect.Width;
    h := ABaseRect.Height;
    arn := (h*w);
    // calculate the ratio (the world forms always an area!)
    Result := ar / arn;
  end;

var
  cn : TLQTNode;
  li : TLQTNodeChildLocation;
  qarterscnt : Integer;
  directitemsum, directexcludedsum : Double;
  remainingitems : Double;
  lhit : TLQTRectOverlapKind;
  d, d0 : Double;
  ncar : TLQTRect;
  lCRect : TLQTRect;
begin
  Result := 0.0;
  if not Assigned(ANode) then Exit;
  if ANode.TotalNodeItemsCount <= 0 then Exit; // Nothing to do here, since no items here or below

  // Get the kind of hit and the overlapping Rect(s) if existing
  // We are using here the FOccupiedRect, to avoid a long search in empty parts of the world
  lhit := RectanglesOverlap(ANode.FOccupiedRect, AUnwrappedRect, lCRect,
                         FRootNode.NodeRect);

  case lhit of
    qroNone :; // This node is not hit, so no items will be found here or below, so good bye
    qroPartly, qroInside :
      begin
        d := FractionOfNodeRectPartly(ANode.FOccupiedRect, lCRect);
        // Partly and Inside needs almost the same action.
        // For each existing child, pass the work to it, for each "non existing" child
        // do the work here.
        // If no children exists or the number of items in this part of the world
        // is below the AItemThreshold then the calculation could be shortened.
        if (not ANode.HasChildNodes) or
           (ANode.TotalNodeItemsCount <= AItemThreshold) then
        begin
          // No children exists, take the fraction of the items
          Result := ANode.TotalNodeItemsCount * d;
        end
        else
        begin
          // At least one child exists, the work has to be devided into the four quadrants.
          // Summation of the direct assigned children of the child node

          d0 := ANode.AssignedNodeItemsCount;
          directitemsum := d0 * d; // Preset with the part of the items here
          directexcludedsum := d0 - directitemsum; // Summation of all not assigned children of the child node
{
          directitemsum := 0.0; // Preset with 0.0
          directexcludedsum := 0.0; // Summation of all not assigned children of the child node
}
          qarterscnt := 0;
          // Iterate through the four child positions
          for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
          begin
            cn := ANode.ChildNodes[li]; // Get the child
            // If the child exists ...
            if Assigned(cn) then
            begin
              // let do it the work
              d := InternalEstimatedCountOfAssignedItemsInRect(cn,AUnwrappedRect,AItemThreshold);
              // We will count two sums
              // The sum of the hit item of the child.
              directitemsum := directitemsum + d;
              // and the remaining items of the child, since these are also out of the race
              directexcludedsum := directexcludedsum + (cn.TotalNodeItemsCount - d);
            end
            else
            begin
              // If no child at this position exists, do the calculation for it here.
              // The calculation is fairly simple.
              // If the child's Rect is none, party or inside (means the test Rect is fully inside the non existing child)
              // then the items in this parts are simple zero.
              // If the opposite is true, the child's Rect is fully part of the test Rect
              // than this quarter of the Rect has to be taken fully into account
              // this means, we have not to deal with fractions of covered child ares here.
              // we can just count the touched quarters.
              ncar := ANode.ChildRects[li];
              lhit := RectanglesOverlap(ncar, AUnwrappedRect, lCRect,
                                     FRootNode.NodeRect);
              case lhit of
                qroEqual, qroFull :
                  // Equal should never occour, but will catched for sanity.
                  // Full means that the Rect of the non existing child lies fully inside the
                  // test Rect. Thus 1/4 of the not directly assigned items of this node has to be counted
                  Inc(qarterscnt);
              else
                // qroPartly :
                //   a part of the child node Rect is covered by the test Rect (which may in total be much bigger)
                //   but since the child does not exists, zero items are located here
                // qroInside :
                //   the test Rect is fully inside the child Rect,
                //   but the child does not exists.
                //   So the sum of the estimated Items must be zero
                // qroNone :
                //   Part of child not hit
                // In all cases the parts Rect sum remains 0.0
              end;
            end;
          end;
          // Take the sum collected from the existing children
          Result := directitemsum;
          // Now we have to calculate the remaining items. These are items who are
          // either are allready "in" or definitly "out".
          remainingitems := ANode.TotalNodeItemsCount-(directitemsum+directexcludedsum);
          // Process the rest.
          // if at least one quarter has to be counted and at least one item is remaining.
          if (qarterscnt > 0) and (remainingitems >= 1.0) then
          begin
            d := (qarterscnt/4.0)*remainingitems; // Adding this remaining fraction of items
            Result := Result + d;
          end;
        end;
      end;
    qroEqual, qroFull :
      begin
        // Equal means, this node is the only node that contains relevant information
        // Full means, the test Rect is larger than this nodes Rect, so all items of this node needs to be counted.
        // Other nodes, but not child node here, has to be queried too.
        Result := ANode.TotalNodeItemsCount;
      end;
  else
  end;
end;

function TLazQuadTree.DoGetItemCoord(AItem: TObject; var AX, AY: Double
  ): Boolean;
begin
  Result := False;
  if not Assigned(FGetItemCoordEvent) then Exit; // The callback is not assigned
  Result := FGetItemCoordEvent(Self,AItem,AX,AY);
end;

function TLazQuadTree.DoGetItemRect(AItem: TObject;
  var AItemRect: TLQTRect): Boolean;
begin
  Result := False;
  if not Assigned(FGetItemRectEvent) then Exit; // The callback is not assigned
  Result := FGetItemRectEvent(Self,AItem,AItemRect);
end;

function TLazQuadTree.DoGetItemCaption(AItem: TObject; var AItemCaption: String
  ): Boolean;
begin
  Result := False;
  if not Assigned(FGetItemCaptionEvent) then Exit; // The callback is not assigned
  Result := FGetItemCaptionEvent(Self,AItem,AItemCaption);
end;

function TLazQuadTree.TreeItemToListEnumerationCallBack(Sender: TObject;
  ANode: TLQTNode; ANodeItem: TObject; ANodeItemIndex: Integer;
  AUserData: Pointer): Boolean;
begin
  Result := Assigned(AUserData);
  if Assigned(ANodeItem) then
    TList(AUserData).Add(ANodeItem);
end;

function TLazQuadTree.TreeNodeToListEnumerationCallBack(Sender: TObject;
  ANode: TLQTNode; AUserData: Pointer): Boolean;
begin
  Result := Assigned(AUserData);
  if Assigned(ANode) then
    TList(AUserData).Add(ANode);
end;

procedure TLazQuadTree.WriteFileStreamStart(const AStream: TStream);
begin
  AStream.WriteDWord(QuadTreeFileStartCode);      // 0: Start Code
end;

procedure TLazQuadTree.WriteStreamTreeHeader(const AStream: TStream;
  const AQuadTreeStreamDataVersion: Cardinal);
begin
  AStream.WriteDWord(AQuadTreeStreamDataVersion); // 1: Tree Data Version
end;

procedure TLazQuadTree.WriteStreamTreeProperties(const AStream: TStream;
  const AQuadTreeStreamDataVersion: Cardinal);
var
  c : Cardinal;
begin
  if FOwnsObjects then                            // 2: OwnsObjects
    AStream.WriteWord(1)
  else
    AStream.WriteWord(0);
  WriteIntegerToStream(AStream,FMaxTreeLevel); // 3: FMaxTreeLevel
  c := SizeOf(FWorldModel);
  AStream.WriteDWord(c);                           // 4: SizeOf FWorldModel
  AStream.WriteBuffer(FWorldModel,c);              // 5: FWorldModel
  WriteIntegerToStream(AStream,FMaxNodeItemCount); // 6: FMaxNodeItemCount
end;

procedure TLazQuadTree.WriteStreamTreeRootNode(const AStream: TStream;
  const AQuadTreeStreamDataVersion: Cardinal);
begin
  FRootNode.WriteNodeToStream(AStream, AQuadTreeStreamDataVersion);         // 7: RootNode (and all Children)
end;

procedure TLazQuadTree.ReadFileStreamStart(const AStream: TStream);
var
  c : Cardinal;
begin
  c := AStream.ReadDWord; // 0: Start Code
  if c <> QuadTreeFileStartCode then
    raise ELQTStreamException.Create('Wrong StartCode, probably wrong file type');
end;

procedure TLazQuadTree.ReadStreamTreeHeader(const AStream: TStream; out
  AQuadTreeStreamDataVersion: Cardinal);
begin
  AQuadTreeStreamDataVersion := AStream.ReadDWord; // 1: Tree Data Version
  // If changes are necessary a
  // case ((AQuadTreeStreamDataVersion and lqtStreamDataVersionUnitMajorMask) shr lqtStreamDataVersionUnitMajorShr) of
  //   implementation needed to catch the changes an stay backward compatible
  if (AQuadTreeStreamDataVersion and lqtStreamDataVersionUnitMajorMask) <> (DefaultQuadTreeStreamDataVersion and lqtStreamDataVersionUnitMajorMask) then
    raise ELQTStreamException.CreateFmt('StreamData tree-version %d not supported! Position: %d',
                                        [AQuadTreeStreamDataVersion, AStream.Position-Sizeof(Cardinal)]);
end;

procedure TLazQuadTree.ReadStreamTreeProperties(const AStream: TStream;
  const AQuadTreeStreamDataVersion: Cardinal);
var
  c : Cardinal;
begin
// If changes are necessary a
// case ((AQuadTreeStreamDataVersion and $00FF0000) shr 16) of
//   implementation needed to catch the changes an stay backward compatible
  c := AStream.ReadWord;            // 2: FOwnsObjects
  case c of
    0 : FOwnsObjects := False;
    1 : FOwnsObjects := True;
  else
    raise ELQTStreamException.CreateFmt('Value %d out of range for OwnsObjects! Position: %d',
                                        [c, AStream.Position-Sizeof(Word)]);
  end;

  FMaxTreeLevel := ReadIntegerFromStream(AStream);   // 3: FMaxTreeLevel

  c := AStream.ReadDWord;                    // 4: SizeOf FWorldModel
  if c <> SizeOf(FWorldModel) then
    raise ELQTStreamException.CreateFmt('SizeOf WorldModel %d differ, expected %d! Position: %d',
                                         [c, SizeOf(FWorldModel), AStream.Position - SizeOf(Cardinal)]);
  AStream.ReadBuffer(FWorldModel,c);  // 5: FWorldModel

  FMaxNodeItemCount := ReadIntegerFromStream(AStream);   // 6: FMaxNodeItemCount
end;

procedure TLazQuadTree.ReadStreamTreeRootNode(const AStream: TStream;
  const AQuadTreeStreamDataVersion: Cardinal);
var
  posi : Int64;
  nt : Cardinal;
  n : TLQTNode;
begin
  posi := AStream.Position;
  nt := AStream.ReadDWord; // Read NodeToken
  if nt <> FRootNode.NodeToken then // A different Class as the current RootNode is in the stream
  begin
    // Try to create the Node, by asking the User (or the drived class)
    n := InternalCreateNode(Nil,FRootNode.NodeRect, nt);
    if not Assigned(n) then
      raise ELQTStreamException.CreateFmt('Unable to create RootNode for QuadNodeToken 0x%8.8x! Position: %d',
                                          [nt, AStream.Position-SizeOf(Cardinal)]);
    // Free the old RootNode and replace the root node by the new node
    FRootNode.Free;
    FRootNode := n;
  end;
  AStream.Position := posi; // Rewind the Stream
  FRootNode.ReadNodeFromStream(AStream, AQuadTreeStreamDataVersion);         // 7: RootNode (and all Children)
end;

function TLazQuadTree.GetOccupiedRect: TLQTRect;
begin
  Result := FRootNode.OccupiedRect;
end;

function TLazQuadTree.GetActualNodeRect(const ANode: TLQTNode): TLQTRect;
begin
  if not Assigned(ANode) then
    Result := FRootNode.NodeRect
  else
    Result := ANode.NodeRect;
end;

function TLazQuadTree.GetActualNodeRectAsTetragonCorners(ANode : TLQTNode; Index: Integer): TLQTPoint;
var
  r : TLQTRect;
begin
  r := GetActualNodeRect(ANode);
  Index := Index mod 4;
  case Index of
    0 : begin
          // Top-Left corner
          Result.X := r.Left;
          Result.Y := r.Top;
        end;
    1 : begin
          // Top-right corner
          Result.X := r.Right;
          Result.Y := r.Top;
        end;
    2 : begin
          // Bottom-right corner
          Result.X := r.Right;
          Result.Y := r.Bottom;
        end;
  else
    // 3
    // Bottom-left corner
    Result.X := r.Left;
    Result.Y := r.Bottom;
  end;
end;

function TLazQuadTree.GetActualOccupiedRect(const ANode: TLQTNode
  ): TLQTRect;
begin
  if not Assigned(ANode) then
    Result := FRootNode.OccupiedRect
  else
    Result := ANode.OccupiedRect;
end;

function TLazQuadTree.GetActualOccupiedRectAsTetragonCorners(
  ANode: TLQTNode; Index: Integer): TLQTPoint;
var
  r : TLQTRect;
begin
  r := GetActualOccupiedRect(ANode);
  Index := Index mod 4;
  case Index of
    0 : begin
          // Top-Left corner
          Result.X := r.Left;
          Result.Y := r.Top;
        end;
    1 : begin
          // Top-right corner
          Result.X := r.Right;
          Result.Y := r.Top;
        end;
    2 : begin
          // Bottom-right corner
          Result.X := r.Right;
          Result.Y := r.Bottom;
        end;
  else
    // 3
    // Bottom-left corner
    Result.X := r.Left;
    Result.Y := r.Bottom;
  end;
end;


function TLazQuadTree.InsertItem(const AItem: TObject;
  const APointValid: Boolean; const ARectValid: Boolean; const APointX: Double;
  const APointY: Double; const ARectLeft: Double; const ARectTop: Double;
  const ARectRight: Double; const ARectBottom: Double): TLQTNode;
var
  lpntvalid, lRectvalid : Boolean;
  lX, lY : Double;
  lItemRect : TLQTRect;
  li : TLQTNodeChildLocation;
  pt : TLQTPoint;
begin
  Result := Nil;
  if not Assigned(AItem) then Exit; // Data item is not valid
  lpntvalid := APointValid;
  lRectvalid := ARectValid;;
  lX := APointX;
  lY := APointY;
  lItemRect.Left := ARectLeft;
  lItemRect.Top := ARectTop;
  lItemRect.Right := ARectRight;
  lItemRect.Bottom := ARectBottom;
  if (not APointValid) and (not ARectValid) then
  begin
    lRectvalid := DoGetItemRect(AItem,lItemRect);  // Get the items Rect, if defined
    lpntvalid := DoGetItemCoord(AItem,lX,lY); // Get items coord
  end;
  if (not lpntvalid) and (not lRectvalid) then Exit;
  if lRectvalid then
    Result := FindRectNode(lItemRect); // Find the node who contains the Rect
  if not Assigned(Result) and lpntvalid then
  begin
    Result:= FindCoordNode(lX,lY); // Find the node who contains the Coord

    // The following block, gives approx. 5% more speed.
    // If the found node is used, the case will happen, that there are populated child nodes
    // but not in the area where the item is located.
    // Lets say NW is wantend, but SW has a lot of inserted items.
    // This will cause the item inserted in the node (which is not a leaf node)
    // and this "fault" is ironed out by a subsequently "SplitNode" call.
    // But this call also try to split all other sibling nodes, where nothing
    // changed during this insertion.
    // So in this case, we do the creation of the ne child node here and the
    // SplitNode call will only affect the new created node and is more or less for free.
    if Assigned(Result) and // We have a node
       Result.HasChildNodes and //The node has children
       (Result.NodeLevel < FMaxTreeLevel) then // and is not a the very bottom of the tree
    begin
      // Let's create a new empty child node
      pt.X := lX;
      pt.Y := lY;
      // Iterate through the 4 areas to find the one we like
      for li := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
      begin
        // Check if this area is the one we have to go in
        if Result.ChildRects[li].ContainsPoint(pt) then // We can use the faster ContainsPoint here, since the FNodeRect is guaranteed to be flat!
        begin
          // yes, create the node
          Result := Result.CreateOrGetChildNode(li);
          Break; // and done
        end;
      end;
    end;
    // End of the speed update block
  end;
  if not Assigned(Result) then Exit; // Should never occour, since the Root-Node covers the whole world
  Result.AddNodeItem(AItem,
                     lpntvalid,lRectvalid,lX,lY,
                     lItemRect.Left,lItemRect.Top,lItemRect.Right,lItemRect.Bottom
                     ); // Add the data
  if Result.SplitNode then // Split the node (if needed, this is decided there)
    Result := FindItemNode(AItem); // If the node was splitted, return the insertion node

  // Iterate upwards to update the OccupiedRect
  if Assigned(Result) then
    Result.UpdateOccupiedRect;
end;

function TLazQuadTree.ExtractItem(const AItem: TObject): TObject;
var
  qn : TLQTNode;
  ndx : Integer;
begin
  Result := Nil;
  qn := FindItemNode(AItem);
  if not Assigned(qn) then Exit;
  ndx := qn.NodeItemIndex(AItem);
  if ndx < 0 then Exit; // should/could never occour, since FindItemNode was successful
  Result := qn.ExtractNodeItem(ndx);
end;

function TLazQuadTree.MoveItemAfterCoordChange(const AItem: TObject): TLQTNode;
var
  qn : TLQTNode;
  lpntvalid, lRectvalid : Boolean;
  lX,lY : Double;
  lItemRect : TLQTRect;
  ndx : Integer;
begin
  Result := Nil;
  lX := 0.0; // Keep the compiler calm
  lY := 0.0;
  lItemRect.Left := 0.0;
  lItemRect.Right := 0.0;
  lItemRect.Top := 0.0;
  lItemRect.Bottom := 0.0;

  qn := FindItemNode(AItem);
  if not Assigned(qn) then Exit;

  // Find the new node, where the new position fits
  lpntvalid := False;
  lRectvalid := DoGetItemRect(AItem,lItemRect);  // Get the items Rect, if defined
  if lRectvalid then
    Result := FindRectNode(lItemRect) // Find the node who contains the Rect
  else
  begin
    lpntvalid := DoGetItemCoord(AItem,lX,lY);
    if not lpntvalid then Exit; // No Rect and no coord of the item could be retrieved, exit
    Result := FindCoordNode(lX,lY); // Find the node who contains the Coord
  end;
  if not Assigned(Result) then Exit; // Should never occour, since the Root-Node covers the whole world
  // if the target node is the same as the current node,
  // no further action is required
  if Result <> qn then
  begin
    // since we know the node already, we can shortcut a bit, rather than calling ExtractItem(AItem);
    ndx := qn.NodeItemIndex(AItem);
    if ndx < 0 then Exit; // should/could never occour, since FindItemNode was successful
    qn.ExtractNodeItem(ndx);
    // since we know the node already, we can shortcut a bit, rather than calling InsertItem(AItem);
    Result.AddNodeItem(AItem,
                       lpntvalid,lRectvalid,lX,lY,
                       lItemRect.Left,lItemRect.Top,lItemRect.Right,lItemRect.Bottom); // Add the data
    if Result.SplitNode then // Split the node (if needed, this is decided there)
      Result := FindItemNode(AItem); // The the node was splitted, return the insertion node
  end;
end;

procedure TLazQuadTree.GetActualNodeTetragon(const ANode: TLQTNode; out
  AActualBoundingtTetragon: TLQTPointArr);
var
  r : TLQTRect;
begin
  AActualBoundingtTetragon := Nil;
  r := GetActualNodeRect(ANode);
  SetLength(AActualBoundingtTetragon,5);
  // Top-Left corner
  AActualBoundingtTetragon[0].X := r.Left;
  AActualBoundingtTetragon[0].Y := r.Top;
  // Top-right corner
  AActualBoundingtTetragon[1].X := r.Right;
  AActualBoundingtTetragon[1].Y := r.Top;
  // Bottom-right corner
  AActualBoundingtTetragon[2].X := r.Right;
  AActualBoundingtTetragon[2].Y := r.Bottom;
  // Bottom-left corner
  AActualBoundingtTetragon[3].X := r.Left;
  AActualBoundingtTetragon[3].Y := r.Bottom;

  // closing leg to top-left corner
  AActualBoundingtTetragon[4] := AActualBoundingtTetragon[0];
end;

function TLazQuadTree.FindRectNode(const ALQTRect: TLQTRect
  ): TLQTNode;
{
  There is always an ordered List of Nodes, beginning from the RootNode,
  down to the smallest rectangle, that contains the passed coordinate.
  If the Tree is empty, than the result will be the RootNode.
}
var
  node : TLQTNode;
  i : TLQTNodeChildLocation;
  found : Boolean;
  hit : TLQTRectOverlapKind;
begin
  // the RootNode contains the world
  Result := FRootNode;
  hit := FRootNode.NodeRectOverlapTest(ALQTRect);
  if (hit = qroEqual) or // node's Rect is equal
     (hit = qroFull) or // test Rect covers the node fully
     (hit = qroPartly) then // test Rect covers the node partly
    Exit; // No chance to find something better

  while True do
  begin
    // Assume no further finding
    found := False;
    // Iterate through the four children (if existing)
    for i := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    begin
      node := Result.ChildNodes[i]; // Get the child
      if Assigned(node) then
      begin
        hit := node.NodeRectOverlapTest(ALQTRect);
        if (hit = qroEqual) then // node's Rect is equal
        begin
          Result := node; // take this child as the (better = smaller rectangle) result
          Exit; // No chance to find something better
        end
        else if (hit = qroInside) then
        begin
          Result := node; // take this child as the (better = smaller rectangle) result
          found := True; // remember we have found something
          Break; // stop searching in the other children, there is no chance to find something
        end
        else if (hit = qroPartly) or // test Rect covers the node partly
                (hit = qroFull) then // test Rect covers the node fully
        begin
          Exit; // stop searching, the Result (parents node) is the best we can find
        end;
      end;
    end;
    if not found then Break; //if nothing further found, than keep the last result
  end;
end;
function TLazQuadTree.FindCoordNode(const AX, AY: Double): TLQTNode;
begin
  // the RootNode contains the world
  Result := FRootNode.FindCoordNode(AX,AY);
end;


function TLazQuadTree.FindItemNode(const AItem: TObject): TLQTNode;
var
  lX, lY : Double;
  qn : TLQTNode;
  ndx : Integer;
begin
  Result := Nil;
  if not Assigned(AItem) then Exit; // Data item is not valid
  qn := Nil; // The Node where the Item is located (Nil = not found)
  ndx := -1; // The Index of the Item within the node
  lX := 0.0; // Keep the compiler calm
  lY := 0.0;
  if DoGetItemCoord(AItem,lX,lY) then
  begin // the function delivered useful data
    qn := FindCoordNode(lX,lY); // Find the node who contains the Coord
    // Now ask the node if it contains the Item, if not climb up to the to the rootnode
    while Assigned(qn) do
    begin
      ndx := qn.NodeItemIndex(AItem);
      if ndx >= 0 then Break;
      qn := qn.ParentNode;
    end;
  end;
  if ndx < 0 then // We have not found the item's node in the smart way
  begin // we have to do it slowly, by iteration through the Tree
    qn := NextNode(Nil);
    while Assigned(qn) do
    begin
      ndx := qn.NodeItemIndex(AItem);
      if ndx >= 0 then Break; //Item and Item's node found!
      qn := NextNode(qn);
    end;
  end;
  if ndx < 0 then Exit; //The Item was not found, neither smart, nor the long way
  Result := qn;
end;

function TLazQuadTree.EnumerateNodesInRect(
  const ALQTRect: TLQTRect;
  const AEnumCallBackFunc: TLQTNodeEnumerationCallBack;
  const AUserData: Pointer): Boolean;
var
  lNormRects : TLQTRectArray; // The normalized Rectangles
  lNormWorld : TLQTRect;
  i : Integer;
begin
  Result := False;
  NormalizeRectangle(ALQTRect, FRootNode.NodeRect, FWorldModel,
                    lNormRects,lNormWorld);
  for i := 0 to High(lNormRects) do
  begin
    if lNormWorld.WorldXAxisLeftPlus <> WorldXAxisLeftPlus then
    begin
      lNormRects[i].Left:= -lNormRects[i].Left;
      lNormRects[i].Right:= -lNormRects[i].Right;
    end;
    if lNormWorld.WorldYAxisUpPlus <> WorldYAxisUpPlus then
    begin
      lNormRects[i].Top:= -lNormRects[i].Top;
      lNormRects[i].Bottom:= -lNormRects[i].Bottom;
    end;
    Result := InternalEnumerateNodesInRect(FRootNode,lNormRects[i],AEnumCallBackFunc, AUserData);
    if not Result then Exit;
  end;
end;

procedure TLazQuadTree.ListNodesInRect(
  const ALQTRect: TLQTRect; const AList: TList);
begin
  EnumerateNodesInRect(ALQTRect,@TreeNodeToListEnumerationCallBack, AList);
end;

function TLazQuadTree.CountNodesInRect(
  const ALQTRect: TLQTRect): Integer;
var
  lNormRects : TLQTRectArray; // The normalized Rectangles
  lNormWorld : TLQTRect;
  i : Integer;
begin
  Result := 0;
  NormalizeRectangle(ALQTRect, FRootNode.NodeRect, FWorldModel,
                    lNormRects,lNormWorld);
  for i := 0 to High(lNormRects) do
  begin
    if lNormWorld.WorldXAxisLeftPlus <> WorldXAxisLeftPlus then
    begin
      lNormRects[i].Left:= -lNormRects[i].Left;
      lNormRects[i].Right:= -lNormRects[i].Right;
    end;
    if lNormWorld.WorldYAxisUpPlus <> WorldYAxisUpPlus then
    begin
      lNormRects[i].Top:= -lNormRects[i].Top;
      lNormRects[i].Bottom:= -lNormRects[i].Bottom;
    end;
    Result := Result + InternalCountNodesInRect(FRootNode,lNormRects[i]);
  end;
end;

function TLazQuadTree.EnumerateItemsInRect(const ALQTRect: TLQTRect;
  const AEnumCallBackFunc: TLQTItemEnumerationCallBack;
  const AUserData: Pointer): Boolean;
var
  lNormRects : TLQTRectArray; // The normalized Rectangles
  lNormWorld : TLQTRect;
  i : Integer;
begin
  Result := False;
  NormalizeRectangle(ALQTRect, FRootNode.NodeRect, FWorldModel,
                    lNormRects,lNormWorld);
  for i := 0 to High(lNormRects) do
  begin
    if lNormWorld.WorldXAxisLeftPlus <> WorldXAxisLeftPlus then
    begin
      lNormRects[i].Left:= -lNormRects[i].Left;
      lNormRects[i].Right:= -lNormRects[i].Right;
    end;
    if lNormWorld.WorldYAxisUpPlus <> WorldYAxisUpPlus then
    begin
      lNormRects[i].Top:= -lNormRects[i].Top;
      lNormRects[i].Bottom:= -lNormRects[i].Bottom;
    end;
    Result := InternalEnumerateItemsInRect(FRootNode,lNormRects[i],AEnumCallBackFunc, AUserData);
    if not Result then Exit;
  end;
end;

procedure TLazQuadTree.ListItemsInRect(const ALQTRect: TLQTRect;
  const AList: TList);
begin
  EnumerateItemsInRect(ALQTRect,@TreeItemToListEnumerationCallBack, AList);
end;

function TLazQuadTree.CountAssignedItemsInRect(
  const ALQTRect: TLQTRect): Integer;
var
  lNormRects : TLQTRectArray; // The normalized Rectangles
  lNormWorld : TLQTRect;
  i : Integer;
begin
  Result := 0;
  NormalizeRectangle(ALQTRect, FRootNode.NodeRect, FWorldModel,
                    lNormRects,lNormWorld);
  for i := 0 to High(lNormRects) do
  begin
    if lNormWorld.WorldXAxisLeftPlus <> WorldXAxisLeftPlus then
    begin
      lNormRects[i].Left:= -lNormRects[i].Left;
      lNormRects[i].Right:= -lNormRects[i].Right;
    end;
    if lNormWorld.WorldYAxisUpPlus <> WorldYAxisUpPlus then
    begin
      lNormRects[i].Top:= -lNormRects[i].Top;
      lNormRects[i].Bottom:= -lNormRects[i].Bottom;
    end;
    Result := Result + InternalCountAssignedNodeItemsInRect(FRootNode,lNormRects[i]);
  end;
end;

function TLazQuadTree.EstimatedCountOfAssignedItemsInRect(
  const ALQTRect: TLQTRect; const AItemThreshold: Integer): Double;
var
  lNormRects : TLQTRectArray; // The normalized Rectangles
  lNormWorld : TLQTRect;
  i : Integer;
begin
  Result := 0;
  NormalizeRectangle(ALQTRect, FRootNode.NodeRect, FWorldModel,
                    lNormRects,lNormWorld);
  for i := 0 to High(lNormRects) do
  begin
    if lNormWorld.WorldXAxisLeftPlus <> WorldXAxisLeftPlus then
    begin
      lNormRects[i].Left:= -lNormRects[i].Left;
      lNormRects[i].Right:= -lNormRects[i].Right;
    end;
    if lNormWorld.WorldYAxisUpPlus <> WorldYAxisUpPlus then
    begin
      lNormRects[i].Top:= -lNormRects[i].Top;
      lNormRects[i].Bottom:= -lNormRects[i].Bottom;
    end;
    Result := Result + InternalEstimatedCountOfAssignedItemsInRect(FRootNode,lNormRects[i],AItemThreshold);
  end;
end;

function TLazQuadTree.AgglomerateNodeAsEnumeration(const ANode: TLQTNode;
  const AAgglomerationDepth: Integer;
  const AMinimumItemNumberInAgglomeration: Integer;
  const AItemEnumFunc: TLQTItemEnumerationCallBack;
  const AItemUserData: Pointer;
  const AAgglomerationEnumFunc: TLQTAgglomerationEnumerationCallBack;
  const AAgglomerationUserData: Pointer): Boolean;
begin
  Result := False;
  if (not Assigned(ANode)) or (ANode.Owner <> Self) then Exit;
  Result := EnumerateItemsInParentNodes(ANode,AItemEnumFunc,AItemUserData); // Enumerate all Items above
  if not Result then Exit;
  // Agglomerate everything below
  Result := ANode.AgglomerateNodeAsEnumeration(
                    AAgglomerationDepth,AMinimumItemNumberInAgglomeration,
                    AItemEnumFunc, AItemUserData,
                    AAgglomerationEnumFunc, AAgglomerationUserData);
end;

procedure TLazQuadTree.AgglomerateNodeAsList(const ANode: TLQTNode;
  const AAgglomerationDepth: Integer;
  const AMinimumItemNumberInAgglomeration: Integer;
  var AAgglomerationArray: TLQTAgglomerationArray;
  const AItemList: TList);
begin
  if (not Assigned(ANode)) or (ANode.Owner <> Self) then Exit;
  AgglomerateNodeAsEnumeration(ANode,AAgglomerationDepth, AMinimumItemNumberInAgglomeration,
                               @TreeItemToListEnumerationCallBack,AItemList,
                               @ANode.NodeAgglomerationToArrayEnumerationCallBack,@AAgglomerationArray);
end;

procedure TLazQuadTree.TreeAsStrings(const AStrings: TStrings);
begin
  AStrings.Clear;
  AStrings.Add('QuadTree');
  AStrings.Add('+ RootNode');
  FRootNode.NodeAsStrings(AStrings);
end;

procedure TLazQuadTree.ListTreeNodes(const AList: TList);
begin
  ListSubNodes(Nil,AList);
end;

{ NextNode will return the next node in the tree, behind the given node.
    If the passed node is empty, the RootNode is returned. If StartNode is the last node,
    than Nil is returned.
  The method of getting the next node is simple, but not easy to understand (maybe / hopefully
  only for me!):
  All *branches* are travelled downwards, all *leaves* upwards.
  So if AStartNode has children, the first Child is always returned.
  If AStartNode has no children, than the point of view is changed. Now we are looking
  to the siblings of AStartNode (since the ParentNode has been returned before!!).
  We look which one of the four (possible) siblings is the "next" behind AStartNode
  and if found one, we return this sibling.
  If AStartNode has no children, and is the last child of its ParentNode, then
  we move in the tree upwards to the GrandParentNode etc. until a sibling found
  or the RootNode is found (this node has no ParentNode).

                         [Root]
              +-------------+-----+-------------+-----------+
           [NE 0]              [SE 0]        [SW 0]       [NW 0]
         +----+---------+                   +---+----+
      [SW 01]       [NW 01]              [NE 01] [SE 01]
     +---+-----+
 [NE 012] [SE 012]

 Calling in a loop
 qn := Nil;
 repeat
   qn := NextNode(qn);
   if not Assigned(qn) then Break;
   // Print qn's Name here
 until False;

 qn would loop 11 times and print:
 Root
 NE 0
 SW 01
 NE 012
 SE 012
 NW 01
 SE 0
 SW 0
 NE 01
 SE 01
 NW 0
 (Nil)

}
function TLazQuadTree.NextNode(const AStartNode: TLQTNode): TLQTNode;
var
  pn, cn, sibling : TLQTNode;
  i : TLQTNodeChildLocation;
  foundThis : Boolean;
begin
  if not Assigned(AStartNode) then
    Exit(FRootNode);
  Result := Nil;
  if AStartNode.Owner <> Self then
    Exit; // Not the right tree
  // Find the first assigned child and return this child
  for i := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
  begin
    cn := AStartNode.ChildNodes[i]; // Get the child
    if Assigned(cn) then
      Exit(cn); // child exists than return the child
  end;
  // This node has no children, so we have to look in the parents child list to find siblings
  // So the Point of View is now from AStartNode beeing a child
  cn := AStartNode;
  pn := AStartNode.ParentNode;
  while Assigned(pn) do
  begin
    foundThis := False;
    for i := Low(TLQTNodeChildLocation) to High(TLQTNodeChildLocation) do
    begin
      sibling := pn.ChildNodes[i]; // Get the child
      if Assigned(sibling) and foundThis then // If we foundThis ourself in one of the loops before, this is the next sibling
        Exit(sibling); // Return this sibling and exit
      if sibling = cn then // this sibling is this node, so we found our self, store this knowledge and the next sibling is our result!
        foundThis := True;
    end;
    if not foundThis then Exit; // Should never happen
    // We found no other sibling than ourself (= we are the last sibling)
    // Than move one level up. Make the new child the current parent, and the parent the parent of the new child
    cn := pn;
    pn := cn.ParentNode;
  end;
end;

function TLazQuadTree.NextItem(out ANextItem: TObject; out
  ANextItemNode: TLQTNode; const AStartItem: TObject; const AStartItemNode: TLQTNode
  ): Boolean;
var
  lItem : TObject;
  lNode : TLQTNode;
  ndx : Integer;
begin
  Result := False;
  lItem := AStartItem;
  lNode := AStartItemNode;
  if not Assigned(lNode) then
    lNode := NextNode(); // A not assigned StartNode will take the RootNode
  if not Assigned(lItem) then // A not assigned StartItem, will take the next item
    ndx := -1
  else
  begin // Item is assigned
    ndx := lNode.NodeItemIndex(lItem); // get the index of the item in the given node
    if ndx < 0 then // The item is not part of this node
    begin
      lNode := FindItemNode(lItem); // find the node of the item
      if not Assigned(lNode) then // item is not part of this tree
      begin
        ANextItem := Nil;
        ANextItemNode := Nil;
        Exit; // No next item found
      end;
      ndx := lNode.NodeItemIndex(lItem); // store the ndx of the item in the found node
    end;
  end;
  // Here lNode is assigned
  // if lItem is assigned, ndx contains the Index in the node items aray
  // if lItem is Nil, ndx is -1
  // Now we can iterate through the tree
  repeat
    // First step (inner loop)
    // Find an existing item index
    repeat
      Inc(ndx); // Try the next index (-1 => 0, 0 => 1, ... n => High(NodeItems) => next node, -1
      if ndx >= lNode.NodeItemsLength then // test if within the current node items array
      begin // No, behind the current item index
        lNode := NextNode(lNode); // get the next node
        if not Assigned(lNode) then // no node found, done
        begin
          ANextItem := Nil;
          ANextItemNode := Nil;
          Exit;
        end;
        // A next node was found
        ndx := -1; // set the index before the first item
      end;
    until ndx >= 0;
    // Second step (outer loop)
    // Get the item from the found position
    lItem := lNode.NodeItems[ndx];
    // If assigned, we found our next item
    // if not assigned, try the next index
  until Assigned(lItem);
  // We found the next item
  ANextItem := lNode.NodeItems[ndx];
  ANextItemNode := lNode;
  Result := True;
end;

procedure TLazQuadTree.RecountTotalNodeItems;
var
  qn : TLQTNode;
  cnt, sum : Integer;

begin
  // First step: Iterate to all Nodes and set the
  // FTotalNodeItemsCount field to 0.
  qn := Nil;
  repeat
    qn := NextNode(qn);
    if not Assigned(qn) then Break;
    qn.FTotalNodeItemsCount := 0; // Set the field variable direct instead of the property
  until False;

  //  + REMOVEEMPTYPROCS + UNUSEDPARA + REGVAR + STACKFRAME + TAILREC + CSE
  {.$optimization noREGVAR}
{
// This loop will never terminate
  qn := FRootNode;
  repeat
    qn.FTotalNodeItemsCount := 0; // Set the field variable direct instead of the property
    qn := NextNode(qn);
  until not Assigned(qn);

// This loops terminates and does never raise an exception
  qn := FRootNode;
  repeat
    qn.FTotalNodeItemsCount := 0; // Set the field variable direct instead of the property
    nn := qn;
    qn := NextNode(qn);
    if qn = nn then
      raise Exception.Create('Why???');
  until not Assigned(qn);
}


  // Second step: Iterate to all Nodes and increment the
  // TotalNodeItemsCount property by the UsedNodeItemsCount
  // of each node.
  // Since modifiying the TotalNodeItemsCount updates all the
  // parent nodes up to the RootNode, the summation is done.
  sum := 0;
  qn := Nil;
  repeat
    qn := NextNode(qn);
    if not Assigned(qn) then Break;
    cnt := qn.AssignedNodeItemsCount;
    qn.TotalNodeItemsCount := qn.FTotalNodeItemsCount + cnt; // Write to the property, to force the action
    sum := sum + cnt;
  until False;

  if FRootNode.FTotalNodeItemsCount <> sum then
    raise ELQTException.Create('Something went terrible wrong in RecountTotalNodeItems');

end;

procedure TLazQuadTree.RebuildOccupiedRect;
begin
  FRootNode.RebuildOccupiedRect;
end;

procedure TLazQuadTree.Clear;
begin
  FRootNode.Clear;
end;

procedure TLazQuadTree.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TLazQuadTree.EndUpdate;
begin
  if FUpdateCount > 0 then
     Dec(FUpdateCount);
  if FUpdateCount <= 0 then
  begin
    FRootNode.SplitNode;
    FRootNode.PackNode;
  end;
end;

procedure TLazQuadTree.WriteTreeToStream(const AStream: TStream);
var
  sh : TLQTStreamWriteHelper;
begin
  AStream.WriteDWord(FTreeToken);
  sh := TLQTStreamWriteHelper.Create(AStream);
  try
    WriteStreamTreeHeader(AStream,DefaultQuadTreeStreamDataVersion);
    WriteStreamTreeProperties(AStream,DefaultQuadTreeStreamDataVersion);
    WriteStreamTreeRootNode(AStream,DefaultQuadTreeStreamDataVersion);
  finally
    sh.Free;
  end;
end;

procedure TLazQuadTree.ReadTreeFromStream(const AStream: TStream; const AStreamReadOptions : Cardinal = 0);
var
  c : Cardinal;
  startposi : Int64;
  delta : Int64;
  lQuadTreeStreamDataVersion : Cardinal;
  len : Int64;
begin
  Clear; // Just for sanity
  c := AStream.ReadDWord;
  if c <> FTreeToken then
    raise ELQTStreamException.CreateFmt('Wrong DataClass to read 0x%8.8x, expected 0x%8.8x! Position: %d',
                                        [c, FTreeToken,AStream.Position-SizeOf(Cardinal)]);
  len := AStream.ReadDWord;
  startposi := AStream.Position;
  ReadStreamTreeHeader(AStream,lQuadTreeStreamDataVersion);
  ReadStreamTreeProperties(AStream,lQuadTreeStreamDataVersion);
  ReadStreamTreeRootNode(AStream,lQuadTreeStreamDataVersion);
  delta := AStream.Position-startposi;
  if delta <> len then
  begin
    // If we read fewer data than expected and the lazy mode is enabled, just seek forward
    if (delta < len) and ((AStreamReadOptions and srofLazyLengthProcessing) <> 0) then
      AStream.Position := startposi + len
    else
      raise ELQTStreamException.CreateFmt('Stream Data corrupted! Read Bytes %d Expected Bytes %d! Position: %d',
                                          [delta, len, AStream.Position]);
  end;
  RecountTotalNodeItems; // Recount all Items and rebuild the TotalItems-Fields
  RebuildOccupiedRect; // Update all OccupiedRect fields of the Nodes
end;

procedure TLazQuadTree.WriteTreeToFile(const AFileName: String);
var
  fs : TFileStream;
begin
  if FileExists(AFileName) then
    DeleteFile(AFileName);
  fs := TFileStream.Create(AFileName,fmCreate);
  try
    WriteFileStreamStart(fs);
    WriteTreeToStream(fs);
  finally
    fs.Free;
  end;
end;

procedure TLazQuadTree.ReadTreeFromFile(const AFileName: String; const AStreamReadOptions : Cardinal = 0);
var
  fs : TFileStream;
begin
  fs := TFileStream.Create(AFileName,fmOpenRead or fmShareDenyWrite);
  try
    fs.Position := 0;
    ReadFileStreamStart(fs);
    ReadTreeFromStream(fs,AStreamReadOptions);
  finally
    fs.Free;
  end;
end;

constructor TLazQuadTree.Create(const AWorldRect : TLQTRect;
                   const AWorldModel : TLQTWorldModel;
                   const AFreeObjects: Boolean = False;
                   const AQuadTreeCreateQuadNodeEvent : TLQTCreateNodeEvent = Nil);
begin
  inherited Create;
  FTreeToken := DefaultQuadTreeToken;
  FOwnsObjects := AFreeObjects;
  FWorldModel := AWorldModel;
  FCreateNodeEvent := AQuadTreeCreateQuadNodeEvent;
  FMaxNodeItemCount := DefaultMaxNodeItemCount;
  FMaxTreeLevel := DefaultMaxTreeLevel;
  FRootNode := InternalCreateNode(Nil,AWorldRect);
end;

destructor TLazQuadTree.Destroy;
begin
  FRootNode.Free;
  inherited Destroy;
end;

{ TLQTStreamWriteHelper }
{ Create will store the current StreamPosition and write a DWord at the
  current position (should be usually the end of the stream) as a placeholder. }
constructor TLQTStreamWriteHelper.Create(const AStream: TStream);
begin
  inherited Create;
  FLengthFieldPosition := -1;
  FStream := AStream;
  FLengthFieldPosition := FStream.Position;
  FStream.WriteDWord($FFFFFFFE); // Illegal Length value
end;
{ Destroy will call WriteLengthField
   (with the needed caution, if the constructor raises an exception). }
destructor TLQTStreamWriteHelper.Destroy;
begin
  WriteLengthField;
  inherited Destroy;
end;

{ WriteLengthField will calculate the length between the current position and the
    stored FLengthFieldPosition, decremented  by 4.
    This value is written at the stored position and the Stream.Position is set
    back to the original value }
procedure TLQTStreamWriteHelper.WriteLengthField;
var
  len : Int64;
  posi : Int64;
begin
  // Make sure the constructor creates some reasonable values
  if (FLengthFieldPosition < 0) or (not Assigned(FStream)) then Exit;
  posi := FStream.Position;
  try
    len := posi-FLengthFieldPosition-SizeOf(Cardinal);
    FStream.Position := FLengthFieldPosition;
    FStream.WriteDWord(len); // the actual length
  finally
    FStream.Position := posi;
  end;
end;


end.

