{
uLazQuadTreeGeometry.pas

  Geometrical functions for the use with (but not limited to) the lazQuadTree

  License: The same modified LGPL as the Free Pascal RTL
           See the file COPYING.modifiedLGPL for more details

  Author: Ekkehard Domning (ekkehard@domning.eu)

  Overview

  The unit operates with 2D-objects on flat and flatened surfaces.
  The used coordinate system for all functions is orthogonal.
  The basic functions assume a default direction of the increment in the X- and Y-Axis.
  The progressive function gives the opportunity to change this.
  There are three word models to use: flat, cylindrical and toroidal.
  The latter two allows a seamless wrap-around from the right to the left edges (and vice versa),
  toroidal also on the top and bottom edges.

  The most specific functions are the point in rectangle, point in polygon and
  the calculation of the kind of overlapping of two rectangles, including the
  covered areas.

  Details are explained in the comments above the specific declarations.

}
unit uLazQuadTreeGeometry;

{$mode ObjFPC}{$H+}
{$modeswitch advancedrecords}

interface

uses
  Classes, SysUtils, Math,
  unextfloat; // Used for the smallest possible increment of an Double value

type
  ELQTGException = class(Exception);

  {TLineRectInteraction defines the three possible interactions between a line and
    a rectangle.}
  TLineRectInteraction = (
    lriOutside,       {The line is located outside the rectangle}
    lriInside,        {The line is located inside the rectangle}
    lriIntersecting); {The line is intersecting with the border of the rectangle }

  { TLQTPoint describes the a single point in the QuadTree-World }
  TLQTPoint = record
    X : Double;
    Y : Double;
  end;
  { TLQTPointArr contains 0 to some TLQTPoint records }
  TLQTPointArr = array of TLQTPoint;

  { TLQTDynDoubleArr an ordinary dynamic double array, to allow out parameters
    to set the length within the called function }
  TLQTDynDoubleArr = array of Double;

  { Introduction into rectangles

    If we are talking about rectangles and worlds, we need to be clear about the
    terms.
    A rectangle is formed by four orthogonal sides,
    named "Left", "Top", "Right" and "Bottom".
    The Terms of Left, Top, Right, Bottom are used as seen on a sheet of paper.
    So "Left" must be left of "Right" and "Top" has to be upper than "Bottom".
    If "Right" and "Left" are equal or "Top" and "Bottom" are equal, than the
    covered Area is zero.
    To allow a seamless und unique tessellation of a greater area, the
    values "Right" and "Bottom" are *not* included in the area,
    while "Left" and "Top" are included!
    Example:

       0 10 20
     0 +--*--+      Left, Top, Right, Bottom
       |A |B |   A:    0,   0,    10,     10
    10 +--+--+   B:   10,   0,    20,     10

    The two areas are not overlapping, even they share the common X-coordinate of "10"!
    Testing the Point (marked as * in the diagram above) X: 10, Y: 0
    to be inside in one of the two areas will: false on A, but true on B.

    The surface of the area is assumed to be flat, or more precise,
    could be flattened without distortion of the surface. Subsequently, the
    distortion of e.g. a curved surface (globe) must be compensated outside
    this unit.

    Rectangles can be malformed, depending on the used world or circumstances.

    The X- and Y-Axis of a rectangle could be setup to have different signs.
    The sign will not touch the "Left" is left from "Right" rule!
  }

  { TLQTWorldModel

    This unit covers three world models: flat, cylindrical and toroidal.

    *On a flat world*
    - Rectangles could be up to infinity size.
      Using infinity is not recommended, due to several limitations.
    - Two rectangles may form only one single overlapping area.
    - A straight line is defined by it's two points, a Polyline is performed by a an
      ordered list of points, a Polygon is closed by the last point of a line having
      the same value as the first.

    *On a cylindrical world*
    - Rectangles could be up to infinity size.
      Using infinity is not recommended, due to several limitations.
    - Rectangles could pass the right and left border of the world, by extent on the
      opposite part of the world. Since (as defined) the upper-left corner is always
      above and left from the bottom-right corner, a rectangle (following the
      world map example as above) with left = +5 and right = -3 is valid.
      Starting with X = 5 and increasing the value to move right(!) until X = 179,
      we will soon pass 180 = -180 and traveling further until the right(!) border at -3.
    - Two rectangles may form one or two overlapping areas.
      Under which circumstances a secondary overlapping area exists?
      Example:
        The Y-Axis could be ignored in the explanation, assuming that the height
        of the two rectangles are same as the world heights.

        The rectangle A covers: Dublin    (X: -7) to Hamburg   (X: +10).
        The rectangle B covers:  Amsterdam (X: +5) to Liverpool (X:  -3).
        Be aware that the rectangle B is very long and reaches around the east-west-border.
        So there exists one overlapping area from Dublin (X: -7) to Liverpool (X: -3) and
        a second one from Amsterdam (X: +5) to Hamburg (X: +10).
                                Dublin  Liverp.   Amstd.  Hamburg
                  -180 =========== -7 == -3 ===== +5 ==== +10 ================ 180
        Rect A                     AAAAAAAAAAAAAAAAAAAAAAAAAA
        Rect B    BBBBBBBBBBBBBBBBBBBBBBBBB       BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB
        Overlap 1                                 11111111111
        Overlap 2                  22222222
      - A straight line is defined by it's two points, a Polygon is performed by a an
        ordered list of points.
        WARNING: In contrast to rectangles, lines between two points can not pass
        the world's border!
        It is not possible to create straight lines (and subsequently polygons)
        passing the worlds border like the rectangles above.
        This is caused by the fact, that e.g. a line between New York and Sydney
        could be drawn in two ways, one across Africa, one across the Pacific Ocean.
        While on a single line it would be possible to define that the first point
        should be always left of the second point, this approach failed in polygons.
        To avoid any ambiguity, line points are only allowed to be within range of
        the world's extension. Thus polygons are also not allowed to pass the
        east-west-border.
        Forming the rectangle B in the example above by a polygon, it is
        necesarry to define two separate polygons.


    *On a toroid-World*
      the capabilities of the cylindrical-World are extent in the Y-axis (Top to Bottom).
    - Rectangles could pass the top and bottom border of the world, by extent on the
      opposite part of the world.
    - Two rectangles may form up to four overlapping rectangles
      Example:

        Diagram 1 shows Rectangle A: (Left, Top, Right, Bottom) 2,2, 7,7
        Diagram 2 shows Rectangle B: (Left, Top, Right, Bottom) 6,6, 3,3

          0123456789      0123456789      0123456789      0123456789
         +----------+    +----------+    +----------+    +----------+
       0 |          |  0 |BBB   BBBB|  0 |...   ....|  0 |          |
       1 |          |  1 |BBB   BBBB|  1 |...   ....|  1 |          |
       2 |  AAAAA   |  2 |BBB   BBBB|  2 |..1***2...|  2 |  1   2   |
       3 |  AAAAA   |  3 |          |  3 |  *****   |  3 |          |
       4 |  AAAAA   |  4 |          |  4 |  *****   |  4 |          |
       5 |  AAAAA   |  5 |          |  5 |  *****   |  5 |          |
       6 |  AAAAA   |  6 |BBB   BBBB|  6 |..3***4...|  6 |  3   4   |
       7 |  AAAAA   |  7 |BBB   BBBB|  7 |...   ....|  7 |          |
       8 |          |  8 |BBB   BBBB|  8 |...   ....|  8 |          |
       9 |          |  9 |BBB   BBBB|  9 |...   ....|  9 |          |
         +----------+    +----------+    +----------+    +----------+
       Diagram 1       Diagram 2       Diagram 3       Diagram 4

         Be aware that going from the upper-left corner (6,6) of the rectangle B to the
         lower right corner, after reaching the point 9,9 (the lower right corner of
         the world), the next more right an more downward(!) point is the
         overflow to 0,0, which is the upper left corner of the world!
       Diagram 3 shows all rectangles (A = '*', B = '.' and the four overlapping
         rectangles '1' to '4'.
       Diagram 4 only the overlapping rectangles  '1' to '4'.

     - The limitations of Lines and Polygons are the same as for the cylindrical world!
  }
  TLQTWorldModel = (lqtwmFlat, lqtwmCylindrical, lqtwmToroidal);


  { TLQTRect describes an area in the QuadTree-World
    Details to the rectangles are described above.
    The TLQTRect could be used in two ways.
    One forming a world description, the other by forming a rectangle within the world.
    If defining a world, the sign of the axis are taken from the values Top vs Bottom or
    Left vs Right.
    To define a world map one would set
      Left = -180 and Right = 180
      Top = 90 and Bottom = -90
    This means that X increases while traveling from left to right and Y increases
    from top to bottom. This is different from the default definition (e.g. on TCanvas).
    A rectangle on the world could be defined by
      Left = -100 and Right = 100
      Top = 45 and Bottom = -45
  }
  TLQTRect = record
    Left   : Double; {or West border}
    Top    : Double; {or North border}
    Right  : Double; {or East border}
    Bottom : Double; {or South border}
    { function Height returns the absolute Value of Top-Bottom.
        Caution: Do not use this function if IsInfinite is true }
    function Height : Double;
    { function Height returns the heigt of the rectangle in the given world and model.
        This overloaded function will treat wrap around rectangles }
    function Height(const AWorldRect: TLQTRect; const AWorldModel : TLQTWorldModel) : Double;
    { function Width returns the absolute Value of Left-Right.
        Caution: Do not use this function if IsInfinite is true }
    function Width : Double;
    { function Width returns the width of the rectangle in the given world and model.
        This overloaded function will treat wrap around rectangles }
    function Width(const AWorldRect: TLQTRect; const AWorldModel : TLQTWorldModel) : Double;

    { function Area returns the Width * Height of the rectangle in the given world and model.
        This overloaded function will treat wrap around rectangles }
    function Area(const AWorldRect: TLQTRect; const AWorldModel : TLQTWorldModel) : Double;

    { procedure Clear sets all fields to zero }
    procedure Clear;

    { function WorldXAxisLeftPlus returns true if the XAxis increases when
        traveling from Right to Left.
      Remark: This function treats the rectangle as been used as a world description. }
    function WorldXAxisLeftPlus : Boolean;inline;
    { function WorldYAxisUpPlus returns true if the YAxis increases when
        traveling from Bottom to Top.
      Remark: This function treats the rectangle as been used as a world description. }
    function WorldYAxisUpPlus : Boolean;inline;
    { function IsPoint returns true if all coords are concentrated on one point, false if not.
       This means that all sides are equal, this includes non-zero value!}
    function IsPoint: Boolean;
    { function IsLine returns true if either right and left or top and bottom has the same value.}
    function IsLine : Boolean;
    { function IsNotArea returns true if one of the functions IsPoint or IsLine would return true }
    function IsNotArea : Boolean;
    { function IfInfinite returns true if one of the Left, Top, Right or Bottom fields
       is Infinite or NegInfinite or
       bigger than LQTRectMaxValue oder smaller than LQTRectMinValue.
       The Limits are choosen to fit the height and width of a rectangle within the
       range of a Double-Value. }
    function IsInfinite : Boolean;
    { function HasInfiniteXAxis returns true if the Left or Right fields
       is Infinite or NegInfinite or
       bigger than LQTRectMaxValue oder smaller than LQTRectMinValue.
       The Limits are choosen to fit the width of a rectangle within the
       range of a Double-Value. }
    function HasInfiniteXAxis : Boolean;
    { function HasInfiniteYAxis returns true if the Top or Bottom fields
       is Infinite or NegInfinite or
       bigger than LQTRectMaxValue oder smaller than LQTRectMinValue.
       The Limits are choosen to fit the height of a rectangle within the
       range of a Double-Value. }
    function HasInfiniteYAxis : Boolean;
    { function IsValidAsWorld validates, that the rectangle could be used as a
       world definition.
       The result is true if the rectangle forms an area and if one side is infinite
       the opposite side is negative infinite, rsp. vice versa }
    function IsValidAsWorld : Boolean;
    {function IsEqual returns true if all fields having the same values as there
      correspondend fields in the compared record. False is not }
    function IsEqual(ALQTRect : TLQTRect): Boolean;
    {function ContainsPoint returns true if the given point is located between
      Left and Right and Top and Bottom, including Left and Top, excluding Bottom and Right.
      APoint : The point to test
      Remark: The function works on all Axis signs. }
    function ContainsPoint(const ALQTPoint : TLQTPoint): Boolean;overload;
    {function ContainsPoint returns true if the given point is located between
      Left and Right and Top and Bottom, including Left and Top, excluding Bottom and Right.
      Paramers: AX, AY the Coords of the point to test.
      Remark: The function works on all Axis signs. }
    function ContainsPoint(const AX, AY : Double): Boolean;overload;
    { function ContainsPoint returns true if the given point is located between
      Left and Right and Top and Bottom, including Left and Top, excluding Bottom and Right.
        APoint : The point to test
        AWorldRect : The rectangle that describes the world size
        AWorldModel : The world model
      Return: True if the point is located inside the rectangle. }
    function ContainsPoint(const APoint : TLQTPoint;
                           const AWorldRect : TLQTRect;
                           const AWorldModel : TLQTWorldModel = lqtwmFlat) : Boolean;overload;

    {procedure NormateToWorld will force the rectangle to be valid in the given World.
       Parameters
         AWorldRect: The rectangle which describes the extensions of the world.
         AWorldModel : The world model (flat, cylindrical or toroidal) to be used. Default: flat
       Remarks
         The passed world must be valid.
         The signs of the axes of the world are used.
         The Right and Bottom values of the rectangle equal to the correspondent
         of the world, are treated as valid in the meaning, that a rectangle may
         have the same size as the world.
         But if the values are bigger than this assumption is not longer used!
       Caution: if AWorldRect is not valid, an exception is thrown. }
    procedure NormateToWorld(const AWorldRect: TLQTRect; const AWorldModel : TLQTWorldModel = lqtwmFlat);
    { procedure Transform modifys the coords, so that the rectangle is located
        on the same position in the target world}
    procedure Transform(const ASourceWorld, ATargetWorld : TLQTRect);
    { procedure Assign copies the rectangle values from the source rectangle }
    procedure Assign(const ALQTRect: TLQTRect);
    { method SetFrom sets all values of the record from the source values }
    procedure SetFrom(const ALeft, ATop, ARight, ABottom : Double);

    { SetFromPoint sets Left,Right,Top and Bottom, so that the Point meets the critera
        to be inside the rectangle using the ContainsPoint functions
       Caution: The rectangle is set using default X- and Y-Axis-Signs (Left < Right, Top < Bottom }
    procedure SetFromPoint(const APoint : TLQTPoint);
    { SetFromPoint sets Left,Right,Top and Bottom, so that the Point meets the critera
        to be inside the rectangle using the ContainsPoint functions
       The Axis signs are taken from the passed World rectangle}
    procedure SetFromPoint(const APoint : TLQTPoint; const AWorldRect: TLQTRect);

    { function TryExtentWith checks if the passed rectangle and this can be combined.
       If so, one Field is updated.
      Parameters: ALQTRect the retectangle who has to be checked and combined
      Result: True if one Field has been updated, false if not.
      Remarks: A rectangle could be merged if one side of the rectangle is identical
        to the opposite side of the other rectangle.
        So if both rectangle have the same height and the
        Right field of one and the Left field of the other are equal
        than the rectangle will be enlarged and the result is true. }
    function TryExtentWith(const ALQTRect: TLQTRect) : Boolean;overload;
    { function TryExtentWith same as before, but able to detect matching rectangles
        on the other end of the world.
      Parameters:
        ALQTRect the retectangle who has to be checked and combined.
        ANormalizeWorld a rectangle which describes the world borders.
        AWorldModel a description of the world model. If flat the function behaves
          like the function above.
      Result: True if one Field has been updated, false if not. }
    function TryExtentWith(const ALQTRect: TLQTRect; const ANormalizeWorld : TLQTRect; const AWorldModel : TLQTWorldModel = lqtwmFlat) : Boolean;overload;
    { function IsXWrappedAround returns True if Left and Right position of this rectangle are exchanged }
    function IsXWrappedAround(const AWorldRect: TLQTRect) : Boolean;
    { function IsXWrappedAround returns True if Top and Bottom position of this rectangle are exchanged }
    function IsYWrappedAround(const AWorldRect: TLQTRect) : Boolean;
    { function Union combines two rectangles. This rectangle will updated
      Parameter
        ALQTRect the rectangle that should united with this rectangle
        AWorldRectangle the rectangle which describes the world
        AWorldModel the WorldModel we are in (flat, cylindrical or toroidal) }
    function Union(const ALQTRect: TLQTRect; const AWorldRect: TLQTRect; const AWorldModel : TLQTWorldModel = lqtwmFlat) : Boolean;
    { function Union extent the rectangle to the given point.
      Caution: The rectangle is set using default X- and Y-Axis-Signs (Left < Right, Top < Bottom }
    function Union(const APoint : TLQTPoint) : Boolean;
    function Union(const APoint : TLQTPoint;const AWorldRect: TLQTRect; const AWorldModel : TLQTWorldModel = lqtwmFlat) : Boolean;
    { function CoordsAsString returns the coordinates of the rectangle in one String
      Parameter:
        AFormatString if not empty it must contain a valid FormatString.
        There will be 4 parametes passed to the format function
        Left,Top,Right,Bottom (in this order).
        Example: CoordsAsString('L %1.2f, T %1.2f, R %1.2f, B %1.2f');
    }
    function CoordsAsString(const AFormatString : String = '') : String;
    { class operator = overwrites the equal operator to ease the usage of comparing two rectangles
    }
    class operator = (a,b : TLQTRect) : Boolean; // Equal
    { class operator + (Union) will combine two rectangles by there max extents
      The result left and top will be the smallest value of the two rectangles,
      while result right and bottom will be the highest value of the two rectangles.
      The function will only work on a flat and native world
    }
    class operator + (a,b: TLQTRect): TLQTRect;
  end;
  { TLQTRectArray an open array of TLQTRect }
  TLQTRectArray = array of TLQTRect;

  { procedure RemoveNotAreaRectanglesFromArray
      will remove rectangles from the passed array, if the area is zero. }
  procedure RemoveNotAreaRectanglesFromArray(var ARectangles: TLQTRectArray);

  { procedure TransformPoint transforms the location of a point in one world
     to the corresponding location in the other world.
    Parameters :
      Inputs
        AInX, AInY the coordinates in the source world
        ASourceWorld, ATargetWorld the extension of the two worlds
      Outputs
        AOutX, AOutY the coordinates in the target world
    Remarks :
      If both worlds are equal the input is copied to the output without any further processing.
      If not equal the following processing is performed:
       - If at least one two worlds is empy an exception is thrown.
       - If one of the worlds in infinite, while the other is not an exception is thrown.
      Care must be taken if points from a complex world (toroidal, cylindric) transformed into a simpler
      world (flat). Creating gemometric objects (Rectangles, Polygons) from those transformed
      points may change the layout dramaticaly.
  }

  procedure TransformPoint(const AInX, AInY : Double;
                           const ASourceWorld, ATargetWorld : TLQTRect;
                           out AOutX, AOutY : Double);

  { procedure NormatePointToWorld will force the point to be valid in the given World.
     Parameters
       Inputs
        AInX, AInY the coordinates to be normated
        AWorldRect: The rectangle which describes the extensions of the world.
        AWorldModel : The world model (flat, cylindrical or toroidal) to be used.
      Outputs
        AOutX, AOutY the coordinates in the target world
      Remarks
        The passed world must be valid.
        The signs of the axes of the world are used.
        The right and bottom values of the rectangle equal to the correspondent
        of the world, are treated as valid in the meaning, that a rectangle may
        have the same size as the world.
        But if the values are bigger than this assumption is not longer used!
      Caution: if AWorldRect is not valid, an exception is thrown. }
  procedure NormatePointToWorld(const AInX, AInY : Double;
                                const AWorldRect: TLQTRect;
                                const AWorldModel : TLQTWorldModel;
                                out AOutX, AOutY : Double);

  {
     All (if not mentioned other) of the following functions work on regular infinite planes:
     Right is bigger than Left (or X is inreasing to the right).
     Top is smaller than Bottom (or X is increasing to the bottom.
  }

  { function LineIntersection checks wether two lines intersects.
    Parameter: AP0X, AP0Y, AP1X, AP1Y, AP2X, AP2Y, AP3X, AP3Y
               The 4 points forming the two lines
               AIX, AIY the coords of the intersection (if found)
    Result:    True if the lines intersect, otherwise False.
    Source: https://stackoverflow.com/questions/563198/how-do-you-detect-where-two-line-segments-intersect
  }
  function LineIntersection(
             const AP0X, AP0Y, AP1X, AP1Y, AP2X, AP2Y, AP3X, AP3Y : Double;
             out AIX, AIY : Double) : Boolean;
  function LineIntersection(
             const AP0, AP1, AP2, AP3 : TLQTPoint;
             out AI : TLQTPoint) : Boolean;

  { function LineRectIntersection will test whether a line intersects or overlap
    with a rectangle.
    The rectangle is seen as a solid block and the right and bottom coord is *not*
    included in rectangles area.
    Thus if the rectangle is (0,0,100,100) a line (100,-1,100,101) which is overlaying
    the right border of the rectangle will tested as false.
    But a line (0,-1,0,101) which is overlaying to the left side is tested as true.

    Parameters
      ALeft, ATop, ARight, ABottom: The coords of the two points forming the rectangle.
      AP0X, AP0Y, AP1X, AP1Y: The coords forming the two points of the line.
    Result: One of the TLineRectInteractio values (lriOutSide, lriInside, lriIntersecting)
    Remark: The used Coordinatesystem is Right bigger Than Left and Bottom bigger than Top!
    CAUTION: If the line is part of a polygon, the function may may give a false
             positive result if the right / bottom side of the polygon is touching
             the left / top side of the rectangle, since the function has no knowledge
             about the orientation of the segment.
    The Code is complexer/larger than the function LineRectInteractionSimple,
    but the performance is 50% faster.
  }
  function LineRectInteraction(
             const ALeft, ATop, ARight, ABottom : Double;
             const AP0X, AP0Y, AP1X, AP1Y : Double) : TLineRectInteraction;
  function LineRectInteraction(
             const ARect : TLQTRect;
             const AP0, AP1 : TLQTPoint) : TLineRectInteraction;

  { function LineRectIntersectionSimple is given for curiosity only.
      It will test whether a line intersects or overlap  with a rectangle.
      Parameters
        ALeft, ATop, ARight, ABottom: The coords of the two points forming the rectangle.
        AP0X, AP0Y, AP1X, AP1Y: The coords forming the two points of the line.
      Result: One of the TLineRectInteractio values (lriOutSide, lriInside, lriIntersecting)
      Remark: The used Coordinatesystem is Right bigger Than Left and Bottom bigger than Top!
      CAUTION: The right and bottom border of the rectangle is treated to be part of the rectangle!
      REMARK: The Code is simpler/smaller than the function LineRectInteraction, but the performance
      is 35% slower.
  }
  function LineRectInteractionSimple(
             const ALeft, ATop, ARight, ABottom : Double;
             const AP0X, AP0Y, AP1X, AP1Y : Double
             ) : TLineRectInteraction;

  { function PolylineRectangleInterAction will test whether the rectangle and polygon
     intersects or overlap
     The function work on regular infinite planes. The sign of X and Y-Axis could be changed
     Parameters
       ALeft, ATop, ARight, ABottom: The coords of the two points forming the rectangle.
       APolyline: vertex points of a polyline or a polygon with V[n]=V[0].
                The Polyline is defined by pairs of Double values.
                The method of passing an untyped reference allows the usage of
                several datatype like packed records etc.
                A dynamic Array is passed by referencing the first item
                The last point of the Polygon must be identical to the first point,
                to close the polygon.
                Self-Overlapping of the polygon is *not* allowed
       APolylinePointCount: The Number of points in the polyline =
                half the number of doubles passed in APolygon.
       AFirstValueIsY: If True, the first double value in APolygon is the Y-Value of the coordinate.
                If False, the first value is the X-Value (default).
       AXAxisLeftPlus : If True the values of X increasing when going direction left.
                If False, they increase to the right (default)
       AYAxisUpPlus : If True the values of Y increasing when going in direction top.
                If false they decreases (default).
       AFullInsideCheck: If True the the Result lriInside is valid for the full polygon.
                In other words, if the Polygon is completely inside the rectangle
                If False, the Result lriInside might be valid only for the first linesegment
                and an intersecting may occour later, but is not checked.
    Result: One of the TLineRectInteractio values (lriOutside, lriInside, lriIntersecting)
             The meaning is as follows
             lriOutside : All parts of the polygon are outside of the rectangle
             lriInside  : If AFullInsideCheck is False (default): At least one part of the
                            polygon is located in the rectangle.
                          If AFullInsideCheck is True: All Parts of the polygon are
                            inside the rectangle.
             lriIntersecting : The polygon intersects with the rectangle
                          (thus, some parts are inside, others outside the rectangle).
    Remark: To reduce the computation load, the function should only be called,
            if the bounding rectangle of the Polygon is overlapping the test rectangle.
            This test is not performed by the function to avoid code duplication.
  }
  function PolylineRectangleInterAction(
              const ALeft, ATop, ARight, ABottom : Double; // The Rectangle
              const APolyline;
              const APolylinePointCount: Integer;
              const AFirstValueIsY: Boolean = False;
              const AXAxisLeftPlus : Boolean = False;
              const AYAxisUpPlus : Boolean = False;
              const AFullInsideCheck : Boolean = False
           ) : TLineRectInteraction;
type
  { TLazQuadNodeAreaOverlap kind of overlapping between two rectangles A and B
    Overlapping:
    None       Partly   Inside   Equal  Fully
    +--+       +---+    +-----+  +---+  +-----+
    |A |       | A+--+  |A+-+ |  | a |  |B... |
    +--+ +--+  +- |B |  | |B| |  | B |  | .a. |
         |B |     +--+  | +-+ |  +---+  | ... |
         +--+           +---- +         +---- +
  }
  TLQTRectOverlapKind = (
    qroNone   {no overlap},
    qroPartly {the A is partly overlayed by B},
    qroInside {the B is located inside A},
    qroEqual  {the A and B are equal},
    qroFull   {the B covers A fully}
  );

  { function RectanglesOverlap checks whether two rectangles are overlapping and if so
      which are the common part.
    Caution: works only for flat (nonwrapping) rectangles
    parameters
      ALeft, ATop, ARight, ABottom : The Rectangle A
      BLeft, BTop, BRight, BBottom : The Rectangle B
    variable
      CLeft, CTop, CRight, CBottom : The overlapping Rectangle
    return value: True if the rectangles overlapping (and the CLeft ...CBottom values are valid)
                  False if the rectangle are not overlapping.
    Remarks: If parts of the right side of one of the rectangle are identical to parts the left side or
             top side and bottom side, the rectangles are *not* overlapping!
  }
  function RectanglesOverlap(const ALeft, ATop, ARight, ABottom : Double; // The Rectangle A
                             const BLeft, BTop, BRight, BBottom : Double; // The Rectangle B
                             out CLeft, CTop, CRight, CBottom : Double; // The overlapping Rectangle
                             const AWorldRect : TLQTRect // The World Rect
                            ) : TLQTRectOverlapKind;
  { function RectanglesOverlap checks whether two rectangles are overlapping and if so
      which are the common part.
    Caution: works only for flat (nonwrapping) rectangles
    parameters
      ARect : The Rectangle A
      BRect : The Rectangle B
    outout
      CRect : The overlapping Rectangle
    return value: True if the rectangles overlapping (and the CRect is valid)
                  False if the rectangle are not overlapping.
    Remarks: If parts of the right side of one of the rectangle are identical to parts the left side or
             top side and bottom side, the rectangles are *not* overlapping!
  }
  function RectanglesOverlap(const ARect, BRect : TLQTRect; // The Rectangles A and B
                             out CRect : TLQTRect; // The overlapping Rectangle
                             const AWorldRect : TLQTRect // The World Rect
                            ) : TLQTRectOverlapKind;


  { The following functions are working on curved surfaces.
    They support two curvations: Cylindrical and toroidal.
    The first one is often used to project a sphere on a flat map.
    The second one is often used in computergames, where the upper
    and lower map borders are connected as well as the right and left side.

    The operation needs a "World" definition, to deal with the overflows.
    For example the Earth might be projected on a cylinder with an circumfence
    of 360 Degree and a height of 180 Degree (or as OSM does 2*85.05112878 Degree).
    Thus the world would be defined as left = -180, right +180.
    All coordinates are limited to this world, but allow rectangles to cover an
    area on both sides of "date border".
  }

  { procedure NormalizeRectangle normalize the passed Rectangle and World rectangle.
      Parameters
        Input parameter
          ARect: The rectangle to be normalized
          AWorldModel: The world model, flat, cylindric or toroidial
          AWorldRect: The rectangle with the world borders
        Output parameters
          ANormalizedRectangles: An array containing the normalized rectangle(s)
          ANormalizedWorldRect: The normalized world rectangle
      Remarks
        The normalized rectangle has smaller values for top and left and bigger for right and bottom.
        The world rectangle defines the world, but also the sign of the X- and Y-Axis.
        The normalization will
          - ensure that the sign of the X- and Y-Axis is positive, in the meaning of
            Increasing values while travel to the right and to the bottom
          - adjust the sign of the borders in coordination with the world rectangle
          - if nesesarry split the rectangle to fit into the worlds border without overlapping
        If the world model ist flat, no splitting will occour.
        If the world model is cylindrical, a split into two rectangles may occour
        If the world model is toroidial, a split into two or four rectangles may occour
  }
  procedure NormalizeRectangle(const ARect: TLQTRect;
                             const AWorldRect: TLQTRect;
                              const AWorldModel: TLQTWorldModel;
                              out ANormalizedRectangles : TLQTRectArray; // The normalized Rectangles
                              out ANormalizedWorldRect : TLQTRect
                              );
  { procedure UnNormalizeRectangles invertes the process of normalization of the rectangles
      Parameters
        Input
          ANormalizedRectangles: An array containing normalized rectangles
                                 (all rectangles must be inside or equal to the normalized word-borders).
          AUnNormalizedWorldModel: The world model to transformed to
          AUnNormalizedWorldRect: The borders of the world to be transformed in
        Output
          AUnNormalizedRectangles : An array of the transformed, unnormalized rectangles
      Remarks
        The input are assumed to be in a flat world with rectangles
        left, top values smaller than right, bottom values. And also all values
        laying within the world borders.
        In a flat world model the signs of the rectangles are adjusted and the result
        placed into the output rectangle array
        In a cylindrical world model the input rectangles are merged, if they share the
        same top and bottom values and having the same values for the left and right
        edges, vice versa. Additionally the left and right border of the world
        are treated as identical, thus rectangle who share the east-west-border are
        combined as well.
        In a toroidial world model additional to the cylidrical world the top and bottom
        values of the rectangles are also merged, if fit. This may collapse up to
        four rectangles in on single bigger rectangle.

  }
  procedure UnNormalizeRectangles(const ANormalizedRectangles : TLQTRectArray;
                                  const AUnNormalizedWorldRect: TLQTRect;
                                  const AUnNormalizedWorldModel: TLQTWorldModel;
                                  out AUnNormalizedRectangles : TLQTRectArray
                              );

  function RectanglesOverlap(const ARect, BRect : TLQTRect; // The Rectangles A and B
                             const AWorldRect : TLQTRect;
                             const AWorldModel : TLQTWorldModel; //
                             out CRectangles : TLQTRectArray // The overlapping Rectangles
                            ) : Boolean;
  { function RectanglesOverlapEx checks whether two rectangles are overlapping and
     if so, what kind of overlapping is found and which areas are shared between
     the two rectangles.
     The function works on all world models (flat, cylindrical and toroidal) and
     on all axis orientations.
    parameters
     Input
       ARect, BRect the two rectangles which may overlapping
       AWorldModel the world model (flat, cylindrical and toroidal)
       AWorldRect the world definition used for the determination of the axis orientation
     Output
       CRectangles an array of rectangles containing the common/shared areas in the
         orientation of the given world.
         The array will has always one of the follwing length
           0: No overlapping area
           1: One overlapping area (all world models)
           2: Two overlapping areas (cylindrical and toroidal worlds)
           4: Four overlapping areas (toroidal worlds)
         Remark: The returned areas in the cylindrical and toroidal worlds may wrap
           around the worlds border.
     Return value is the TLQTRectOverlapKind type.
  }
  function RectanglesOverlapEx(const ARect, BRect : TLQTRect;
                             const AWorldRect : TLQTRect;
                             const AWorldModel : TLQTWorldModel;
                             out CRectangles : TLQTRectArray
                            ) : TLQTRectOverlapKind;

{ function PointInRectangle tests whether a point is inside a rectangle on a curved world.
    It decapsulate the function ARect.ContainsPoint(....).
    See details there.
}
function PointInRectangle(const ARect: TLQTRect;
                          const APoint : TLQTPoint;
                          const AWorldRect : TLQTRect;
                          const AWorldModel : TLQTWorldModel) : Boolean;


const
  LQTWorldModelString  : array[TLQTWorldModel ] of String = ('Flat', 'Cylindrical', 'Toroidal');
  LineRectInteractionString : array[TLineRectInteraction] of String = ('Outside','Inside','Intersecting');
  { Definition of an empty rectangle }
  LQTRectEmpty : TLQTRect = (Left: 0.0; Top : 0.0; Right : 0.0; Bottom : 0.0);
  { LQTRectInfinite of an infinite rectangle }
  LQTRectInfinite : TLQTRect = (
    Left: NegInfinity;
    Top : NegInfinity;
    Right : Infinity;
    Bottom : Infinity);
  { LQTRectMinValue the absolute minimum value that could be used within this unit }
  LQTRectMinValue = -(MaxDouble/2.0);
  { LQTRectMaxValue the absolute maximum value that could be used within this unit }
  LQTRectMaxValue = LQTRectMinValue+MaxDouble;
  { LQTRectInfinite of an maximum rectangle }
  LQTRectMaximum : TLQTRect = (
    Left: LQTRectMinValue;
    Top : LQTRectMinValue;
    Right : LQTRectMaxValue;
    Bottom : LQTRectMaxValue);

  RectOverlapKindString : array[TLQTRectOverlapKind] of String = (
    'None',    // no overlap
    'Partly',  // the A is partly overlayed by B
    'Inside',  // the B is located inside A
    'Equal',   // the A and B are equal
    'Full');   // the B covers A fully


implementation

procedure TransformPoint(const AInX, AInY: Double; const ASourceWorld,
  ATargetWorld: TLQTRect; out AOutX, AOutY: Double);
var
  lsrcXLeftPlus : Boolean;
  lsrcYUpPlus : Boolean;
  ltarXLeftPlus : Boolean;
  ltarYUpPlus : Boolean;
  srcinf, tarinf : Boolean;
  lX, lY : Double;
  lNormSrcWorld, lNormTargetWorld : TLQTRect;
  wsrc, wtar, hsrc, htar : Double;
  hratio, wratio : Double;
begin
  if ASourceWorld.IsEqual(ATargetWorld) then
  begin
    AOutX := AInX;
    AOutY := AInY;
    Exit;
  end;
  if ASourceWorld.IsNotArea or ATargetWorld.IsNotArea then
    raise ELQTGException.Create('World extension is empty!');
  srcinf := ASourceWorld.IsInfinite;
  tarinf := ATargetWorld.IsInfinite;
  if srcinf xor tarinf then
    raise ELQTGException.Create('World extension mixed infinite and non-infinite!');
  // The sign changes is independent from the extension, this works also on infinite worlds
  lsrcXLeftPlus := ASourceWorld.WorldXAxisLeftPlus;
  lsrcYUpPlus := ASourceWorld.WorldYAxisUpPlus;
  ltarXLeftPlus := ATargetWorld.WorldXAxisLeftPlus;
  ltarYUpPlus := ATargetWorld.WorldYAxisUpPlus;
  lX := AInX;
  lY := AInY;
  if lsrcXLeftPlus then
    lX := -lX;
  if lsrcYUpPlus then
    lY := -lY;
  // Scaling works only if both worlds are within calculations limits
  if (not srcinf) and (not tarinf) then
  begin
    lNormSrcWorld := ASourceWorld;
    if lsrcXLeftPlus then
    begin
      lNormSrcWorld.Left := -lNormSrcWorld.Left;
      lNormSrcWorld.Right := -lNormSrcWorld.Right;
    end;
    if lsrcYUpPlus then
    begin
      lNormSrcWorld.Top := -lNormSrcWorld.Top;
      lNormSrcWorld.Bottom := -lNormSrcWorld.Bottom;
    end;
    lNormTargetWorld := ATargetWorld;
    if ltarXLeftPlus then
    begin
      lNormTargetWorld.Left := -lNormTargetWorld.Left;
      lNormTargetWorld.Right := -lNormTargetWorld.Right;
    end;
    if ltarYUpPlus then
    begin
      lNormTargetWorld.Top := -lNormTargetWorld.Top;
      lNormTargetWorld.Bottom := -lNormTargetWorld.Bottom;
    end;
    wsrc := lNormSrcWorld.Width;
    hsrc := lNormSrcWorld.Height;
    wtar := lNormTargetWorld.Width;
    htar := lNormTargetWorld.Height;
    wratio := wsrc / wtar;
    hratio := hsrc / htar;
    lX := ((lX-lNormSrcWorld.Left) / wratio) + lNormTargetWorld.Left;
    lY := ((lY-lNormSrcWorld.Top) / hratio) + lNormTargetWorld.Top;
  end;
  if ltarXLeftPlus then
    lX := -lX;
  if ltarYUpPlus then
    lY := -lY;
  AOutX := lX;
  AOutY := lY;
end;

procedure NormatePointToWorld(const AInX, AInY: Double;
  const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel;
  out AOutX, AOutY: Double);
var
  lX, lY : Double;
  lXAxisLeftPlus, lYAxisUpPlus : Boolean;
  d : Double;
  w,h : Double;
  lWorldRect : TLQTRect;
begin
  if not AWorldRect.IsValidAsWorld then
    raise ELQTGException.Create('Could not normate to an invalid world');
  lXAxisLeftPlus := AWorldRect.WorldXAxisLeftPlus;
  lYAxisUpPlus := AWorldRect.WorldYAxisUpPlus;
  // To get rid of the complicated different axis orientations
  // we copy the rectangles into local variables and normalize them
  lX := AInX;
  lY := AInY;
  lWorldRect.Assign(AWorldRect);

  if lXAxisLeftPlus then
  begin
    lX := -lX;
    lWorldRect.Left := -lWorldRect.Left;
    lWorldRect.Right := -lWorldRect.Right;
  end;
  if lYAxisUpPlus then
  begin
    lY := -lY;
    lWorldRect.Top := -lWorldRect.Top;
    lWorldRect.Bottom := -lWorldRect.Bottom;
  end;
  if AWorldModel <> lqtwmFlat then
  begin
    // The world is either cylindical or toroidal.
    // In both cases the X-Axis may wrap around.
    // If the X-Axis is infinite, nothing could be done here ...
    if not AWorldRect.HasInfiniteXAxis then
    begin
      // ... but in all other cases values larger / smaller than the height
      // of the world extent could be wrapped to be located inside
      w := AWorldRect.Width;
      if (lX < lWorldRect.Left) or
        (lX > lWorldRect.Right) then
      begin
        d := (lX-lWorldRect.Left) / w;
        d := Frac(d);
        if d < 0.0 then
          d := d + 1.0;
        d := d * w;
        lX := d+lWorldRect.Left;
      end;
    end;
    // Ff the world is toroidal the Y-Axis may wrap around.
    // If the Y-Axis is infinite, nothing could be done here ...
    if (AWorldModel = lqtwmToroidal) and
       (not AWorldRect.HasInfiniteYAxis) then
    begin
      // ... but in all other cases values larger / smaller than the width
      // of the world extent could be wrapped to be located inside
      h := AWorldRect.Height;
      if (lY < lWorldRect.Top) or
         (lY > lWorldRect.Bottom) then
      begin
        d := (lY-lWorldRect.Top) / h;
        d := Frac(d);
        if d < 0.0 then
          d := d + 1.0;
        d := d * h;
        lY := d+lWorldRect.Top;
      end;
    end;
   end;
  // In all cases limit the values to be inside the world.
  // Actual, in the toroidal world all values are inside here.
  // Limit the overall values
  if (AWorldModel = lqtwmFlat) then
  begin  // the X value needs only to be fixed in the flat world
    if lX < lWorldRect.Left then
      lX := lWorldRect.Left
    else if lX > lWorldRect.Right then
      lX := lWorldRect.Right;
  end;
  if (AWorldModel <> lqtwmToroidal) then
  begin // the Y values need to be fixed in the flat and cylindrical world
    if lY < lWorldRect.Top then
      lY := lWorldRect.Top
    else if lY > lWorldRect.Bottom then
      lY := lWorldRect.Bottom;
  end;
  // set the sign of values to the original orientation
  if lXAxisLeftPlus then
    lX := -lX;
  if lYAxisUpPlus then
    lY := -lY;
  AOutX := lX;
  AOutY := lY;
end;

type
  TLineKeyValue = record
    Value : Double;
    Key : Byte;
  end;

{ BestLineUnion, return the smallest length, covering all points of the two lines formed
   by the parameters.
   CAUTION: All values expected to be normalized (Left < Right, Top < Bottom)
            Wrap around is allowed! (Otherwise  Out1 := Min(InA1, InB1);
            Out2 := Max(InA2, InB2); would do the job faster!)
            Never mix Left/Right with Top/Bottom!!
   For a real rectangle this procedure should be called two times.
   One time for left and right values and the other time for top and bottom values.
   Parameter
    InA1: Rectangle A Left  (or in a different call Top)
    InA2: Rectangle A Right (or in a different call  Bottom)
    InB1: Rectangle B Left  (or in a different call  Top)
    InB2: Rectangle B Right (or in a different call  Bottom)
    AWorld1: World Left  (or in a different call  Top)
    AWorld2: World Right (or in a different call  Bottom)
   Out
    Out1: The Left or Top value
    Out2: The Right or Bottom value

   Example
      W1  A2 B1 B2 A1  W2
   A          +-+
   B  >---+         +--->
   C0 >---------+   +--->  Possibility one
   C1 >---+   +--------->  Possibility two
   Out1 and Out2 will be the shorter value-pair out of [C0, C1]

}
procedure BestLineUnion(const InA1,InA2,InB1,InB2,AWorld1,AWorld2 : Double;
                        const AWrapAllowed : Boolean;
                        out Out1, Out2 : Double);
var
  // A sorted array for the Values InA1, InA2, InB1, InB2
  lAppearance : array[0..3] of TLineKeyValue;
  b : Boolean;
  i : Integer;
  kv : TLineKeyValue;
  d0, d1 : Double;
  c : Cardinal;
begin
  // Drop the parameters unsorted into the array
  lAppearance[0].Key   := $A1; // see the Hex-Notation for the key
  lAppearance[0].Value := InA1;
  lAppearance[1].Key := $A2;
  lAppearance[1].Value := InA2;
  lAppearance[2].Key := $B1;
  lAppearance[2].Value := InB1;
  lAppearance[3].Key := $B2;
  lAppearance[3].Value := InB2;
  // Bubble Sort: Sort alle values from small to big.
  // For the 4 values this method seems to be ok.
  repeat
    b := False; // reset to "No item swapped"
    for i := 1 to 3 do
    begin
      // Check if we have to swap a value
      if (lAppearance[i-1].Value > lAppearance[i].Value) or // The previous value is bigger than this one, swap them
         ((lAppearance[i-1].Value = lAppearance[i].Value) and // Swap Begin and End on the same coord, so that End is prior begin
          ((lAppearance[i-1].Key and $0F) < (lAppearance[i].Key and $0F))) then
      begin
        b := True; // Set the "we swapped a pair"
        // Save the target
        kv := lAppearance[i-1];
        // Move the value one field
        lAppearance[i-1] := lAppearance[i];
        // Update the new backward field
        lAppearance[i] := kv;
      end;
    end;
  until not b; // finish, if one loop does not touch the order.
  // Now we have all value in the right order
  // 4 fields, we have 24 possible permutations
  // https://en.wikipedia.org/wiki/Permutation
  // Create the case value by shift and oring the four keys together
  c := lAppearance[0].Key;
  for i := 1 to 3 do
  begin
    c := c shl 8;
    c := c or lAppearance[i].Key;
  end;
  // Now the key shows the order of increasing values and which value sit in
  // Caution: adjacent values could be equal!
  case c of
    $A1B2B1A2, // The whole world, in your hands...
    $A2A1B2B1, //
    $B1A2A1B2, //
    $B2B1A2A1: //
      begin
        // The bounding box is the world, if both rectangles
        // are overlapping each other two times
        //   W1  A1 B2 B1 A2  W2
        // A      +-------+
        // B >------+   +------>
        // C >----------------->
        Out1 := AWorld1;
        Out2 := AWorld2;
      end;
    $A1B1B2A2, // A covers B
               //   W1  A1 B1 B2 A2  W2
               // A      +-------+
               // B         +-+
               // C      +-------+
    $A2A1B1B2, // A covers B (A wrapped)
    $B2A2A1B1, // A covers B (A and B wrapped)
    $B1B2A2A1: // A covers B (A wrapped)
      begin
        Out1 := InA1;
        Out2 := InA2;
      end;
    $A1A2B2B1, // B covers A (B wrapped)
    $B1A1A2B2, // B covers A
    $B2B1A1A2, // B covers A (B wrapped)
    $A2B2B1A1: // B covers A (A and B wrapped)
      begin
        Out1 := InB1;
        Out2 := InB2;
      end;
    $A2B1B2A1, // side by side (A wrapped)
    $B2A1A2B1: // side by side (B wrapped)
      begin    // Caution two possibitlities
        // Explaining for the first value
        //    W1  A2 B1 B2 A1  W2
        // A          +-+
        // B  >---+         +--->
        // C0 >---------+   +--->  Possibility one
        // C1 >---+   +--------->  Possibility two
        // Calculate the width of the two options
        if not AWrapAllowed then
          ELQTGException.Create('Wrap around detected, but allowed!');
        d0 := Abs(AWorld2-InA1) + Abs(InB2-AWorld1);
        d1 := Abs(AWorld2-InB1) + Abs(InA2-AWorld1);
        if d0 < d1 then // take the smaller one
        begin
          Out1 := InA1;
          Out2 := InB2;
        end
        else
        begin
          Out1 := InB1;
          Out2 := InA2;
        end;
      end;
    $A1B1A2B2,  // overlapping, A followed by B
    $B2A1B1A2,  // overlapping (B wrapped), A followed by B
    $A2B2A1B1,  // overlapping (A and B wrapped), A followed by B
    $B1A2B2A1,  // overlapping (A wrapped), A followed by B
    $A1A2B1B2:  // side by side, A followed by B
      begin
        Out1 := InA1;
        Out2 := InB2;
      end;
    $A1B2A2B1,  // overlapping (B wrapped), B followed by A
    $B1A1B2A2,  // overlapping, B followed by A
    $A2B1A1B2,  // overlapping (A wrapped), B followed by A
    $B2A2B1A1,  // overlapping (A and B wrapped), B followed by A
    $B1B2A1A2:  // side by side, B followed by A
      begin
        Out1 := InB1;
        Out2 := InA2;
      end;
  end;
end;

{ BestLineOverlap, return the overlapping parts of the lines from the parameters.
   CAUTION: All values expected to be normalized (Left < Right, Top < Bottom)
            Wrap around is allowed!
            Never mix Left/Right with Top/Bottom!!
   For a real rectangle this procedure should be called two times.
   One time for left and right values and the other time for top and bottom values.
   Parameter
    InA1: Rectangle A Left  (or in a different call Top)
    InA2: Rectangle A Right (or in a different call  Bottom)
    InB1: Rectangle B Left  (or in a different call  Top)
    InB2: Rectangle B Right (or in a different call  Bottom)
    AWorld1: World Left  (or in a different call  Top)
    AWorld2: World Right (or in a different call  Bottom)
   Out
     OutOverlapArr containing 0, 2 or 4 entries, each two values to be interpreted
       as BeginOfOverlap; EndOfOverlap.
       Length = 0 means no overlap-region exists,
       Lenght = 2 means one overlap-region exists,
       Length = 4 means two separate overlapping regions exists
}
procedure BestLineOverlap(const InA1,InA2,InB1,InB2 : Double;
                          out OutOverlapArr : TLQTDynDoubleArr);
var
  // A sorted array for the Values InA1, InA2, InB1, InB2
  lAppearance : array[0..3] of TLineKeyValue;
  b : Boolean;
  i : Integer;
  kv : TLineKeyValue;
  c : Cardinal;
  ndx : Integer;
begin
  OutOverlapArr := Nil;
  // Drop the parameters unsorted into the array
  lAppearance[0].Key   := $A1; // see the Hex-Notation for the key
  lAppearance[0].Value := InA1;
  lAppearance[1].Key   := $A2;
  lAppearance[1].Value := InA2;
  lAppearance[2].Key   := $B1;
  lAppearance[2].Value := InB1;
  lAppearance[3].Key   := $B2;
  lAppearance[3].Value := InB2;
  // Bubble Sort: Sort alle values from small to big.
  // For the 4 values this method seems to be ok.
  repeat
    b := False; // reset to "No item swapped"
    for i := 1 to 3 do
    begin
      // Check if we have to swap a value
      if (lAppearance[i-1].Value > lAppearance[i].Value) or // The previous value is bigger than this one, swap them
         ((lAppearance[i-1].Value = lAppearance[i].Value) and // Swap Begin and End on the same coord, so that End is prior begin
          ((lAppearance[i-1].Key and $0F) < (lAppearance[i].Key and $0F))) then
      begin
        b := True; // Set the "we swapped a pair"-Flag
        // Swap the two fiels
        // Save the target
        kv := lAppearance[i-1];
        // Move the value by one field
        lAppearance[i-1] := lAppearance[i];
        // Update the new backward field
        lAppearance[i] := kv;
      end;
    end;
  until not b; // finish, if one loop does not touch the order.
  // Now we have all value in the right order
  // For the for 4 fields, there are 24 permutations possible.
  // https://en.wikipedia.org/wiki/Permutation
  // Create the "case c of"-Value by shift and or-ing the four keys together.
  c := lAppearance[0].Key;
  for i := 1 to 3 do
  begin
    c := c shl 8;
    c := c or lAppearance[i].Key;
  end;
  // Now the key shows the order of increasing values and which value sit in
  // Caution: adjacent values could be equal!
  //   We will suppress to add zero size overlapping lines to the result
  case c of
    $A1B2B1A2, // Two overlapping areas
    $B1A2A1B2,
    $A2A1B2B1,
    $B2B1A2A1:
      begin
        //   W1  A1 B2 B1 A2  W2
        // A >------+   +------>
        // B      +-------+
        // C0     +-+
        // C1           +-+

        //   W1  B1 A2 A1 B2  W2
        // A >------+   +----->
        // B      +-------+
        // C0     +-+
        // C1           +-+

        //   W1  A2 A1 B2 B1  W2
        // A >---------+   +-->
        // B >---+  +--------->
        // C0       +--+
        // C1>---+         +-->

        //   W1  B2 B1 A2 A1  W2
        // A >---------+   +-->
        // B >---+  +--------->
        // C0       +--+
        // C1>---+         +-->
        ndx := 0;
        if (InA1 <> InB2) then
        begin
          SetLength(OutOverlapArr,ndx+2);
          OutOverlapArr[ndx] := InA1;
          OutOverlapArr[ndx+1] := InB2;
          ndx := ndx+2;
        end;
        if (InB1 <> InA2) then
        begin
          SetLength(OutOverlapArr,ndx+2);
          OutOverlapArr[ndx] := InB1;
          OutOverlapArr[ndx+1] := InA2;
          ndx := ndx+2;
        end;
      end;

    $A1B1B2A2, // A covers B
    $A2A1B1B2, // A covers B (A wrapped)
    $B2A2A1B1, // A covers B (A and B wrapped)
    $B1B2A2A1: // A covers B (A wrapped)
      begin
        // Overlapping is B
        //   W1  A1 B1 B2 A2  W2
        // A      +-------+
        // B         +-+
        // C         +-+
        if InB1 <> InB2 then
        begin
          SetLength(OutOverlapArr,2);
          OutOverlapArr[0] := InB1;
          OutOverlapArr[1] := InB2;
        end;
      end;


    $A1A2B2B1, // B covers A (B wrapped)
    $B1A1A2B2, // B covers A
    $B2B1A1A2, // B covers A (B wrapped)
    $A2B2B1A1: // B covers A (A and B wrapped)
      begin
        // Overlapping is A
        //   W1  B1 A1 A2 B2  W2
        // A         +-+
        // B      +-------+
        // C         +-+
        if InA1 <> InA2 then
        begin
          SetLength(OutOverlapArr,2);
          OutOverlapArr[0] := InA1;
          OutOverlapArr[1] := InA2;
        end;
      end;

    $A1B1A2B2, // overlapping, A followed by B
    $B2A1B1A2, // overlapping (B wrapped), A followed by B
    $A2B2A1B1, // overlapping (A and B wrapped), A followed by B
    $B1A2B2A1: // overlapping (A wrapped), A followed by B
      begin
        //   W1  A1 B1 A2 B2  W2
        // A      +----+
        // B         +----+
        // C         +-+

        //   W1  B2 A1 B1 A2  W2
        // A         +----+
        // B >---+      +------>
        // C            +-+

        //   W1  A2 B2 A1 B1  W2
        // A >---+      +------>
        // B >------+     +---->
        // C >---+        +---->

        //   W1  B1 A2 B2 A1   W2
        // A >------+     +---->
        // B      +----+
        // C      +-+
        if InB1 <> InA2 then
        begin
          SetLength(OutOverlapArr,2);
          OutOverlapArr[0] := InB1;
          OutOverlapArr[1] := InA2;
        end;
      end;
    $A1B2A2B1, // overlapping (B wrapped), B followed by A
    $B1A1B2A2, // overlapping, B followed by A
    $A2B1A1B2, // overlapping (A wrapped), B followed by A
    $B2A2B1A1: // overlapping (A and B wrapped), B followed by A
      begin
        //   W1  A1 B2 A2 B1  W2
        // A      +----+
        // B >------+      +--->
        // C      +-+

        //   W1  B1 A1 B2 A2   W2
        // A         +----+
        // B      +----+
        // C         +-+

        //   W1  A2 B1 A1 B2   W2
        // A >---+      +------->
        // B         +----+
        // C            +-+

        //   W1  B2 A2 B1 A1   W2
        // A >------+      +---->
        // B >---+      +------->
        // C >---+         +---->
        if InA1 <> InB2 then
        begin
          SetLength(OutOverlapArr,2);
          OutOverlapArr[0] := InA1;
          OutOverlapArr[1] := InB2;
        end;
      end;
  else
    { // No overlap
      $B1B2A1A2, // side by side, B followed by A
      $A2B1B2A1, // side by side, B followed by A (A wrapped)
      $B2A1A2B1, // side by side, A followed by B(B wrapped)
      $A1A2B1B2: // side by side, A followed by B
    }
  end;
end;

function LineIntersection(
           const AP0X, AP0Y, AP1X, AP1Y, AP2X, AP2Y, AP3X, AP3Y : Double;
           out AIX, AIY : Double) : Boolean;
var
  s1x, s1y, s2x, s2y : Double;
  s, t : Double;
  det, eps : Double;
begin
  Result := False;
  s1x := AP1X - AP0X;
  s1y := AP1Y - AP0Y;
  s2x := AP3X - AP2X;
  s2y := AP3Y - AP2Y;

  det := (-s2x * s1y + s1x * s2y);
  // The determinate has to be checked versus 0.0
  // we use an epsilon, that is +- one Bit around
  eps := ULP(det);
  if Abs(det) <= eps then
    Exit;

  s := (-s1y * (AP0X - AP2X) + s1x * (AP0Y - AP2Y)) / det;
  if (s >= 0) and (s <= 1) then
  begin
    t := ( s2x * (AP0Y - AP2Y) - s2y * (AP0X - AP2X)) / det;
    if  (t >= 0) and (t <= 1) then
    begin
      // Collision detected
      AIX := AP0X + (t * s1x);
      AIY := AP0Y + (t * s1y);
      Result := True;
    end;
  end;
end;
function LineIntersection(
           const AP0, AP1, AP2, AP3 : TLQTPoint;
           out AI : TLQTPoint) : Boolean;
begin
  Result := LineIntersection(AP0.X, AP0.Y, AP1.X, AP1.Y, AP2.X, AP2.Y, AP3.X, AP3.Y,
                             AI.X, AI.Y);
end;

function LineRectInteractionSimple(
           const ALeft, ATop, ARight, ABottom : Double;
           const AP0X, AP0Y, AP1X, AP1Y : Double
           ) : TLineRectInteraction;
var
  ix, iy : Double;
  rc : Boolean;
begin
  // Catch the special case that the line is fully inside the rectangle
  if ((AP0X >= ALeft) and (AP0X <= ARight) and
      (AP0Y >= ATop) and (AP0Y <= ABottom) and
      (AP1X >= ALeft) and (AP1X <= ARight) and
      (AP1Y >= ATop) and (AP1Y <= ABottom)) then
  begin
    Result := lriInside;
    Exit;
  end;
  // Check all four edges if one of them intersects with the line
  ix := 0.0;
  iy := 0.0;
  rc := LineIntersection(ALeft,ATop,ARight,ATop, // Upper Line
                  AP0X, AP0Y, AP1X, AP1Y,
                  ix,iy) or
        LineIntersection(ALeft,ATop,ALeft,ABottom, // Left Line
                  AP0X, AP0Y, AP1X, AP1Y,
                  ix,iy) or
        LineIntersection(ARight,ATop,ARight,ABottom, // Right Line
                  AP0X, AP0Y, AP1X, AP1Y,
                  ix,iy) or
        LineIntersection(ALeft,ABottom,ARight,ABottom, // Bottom Line
                  AP0X, AP0Y, AP1X, AP1Y,
                  ix,iy);
  if rc then
    Result := lriIntersecting
  else
    Result := lriOutside;
end;


function LineRectInteraction(
           const ALeft, ATop, ARight, ABottom : Double;
           const AP0X, AP0Y, AP1X, AP1Y : Double) : TLineRectInteraction;
var
  lAreaA, lAreaB : Integer;
  lAreaAB : Integer;
  n : Integer;
  lP0X, lP0Y, lP1X, lP1Y : Double;
  lPhi0, lPhi1, lPhiLine : Double;
begin
  Result := lriOutside;
  // For both line endings, calculate the "Area"-Code where they are located
  // You may imagine a 3 by 3 checker board (tic-tac-toe),
  // the tested rectangle is located in/as "B2"
  //      |    |
  //   A1 | A2 | A3
  //  ----+----+-----
  //   B1 | B2 | B3
  //  ----+----+-----
  //   C1 | C2 | C3
  //      |    |
  // A lot of cases could be sorted out prior any complicated math calculation
  // - if both of the line points fall into B2, the rectangle is hit (but not intersected).
  // - If one of the line points fall into B2 and the other not, an intersection is found.
  // - if both points fall in the same of the remaining areas, no intersection is possible.
  // - If one of the line points fall into A2 and the other into C2, or alternative
  //     one into B1 and the other into B3, an intersection is found, since B2 must be passed.
  // - A1 to B1, A1 to C1 etc. are also outside, but it is better to
  //   test the cases where a calclulation of the intersection has to be performed
  //   (e.g. B1-C2, B1-C3) and treat the remaining as "outside".

  // Location of Point 0
  if AP0X < ALeft then
    lAreaA := $01
  else if AP0X < ARight then // Right border *not* included
    lAreaA := $02
  else
    lAreaA := $03;
  if AP0Y < ATop then
    lAreaA := lAreaA or $A0
  else if AP0Y < ABottom then // Bottom border *not* included
    lAreaA := lAreaA or $B0
  else
    lAreaA := lAreaA or $C0;
  // Location of Point 1
  if AP1X < ALeft then
    lAreaB := $01
  else if AP1X < ARight then // Right border *not* included
    lAreaB := $02
  else
    lAreaB := $03;
  if AP1Y < ATop then
    lAreaB := lAreaB or $A0
  else if AP1Y < ABottom then // Bottom border *not* included
    lAreaB := lAreaB or $B0
  else
    lAreaB := lAreaB or $C0;
  // Simple cases, where B2 = TestRectangle is in charge
  if (lAreaA = $B2) or (lAreaB = $B2) then
  begin // At least one end are inside the rectangle
    if lAreaA = lAreaB then  // both endings are in the rectangle
      Result := lriInside
    else
      Result := lriIntersecting;
    Exit;
  end;
  // The second simple case
  if lAreaA = lAreaB then
    Exit; // endings are in the same area, so no intersection is possible

  // Sort the area order, to avoid duplicate cases
  if lAreaA > lAreaB then
  begin // Swap the order of the areas, so that always A comes before B etc.
    n := lAreaA;
    lAreaA := lAreaB;
    lAreaB := n;
    // Swap the points too
    lP0X := AP1X;
    lP0Y := AP1Y;
    lP1X := AP0X;
    lP1Y := AP0Y;
  end
  else
  begin
    // Copy the points
    lP0X := AP0X;
    lP0Y := AP0Y;
    lP1X := AP1X;
    lP1Y := AP1Y;
  end;
  lAreaAB := (lAreaA shl 8) or lAreaB; // Stitch the area codes together
  // Now we can distinguish the different cases
  case lAreaAB of
    $A2C2, $B1B3 :
      begin // From the squares top to bottom, or left to right, always B2 are crossed
        //      |    |
        //      | A2 |
        //  ----+----+-----
        //   B1 | B2 | B3
        //  ----+----+-----
        //      | C2 |
        //      |    |
        Result := lriIntersecting;
      end;
    $A1B3,$A1C2,$A1C3 :
      begin
        // A1 means the Point 0 is top and left of the top-left corner of the rectangle.
        // An intersection will happen, if the line's angle is between the angle
        // to the lower-left and top-right rectangle (marked with X).
        // Both corners are excluded from the rectangle.
        //      |    |
        //   A1 |    |
        //  ----+----X-----
        //      | B2 | B3
        //  ----X----+-----
        //      | C2 | C3
        //      |    |
        lPhiLine := (lP0x-lP1x)/(lP0y-lP1y); //(lP0y-lP1y) is never zero
        lPhi0 := (lP0x-ALeft)/(lP0y-ABottom);
        lPhi1 := (lP0x-ARight)/(lP0y-ATop); // (lP0y-ATop) is never zero, since this is in B1
        if (lPhiLine > lPhi0) and (lPhiLine < lPhi1) then
           Result := lriIntersecting;
      end;
    $A2B1,$A2B3,$A2C1,$A2C3 :
      begin // From A2 only the upper edge of the rectangle must be tested
        // An intersection will happen, if the line's angle is between the angle
        // between the top-left and top-right rectangle.
        // The top-left corner is included, the top-right corner excluded from the rectangle.
        //      |    |
        //      | A2 |
        //  ----X----X-----
        //   B1 | B2 | B3
        //  ----+----+-----
        //   C1 |    | C3
        //      |    |
        lPhi0 := (lP0x-ALeft)/(lP0y-ATop); // lP0y can not equal to ATop, since than it would be inside A2
        lPhi1 := (lP0x-ARight)/(lP0y-ATop);
        lPhiLine := (lP0x-lP1x)/(lP0y-lP1y);
        if (lPhiLine >= lPhi0) and (lPhiLine < lPhi1) then
          Result := lriIntersecting;
      end;
    $A3B1,$A3C1,$A3C2 :
      begin // From A3 only the upper and the right edge of the rectangle must be tested
        // An intersection will happen, if the line's angle is between the angle
        // between the top-left and bottom-right rectangle.
        // The top-left corner is included, the bottom-right corner excluded from the rectangle.
        //      |    |
        //      |    | A3
        //  ----X----+-----
        //   B1 | B2 |
        //  ----+----X-----
        //   C1 | C2 |
        //      |    |

        lPhi0 := (lP0x-ALeft)/(lP0y-ATop);
        lPhi1 := (lP0x-ARight)/(lP0y-ABottom);
        lPhiLine := (lP0x-lP1x)/(lP0y-lP1y);
        if (lPhiLine >= lPhi0) and (lPhiLine < lPhi1) then
          Result := lriIntersecting;
      end;
    $B1C2,$B1C3 :
      begin // From B1 only the left edge of the rectangle must be tested
        // An intersection will happen, if the line's angle is between the angle
        // between the top-left and bottom-left rectangle.
        // The lower corner is not included in the rectangle, the upper corner
        // could never be reached.
        //      |    |
        //      |    |
        //  ----X----+-----
        //   B1 | B2 |
        //  ----X----+-----
        //      | C2 | C3
        //      |    |
        lPhi0 := (lP0y-ATop)/(lP0x-ALeft);
        lPhi1 := (lP0y-ABottom)/(lP0x-ALeft);
        lPhiLine := (lP0y-lP1y)/(lP0x-lP1x);
        if (lPhiLine >= lPhi0) and (lPhiLine < lPhi1) then
          Result := lriIntersecting;

      end;
    $B3C1,$B3C2 :
      begin // From B3 only the right edge of the rectangle must be tested
        // An intersection will happen, if the line's angle is between the angle
        // between the top-right and bottom-right rectangle.
        // Both corners are not included in the rectangle!
        //      |    |
        //      |    |
        //  ----+----X-----
        //      | B2 | B3
        //  ----+----X-----
        //   C1 | C2 |
        //      |    |

        // Catch a special case. If the point of the line is on the right edge
        // of the rectangle (which belongs to B3 and not to B2) we can assume
        // an intersection of the rectangle.
        if (lP0x = ARight) then
        begin
          Result := lriIntersecting;
          Exit;
        end;
        lPhi1 := (lP0y-ATop)/(lP0x-ARight);
        lPhi0 := (lP0y-ABottom)/(lP0x-ARight);
        lPhiLine := (lP0y-lP1y)/(lP0x-lP1x);
        if (lPhiLine > lPhi0) and (lPhiLine < lPhi1) then
          Result := lriIntersecting;
       end;
  else
    // All other combinations will not hit B2 (=the rectangle) and are outside
    Exit;
  end;
end;

function LineRectInteraction(const ARect: TLQTRect; const AP0, AP1: TLQTPoint
  ): TLineRectInteraction;
begin
  Result := LineRectInteraction(ARect.Left, ARect.Top, ARect.Right, ARect.Bottom,
                                AP0.X, AP0.Y, AP1.X, AP1.Y);
end;

function PolylineRectangleInterAction(const ALeft, ATop, ARight,
  ABottom: Double; const APolyline; const APolylinePointCount: Integer;
  const AFirstValueIsY: Boolean; const AXAxisLeftPlus: Boolean;
  const AYAxisUpPlus: Boolean; const AFullInsideCheck: Boolean
  ): TLineRectInteraction;
type
  TStaticDoubleArray = array[0..(MaxInt div SizeOf(Double))-1] of Double;
var
  rc : TLineRectInteraction;
  i : Integer;
  pp : TStaticDoubleArray absolute APolyline; //A kind of TypeCast of the passed Polygon data
  lLeft, lTop, lRight, lBottom : Double;
  lP0X, lP0Y, lP1X, lP1Y: Double;
  ndx : Integer;
begin
  if APolylinePointCount < 4 then  // At least a closed triangle is needed to have any point inside
    raise Exception.Create('APolygonPointCount < 4');
  Result := lriOutside;
  // Prepare the possible inverse meaning of the Y-Axis
  lTop := ATop;
  lBottom := ABottom;
  if AYAxisUpPlus then
  begin
    lTop := -lTop;
    lBottom := -lBottom;
  end;
  lLeft := ALeft;
  lRight := ARight;
  if AXAxisLeftPlus then
  begin
    lLeft := -lLeft;
    lRight := -lRight;
  end;
  // If the first value in the points is a Y then we have to swap
  // X and Y values for the LineRectInteraction function call
  // Since each segment must be tested, we can avoid doing this exchange
  // twice for every point.
  // Place the very first point coords in the array on the second place
  if AFirstValueIsY then
  begin
    lP1X := pp[1];
    lP1Y := pp[0];
  end
  else
  begin
    lP1X := pp[0];
    lP1Y := pp[1];
  end;
  // If the X-axis is inverted, do so with the value
  if AXAxisLeftPlus then
    lP1X := -lP1X;
  // If the Y-axis is inverted, do so with the value
  if AYAxisUpPlus then
    lP1Y := -lP1Y;

  // For all values except the first (starting with the second one)
  for i := 1 to APolylinePointCount-1 do
  begin
    // Copy the coords of the second point of the last segment to the first point of this test
    lP0X  := lP1X;
    lP0Y  := lP1Y;
    // Copy the testing values from the Parameters array in the right order to the second point
    ndx := i*2;
    if AFirstValueIsY then
    begin
      lP1X := pp[ndx+1];
      lP1Y := pp[ndx];
    end
    else
    begin
      lP1X := pp[ndx];
      lP1Y := pp[ndx+1];
    end;
    // If the X-axis is inverted, do so with the value
    if AXAxisLeftPlus then
      lP1X := -lP1X;
    // If the Y-axis is inverted, do so with the value
    if AYAxisUpPlus then
      lP1Y := -lP1Y;

    // Now we have the Points in the correct order
    rc := LineRectInteraction(lLeft, lTop, lRight, lBottom,
                              lP0X, lP0Y, lP1X, lP1Y);
    if rc <> lriOutside then
    begin // We got a hit
      Result := rc; // save the result
      if (Result = lriIntersecting) or // If an Intersection is found
         (not AFullInsideCheck) then // or an Inside is found but we don't have to perform a full check
        Exit; //Then we are done
    end;
  end;
end;

function RectanglesOverlap(const ARect, BRect: TLQTRect;
                           out CRect: TLQTRect;
                           const AWorldRect : TLQTRect
                           ): TLQTRectOverlapKind;
var
  lARect, lBRect : TLQTRect;
  boundRect : TLQTRect;
  hA,hB,hbR,hC : Double;
  wA,wB,wbR,WC : Double;
  lARectArea, lBRectArea, lCRectArea : Double;
  eps : Double;
begin
  lARect.Assign(ARect);
  lBRect.Assign(BRect);
  if AWorldRect.WorldXAxisLeftPlus then
  begin
    lARect.Left  := -lARect.Left;
    lARect.Right := -lARect.Right;
    lBRect.Left  := -lBRect.Left;
    lBRect.Right := -lBRect.Right;
  end;
  if AWorldRect.WorldYAxisUpPlus then
  begin
    lARect.Top    := -lARect.Top;
    lARect.Bottom := -lARect.Bottom;
    lBRect.Top    := -lBRect.Top;
    lBRect.Bottom := -lBRect.Bottom;
  end;
  try
    // Calculate the bounding Rectangle of both rectangles A and B
    boundRect.Left := Min(lARect.Left, lBRect.Left);
    boundRect.Top := Min(lARect.Top,lBRect.Top);
    boundRect.Right := Max(lARect.Right, lBRect.Right);
    boundRect.Bottom := Max(lARect.Bottom,lBRect.Bottom);

    // Sort out the simple cases
    if boundRect.IsEqual(lARect) then
    begin
      Result := qroInside;
      CRect.Assign(lBRect);
      Exit;
    end;
    // Sort out the simple cases
    if boundRect.IsEqual(lBRect) then
    begin
      Result := qroFull;
      CRect.Assign(lARect);
      Exit;
    end;
    // Calculate the width and height of the rectangles A, B and the bounding Rectangle
    wA := lARect.Right - lARect.Left;
    wB := lBRect.Right - lBRect.Left;
    wbR := boundRect.Right - boundRect.Left;
    hA := lARect.Bottom - lARect.Top;
    hB := lBRect.Bottom - lBRect.Top;
    hbR := boundRect.Bottom - boundRect.Top;

    // Calculate the width and heigth of the overlapping rectangle
    // This is the difference between the sum of the width/height of the
    // rectangles A and B and the width/height of the bounding rectangle
    // if both values > 0 they are overlapping
    //
    //  012345678901234
    //  +------bR-----+
    //  +--A--+  +--B-+
    //
    //  wA = 7, wB = 6, wbR = 15 => (7+6)-15 = -2 => not Overlapping
    //
    //  0123456789
    //  +---bR---+
    //  +--A--+
    //      +--B-+
    //
    //  wA = 7, wB = 6, wbR = 10 => (7+6)-13 = 3 => Overlapping
    //
    wC := (wA+wB) - wbR;
    hC := (hA+hB) - hbR;
    if (wC > 0) and (hC > 0) then
{   // comparing big with small rectangles will make problems due to rounding
    // The test of equal rectangles above will fix this problem,
    // so this proble could be ignored here
    if ((wC > 0) or
        ((wC = 0) and (((wA = wbR) and (wB > 0)) or ((wB = wbR) and (wA > 0))))
       ) and (
        (hC > 0) or
        ((hC = 0) and (((hA = hbR) and (hB > 0)) or ((hB = hbR) and (hA > 0))))
       ) then }
    begin
      // Calculate the coords of the overlapping rectangle
      if lARect.Left = boundRect.Left then
        CRect.Left := lBRect.Left
      else
        CRect.Left := lARect.Left;
      CRect.Right := CRect.Left + wC;
      if lARect.Top = boundRect.Top then
        CRect.Top := lBRect.Top
      else
        CRect.Top := lARect.Top;
      CRect.Bottom := CRect.Top + hC;


      // Calculate the area size for all of the A, B and C rectangles
      lARectArea := wA * hA;
      lBRectArea := wB * hB;
      lCRectArea := Abs(CRect.Left-CRect.Right) * Abs(CRect.Top-CRect.Bottom);

      // The sizes of the area of the three parts will help us sorting out the
      // cases full, inside, so that partly will remain.
      // Since the overlapping area could not be greater than the sum of overlapping
      // rectangles, subsequently the equality to one of the A or B rectangles means,
      // that one of them is fully covered by the other.
      // If all three parts having the same area size, they must be equal.
      // (Remark: one may say, this case is catched above, but due to rounding this
      // is still here possible).
      // Comparing to equal is dangerous, because in the flow of calculation,
      // the equal values from the beginning may change in one Bit by rounding "errors".
      // To prevent this problem, it is recommended to use an comparism
      // of the Abs(delta)-Values to a small epsilon value.
      // The problem with this approach is, that an secondary error is constructed,
      // in the range of epsilon.
      // So epsilon must be choosen very carefully.
      // What we are doing here is, to use the smallest possible change on the
      // Double value of the greatest of the three areas (ensuring the greatest of the
      // smallest epsilons) and comparing to this.
      // Since rounding is limited to one Bit, and epsilon contain one Bit, this
      // should ensure, that no wrong detection will be made.
      // The ULP does exactly this, returning the step to the next possible Double value.
      if (lCRectArea > lARectArea) and
         (lCRectArea > lBRectArea) then
        eps := ULP(lCRectArea) // Should never be the case
      else if lBRectArea > lARectArea then
        eps := ULP(lBRectArea)
      else
        eps := ULP(lARectArea);
      // Depending of the sizes of the rectangles the values will be very small
      // In the Demo-Applications around 1e-12
      if (Abs(lARectArea-lCRectArea) <= eps) and // The allowed delat is 0.0 and +- one Bit
         (Abs(lBRectArea-lCRectArea) <= eps) then
      begin
        Result := qroEqual;
        Exit;
      end;
      // If the ARect area size is identical to the overlapping
      if Abs(lARectArea-lCRectArea) <= eps then
      begin
        Result := qroFull; // the B covers A fully
        Exit;
      end;
      // If the BRect area size is identical to the overlapping
      if Abs(lBRectArea-lCRectArea) <= eps then
      begin
        Result := qroInside; // the B is located inside A
        Exit;
      end;
      // All other cases has been sorted out
      Result := qroPartly;
    end
    else
      Result := qroNone;
  finally
    if AWorldRect.WorldXAxisLeftPlus then
    begin
      CRect.Left  := -CRect.Left;
      CRect.Right := -CRect.Right;
    end;
    if AWorldRect.WorldYAxisUpPlus then
    begin
      CRect.Top  := -CRect.Top;
      CRect.Bottom := -CRect.Bottom;
    end;
  end;
end;

function RectanglesOverlap(const ALeft, ATop, ARight, ABottom: Double;
                           const BLeft, BTop, BRight, BBottom: Double;
                           out CLeft, CTop, CRight, CBottom: Double;
                           const AWorldRect : TLQTRect // The World Rect
                          ): TLQTRectOverlapKind;
var
  lARect, lBRect, lCRect : TLQTRect;
begin
  lARect.SetFrom(ALeft, ATop, ARight, ABottom);
  lBRect.SetFrom(BLeft, BTop, BRight, BBottom);
  Result := RectanglesOverlap(ALeft, ATop, ARight, ABottom,
                              BLeft, BTop, BRight, BBottom,
                              lCRect.Left, lCRect.Top, lCRect.Right, lCRect.Bottom,
                              AWorldRect);
  CLeft := lCRect.Left;
  CTop  := lCRect.Top;
  CRight := lCRect.Right;
  CBottom := lCRect.Bottom;
end;

procedure RemoveNotAreaRectanglesFromArray(var ARectangles: TLQTRectArray);
var
  i : Integer;
begin
  for i := High(ARectangles) downto 0 do // Running from the end, allows a simple delete of the elements
  begin
    if ARectangles[i].IsNotArea then
      Delete(ARectangles,i,1);
  end;
end;

procedure NormalizeRectangle(const ARect: TLQTRect;
  const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel;
  out ANormalizedRectangles: TLQTRectArray; out ANormalizedWorldRect: TLQTRect);
var
  lARect : TLQTRect;
  i : Integer;
  lXLeftPlus, lYUpPlus : Boolean;
  cnt : Integer;
begin
  lARect := ARect;
  lARect.NormateToWorld(AWorldRect,AWorldModel);
  ANormalizedWorldRect := AWorldRect;
  lXLeftPlus := AWorldRect.WorldXAxisLeftPlus;
  lYUpPlus := AWorldRect.WorldYAxisUpPlus;
  if lXLeftPlus then
  begin
    lARect.Left :=  -lARect.Left;
    lARect.Right := -lARect.Right;
    ANormalizedWorldRect.Left :=  -ANormalizedWorldRect.Left;
    ANormalizedWorldRect.Right := -ANormalizedWorldRect.Right;
  end;
  if lYUpPlus then
  begin
    lARect.Top := -lARect.Top;
    lARect.Bottom := -lARect.Bottom;
    ANormalizedWorldRect.Top :=  -ANormalizedWorldRect.Top;
    ANormalizedWorldRect.Bottom := -ANormalizedWorldRect.Bottom;
  end;
  ANormalizedRectangles := Nil;
  SetLength(ANormalizedRectangles,1);
  ANormalizedRectangles[0] := lARect;
  if lARect.IsPoint then Exit;
  if AWorldModel = lqtwmFlat then Exit; // Nothing further to do
  // AWorldModel = lqtwmCylindrical, lqtwmToroidal
  if lARect.Right < lARect.Left then
  begin  // A wrap around has been detected, split into two rectangles
    //  0 ... +180/-180 ... 0
    //    AAAAAAAAAAAAAAAAA
    // Left  +100
    // Right -100
    // Will be splitted into
    // -180 ...  0 ... +180
    //      AAA    BBB
    // A: Left +100, Right +180
    // B: Left -180, Right -100

    SetLength(ANormalizedRectangles,2);
    ANormalizedRectangles[0].SetFrom(              lARect.Left,lARect.Top,ANormalizedWorldRect.Right,lARect.Bottom);
    ANormalizedRectangles[1].SetFrom(ANormalizedWorldRect.Left,lARect.Top,              lARect.Right,lARect.Bottom);
    // If zero sized rectangles has been included, remove them
    RemoveNotAreaRectanglesFromArray(ANormalizedRectangles);
  end;

  if AWorldModel = lqtwmCylindrical then Exit; // Nothing further to do
  // From here AWorldModel = lqtwmToroidal
  if lARect.Bottom < lARect.Top then
  begin  // A wrap around has been detected, split rectangles
    // Same as above but for Top and Bottom
    cnt := Length(ANormalizedRectangles);
    SetLength(ANormalizedRectangles,cnt*2); // Duplicate existing rectangle(s)
    for i := 0 to cnt-1 do
    begin
      lARect.Assign(ANormalizedRectangles[i]);
      ANormalizedRectangles[i].SetFrom(    lARect.Left,ANormalizedWorldRect.Top,lARect.Right,              lARect.Bottom);
      ANormalizedRectangles[cnt+i].SetFrom(lARect.Left,              lARect.Top,lARect.Right,ANormalizedWorldRect.Bottom);
    end;
    // If zero sized rectangles has been included, remove them
    RemoveNotAreaRectanglesFromArray(ANormalizedRectangles);
  end;
end;

procedure UnNormalizeRectangles(const ANormalizedRectangles: TLQTRectArray;
  const AUnNormalizedWorldRect: TLQTRect;
  const AUnNormalizedWorldModel: TLQTWorldModel;
  out AUnNormalizedRectangles: TLQTRectArray);
var
  i,j,k : Integer;
  lXLeftPlus, lYUpPlus : Boolean;
  cnt : Integer;
  chgd : Boolean;
  lNormalizedWorld : TLQTRect;
begin
  lNormalizedWorld := AUnNormalizedWorldRect;
  lXLeftPlus := AUnNormalizedWorldRect.WorldXAxisLeftPlus;
  lYUpPlus := AUnNormalizedWorldRect.WorldYAxisUpPlus;
  if lXLeftPlus then
  begin
    lNormalizedWorld.Left :=  -lNormalizedWorld.Left;
    lNormalizedWorld.Right := -lNormalizedWorld.Right;
  end;
  if lYUpPlus then
  begin
    lNormalizedWorld.Top := -lNormalizedWorld.Top;
    lNormalizedWorld.Bottom := -lNormalizedWorld.Bottom;
  end;

  AUnNormalizedRectangles := Nil; // Keep the compiler calm
  // Copy all rectangles to the output, assuming no combination is possible
  cnt := Length(ANormalizedRectangles);
  SetLength(AUnNormalizedRectangles,cnt);
  for i := 0 to cnt-1 do
    AUnNormalizedRectangles[i] := ANormalizedRectangles[i];
  // Now try to combine until no mor combinations found
  repeat
    chgd := False; // Assume nothing found
    // Now try to combine from the highest downwards (makes it much(!) easier to remove combined rectangles)
    cnt := Length(AUnNormalizedRectangles);
    for i := cnt-1 downto 1 do
    begin
      // choose the possible combination partner, one lower to zero
      for j := i-1 downto 0 do
      begin
        if AUnNormalizedRectangles[j].TryExtentWith(AUnNormalizedRectangles[i],lNormalizedWorld,AUnNormalizedWorldModel) then
        begin
          // Being here means that the rectangle[j] has been modified
          // and the rectangle[i] is a dublicate and must be removed
          chgd := True;
          // Copy all rectangles to their left position, overwriting the duplicate [i}
          for k := i to cnt-2 do
            AUnNormalizedRectangles[k] := AUnNormalizedRectangles[k+1];
          // Now reduce the array length by one
          SetLength(AUnNormalizedRectangles,cnt-1);
          Break; // done here
        end;
      end;
      if chgd then Break; //done here too
    end;
  until not chgd;
  // Now change the signs back to the original world
  cnt := Length(AUnNormalizedRectangles);
  for i := 0 to cnt-1 do
  begin
    if lXLeftPlus then
    begin
      AUnNormalizedRectangles[i].Left :=  -AUnNormalizedRectangles[i].Left;
      AUnNormalizedRectangles[i].Right := -AUnNormalizedRectangles[i].Right;
    end;
    if lYUpPlus then
    begin
      AUnNormalizedRectangles[i].Top := -AUnNormalizedRectangles[i].Top;
      AUnNormalizedRectangles[i].Bottom := -AUnNormalizedRectangles[i].Bottom;
    end;
  end;
end;

{ function RectanglesOverlap checks whether two rectangles overlap and returns
    the overlapping rectangle(s).
  Parameters
    ARect, BRect: the two rectangles whos has to be tested
    AWorldModel: the world model simple plane, cylindric or toroid
    AWorldRect: the rectangle with the world borders
    Out CRectangles: an array containing the overlapping rectangles
  Result: True if the rectangles overlapping, false if not
  Remarks: The two rectangles could be placed on any position of the world
    with any allowed size (e.g. not empty and mot larger than the world).
}
function RectanglesOverlap(const ARect, BRect: TLQTRect;
  const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel;
  out CRectangles: TLQTRectArray): Boolean;
var
  ovl : TLQTRectOverlapKind;
begin
  ovl := RectanglesOverlapEx(ARect, BRect, AWorldRect, AWorldModel, CRectangles);
  Result := (ovl <> qroNone);
end;
{ RectanglesOverlapEx the function will return the state of overlapping rectangles
   including the overlapping rectangle, if some exists.
   The calculation is not trivial, since a lot of strange cases exists.
   To ease the stuff, in a first step the simple cases (equality, worldsize) are
   removed, since they need only comparism.
   The next step is to normalize the signs of the axis'.
   Then the overlapping of the "Lines" in horizontal and vertical direction is
    performed. This process acts like a projection / shadow to the X or Y axis.
   If one of this two projection does not show an overlap, than non exists.
   If overlapping exists (= overlapping in X and Y axis) then this overlapping
    rectangles could be created by the compinations of the projection / shadows.
   The next step is to determine the kind of overlapping. This is performed using
    the sizes of the areas of the rectangle and the sum of the overlapping.
    If the overlapping and the rectangles are equal, then the result is equal
    If the overlapping is equal to one of the Input rectangles, this one must be
    covered fully by the other.
   Caution
   The result of this function is dependend from the rounding in the
    hardware, thus it could be possible to have a different result , by comparing
    the fully qualified coords to the result.
}

function RectanglesOverlapEx(const ARect, BRect: TLQTRect;
  const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel;
  out CRectangles: TLQTRectArray): TLQTRectOverlapKind;
var
  lNormWorld : TLQTRect;
  lNormRectA, lNormRectB : TLQTRect;
  i,j : Integer;
  lNormRectC : TLQTRect;
  lCRectangles : TLQTRectArray = Nil;
  ndx : Integer;
  lARectArea, lBRectArea, lCRectArea : Extended;
  eps : Double;
  lHorizOverlaps : TLQTDynDoubleArr = Nil;
  lVertOverlaps : TLQTDynDoubleArr = Nil;
  cntH, cntV : Integer;
// d : Double; // Is used by debugging the one Bit rounding error catching
begin
  Result := qroNone;
  CRectangles := Nil;
  // Sort out the simple cases
  if ARect.IsEqual(BRect) then
  begin
    Result := qroEqual;
    // Rectangles are equal
    SetLength(CRectangles,1);
    CRectangles[0] := ARect;
    Exit;
  end;
  if ARect.IsEqual(AWorldRect) or
     BRect.IsNotArea then
  begin
    Result := qroInside;
    // ARect is the world, so BRect must be inside
    SetLength(CRectangles,1);
    CRectangles[0] := BRect;
    Exit;
  end;
  if BRect.IsEqual(AWorldRect) or
     ARect.IsNotArea then
  begin
    Result := qroFull;
    // BRect is the world, so ARect must be inside
    SetLength(CRectangles,1);
    CRectangles[0] := ARect;
    Exit;
  end;

  // From here the Axis are not default, or if, the world is not flat.
  // Prepare the local normed World and input Rectangles
  lNormWorld.Assign(AWorldRect);
  lNormRectA.Assign(ARect);
  lNormRectB.Assign(BRect);
  lNormRectC.Clear;
  // Norm the sign, so that left is smaller than right, and top is smaller than bottom
  if AWorldRect.WorldXAxisLeftPlus then
  begin
    lNormWorld.Left := -lNormWorld.Left;
    lNormWorld.Right := -lNormWorld.Right;
    lNormRectA.Left := -lNormRectA.Left;
    lNormRectA.Right := -lNormRectA.Right;
    lNormRectB.Left := -lNormRectB.Left;
    lNormRectB.Right := -lNormRectB.Right;
  end;
  if AWorldRect.WorldYAxisUpPlus then
  begin
    lNormWorld.Top := -lNormWorld.Top;
    lNormWorld.Bottom := -lNormWorld.Bottom;
    lNormRectA.Top := -lNormRectA.Top;
    lNormRectA.Bottom := -lNormRectA.Bottom;
    lNormRectB.Top := -lNormRectB.Top;
    lNormRectB.Bottom := -lNormRectB.Bottom;
  end;
  try // the reverse transformation will be done in the finally block

{
    How it works. The BestLineOverlap procedure produce an projection / shadow
    to find overlapping regions within a single line.
    This method is used for the horizontal and vertical coordinates.
    A and B are the projections / shadows for the input rectangle to the horiz and
    vertical axis.
    C is the overlapping of the projection
    In the following example, the shadow produce only one horizontal overlapping
    region, while in the vertical one the rectangles are not overlapping.
    Since 0 * 1 = 0, there is no overlapping found.

      |                         |  A B C
      |  +-------+              |  +
      |  |  A    |              |  |
      |  |       |  +-------+   |  |   +
      |  |       |  |   B   |   |  | + |
      |  +-------+  |       |   |  + | +
      |             |       |   |    |
      |             |       |   |    |
      |             +-------+   |    +
      |                         |
      +-------------------------+  A B C (one field)
    A  +-------+                 A
    B             +-------+      B
    C                            C (no overlapping)
    No overlapping found

    In the next example each horizontal and vertical one overlapping is found.
    Since 1 * 1 = 1, there is one overlapping found.

       |                        |  A B C
       |   +-------+            |  +
       |   |  A    |            |  |
       |   |     +-------+      |  |   +
       |   |     |.. B   |      |  | + |
       |   +-----|..     |      |  + | +
       |         |       |      |    |
       |         |       |      |    |
       |         +-------+      |    +
       |                        |
       +------------------------+  A B C (one field)
   A       +-------+             A
   B             +-------+       B
   C             +-+             C (one field)
   1 (one) overlapping found

    In the last example one horizontal and two vertical overlappings found.
    This is caused by a wrap around of Rectangle A.
    Since 2 * 1 = 2, there are two overlapping founds.


       |                        |  A B C
       >-----------+  +--------->  +
       |      A    |  |         |  |
       |         +-------+      |  |   +
       |         |.. B...|      |  | + |
       >---------|..  ...|------>  + | +
       |         |       |      |    |
       |         |       |      |    |
       |         +-------+      |    +
       |                        |  A B C (one field)
       +------------------------+
   A   >-----------+  +---------> A
   B             +-------+        B
   C             +-+  +--+        C (two fields)
   2 (two) overlappings found
}
    BestLineOverlap(lNormRectA.Left,lNormRectA.Right,lNormRectB.Left,lNormRectB.Right,
                    lHorizOverlaps);
    if Length(lHorizOverlaps) <= 0 then Exit; // No overlappings possible
    BestLineOverlap(lNormRectA.Top,lNormRectA.Bottom,lNormRectB.Top,lNormRectB.Bottom,
                    lVertOverlaps);
    if Length(lVertOverlaps) <= 0 then Exit; // No overlappings possible

    cntH := Length(lHorizOverlaps) div 2; // The arrays contains two Doubles per overlapping
    cntV := Length(lVertOverlaps) div 2;
    SetLength(lCRectangles,cntH*cntV); // The amount of overlapping rectangles
    ndx := 0; // use an index to ease the access
    for i := 0 to cntH-1 do
    begin
      for j := 0 to cntV-1 do
      begin
        lCRectangles[ndx].Left:=lHorizOverlaps[i*2];
        lCRectangles[ndx].Right:=lHorizOverlaps[i*2+1];
        lCRectangles[ndx].Top:=lVertOverlaps[j*2];
        lCRectangles[ndx].Bottom:=lVertOverlaps[j*2+1];
        Inc(ndx); // next rectangle
      end;
    end;
    // Calculate the area size for all of the A, B and C rectangles
    lARectArea := lNormRectA.Area(lNormWorld,AWorldModel);
    lBRectArea := lNormRectB.Area(lNormWorld,AWorldModel);
    lCRectArea := 0.0;
    for i := 0 to High(lCRectangles) do
      lCRectArea := lCRectArea + lCRectangles[i].Area(lNormWorld,AWorldModel);

    // The sizes of the area of the three parts will help us sorting out the
    // cases full, inside, so that partly will remain.
    // Since the overlapping area could not be greater than the sum of overlapping
    // rectangles, subsequently the equality to one of the A or B rectangles means,
    // that one of them is fully covered by the other.
    // If all three parts having the same area size, they must be equal.
    // (Remark: one may say, this case is catched above, but due to rounding this
    // is still here possible).
    // Comparing to equal is dangerous, because in the flow of calculation,
    // the equal values from the beginning may change in one Bit by rounding "errors".
    // To prevent this problem, it is recommended to use an comparism
    // of the Abs(delta)-Values to a small epsilon value.
    // The problem with this approach is, that an secondary error is constructed,
    // in the range of epsilon.
    // So epsilon must be choosen very carefully.
    // What we are doing here is, to use the smallest possible change on the
    // Double value of the greatest of the three areas (ensuring the greatest of the
    // smallest epsilons) and comparing to this.
    // Since rounding is limited to one Bit, and epsilon contain one Bit, this
    // should ensure, that no wrong detection will be made.
    // The ULP does exactly this, returning the step to the next possible Double value.
    if (lCRectArea > lARectArea) and
       (lCRectArea > lBRectArea) then
      eps := ULP(lCRectArea) // Should never be the case
    else if lBRectArea > lARectArea then
      eps := ULP(lBRectArea)
    else
      eps := ULP(lARectArea);
    // Depending of the sizes of the rectangles the values will be very small
    // In the Demo-Applications around 1e-12
    if (Abs(lARectArea-lCRectArea) <= eps) and // The allowed delat is 0.0 and +- one Bit
       (Abs(lBRectArea-lCRectArea) <= eps) then
    begin
{  // For debugging purpose this may help to validate the correct behaviour
      if (lARectArea <> lCRectArea) or
         (lBRectArea <> lCRectArea )then
      begin // for debugging only
        d := lARectArea-lCRectArea;
        d := lBRectArea-lCRectArea;
      end;
}
      Result := qroEqual;
      Exit;
    end;
    // If the ARect area size is identical to the overlapping
    if Abs(lARectArea-lCRectArea) <= eps then
    begin
{  // For debugging purpose this may help to validate the correct behaviour
      if lARectArea <> lCRectArea then
        d := lBRectArea-lCRectArea; // for debugging only
}
      Result := qroFull; // the B covers A fully
      Exit;
    end;
    // If the BRect area size is identical to the overlapping
    if Abs(lBRectArea-lCRectArea) <= eps then
    begin
{  // For debugging purpose this may help to validate the correct behaviour
      if lBRectArea <> lCRectArea then
        d := lBRectArea-lCRectArea;  // for debugging only
}
      Result := qroInside; // the B is located inside A
      Exit;
    end;
    // All other cases has been sorted out
    Result := qroPartly;
  finally
    // Revert the Axis' of the returned values
    SetLength(CRectangles,Length(lCRectangles));
    for i := 0 to High(lCRectangles) do
    begin
      if AWorldRect.WorldXAxisLeftPlus then
      begin
        lCRectangles[i].Left := -lCRectangles[i].Left;
        lCRectangles[i].Right := -lCRectangles[i].Right;
      end;
      if AWorldRect.WorldYAxisUpPlus then
      begin
        lCRectangles[i].Top := -lCRectangles[i].Top;
        lCRectangles[i].Bottom := -lCRectangles[i].Bottom;
      end;
      CRectangles[i] := lCRectangles[i];
    end;
  end;
end;

function PointInRectangle(const ARect: TLQTRect; const APoint: TLQTPoint;
  const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel): Boolean;
begin
  Result := ARect.ContainsPoint(APoint,AWorldRect,AWorldModel);
end;

{ TLQTRect }

function TLQTRect.Height: Double;
begin
  Result := Abs(Top-Bottom);
end;

function TLQTRect.Height(const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel): Double;
var
  r, rw : TLQTRect;
begin
  Result := Height();
  if (AWorldModel = lqtwmFlat) or (AWorldModel = lqtwmCylindrical) then Exit;
  r.Assign(Self);
  rw.Assign(AWorldRect);
  if AWorldRect.WorldXAxisLeftPlus then
  begin
    r.Top := -r.Top;
    r.Bottom := -r.Bottom;
    rw.Top := -rw.Top;
    rw.Bottom := -rw.Bottom;
  end;
  if r.Bottom < r.Top then
    Result := (rw.Bottom-r.Top) + (r.Bottom-rw.Top);
end;


function TLQTRect.Width: Double;
begin
  Result := Abs(Left-Right);
end;

function TLQTRect.Width(const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel): Double;
var
  r, rw : TLQTRect;
begin
  Result := Width();
  if AWorldModel = lqtwmFlat then Exit;
  r.Assign(Self);
  rw.Assign(AWorldRect);
  if AWorldRect.WorldXAxisLeftPlus then
  begin
    r.Left := -r.Left;
    r.Right := -r.Right;
    rw.Left := -rw.Left;
    rw.Right := -rw.Right;
  end;
  if r.Left > r.Right then
    Result := (rw.Right-r.Left) + (r.Right-rw.Left);
end;

function TLQTRect.Area(const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel): Double;
var
  r, rw : TLQTRect;
  w, h : Double;
begin
  w := Width();
  h := Height();
  if AWorldModel <> lqtwmFlat then
  begin
    r.Assign(Self);
    rw.Assign(AWorldRect);
    if AWorldRect.WorldXAxisLeftPlus then
    begin
      r.Left := -r.Left;
      r.Right := -r.Right;
      rw.Left := -rw.Left;
      rw.Right := -rw.Right;
    end;
    if AWorldRect.WorldYAxisUpPlus then
    begin
      r.Top := -r.Top;
      r.Bottom := -r.Bottom;
      rw.Top := -rw.Top;
      rw.Bottom := -rw.Bottom;
    end;

    if r.Left > r.Right then
      w := (rw.Right-r.Left) + (r.Right-rw.Left);
    if r.Bottom < r.Top then
      h := (rw.Bottom-r.Top) + (r.Bottom-rw.Top);
  end;
  Result := w * h;
end;

procedure TLQTRect.Clear;
begin
  SetFrom(0.0,0.0,0.0,0.0);
end;

function TLQTRect.WorldXAxisLeftPlus: Boolean;
begin
  Result := (Left > Right);
end;

function TLQTRect.WorldYAxisUpPlus: Boolean;
begin
  Result := (Top > Bottom);
end;

function TLQTRect.IsPoint: Boolean;
begin
  Result := (Top = Left) and
            (Right = Left) and
            (Bottom = Left);
end;

function TLQTRect.IsLine: Boolean;
begin
  Result := IsPoint;
  if not Result then
    Result := (Left = Right) or (Top = Bottom);
end;

function TLQTRect.IsNotArea: Boolean;
begin
  Result := IsPoint or IsLine;
end;

function TLQTRect.IsValidAsWorld: Boolean;
begin
  Result := False;
  if IsNotArea then Exit; // A world must cover an area
  if HasInfiniteXAxis then
  begin // If the X-Axis infinite, both left and right has be infinite and -infinite
    if ((Left = NegInfinity) and (Right <> Infinity)) or
       ((Left = Infinity) and (Right <> NegInfinity)) or
       ((Right = Infinity) and (Left <> NegInfinity)) or
       ((Right = NegInfinity) and (Left <> Infinity)) then
      Exit;
  end;
  if HasInfiniteYAxis then
  begin // If the Y-Axis infinite, both top and and has be infinite and -infinite
    if ((Top = Infinity) and (Bottom <> NegInfinity)) or
       ((Top = NegInfinity) and (Bottom <> Infinity)) or
       ((Bottom = Infinity) and (Top <> NegInfinity)) or
       ((Bottom = NegInfinity) and (Top <> Infinity)) then
      Exit;
  end;
  Result := True;
end;

function TLQTRect.IsInfinite: Boolean;
begin
  Result := HasInfiniteXAxis or
            HasInfiniteYAxis;
end;

function TLQTRect.HasInfiniteXAxis: Boolean;
begin
  Result := (Left   = Infinity) or
            (Left   = NegInfinity) or
            (Right  = Infinity) or
            (Right  = NegInfinity) or
            (Left <  LQTRectMinValue) or
            (Left > LQTRectMaxValue) or
            (Right <  LQTRectMinValue) or
            (Right > LQTRectMaxValue);
end;

function TLQTRect.HasInfiniteYAxis: Boolean;
begin
  Result := (Top    = Infinity) or
            (Top    = NegInfinity) or
            (Bottom = Infinity) or
            (Bottom = NegInfinity) or
            (Top <  LQTRectMinValue) or
            (Top > LQTRectMaxValue) or
            (Bottom <  LQTRectMinValue) or
            (Bottom > LQTRectMaxValue);
end;

function TLQTRect.IsEqual(ALQTRect: TLQTRect): Boolean;
begin
  Result := (Left = ALQTRect.Left) and
            (Top = ALQTRect.Top) and
            (Right = ALQTRect.Right) and
            (Bottom = ALQTRect.Bottom);
end;

function TLQTRect.ContainsPoint(const ALQTPoint: TLQTPoint): Boolean;
var
  w2, h2 : Double;
  cx, cy : Double;
begin
  // the center of the node is the average of the borders, regardless of the signs
  cx := (Right+Left) / 2.0;
  cy := (Bottom+Top) / 2.0;
  // The width and the height is the Abs() of the delta between the sides, regardless of the signs
  w2 := Abs(Right-Left) / 2.0;
  h2 := Abs(Top-Bottom) / 2.0;
  // Inside is if the distance in X and Y are smaller than half the width / height
  // or Coords laying on the left or top border
  Result := ((Abs(cx-ALQTPoint.X) < w2) or (ALQTPoint.X = Left)) and
           ((Abs(cy-ALQTPoint.Y) < h2) or (ALQTPoint.Y = Top));
end;

function TLQTRect.ContainsPoint(const AX, AY: Double): Boolean;
var
  pt : TLQTPoint;
begin
  pt.X := AX;
  pt.Y := AY;
  Result := ContainsPoint(pt);
end;

function TLQTRect.ContainsPoint(const APoint: TLQTPoint;
  const AWorldRect: TLQTRect; const AWorldModel: TLQTWorldModel): Boolean;
var
  lARectangles : TLQTRectArray = Nil;
  lWorldRect: TLQTRect;
  lPoint: TLQTPoint;
  i : Integer;
begin
  if (not IsXWrappedAround(AWorldRect)) and
     (not IsYWrappedAround(AWorldRect)) then
  begin // If this rectangle is an ordinary rectangle (not wraping around the world borders)
    Result := ContainsPoint(APoint);
    Exit;
  end;
  Result := False;
  lPoint := APoint;
  // Normalize the Rectangle, this may include spliting into up to four regular rectangles
  // including the sign switching
  NormalizeRectangle(Self,AWorldRect,AWorldModel,lARectangles,lWorldRect);
  // Check if the signs of the Point-Coords must be inverted too
  if WorldXAxisLeftPlus <> lWorldRect.WorldXAxisLeftPlus then
    lPoint.X := -lPoint.X;
  if WorldYAxisUpPlus <> lWorldRect.WorldYAxisUpPlus then
    lPoint.Y := -lPoint.Y;
  // Now check if one of the rectangle contains the point
  for i := 0 to High(lARectangles) do
  begin
    Result := lARectangles[i].ContainsPoint(lPoint);
    if Result then Exit;
  end;
end;

procedure TLQTRect.NormateToWorld(const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel);
var
  lXAxisLeftPlus, lYAxisUpPlus : Boolean;
  d : Double;
  w,h : Double;
  lRect : TLQTRect;
  lWorldRect : TLQTRect;
begin
  if not AWorldRect.IsValidAsWorld then
    raise ELQTGException.Create('Could not normate to an invalid world');
  lXAxisLeftPlus := AWorldRect.WorldXAxisLeftPlus;
  lYAxisUpPlus := AWorldRect.WorldYAxisUpPlus;
  // To get rid of the complicated different axis orientations
  // we copy the rectangles into local variables and normalize them
  lRect.Assign(Self);
  lWorldRect.Assign(AWorldRect);

  if lXAxisLeftPlus then
  begin
    lRect.Left := -lRect.Left;
    lRect.Right := -lRect.Right;
    lWorldRect.Left := -lWorldRect.Left;
    lWorldRect.Right := -lWorldRect.Right;
  end;
  if lYAxisUpPlus then
  begin
    lRect.Top := -lRect.Top;
    lRect.Bottom := -lRect.Bottom;
    lWorldRect.Top := -lWorldRect.Top;
    lWorldRect.Bottom := -lWorldRect.Bottom;
  end;
  try
    if AWorldModel <> lqtwmFlat then
    begin
      // The world is either cylindical or toroidal.
      // In both cases the X-Axis may wrap around.
      // If the X-Axis is infinite, nothing could be done here ...
      if not AWorldRect.HasInfiniteXAxis then
      begin
        // ... but in all other cases values larger / smaller than the height
        // of the world extent could be wrapped to be located inside
        w := AWorldRect.Width;
        if (lRect.Left < lWorldRect.Left) or
          (lRect.Left > lWorldRect.Right) then
        begin
          d := (lRect.Left-lWorldRect.Left) / w;
          d := Frac(d);
          if d < 0.0 then
            d := d + 1.0;
          d := d * w;
          lRect.Left := d+lWorldRect.Left;
        end;
        if (lRect.Right < lWorldRect.Left) or
          (lRect.Right > lWorldRect.Right) then
        begin
          d := (lRect.Right-lWorldRect.Left) / w;
          d := Frac(d);
          if d < 0.0 then
            d := d + 1.0;
          d := d * w;
          lRect.Right := d+lWorldRect.Left;
        end;
      end;
      // Ff the world is toroidal the Y-Axis may wrap around.
      // If the Y-Axis is infinite, nothing could be done here ...
      if (AWorldModel = lqtwmToroidal) and
         (not AWorldRect.HasInfiniteYAxis) then
      begin
        // ... but in all other cases values larger / smaller than the width
        // of the world extent could be wrapped to be located inside
        h := AWorldRect.Height;
        if (lRect.Top < lWorldRect.Top) or
           (lRect.Top > lWorldRect.Bottom) then
        begin
          d := (lRect.Top-lWorldRect.Top) / h;
          d := Frac(d);
          if d < 0.0 then
            d := d + 1.0;
          d := d * h;
          lRect.Top := d+lWorldRect.Top;
        end;
        if (lRect.Bottom < lWorldRect.Top) or
           (lRect.Bottom > lWorldRect.Bottom) then
        begin
          d := (lRect.Bottom-lWorldRect.Top) / h;
          d := Frac(d);
          if d < 0.0 then
            d := d + 1.0;
          d := d * h;
          lRect.Bottom := d+lWorldRect.Top;
        end;
      end;
     end;
    // In all cases limit the values to be inside the world.
    // Actual, in the toroidal world all values are inside here.
    // Limit the overall values
    if (AWorldModel = lqtwmFlat) then
    begin  // the right an left values needs only to be fixed in the flat world
      if (lRect.Left > lRect.Right) then
      begin
        d := lRect.Left;
        lRect.Left := lRect.Right;
        lRect.Right := d;
      end;
      if lRect.Left < lWorldRect.Left then
        lRect.Left := lWorldRect.Left;
      if lRect.Right > lWorldRect.Right then
        lRect.Right := lWorldRect.Right;
    end;
    if (AWorldModel <> lqtwmToroidal) then
    begin // the top and bottom values needs to be fixed in the flat and cylindrical world
      if (lRect.Top > lRect.Bottom) then
      begin
        d := lRect.Top;
        lRect.Top := lRect.Bottom;
        lRect.Bottom := d;
      end;
      if lRect.Top < lWorldRect.Top then
        lRect.Top := lWorldRect.Top;
      if lRect.Bottom > lWorldRect.Bottom then
        lRect.Bottom := lWorldRect.Bottom;
    end;
  finally
    // set the sign of values to the original orientation
    if lXAxisLeftPlus then
    begin
      lRect.Left := -lRect.Left;
      lRect.Right := -lRect.Right;
    end;
    if lYAxisUpPlus then
    begin
      lRect.Top := -lRect.Top;
      lRect.Bottom := -lRect.Bottom;
    end;
    Assign(lRect);
  end;
end;

procedure TLQTRect.Transform(const ASourceWorld, ATargetWorld: TLQTRect);
var
  lRect : TLQTRect;
begin
  TransformPoint(Left, Top,
                 ASourceWorld,ATargetWorld,
                 lRect.Left, lRect.Top);
  TransformPoint(Right, Bottom,
                 ASourceWorld,ATargetWorld,
                 lRect.Right, lRect.Bottom);
  Assign(lRect);
end;

procedure TLQTRect.Assign(const ALQTRect: TLQTRect);
begin
  Left := ALQTRect.Left;
  Top := ALQTRect.Top;
  Right := ALQTRect.Right;
  Bottom := ALQTRect.Bottom;
end;

procedure TLQTRect.SetFrom(const ALeft, ATop, ARight, ABottom: Double);
begin
  Left := ALeft;
  Top := ATop;
  Right := ARight;
  Bottom := ABottom;
end;

procedure TLQTRect.SetFromPoint(const APoint: TLQTPoint);
begin
  Left := APoint.X;
  Right := NextAfter(Left, MaxDouble);
  Top := APoint.Y;
  Bottom := NextAfter(Top, MaxDouble);
end;

procedure TLQTRect.SetFromPoint(const APoint: TLQTPoint;
  const AWorldRect: TLQTRect);
begin
  Left := APoint.X;
  Top := APoint.Y;
  if AWorldRect.WorldXAxisLeftPlus then
    Right := NextAfter(Left, -MaxDouble)
  else
    Right := NextAfter(Left, MaxDouble);
  if AWorldRect.WorldYAxisUpPlus then
    Bottom := NextAfter(Top, -MaxDouble)
  else
    Bottom := NextAfter(Top, MaxDouble)
end;

function TLQTRect.TryExtentWith(const ALQTRect: TLQTRect): Boolean;
begin
  Result := False;
  if ((Left = ALQTRect.Left) and (Right = ALQTRect.Right)) then
  begin // Both rectangles have the same width
    if (Top = ALQTRect.Bottom) then
    begin // The passed rectangle fits on the top
      Top := ALQTRect.Top;
      Result := True;
    end
    else if (Bottom = ALQTRect.Top) then
    begin // The passed rectangle fits on the bottom
      Bottom := ALQTRect.Bottom;
      Result := True;
    end;
  end
  else if ((Top = ALQTRect.Top) and (Bottom = ALQTRect.Bottom)) then
  begin // both rectangles have the same height
    if (Left = ALQTRect.Right) then
    begin // The passed rectangle fits on the left
      Left := ALQTRect.Left;
      Result := True;
    end
    else if (Right = ALQTRect.Left) then
    begin // The passed rectangle fits on the right
      Right := ALQTRect.Right;
      Result := True;
    end;
  end;
end;

function TLQTRect.TryExtentWith(const ALQTRect: TLQTRect;
  const ANormalizeWorld: TLQTRect; const AWorldModel: TLQTWorldModel): Boolean;
begin
  Result := TryExtentWith(ALQTRect);
  if Result or (AWorldModel = lqtwmFlat) then Exit; // If something found or flat world, exit
  // Now we no matching rectangles so far and a non flat world
  // In the cylindrical world right an left of the world could be extended
  if ((Top = ALQTRect.Top) and (Bottom = ALQTRect.Bottom)) then
  begin // both rectangles have the same height
    if (Left = ANormalizeWorld.Left) and
       (ALQTRect.Right = ANormalizeWorld.Right) then
    begin // The passed rectangle fits on the left
      Left := ALQTRect.Left;
      Result := True;
    end
    else if (Right = ANormalizeWorld.Right) and
       (ALQTRect.Left = ANormalizeWorld.Left) then
    begin // The passed rectangle fits on the right
      Right := ALQTRect.Right;
      Result := True;
    end;
  end;
  if Result or (AWorldModel = lqtwmCylindrical) then Exit; // If something found or cylindrical world, exit
  // Now only the toroidial world is left, so try to extent the top or bottom
  if ((Left = ALQTRect.Left) and (Right = ALQTRect.Right)) then
  begin // Both rectangles have the same width
    if (Top = ANormalizeWorld.Top) and
       (ALQTRect.Bottom = ANormalizeWorld.Bottom) then
    begin // The passed rectangle fits on the top
      Top := ALQTRect.Top;
      Result := True;
    end
    else if (Bottom = ANormalizeWorld.Bottom) and
         (ALQTRect.Top = ANormalizeWorld.Top) then
    begin // The passed rectangle fits on the bottom
      Bottom := ALQTRect.Bottom;
      Result := True;
    end;
  end;
end;

function TLQTRect.IsXWrappedAround(const AWorldRect: TLQTRect): Boolean;
begin
  if AWorldRect.WorldXAxisLeftPlus then
    Result := (Left < Right)
  else
    Result := (Left > Right);
end;

function TLQTRect.IsYWrappedAround(const AWorldRect: TLQTRect): Boolean;
begin
  if AWorldRect.WorldYAxisUpPlus then
    Result := (Top < Bottom)
  else
    Result := (Top > Bottom);
end;

function TLQTRect.Union(const ALQTRect: TLQTRect; const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel): Boolean;
var
  r : TLQTRect;
  lNormWorld : TLQTRect;
  lNormSelf, lNormRect : TLQTRect;
begin
  // Special case, Rectangles are equal
  Result := False;
  if IsEqual(ALQTRect) then Exit;

  // The case that the world is flat and default axis are used
  // Speed up the things
  if (AWorldModel = lqtwmFlat) and
     (not AWorldRect.WorldXAxisLeftPlus) and
     (not AWorldRect.WorldYAxisUpPlus) then
  begin
    if IsYWrappedAround(AWorldRect) or ALQTRect.IsYWrappedAround(AWorldRect) or
       IsXWrappedAround(AWorldRect) or ALQTRect.IsXWrappedAround(AWorldRect) then
      raise ELQTGException.Create('Rectangle malformed!');
    r := Self;
    r := r + ALQTRect; // use the overwritten class operator TLQTRect.+(a, b: TLQTRect): TLQTRect;
    Result := not IsEqual(r); // see if something changes
    if Result then // if something changed
      Assign(r); // then assign the new values
    Exit;
  end;

  // From here the Axis are not default, or if then the world are not flat
  // Prepare the local normed Self, World and input Rectangle
  lNormWorld.Assign(AWorldRect);
  lNormRect.Assign(ALQTRect);
  lNormSelf.Assign(Self);
  // Norm the sign, so that left is smaller than right, and top is smalle than bottom
  if AWorldRect.WorldXAxisLeftPlus then
  begin
    lNormWorld.Left := -lNormWorld.Left;
    lNormWorld.Right := -lNormWorld.Right;
    lNormRect.Left := -lNormRect.Left;
    lNormRect.Right := -lNormRect.Right;
    lNormSelf.Left := -lNormSelf.Left;
    lNormSelf.Right := -lNormSelf.Right;
  end;
  if AWorldRect.WorldYAxisUpPlus then
  begin
    lNormWorld.Top := -lNormWorld.Top;
    lNormWorld.Bottom := -lNormWorld.Bottom;
    lNormRect.Top := -lNormRect.Top;
    lNormRect.Bottom := -lNormRect.Bottom;
    lNormSelf.Top := -lNormSelf.Top;
    lNormSelf.Bottom := -lNormSelf.Bottom;
  end;
  try // the reverse transformation will be done in the finally block
    // In the flat world, ab lot of calculation could be shortend.
    if (AWorldModel = lqtwmFlat) then
    begin
      if IsYWrappedAround(AWorldRect) or ALQTRect.IsYWrappedAround(AWorldRect) or
         IsXWrappedAround(AWorldRect) or ALQTRect.IsXWrappedAround(AWorldRect) then
        raise ELQTGException.Create('Rectangle malformed!');
      lNormSelf := lNormSelf + lNormRect; // use the overwritten class operator TLQTRect.+(a, b: TLQTRect): TLQTRect;
      Exit;
    end;
    // Validate the Y-Axis for the cylindrical world
    if (AWorldModel = lqtwmCylindrical) and
       (IsYWrappedAround(AWorldRect) or ALQTRect.IsYWrappedAround(AWorldRect)) then
        raise ELQTGException.Create('Rectangle malformed!');

    // Find the shortest horizontal length, that covers both rectangles
    BestLineUnion(lNormSelf.Left,lNormSelf.Right,
                        lNormRect.Left, lNormRect.Right,
                        lNormWorld.Left, lNormWorld.Right,
                        AWorldModel <> lqtwmFlat, // On the X-axis wrapping is allowed in cylindrical and toroidal worlds
                        r.Left, r.Right);
    // Find the shortest vertical length, that covers both rectangles
    BestLineUnion(lNormSelf.Top,lNormSelf.Bottom,
                        lNormRect.Top, lNormRect.Bottom,
                        lNormWorld.Top, lNormWorld.Bottom,
                        AWorldModel = lqtwmToroidal, // On the Y-axis wrapping is allowed in toroidal worlds only
                        r.Top, r.Bottom);
    lNormSelf.Assign(r); // Copy the overlapping rectangle
  finally
    // Revert the Axis'
    if AWorldRect.WorldXAxisLeftPlus then
    begin
      lNormSelf.Left := -lNormSelf.Left;
      lNormSelf.Right := -lNormSelf.Right;
    end;
    if AWorldRect.WorldYAxisUpPlus then
    begin
      lNormSelf.Top := -lNormSelf.Top;
      lNormSelf.Bottom := -lNormSelf.Bottom;
    end;
    Result := not IsEqual(lNormSelf); // see if something changes
    if Result then // if something changed
      Assign(lNormSelf); // then assign the new values
  end;
end;

function TLQTRect.Union(const APoint: TLQTPoint): Boolean;
var
  w2, h2 : Double;
  cx, cy : Double;
  b : Boolean;
  dl, dr, dt, db : Double;
  dirx, diry : Double;
begin
  Result := False; // Preset with no change
  // the center of the node is the average of the borders, regardless of the signs
  cx := (Right+Left) / 2.0;
  cy := (Bottom+Top) / 2.0;
  // The width and the height is the Abs() of the delta between the sides, regardless of the signs
  w2 := Abs(Right-Left) / 2.0;
  h2 := Abs(Top-Bottom) / 2.0;
  // Inside is if the distance in X and Y are smaller than half the width / height
  // or Coords laying on the left or top border
  b := ((Abs(cx-APoint.X) < w2) or (APoint.X = Left)) and
       ((Abs(cy-APoint.Y) < h2) or (APoint.Y = Top));
  if b then Exit; // The point is in the rectangle
  // The point has to be included
  if Right > Left then
    dirx := MaxDouble
  else
    dirx := -MaxDouble;
  if Bottom > Top then
    diry := MaxDouble
  else
    diry := -MaxDouble;
  dl := Abs(APoint.X-Left);
  dr := Abs(APoint.X-Right);
  dt := Abs(APoint.Y-Top);
  db := Abs(APoint.Y-Bottom);
  if (dl+dr > w2*2.0) or // If the X is outside the rectangle,
     (APoint.X = Right) then // or X on the right border
  begin
    if dl < dr then
      Left := APoint.X  // Left side includes the point
    else
      Right := NextAfter(APoint.X,dirx); // Right side needs an incremented
  end;
  if (dt+db > h2*2.0) or // If the Y is outside the rectangle,
     (APoint.Y = Bottom)  then // or if Y on the bottom border
  begin
    if dt < db then
      Top := APoint.Y // Top side includes the point
    else
      Bottom := NextAfter(APoint.Y,diry); // Bottom side needs an increment
  end;
  Result := True;
end;

function TLQTRect.Union(const APoint: TLQTPoint; const AWorldRect: TLQTRect;
  const AWorldModel: TLQTWorldModel): Boolean;
  { BestLinePointExtent, return the smallest length, covering all three points
     of the line and the point formed by the parameters.
     CAUTION: All values expected to be normalized (Left < Right, Top < Bottom)
              Wrap around is allowed!
              Never mix Left/Right with Top/Bottom!!
     For a real rectangle this procedure should be called two times.
     One time for left and right values and the other time for top and bottom values.
     Parameter
      InA: Rectangle Left  (or in a different call Top)
      InB: Rectangle Right (or in a different call Bottom)
      InC: Point X (or in a different call Y)
      AWorld1: World Left  (or in a different call  Top)
      AWorld2: World Right (or in a different call  Bottom)
     Out
      Out1: The Left or Top value
      Out2: The Right or Bottom value

     Example
        W1  B   C   A  W2
        AB  >---+       +--->
     C          +
     Oa >-------+   +--->  Possibility one
     Ob >---+   +------->  Possibility two
     Out1 and Out2 will be the shorter value-pair out of [C0, C1]

  }
  procedure BestLinePointExtent(const InA,InB,InC,AWorld1,AWorld2 : Double;
                                const AWrapAllowed : Boolean;
                                out Out1, Out2 : Double);
  var
    // A sorted array for the Values InA1, InA2, InB1, InB2
    lAppearance : array[0..2] of TLineKeyValue;
    i : Integer;
    // kv : TLineKeyValue;
    d0, d1 : Double;
    c : Cardinal;
    ndxC, ndxA, ndxB  : Integer;
  begin
    // Sort ABC in the correct order (avoiding the usage of Bubble Sort)
    // First sort A and B
    if InA > InB then
    begin
      ndxA := 1;
      ndxB := 0;
    end
    else
    begin
      ndxA := 0;
      ndxB := 1;
    end;
    lAppearance[ndxA].Key   := $AA; // note the Hex-Notation for the key
    lAppearance[ndxA].Value := InA;
    lAppearance[ndxB].Key   := $BB; // note the Hex-Notation for the key
    lAppearance[ndxB].Value := InB;

    // Now the Point C. May sit at the end, in the middle or in the beginning
    if (InC >= lAppearance[1].Value) then // C is biggest, or equal value
    begin
      // If C is equal, than it should stay here too, because
      // if equal to B (the end), this position is out of the rectangle
      // if equal to A (the beginning), this position is in the rectangle
      ndxC := 2;
    end
    else if (InC >= lAppearance[0].Value) then
    begin
      // Make room in the middle
      lAppearance[2] := lAppearance[1];
      ndxC := 1;
    end
    else
    begin
      // Make room at the beginning
      lAppearance[2] := lAppearance[1];
      lAppearance[1] := lAppearance[0];
      ndxC := 0;
    end;
    lAppearance[ndxC].Key   := $CC;
    lAppearance[ndxC].Value := InC;

(*

    // Drop the parameters unsorted into the array
    lAppearance[0].Key   := $AA; // see the Hex-Notation for the key
    lAppearance[0].Value := InA;
    lAppearance[1].Key   := $BB;
    lAppearance[1].Value := InB;
    lAppearance[2].Key   := $CC;
    lAppearance[2].Value := InC;
    // Bubble Sort: Sort alle values from small to big.
    // For the 3 values this method seems to be ok.
    repeat
      b := False; // reset to "No item swapped"
      for i := 1 to 2 do
      begin
        // Check if we have to swap a value
        if (lAppearance[i-1].Value > lAppearance[i].Value) or // The previous value is bigger than this one, swap them
           ((lAppearance[i-1].Value = lAppearance[i].Value) and // On equal values
            // $AA is left of $BB, the beginning of the rectangle should be left from the ending
            // $CC if right form $AA, since the point is included (Left or Top is part of the rectangle)
            // $CC is right from $BB, since the point in not included (Right or Top is not part of the rectangle)
            ((lAppearance[i-1].Key = $BB) and (lAppearance[0].Key = $AA))) then
        begin
          b := True; // Set the "we swapped a pair"
          // Save the target
          kv := lAppearance[i-1];
          // Move the value one field
          lAppearance[i-1] := lAppearance[i];
          // Update the new backward field
          lAppearance[i] := kv;
        end;
      end;
    until not b; // finish, if one loop does not touch the order.
*)
    // Now we have all value in the right order
    // 4 fields, we have 6 possible permutations
    // https://en.wikipedia.org/wiki/Permutation
    // Create the case value by shift and oring the four keys together
    c := lAppearance[0].Key;
    for i := 1 to 2 do
    begin
      c := c shl 8;
      c := c or lAppearance[i].Key;
    end;
    // Now the key shows the order of increasing values and which value sit in
    // Caution: adjacent values could be equal!
    case c of
      $AABBCC : // Rectangle followed by Point
        begin
          //    W1  A   B   C  W2
          // AB     +---+
          // C              +
          // Oa >-------+   +--->  Possibility one
          // Ob     +-------+      Possibility two
          Out1 := InA;
          Out2 := NextAfter(InC, AWorld2); // NextAfter() ensures, that the Point is inside the Rectangle
          if AWrapAllowed then
          begin
            d0 := Abs((InB - AWorld1) + (AWorld2 - InC));
            d1 := Abs(InC - InA);
            if d0 <= d1 then
            begin
              Out1 := InC;
              Out2 := InB;
            end;
          end;
        end;
      $AACCBB, // Point inside Rectangle
      $BBAACC, // Point inside Rectangle (wrapped)
      $CCBBAA: // Point inside Rectangle (wrapped)
        begin
          //    W1  A   C   B  W2
          // AB     +-------+
          // C          +
          // O      +-------+

          //    W1  B   A   C  W2
          // AB >---+   +------->
          // C              +
          // O  >---+   +------->

          //    W1  C   B   A  W2
          // AB >-------+   +--->
          // C      +
          // O  >-------+   +--->

          Out1 := InA;
          Out2 := InB;
        end;
      $CCAABB :  // Point followed by rectangle
        begin
          //    W1  C   A   B  W2
          // AB         +---+
          // C      +
          // Oa     +-------+      Possibility one (no wrap)
          // Ob >---+   +------->  Possibility two (wrapped)
          Out1 := InC;
          Out2 := InB;
          if AWrapAllowed then
          begin
            d0 := Abs(InC - InB);
            d1 := Abs((InC - AWorld1) + (AWorld2 - InA));
            if d1 < d0 then
            begin
              Out1 := InA;
              Out2 := NextAfter(InC,AWorld2); // NextAfter() ensures, that the Point is inside the Rectangle
            end
          end;
        end;
      $BBCCAA : // Point followed by rectangle (wrapped)
        begin
          if not AWrapAllowed then
            ELQTGException.Create('Wrap around detected, but allowed!');
          //    W1  B   C   A  W2
          // AB >---+       +--->
          // C          +
          // Oa >---+   +------->  Possibility one
          // Ob >-------+   +--->  Possibility two
          d0 := Abs((AWorld2 - InC) + (InB - AWorld1));
          d1 := Abs((AWorld2 - InA) + (InC - AWorld1));
          if d0 < d1 then
          begin
            Out1 := InC;
            Out2 := InB;
          end
          else
          begin
            Out1 := InA;
            Out2 := NextAfter(InC, AWorld2); // NextAfter() ensures, that the Point is inside the Rectangle
          end;
        end;
    else
    end;
  end;


var
  lNormWorld,lNormSelf : TLQTRect;
  lPoint: TLQTPoint;
  r : TLQTRect;
begin

  if (AWorldModel = lqtwmFlat) (* and
     (not AWorldRect.WorldXAxisLeftPlus) and
     (not AWorldRect.WorldYAxisUpPlus) *) then
  begin
    Result := Union(APoint);
    Exit;
  end;

  Result := False;
  lPoint.X := APoint.X;
  lPoint.Y := APoint.Y;
  lNormSelf.Assign(Self);
  lNormWorld.Assign(AWorldRect);
  if AWorldRect.WorldXAxisLeftPlus then
  begin
    lNormSelf.Left    := -lNormSelf.Left;
    lNormSelf.Right := -lNormSelf.Right;
    lNormWorld.Left    := -lNormWorld.Left;
    lNormWorld.Right := -lNormWorld.Right;
    lPoint.X := -lPoint.X;
  end;
  if AWorldRect.WorldYAxisUpPlus then
  begin
    lNormSelf.Top    := -lNormSelf.Top;
    lNormSelf.Bottom := -lNormSelf.Bottom;
    lNormWorld.Top    := -lNormWorld.Top;
    lNormWorld.Bottom := -lNormWorld.Bottom;
    lPoint.Y := -lPoint.Y;
  end;
  try
    r.Assign(lNormSelf);
    BestLinePointExtent(lNormSelf.Left,lNormSelf.Right,lPoint.X,
                        lNormWorld.Left, lNormWorld.Right,
                        (AWorldModel <> lqtwmFlat),
                        r.Left, r.Right);
    BestLinePointExtent(lNormSelf.Top,lNormSelf.Bottom,lPoint.Y,
                        lNormWorld.Top, lNormWorld.Bottom,
                        (AWorldModel = lqtwmToroidal),
                        r.Top, r.Bottom);
    lNormSelf.Assign(r);
  finally
    if AWorldRect.WorldXAxisLeftPlus then
    begin
      lNormSelf.Left    := -lNormSelf.Left;
      lNormSelf.Right := -lNormSelf.Right;
    end;
    if AWorldRect.WorldYAxisUpPlus then
    begin
      lNormSelf.Top    := -lNormSelf.Top;
      lNormSelf.Bottom := -lNormSelf.Bottom;
    end;
    Result := not IsEqual(lNormSelf);
    if Result then
      Assign(lNormSelf);
  end;
end;

function TLQTRect.CoordsAsString(const AFormatString: String): String;
var
  s : String;
begin
  s := AFormatString;
  if Length(s) <= 0 then
    s := 'L: %g; T: %g; R: %g; B: %g';
  Result := Format(s,[Left,Top,Right,Bottom]);
end;

class operator TLQTRect.=(a, b: TLQTRect): Boolean;
begin
  Result := a.IsEqual(b);
end;

class operator TLQTRect.+(a, b: TLQTRect): TLQTRect;
begin
  Result.Left := Min(a.Left, b.Left);
  Result.Top := Min(a.Top, b.Top);
  Result.Right := Max(a.Right, b.Right);
  Result.Bottom := Max(a.Bottom, b.Bottom);
end;


end.

