{
uMouseRectShifter.pas

A helper class for shifting rectangles on a paintbox or simmilar components

License: The same modified LGPL as the Free Pascal RTL
         See the file COPYING.modifiedLGPL for more details

Author: Ekkehard Domning (ekkehard@domning.eu)

}
unit uMouseRectShifter;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, ExtCtrls,
  Contnrs;
type
  { class TMouseHitItem
    A helper class to hold a single detectable Item, Like corner, edge, etc.
  }

  TMouseHitItem = class(TObject)
  private
    FTag : Integer;
    FUserData : Pointer;
    FZOrder : Integer;
    FActiveRect : TRect;
    FCursor : TCursor;
    FItemHit : Boolean;
    FMouseDownFlag : Boolean;
    FOrgX, FOrgY : Integer;
    FMouseDownX, FMouseDownY : Integer;
    FMouseDeltaX, FMouseDeltaY : Integer;
  public
    { Tag is almost free usable by the user
      In the used examples and in the example AddRectangle below
      its partly used, see also tagRectXXX const definition }
    property Tag : Integer read FTag write FTag;
    { UserDate is free usable by the user }
    property UserData : Pointer read FUserData write FUserData;
    { ZOrder is used to priorize between overlaping items }
    property ZOrder : Integer read FZOrder write FZOrder;
    { ActiveRect the rectangle who is sensitiv }
    property ActiveRect : TRect read FActiveRect write FActiveRect;
    { The cursor for this item crSize, crSizeNS etc. }
    property Cursor : TCursor read FCursor  write FCursor;
    { The original X position of the items (usually the center of the ActiveRect) }
    property OrgX : Integer read FOrgX;
    { The original Y position of the items (usually the center of the ActiveRect) }
    property OrgY : Integer read FOrgY;
    { ItemHit becomes true if in the MouseMove or MouseDown method, the X,Y values
      of the Mouse are contained in the ActiveRect.
      If the MouseDownFlag is true, the ItemHit stayes true until MouseUp }
    property ItemHit : Boolean read FItemHit;
    { MouseDownFlag, true if the item is selected by the MouseDown event }
    property MouseDownFlag : Boolean read FMouseDownFlag write FMouseDownFlag;
    { The X value of the Mouse during selection }
    property MouseDownX : Integer read FMouseDownX;
    { The Y value of the Mouse during selection }
    property MouseDownY : Integer read FMouseDownY;
    { MouseDeltaX contains the delta between the original X Mouse Value and the current MouseMove X-Value }
    property MouseDeltaX : Integer read FMouseDeltaX;
    { MouseDeltaY contains the delta between the original Y Mouse Value and the current MouseMove Y-Value }
    property MouseDeltaY : Integer read FMouseDeltaY;
    { OnMouseDown: To be called within the original OnMouseDown event
        if hit the MouseDownFlag is set and the passed X, Y values are stored in the
        MouseDownX, MouseDownY properties
        After testing all items, the MouseDownFlag must be reset of all but the highest Z-Order
    }
    procedure OnMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    { OnMouseMove: Call within the original OnMouseDown event
        The method behaves different depending of the MouseDown property value.
        If MouseDown is true than the MouseDeltaX and MouseDeltaY properties are updated.
        If MouseDown is false the ItemHit property are updated.
    }
    procedure OnMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    { OnMouseUp: Resets the properties MouseDown, MouseDeltaX, MouseDeltaY }
    procedure OnMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    constructor Create;
    { constructor Create with extended parameters to initialize the fields direct on creation }
    constructor Create(const AActiveRect : TRect;
                       const AOrgX : Integer;
                       const AOrgY : Integer;
                       const AZOrder : Integer = 0;
                       const ACursor : TCursor = crSize;
                       const ATag : Integer = 0;
                       const AUserData : Pointer = Nil);

  end;

  { TMouseRectShifter a class which holds several clickable items and select one of them  }

  TMouseRectShifter = class(TObject)
  private
    FMouseHitItems : TObjectList; // The list with the clickable items
    function GetCurrentItem : TMouseHitItem;
    function GetItemsCount: Integer;
    function GetItems(Index : Integer) : TMouseHitItem;
  public
    { property ItemsCount, returns the number of items in the list }
    property ItemsCount : Integer read GetItemsCount;
    { property Items, allows the reading access to the items. Caution: never free a retrieved item!! }
    property Items[Index : Integer] : TMouseHitItem read GetItems;
    { property CurrentItem returns either the Item where the Mouse is down or the one where the mouse is hovering above }
    property CurrentItem : TMouseHitItem read GetCurrentItem;
    { procedure OnMouseDown must be called on the MouseDown event of the GUI-Component, most likely a PaintBox }
    procedure OnMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    { procedure OnMouseMove must be called on the MouseDown event of the GUI-Component, most likely a PaintBox.
      ACurrentItem will be filled with the active item, or Nil if none }
    procedure OnMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer; out ACurrentItem : TMouseHitItem);
    { procedure OnMouseMove must be called on the MouseDown event of the GUI-Component, most likely a PaintBox. }
    procedure OnMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    { procedure Clear the list of all items. Usually called in MouseUp and prior the rebuilding of a new setup }
    procedure Clear;
    { procedure AddRectangle adds all needed items for a full rectangle.
        Those parts are the area, the borders and the corners.
        Each Part is "tagged" with on of the tagRectXXX const, so that they
        could be identified later in the OnMouseMove event or by accessing
        the items by iterating through the items list
      Parameter
        ARect the rectangle that must be added
        AInflate the value to increase the areas so that they are clickable.
        AZOrder to prior the different parts. The value is used for the Area,
          AZOrder+1 for the edges, AZOrder+2 for the corners.
          If several different Rectangles are used one may use Z-Orders
          separated by 3 or better 10 for each of them.
        ATag a user value to distingiush between the Items.
          The lowest 4 bits of the Tag overwritten by the tagRectXXX const.
          This (Item.Tag and $F) will give the tagRectXXX  value back.
        AUserData a fully user controlled value.
    }
    procedure AddRectangle(const ARect : TRect;
                           const AInflate : Integer;
                           const AZOrder : Integer;
                           const ATag : Integer = 0;
                           const AUserData : Pointer = Nil);
    { procedure AddRectangleEx adds specific items for a full rectangle.
        For all Parameters except ARectParts, see the AddRectangle method.
      Extended Parameter
        ARectParts is an dynamic array which should contain only tagRectXXX const.
        For each of those values the specific item is created.
        This allows the generation partly clickable rectangles.
        Example: Passing
        [tagRectArea, tagRectRight, tagRectBottom, tagRectBottomRight]
        will create a rectangle who's area could be hit and the right and bottom
        border and the bottom-right corner too.
        All other items (corners and left and top borders) are not clickable.
    }
    procedure AddRectangleEx(const ARect : TRect;
                           const ARectParts : array of Integer;
                           const AInflate : Integer;
                           const AZOrder : Integer;
                           const ATag : Integer = 0;
                           const AUserData : Pointer = Nil);
    constructor Create;
    destructor Destroy;override;
  end;

const
  tagRectArea        = 1;
  tagRectLeft        = 2;
  tagRectTop         = 3;
  tagRectRight       = 4;
  tagRectBottom      = 5;
  tagRectTopLeft     = 6;
  tagRectTopRight    = 7;
  tagRectBottomRight = 8;
  tagRectBottomLeft  = 9;

implementation

{ function IsInRectangle returns true, if the passed X and Y coords are inside the rectangle
   Parameter:
     ARect : The rectangle to test
     AX, AY : The coords to test
     AInflate : The amount of inflating the rectangle
   Remarks: The function works different from the definition in uLazQuadTreeGeometry.
     The rectangle here is treated having the right and bottom border as part of the
     rectangle! This is needed, because the right and bottom border must be hit for
     resizing.
}
function IsInRectangle(const ARect : TRect; const AX, AY : Integer; const AInflate : Integer = 0): Boolean;
var
  r : TRect;
begin
  r := ARect;
  if AInflate > 0 then
    r.Inflate(AInflate,AInflate);
  Result := (AX >= r.Left) and (AX <= r.Right) and
            (AY >= r.Top) and (AY <= r.Bottom);
end;

{ TMouseHitItem }

procedure TMouseHitItem.OnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if IsInRectangle(FActiveRect, X, Y, 0) then
  begin
    FMouseDownFlag := True;
    FMouseDeltaX := 0;
    FMouseDeltaY := 0;
    FMouseDownX := X;
    FMouseDownY := Y;
  end;
end;

procedure TMouseHitItem.OnMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if FMouseDownFlag then
  begin
    FMouseDeltaX := X-FMouseDownX;
    FMouseDeltaY := Y-FMouseDownY;
  end
  else
  begin
    FMouseDeltaX := 0;
    FMouseDeltaY := 0;
    FItemHit := IsInRectangle(FActiveRect, X, Y, 0);
  end;
end;

procedure TMouseHitItem.OnMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if FMouseDownFlag then
  begin
    FMouseDeltaX := X-FMouseDownX;
    FMouseDeltaY := Y-FMouseDownY;
    FMouseDownFlag := False;
  end
  else
  begin
    FMouseDeltaX := 0;
    FMouseDeltaY := 0;
  end;
end;

constructor TMouseHitItem.Create;
begin
  inherited Create;
end;

constructor TMouseHitItem.Create(const AActiveRect: TRect;
  const AOrgX: Integer; const AOrgY: Integer; const AZOrder: Integer;
  const ACursor: TCursor; const ATag: Integer; const AUserData: Pointer);
begin
  Create();
  FActiveRect := AActiveRect;
  FOrgX :=  AOrgX;
  FOrgY :=  AOrgY;
  FZOrder :=  AZOrder;
  FCursor :=  ACursor;
  FTag :=  ATag;
  FUserData :=  AUserData;
end;

function TMouseRectShifter.GetCurrentItem: TMouseHitItem;
var
  i : Integer;
  lItem : TMouseHitItem;
begin
  Result := Nil;
  for i := 0 to ItemsCount-1 do
  begin
    lItem := TMouseHitItem(FMouseHitItems[i]);
    if lItem.FMouseDownFlag then
    begin
      if (not Assigned(Result)) or
         (Result.ZOrder < lItem.ZOrder) then
        Result := lItem;
    end;
  end;
  if Assigned(Result) then Exit;
  for i := 0 to ItemsCount-1 do
  begin
    lItem := TMouseHitItem(FMouseHitItems[i]);
    if lItem.ItemHit then
    begin
      if (not Assigned(Result)) or
         (Result.ZOrder < lItem.ZOrder) then
        Result := lItem;
    end;
  end;
end;

function TMouseRectShifter.GetItemsCount: Integer;
begin
  Result := FMouseHitItems.Count;
end;

function TMouseRectShifter.GetItems(Index: Integer): TMouseHitItem;
begin
  Result := TMouseHitItem(FMouseHitItems[Index]);
end;

procedure TMouseRectShifter.OnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  i : Integer;
  lItem, lCurrItem : TMouseHitItem;
begin
  // Forward the MouseDown-Event to all Items
  for i := 0 to ItemsCount-1 do
  begin
    lItem := TMouseHitItem(FMouseHitItems[i]);
    lItem.OnMouseDown(Sender,Button,Shift,X,Y);
  end;

  // Resetting all other Item's MouseDown-Flags but not from the current item.
  // Might not be necessary, but double tap ;-)
  lCurrItem := GetCurrentItem;
  if Assigned(lCurrItem) then
  begin
    for i := 0 to ItemsCount-1 do
    begin
      lItem := TMouseHitItem(FMouseHitItems[i]);
      if lItem = lCurrItem then Continue;
      if lItem.MouseDownFlag then
        lItem.FMouseDownFlag := False;
    end;
  end;
end;

procedure TMouseRectShifter.OnMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer; out ACurrentItem: TMouseHitItem);
var
  i : Integer;
  lItem : TMouseHitItem;
begin
  for i := 0 to ItemsCount-1 do
  begin
    lItem := TMouseHitItem(FMouseHitItems[i]);
    lItem.OnMouseMove(Sender,Shift,X,Y);
  end;
  ACurrentItem := GetCurrentItem;
end;

procedure TMouseRectShifter.OnMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  i : Integer;
  lItem : TMouseHitItem;
begin
  for i := 0 to ItemsCount-1 do
  begin
    lItem := TMouseHitItem(FMouseHitItems[i]);
    lItem.OnMouseUp(Sender,Button,Shift,X,Y);
  end;
end;

procedure TMouseRectShifter.Clear;
begin
  FMouseHitItems.Clear;
end;

procedure TMouseRectShifter.AddRectangle(const ARect: TRect;
  const AInflate: Integer; const AZOrder: Integer; const ATag: Integer;
  const AUserData: Pointer);
begin
  AddRectangleEx(ARect,
                 [tagRectArea,tagRectLeft, tagRectTop, tagRectRight, tagRectBottom,
                  tagRectTopLeft, tagRectTopRight, tagRectBottomRight, tagRectBottomLeft],
                 AInflate,AZOrder,ATag,AUserData);
end;

procedure TMouseRectShifter.AddRectangleEx(const ARect: TRect;
  const ARectParts: array of Integer; const AInflate: Integer;
  const AZOrder: Integer; const ATag: Integer; const AUserData: Pointer);
var
  i : Integer;
  llr : TRect;
  lItem : TMouseHitItem = Nil;
  lTag : Integer;
begin
  lTag := ATag and (not $F); // Clear lowest 4 bits
  for i := 0 to High(ARectParts) do
  begin
    case ARectParts[i] of
      tagRectArea :
        begin
          lItem := TMouseHitItem.Create(ARect,ARect.Left,ARect.Top,AZOrder+0,crSize,
                                      lTag or tagRectArea,AUserData);
        end;
      tagRectLeft :
        begin
          // Left Border
          llr.Left:=ARect.Left-AInflate;
          llr.Top:=ARect.Top;
          llr.Right:=ARect.Left+AInflate;
          llr.Bottom:=ARect.Bottom;
          lItem := TMouseHitItem.Create(llr,ARect.Left,ARect.Top,AZOrder+1,crSizeWE,
                                        lTag or tagRectLeft,AUserData);
        end;
      tagRectTop :
        begin
          // Top Border
          llr.Left:=ARect.Left;
          llr.Top:=ARect.Top-AInflate;
          llr.Right:=ARect.Right;
          llr.Bottom:=ARect.Top+AInflate;
          lItem := TMouseHitItem.Create(llr,ARect.Left,ARect.Top,AZOrder+1,crSizeNS,
                                        lTag or tagRectTop,AUserData);
        end;
      tagRectRight :
        begin
          // Right Border
          llr.Left:=ARect.Right-AInflate;
          llr.Top:=ARect.Top;
          llr.Right:=ARect.Right+AInflate;
          llr.Bottom:=ARect.Bottom;
          lItem := TMouseHitItem.Create(llr,ARect.Right,ARect.Top,AZOrder+1,crSizeWE,
                                        lTag or tagRectRight,AUserData);
        end;
      tagRectBottom :
        begin
          // Bottom border
          llr.Left:=ARect.Left;
          llr.Top:=ARect.Bottom-AInflate;
          llr.Right:=ARect.Right;
          llr.Bottom:=ARect.Bottom+AInflate;
          lItem := TMouseHitItem.Create(llr,ARect.Left,ARect.Bottom,AZOrder+1,crSizeNS,
                                        lTag or tagRectBottom,AUserData);
        end;
      tagRectTopLeft :
        begin
          // Northwest edge
          llr.Left := ARect.Left;
          llr.Top := ARect.Top;
          llr.Right := ARect.Left;
          llr.Bottom := ARect.Top;
          llr.Inflate(AInflate,AInflate);
          lItem := TMouseHitItem.Create(llr,ARect.Left,ARect.Top,AZOrder+2,crSizeNWSE,
                                        lTag or tagRectTopLeft,AUserData);
        end;
      tagRectTopRight :
        begin
          // NorthEast edge
          llr.Left := ARect.Right;
          llr.Top := ARect.Top;
          llr.Right := ARect.Right;
          llr.Bottom := ARect.Top;
          llr.Inflate(AInflate,AInflate);
          lItem := TMouseHitItem.Create(llr,ARect.Right,ARect.Top,AZOrder+2,crSizeNESW,
                                        lTag or tagRectTopRight,AUserData);
        end;
      tagRectBottomRight :
        begin
          // SouthEast edge
          llr.Left := ARect.Right;
          llr.Top := ARect.Bottom;
          llr.Right := ARect.Right;
          llr.Bottom := ARect.Bottom;
          llr.Inflate(AInflate,AInflate);
          lItem := TMouseHitItem.Create(llr,ARect.Right,ARect.Bottom,AZOrder+2,crSizeNWSE,
                                        lTag or tagRectBottomRight,AUserData);
        end;
      tagRectBottomLeft :
        begin
          // SouthWest edge
          llr.Left := ARect.Left;
          llr.Top := ARect.Bottom;
          llr.Right := ARect.Left;
          llr.Bottom := ARect.Bottom;
          llr.Inflate(AInflate,AInflate);
          lItem := TMouseHitItem.Create(llr,ARect.Left,ARect.Bottom,AZOrder+2,crSizeNESW,
                                        lTag or tagRectBottomLeft,AUserData);
        end;
    end;
    if Assigned(lItem) then
      FMouseHitItems.Add(lItem);
    lItem := Nil;
  end;
end;

constructor TMouseRectShifter.Create;
begin
  inherited Create;
  FMouseHitItems := TObjectList.Create(True);
end;

destructor TMouseRectShifter.Destroy;
begin
  FreeAndNil(FMouseHitItems);
  inherited Destroy;
end;


end.

